// =====================================================================================================================================================================================================
// (c) 2022 Lynn Hansen, KU7Q															                                                                                                               |
// This Source Code Form is subject to the terms of the GNU GENERAL PUBLIC LICENSE, Version 3, 29 June 2007. A copy of this license can be found here: https://choosealicense.com/licenses/gpl-3.0/|
// =====================================================================================================================================================================================================



/*
 Name:		CTR2_Mini.ino
 Created:	12/9/2021 2:04:23 PM
 Author:	Lynn Hansen, KU7Q
*/

/*
* NOTE: The WIO Terminal does not have a backup battery to keep the clock alive - this doesn't serve any purpose :(
#include <RTC_SAMD51.h> //from https://github.com/Seeed-Studio/Seeed_Arduino_RTC
//#include <RTC_SAMD21.h>
#include <DateTime.h>
*/
#include <Crc16.h>
#include <seeed_rpcUnified.h>
#include <rpc_api_utils.h>
#include <rpcBLEDevice.h>
//#include <WiFiClient.h>
//#include <WiFi.h>
#include <rpcWiFi.h>
#include <WiFiUdp.h>
//#include <WiFiServer.h>


#include <SD.h>
#include <SAMD_ISR_Timer.h> //download from https://github.com/khoih-prog/SAMD_TimerInterrupt#why-do-we-need-this-samd_timerinterrupt-library
#include <SAMDTimerInterrupt.h>
#include <TFT_eSPI.h> //refer to https://wiki.seeedstudio.com/Wio-Terminal-LCD-Overview/
#include "Free_Fonts.h" //refer to https://wiki.seeedstudio.com/Wio-Terminal-LCD-Fonts/

String version = "Rev: 1.03.04"; //this appears in PrtSplash

File myFile;

TFT_eSPI tft;

SAMDTimer ITimer0(TIMER_TC3); //set ISR hardware timer
// Init SAMD_ISR_Timer
// Each SAMD_ISR_Timer can service 16 different ISR-based timers
//SAMD_ISR_Timer ISR_Timer;
//start timer


//set ISR interrupt timer here
#define HW_TIMER_INTERVAL_1MS         1000L		//ISR at 1 mS rate

//#define TIMER_INTERVAL_2MS           2000L      // this is in uSec
//#define TIMER_INTERVAL_1000HZ		 1000L	     // period for 1000 hz for sidetone

/*
#ifdef _SAMD21_
#define SDCARD_SS_PIN 1
#define SDCARD_SPI SPI
#endif 
*/

#define RADIO   Serial1  //use for radio protocol


//----------------------------------------------------------------------------------------
//Debug switches - uncomment to send debug data to Serial

//comment this out to turn off Serial.print/prnitln debug outputs 
#define enabSerialOut 1 

//uncomment to display flex api transactions
//#define DEBUG_FLEX_API     1


//----------------------------------------------------------------------------------------


//hardware pin assingments 
#define PTT_OUT			D0 //active HIGH - D0 is blown on Dev 1 Unit
#define ROTARY_SW       D1 
#define KEY_OUT			D2 //active HIGH - D2 is blown on Dev 1 Unit
#define SIDETONE		D3 
#define RMT_BTNS		A4 //use ADC for resistive ladder for multiple buttons
#define PDL_LEFT		D5
#define ROTARY_A		D8
#define PDL_RIGHT		D7 
#define ROTARY_B		D6

//button enumeration
#define BTN_NONE        0
#define BTN_A           1
#define BTN_B           2
#define BTN_C           3
#define BTN_UP          4
#define BTN_DOWN        5
#define BTN_LEFT        6
#define BTN_RIGHT       7
#define BTN_SW          8

//HMI modes
#define HMI_HOME        1 //home page
#define HMI_MENU		2 //menus active
#define HMI_HELP		3 //help displayed
#define HMI_VALUE		4 //encoder used to change value
#define HMI_TXTINPUT    5 //encoder or keyer used to change txtInput[]

//HMI Themes
#define HMI_THEME_GREY		0
#define HMI_THEME_VIOLET	1
#define HMI_THEME_BLUE		2
#define HMI_THEME_MAROON	3
#define HMI_THEME_CYAN		4
#define HMI_THEME_GREEN		5
#define HMI_THEME_OLIVE		6

//Start of unique menu options

//Menu ID - all menus must have a unique ID#

//NOTE: Change NextMnu() if changing order of menus

//NOTE: *** Add new menus to GetMenuName() so they appear properly in Terminal mode
//      *** Add help pages in Help.ino when adding new menus

//Main Menu definitions
#define MNU_NONE        0 //home page
#define MNU_MAIN        1
#define MNU_FREQ		2
#define MNU_BAND        3
#define MNU_MODE        4
#define MNU_VFO			5
#define MNU_TXMSG		6
#define MNU_TXMSG_EDIT	7
#define MNU_KEYER		8
#define MNU_WIFI		9
#define MNU_LINK        10
#define MNU_ANT_SW		11 //only appears on Mini-A if ASC is online
#define MNU_RADIOPORT	12
#define MNU_CONFIG	    13 //end of main menu items
#define MNU_RADIO		14 //only used for Flex, PCR1000 and future advanced radio settings

//Start of sub menus items that have sub menus or use the GetTxtInput() function
//These must be unique IDs so the program knows what to do with the selected values

#define MNU_FREQ_LD     100
#define MNU_FREQ_SV     101
#define MNU_FREQ_REN    102
						
//config menu options
#define MNU_C_BTN		120
#define MNU_FUNC_BTN	121
#define MNU_THEME		122
#define MNU_TAGLINE		123
#define MNU_RADIOCAT	124
#define MNU_RADIOBAUD	125
#define MNU_AMPSEQ		126

//custom radios appear on Main menu when they are selected
#define MNU_FLEX		140
#define MNU_PCR1000		141

//Menu IDs for submenus

//CW Setting options requiring a menu ID
#define MNU_CW_KEYERS   160

//WiFi Settings
#define MNU_WIFI_SSID_EDIT		180
#define MNU_WIFI_STAKEY_EDIT	181

//Link settings
#define MNU_UDP_ADRS_EDIT		190
#define MNU_UDP_ADRS_ASC_EDIT   191

//Flex Setting
#define MNU_FLEX_IP_EDIT		200
#define MNU_FLEX_PORT_EDIT		201
#define MNU_FLEX_SLICE_EDIT		202

//Contest Keyer setting
#define MNU_KEYER_CALL_EDIT		220
#define MNU_KEYER_EXCH_EDIT		221

//Tag line setting
#define MNU_CONFIG_TAGLINE_EDIT 240

//Function button menu
#define MNU_FUNC_BTN_OPTIONS	250

//ant switch menu
#define MNU_ANT_SW_ANT_A_EDIT	260
#define MNU_ANT_SW_ANT_B_EDIT	261
#define MNU_ANT_SW_LBL_EDIT		262

//rigctld ip 
#define MNU_RIGCTLD_IP_EDIT		270

//End of unique menu options

//The following are submenu IDs used to locate the submenu items in the parent menu
//The ID is the order in which each menu appears in the submenu

//Frequency menu definitions
#define MNU_FREQ_ZERO       0
#define MNU_FREQ_RIT		1
#define MNU_FREQ_DB         2
#define MNU_FREQ_LOAD       3 //this is the freq load list under MNU_FREQ_LD main menu
#define MNU_FREQ_SAVE       4 //this is the freq save list under MNU_FREQ_SV main menu
#define MNU_FREQ_RENAME     5 //this is the freq rename list under the MNU_FREQ_REN main menu
#define MNU_FREQ_PREV		6

//vfo menu definitions
#define MNU_VFO_A			0
#define MNU_VFO_B			1 
#define MNU_VFO_SPLIT		2 //vfo-a=rx, vfo-b=tx
#define MNU_VFO_MEM			3 //tune using memory
#define MNU_VFO_MODE		4 //tune using a or b vfo
#define MNU_VFO_LOCK		5
#define MNU_VFO_FUTURE		6
#define MNU_VFO_COPYAB		7
#define MNU_VFO_COPYBA      8
#define MNU_VFO_SWAPAB		9
#define MNU_VFO_COPY2A		10
#define MNU_VFO_PUSH		11 //push freq & mode to other Mini (Basic Link mode only)
#define MNU_VFO_PULL		12 //pull freq & mode from other Mini (Basic Link mode only)

//Config menu definitions
#define MNU_CONFIG_C_BTN		0 
#define MNU_CONFIG_FUNC			1
#define MNU_CONFIG_THEME		2
#define MNU_CONFIG_TAGLINE		3
#define MNU_CONFIG_CAT			4
#define MNU_CONFIG_BAUD			5
#define MNU_CONFIG_AMPSEQ		6


//[C] Btn Select menu definitions
#define MNU_C_BTN_PTT			0
#define MNU_C_BTN_RIT			1
#define MNU_C_BTN_SPEED			2
#define MNU_C_BTN_FREQ			3
#define MNU_C_BTN_MODE			4
#define MNU_C_BTN_BAND			5
#define MNU_C_BTN_VFO			6
#define MNU_C_BTN_TXMSG			7
#define MNU_C_BTN_EDIT_TXMSG	8
#define MNU_C_BTN_KEYER			9
#define MNU_C_BTN_WIFI			10
#define MNU_C_BTN_LINK			11
#define MNU_C_BTN_ASC			12
#define MNU_C_BTN_FUNC_BTN		13
#define MNU_C_BTN_PUSH			14 //Push info to other Mini in Basic sync mode
#define MNU_C_BTN_PULL			15 //Pull info from other Mini in Basic sync mode
#define MNU_C_BTN_RADIO			16

//Function button option menu definitions
#define MNU_FUNC_OPTION_TXMSG1		0
#define MNU_FUNC_OPTION_TXMSG2		1
#define MNU_FUNC_OPTION_TXMSG3		2
#define MNU_FUNC_OPTION_TXMSG4		3
#define MNU_FUNC_OPTION_TXMSG5		4
#define MNU_FUNC_OPTION_TXMSG6		5
#define MNU_FUNC_OPTION_TXMSG7		6
#define MNU_FUNC_OPTION_RIT			7
#define MNU_FUNC_OPTION_SPEED		8
#define MNU_FUNC_OPTION_TXMSG		9
#define MNU_FUNC_OPTION_BTNA		10
#define MNU_FUNC_OPTION_BTNB		11
#define MNU_FUNC_OPTION_BTNC		12
#define MNU_FUNC_OPTION_ANT_SW		13 //only if ant sw is online

//link menu definitions
#define MNU_LINK_CONNECT		0 //Link menu options
#define MNU_LINK_SVR_CLIENT		1 //server/client
#define MNU_LINK_MODE			2 //sync options only visible on server
#define MNU_LINK_TYPE			3 //toggles gLinkTypeUDP, false= BLE, true=UDP
#define MNU_LINK_AUTO			4 //autoconnect 
#define MNU_LINK_UDP_ADRS		5 //to set the WiFi UDP address for Link
//the next menu item is also used for ASC IP adrs on Mini-A
#define MNU_LINK_OPTION6		6 //allow user to offset radio B freq for zero-beat

//Ant switch menu definitions
#define MNU_ANT_SW_PORT_A		0
#define MNU_ANT_SW_PORT_B		1
#define MNU_ANT_SW_SWAP			2
#define MNU_ANT_SW_LOCK			3
#define MNU_ANT_SW_ENABLE		4
#define MNU_ANT_SW_LABELS		5
#define MNU_ANT_SW_SEND_BAND	6

//Keyer menu defintions
#define MNU_KEYER_SPEED		0
#define MNU_KEYER_FARNS		1
#define MNU_KEYER_SIDETONE	2
#define MNU_KEYER_TYPE		3
#define MNU_KEYER_PDL		4
#define MNU_KEYER_PTT		5 //three modes, normal PTT, latching PTT, and straight key
#define MNU_KEYER_SPARE		6 //just display a blank space - for future option - this puts SN and Exch on right column
#define MNU_KEYER_CALL		7
#define MNU_KEYER_SN		8 //contest SN, enter # into buffer to send
#define MNU_KEYER_EXCH		9 //contest Exchange, enter % into buffer to send

#define MNU_RMT_NORM		0 //Rmt PTT modes
#define MNU_RMT_LATCH       1
#define MNU_RMT_KEY			2

//Icom adrs edit
#define MNU_ICOM_ADRS_EDIT	0

//WiFi Menu definitions
#define MNU_WIFI_CONNECT	0
#define MNU_WIFI_SSID		1
#define MNU_WIFI_STAKEY		2 //password
#define MNU_WIFI_AUTO		3 //auto connect


//radio port copy/erase options
#define MNU_PORT_COPY		17
#define MNU_PORT_RENAME		18
#define MNU_PORT_ERASE		20


//Flex Menu definitions
#define MNU_FLEX_IP			0 //radio IP adrs
#define MNU_FLEX_PORT		1 //ip port - normally 4992 for local radio, change for remote
#define MNU_FLEX_PAN        2 //panadapter to use for selected slice
#define MNU_FLEX_SLICE		3 //slice to create/connect to
#define MNU_FLEX_ANT		4 //ant 0 or 1
#define MNU_FLEX_TOGGLE		5 //atu tune
#define MNU_FLEX_VOL		6 //volume
#define MNU_FLEX_PO			7 //output power
#define MNU_FLEX_TO			8 //tune power
#define MNU_FLEX_WNB		9
#define MNU_FLEX_NB			10
#define MNU_FLEX_NR			11
#define MNU_FLEX_AF 		12
#define MNU_FLEX_FUTURE		13
#define MNU_FLEX_PTT		14 //hardware or api ptt
#define MNU_FLEX_KEYER		15 //0=hardware, 1=cw key (network)
#define MNU_FLEX_DELAY		16 //key delay - use more for low latency networks when using cw key 


//#define MNU_FLEX_ATU_STAT	14 //not used, just display status result in status window

//PCR-1000 Menu definitions
#define MNU_PCR_VOL			0
#define MNU_PCR_SQ			1
#define MNU_PCR_IFSHIFT		2
#define MNU_PCR_ZEROIF		3
#define MNU_PCR_DSP			4
#define MNU_PCR_DSPNR		5
#define MNU_PCR_ANch        6
#define MNU_PCR_BW			7
#define MNU_PCR_AGC			8 
#define MNU_PCR_ATTEN		9
#define MNU_PCR_NB			10


//RADIO TYPES*********************************************************************************************************************************
#define RADIO_CAT_NONE			0
#define RADIO_CAT_FLEX			1
#define RADIO_CAT_PCR1000		2
#define RADIO_CAT_ICOM			3
#define RADIO_CAT_KENWOOD1		4
#define RADIO_CAT_KENWOOD2		5
#define RADIO_CAT_YAESUFTDX		6
#define RADIO_CAT_YAESU100		7
#define RADIO_CAT_YAESU757		8 
#define RADIO_CAT_YAESU8x7		9
#define RADIO_CAT_YAESU890		10
#define RADIO_CAT_YAESU1000		11
#define RADIO_CAT_RIGCTLD		12 //used for Hamlib rigctld daemon - default port is 4532
#define RADIO_CAT_IPADRS		13 //only visible when rigctld is selected
#define RADIO_CAT_NUMRADIOS		14 //this must match the # of entries in the radioList[] array

//MODES **************************************************************************************************************************************
#define RADIO_CW			0 //coorlates to Mode page Radio Mode options
#define RADIO_CW_R			1
#define RADIO_LSB			2
#define RADIO_USB			3
#define RADIO_FM_N			4
#define RADIO_FM_W			5
#define RADIO_AM			6
#define RADIO_DIG_L			7
#define RADIO_DIG_H			8

//transmit modes
#define RADIO_TX_OFF		0 //transmit modes
#define RADIO_TX_ENAB		1 //tx enabled
#define RADIO_TX_KEY		2 //transmitting using paddle ptt
#define RADIO_TX_PTT		3 //transmitting using ptt
#define RADIO_TX_WDOG		4 //watchdog timer expired - print yellow warning label

//keyer mode enumeration
#define KEYER_NONE			0 //keyer modes
#define KEYER_STRAIGHT		1
#define KEYER_PASSTHRU      2 //passes tip to Key and ring to PTT toward radio
#define KEYER_IAMBIC_A		3
#define KEYER_IAMBIC_B		4
#define KEYER_ULTIMATIC     5
#define KEYER_BUG			6

#define KEYER_MODE_STOP     0 //keyer modes
#define KEYER_MODE_RUN      1
#define KEYER_MODE_PAUSE    2

//esp tx enumeration
#define WIFI_FLEX			0 //tx to flex
#define ESP_WEB				1 //tx to web server

//comm mode enumeration
#define COM_MODE_DEBUG      0 //idle, print out serial messages
#define COM_MODE_CAT        1 //process cat commands
#define COM_MODE_TERM		2 //process keyboard commands from a terminal program

#define FLEX_CON_GUI		1
#define FLEX_CON_NON_GUI	2
#define FLEX_CON_CONNECTING 3

#define TCP_CON_OFF			0
#define TCP_CON_ACTIVE		1

#define LONG_PRESS_LAST		0
#define LONG_PRESS_NO		1
#define LONG_PRESS_YES		2

//remote PTT ADC functions
#define RMT_FUNC_PTT		0 //Remote PTT pushed
#define RMT_FUNC_F1			1 //function keys from resistive ladder
#define RMT_FUNC_F2			2
#define RMT_FUNC_F3			3
#define RMT_FUNC_F4			4
#define RMT_FUNC_F5			5
#define RMT_FUNC_OFF		255


//link message pre and postambles - set in first and last byte of link message
#define LINK_PREAMBLE_ASC		0x6e
#define LINK_POSTAMBLE_ASC		0x6f

#define LINK_PREAMBLE_SVR		0x7e
#define LINK_POSTAMBLE_SVR		0x7f

#define LINK_PREAMBLE_CLIENT	0x3e
#define LINK_POSTAMBLE_CLIENT	0x3f

//ary[] indexes (byte mapping) for Link mode - 0 and last byte are sync bytes that flag server/client as source	
// set gNumLinkBytes to match # of bytes below
#define LINK_PREAMBLE		0
#define LINK_FLAGS			1
#define LINK_FREQ			2 //2-5 = freq
#define LINK_BAND			6
#define LINK_RMODE			7
#define LINK_RIT			8 //8-11 = RIT freq
#define LINK_CW_SPEED		12
#define LINK_FARNS_SPEED	13
#define LINK_KEY_TYPE		14
#define LINK_KEYER_SN		15 //15-16
#define LINK_FUTURE1		17 //was status of ant sw B which is now sent in a full ASC update from A to B
#define LINK_AUX1			18 //Aux variables for PCR and Flex values
#define LINK_AUX2			19
#define LINK_AUX3			20
#define LINK_AUX4			21
#define LINK_AUX5			22
#define LINK_AUX6			23
#define LINK_AUX7			24
#define LINK_AUX8			25
#define LINK_AUX9			26
#define LINK_AUX10			27
#define LINK_AUX11			28
#define LINK_AUX12			29
#define LINK_AUX13			30 
//crc MSB and LSB follow this byte
//crc MSB = byte 31
//crc LSB = byte 32
#define LINK_POSTAMBLE		33 //last byte

//data locations in ASC link message
#define LINK_ASC_PREAMBLE	0
#define LINK_ASC_SWACTIVE	1
#define LINK_ASC_SWDUAL		2
#define LINK_ASC_SW8PORT	3
#define LINK_ASC_SWLOCK		4
#define LINK_ASC_SWA		5
#define LINK_ASC_SWB		6
#define LINK_ASC_STARTLBL	7

//gSyncMode enumeration
#define SYNC_MODE_BASIC		0
#define SYNC_MODE_FREQ		1
#define SYNC_MODE_TRACK		2
#define SYNC_MODE_RMT		3

//flags for Basic Link mode push/pull operation
#define SYNC_MODE_BASIC_NONE	0
#define SYNC_MODE_BASIC_PUSH	1
#define SYNC_MODE_BASIC_PULL	2

//data locations in ASC command msg
#define LINK_ASC_CMD		1
#define LINK_ASC_AUX1		2 //used for miniSel or gTxMode
#define LINK_ASC_AUX2		3 //used for antSel or gTxModeRmt

//Antenna Switch Controller Commands
#define ASC_CMD_A_PORT		100 //set A port ant
#define ASC_CMD_B_PORT		101 //set B port ant
#define ASC_CMD_ANT_SWAP	102 //swap A & B Ports
#define ASC_CMD_LOCK		103 //set lock mode
#define ASC_CMD_ENABLE		104 //set enabled mode
#define ASC_CMD_STATUS		105 //resent current status
#define ASC_CMD_BAND		106 //band info from Mini to ASC - used to select ant by band - multiple bands can be set for each antenna
#define ASC_CMD_LABEL		107 //new label for ant - includes point# and 10 chr label
#define ASC_CMD_TXMODE		108 //send A and B gTxMode to ASC - it will open antenna on radio that is inactive if the other has Tx enabled
//end of asc commands - always leave this at the end when adding new commands
#define ASC_CMD_END			109 

//========================================================================================================

//BLE CONFIGURATIONS
//Service UUID
// Use 16 bit numbers for development, change to these on release
// 

//BLEUUID servUUID(0x1101);
//BLEUUID charUUID(0x2101);


//unique Characteristic UUIDs for release

BLEUUID servUUID("b44cb851-83ea-4c65-8f1a-8689c7897685");
BLEUUID charUUID("643fb00d-36de-43a1-b04e-9c3a6bba10ed");


BLEServer* pServer;
BLEService* pService;
//values for server
BLECharacteristic* pLocalParams; //parameters from local client

//values for client
BLEClient* pClient;
BLERemoteService* pRmtService;
BLERemoteCharacteristic* pRmtParams; //parameters from remote server
BLERemoteCharacteristic* pRmtParamsPrev; //look for change
BLEAdvertisedDevice* pRmtDevice; //server info
BLEAdvertising* pAdvertising;
BLEScan* pBLEScan;

//BLE/UDP LINK options
const int gNumLinkBytes = 34; //determines # of bytes sent on link
const int gNumASCLinkBytes = 32; //# of bytes sent from ASC w/ start & end sync + crc msb & lsb

bool gBLEConnect = false; //command state
bool gBLEConnectPrev = false;
bool gLinkMiniA = true; //true=Mini A (server), false=Mini B (client)
bool gLinkRxData = false; // goes true when we receive an update from the other end
bool gLinkAutoConnect = false; //true to auto connect Link on boot - wifi autoconnect must be set for UDP to auto connect
bool gLinkOnline = false; //goes true when we rx link packets - used to detect first connection so we can reset previous link offsets

volatile bool gBLEConnected = false; //connection state
bool gBLEConnectedPrev = false;
uint16_t gBLEHandle = 0; //session handle
volatile bool gBLEKillScan = false; //set to true to stop scanning after we found our ctr2 server

//variables for Link using WiFi UDP
bool gLinkTypeUDP = false; //false=BLE, true=WiFi UDP for Link connection
bool gUDPConnected = false;
String gUDPAdrs = ""; //wifi address of other Mini for Link UDP connection
String gUDPAdrsASC = ""; //adrs of ASC - Mini-A only
IPAddress udpAdrs = { 0,0,0,0 };
IPAddress udpAdrsASC = {0, 0, 0, 0};
int udpPort = 42022; //month and year UDP was implimented, for want of another port #

uint8_t gSyncMode = SYNC_MODE_BASIC; //basic=0, freq=1, track=2, rmt=3
uint8_t gSyncModePrev = gSyncMode;

bool gSyncOffset = false; //goes true if our freq is offset if we're in Freq Sync mode and other Mini Tx is enabled
int gLinkTrackOffset = 0;//holds the calculated offset when gLinkTrack first goes true
int gLink_ZeroBeat = 0; //holds the freq offset for Mini-B so you can zero-beat radios
std::string gLinkMsgPrev = ""; //used in GetLinkParams() to check for link msg change
std::string gLinkASCMsgPrev = ""; //used for ASC link 

const uint32_t gLinkCommTimeoutDefault = 5000; //grayout the link icon if we don't get an update after 5 seconds (they're sent every 3/4 second)
uint32_t gLinkCommTimeout = 0; //holds the millis() of the last valid link packet from the far end
uint32_t gLinkASCCommTimeout = 0;
bool gLinkCommFail = false; //goes true if no link packets after timeout

const uint32_t gLinkChngTimerDefault = 260; //wait x mS after Link Tx or Rx freq so we track properly
uint32_t gLinkTxChngTimer = 0; //blocks Link Rx for timer default after sending freq change
uint32_t gLinkRxChngTimer = 0; //blocks Link Tx for timer default after receiving a freq change from other Mini

//========================================================================================================

//global varibles

//terminal movement enumeration - can't sent these in #define above
//see https://www.csie.ntu.edu.tw/~r92094/c++/VT100.html for additional controls

String TERM_CURSOR_UP = "\033[1A";  //move cursor up one line
String TERM_CUROSR_DOWN = "\033[1B";  //move cursor down one line
String TERM_CURSOR_RIGHT = "\033[1C";  //move cursor right one column
String TERM_CURSOR_LEFT = "\033[1D";   //move cursor left one column
String TERM_CLS = "\033[2J"; //clear terminal screen
String TERM_NORM = "\033[0m"; //set normal text
String TERM_INV = "\033[7m"; //invert text
String TERM_HIDE_CURSOR = "\033[?25l";
String TERM_SHOW_CURSOR = "\033[?25h";

bool gInit = false; //goes true after all the setup files are loaded - blocks looking at arrays until they're loaded

bool blockISR = false; //block ISR during save operations
bool ISRActive = false; //goes true during ISR

int hmiMode = HMI_HOME; //tracks state of hmi, home =1, menus active=2, value input = 3

int gStuckBtn = 0; //ignore rmt func buttons if any button pressed > 5 seconds

//HMI Theme variables = default to HMI_THEME_BLUE

uint16_t hmiFGnd = TFT_WHITE; //control text
uint16_t hmiBGnd = TFT_BLUE; //control background
uint16_t hmiBase = TFT_BLACK; //screen color

bool gBtnSwLongPress = false; //for 5-way switch
bool gBtnALongPress = false; //short press, set true if btnA pressed longer than 1 sec
bool gBtnBLongPress = false; //for b btn
bool gBtnCLongPress = false; //for c btn
volatile uint8_t gBtnPress = 0; //holds button press from ISR so we can process it in Loop

volatile int gMnuValue = 0; //holds value for HMI_VALUE mode
int gMnuValuePrev = 0; //hold previous value from long-press, if !0, restore value on next long press
int gMnuValueMin = 0; //holds minimum value
int gMnuValueMax = 0; //holds max value
uint8_t gLongPress = LONG_PRESS_NO; //if set to yes, allows short press
bool gLongPressToggle = false; //goes true if toggling value to with long-press
String lastStatusMsg = ""; //holds the last menu's status info so we can redisplay in termnal when we use up/down arrows to change selection

bool gZeroDigits = false; //zero all digits below step digit if this is true
uint8_t gMSel = 0; //memSel status - 1s based -- holds selected memory location from VFO Mem page - reset to 0 once changed
uint8_t gMSelPtr = 0; //points to current memory selection
uint8_t gMSelPrev = 0; //tracks changes
bool gMSelMode = false; //false=tune VFO-A memory tuning, true=tune VFO-B memory tuning

uint8_t gWDog = 0; //sent to 1 when we get gSize and gRadioPort from hmi - this starts initial load process - toggle between 1 and 2 every update of Home page - if hmi doesn't see the change it raises a comm fail warning

uint8_t gRadioCAT = RADIO_CAT_ICOM; //index of selected radio - from radioList[]

uint8_t gRadioPort = 0; //selected radio port#, 0 based

uint32_t gFreqRadio = 10000000;
uint32_t gFreqRadioPrev = 0; //stores last freq set to detect changes in digits

uint8_t gVFO = 0; //a=0, b=1, split=2, Memory=3, 4=Toggle Lock, 5=Copy A>B, 6=Copy B>A, 7=Swap A/B, 8- Copy Mem->A (only if in MEM mode)

uint32_t gVFO_Freq[4] = { 0,0,0,0 }; //holds vfo freq, 0=a, 1=b, 2=a memory, 3=b memory
uint8_t gVFO_Mode[4] = { 0,0,0,0 }; //holds radio mode
uint8_t gVFO_BW[4] = { 0,0,0,0 }; //hold PCR bandwidth
uint8_t gVFO_PushPull = SYNC_MODE_BASIC_NONE; //1= push to other Mini, 2= pull from other Mini (Link Basic mode)

int gRIT = 0; //rx inc tuning - 0-99999
int gRITPrev = 0;

volatile uint32_t gUnit = 100; //freq step changed during ISR
uint32_t gUnitPrev = gUnit; //detect COS

uint8_t gRMode = 0; //radio mode, 0=cw, 1=cw_r, 2=lsb, 3=usb, 4=fmn, 5=fmw, 6=am, 7=digL, 8=digH, mnuMode[] is one's based
uint8_t gRModePrev = 0; //detect changes

uint8_t gComMode = COM_MODE_DEBUG; //mode of USB com port - NONE, CAT, or TERMinal

bool gSaveRadio = false; //if == true changes need to be save to the radio file
int gSaveRadioCtr = 0; // counts # of polls after gSaveRadio goes true - save radio data after 30 seconds so we don't save on every change

uint8_t gBand = 0; //load new band freq if this changes

String hmiTagLine = ""; //sticky message for status bar - it will be restored after alert msg is posted
int hmiTagLineRefreshMax = 1500;
int hmiTagLineRefresh = 0; //if !=0 inc to 1500 and refresh hmiTagLine msg with sticky

String gMemLabel = " "; //holds last memory label - used if we're copying between shared and normal memory

bool gWiFi = false; //0=off, 1=active
bool gWiFiAutoConnect = false; //true to auto connect on boot
String gSSID = ""; //SSID
String gStaKey = ""; //wifi passkey
String gStaAdrs = ""; //station wifi address
String gSubnet = "255.255.255.0"; //station sub net
String gGW = ""; //station gateway address

//tcp adrs
String gIPAdrsDefault = "IP Adrs";
String gIPAdrs = gIPAdrsDefault; 
uint8_t gTCPCon = TCP_CON_OFF; //0=not connected, 1=connected as GUI client, 2=connected as non-GUI client (SmartSDR is running)

//flex radio variables
String gFlexPort = "4992"; //default IP port
String gFlexHandle = ""; //unique handle returned by flex on connection - used to identify data from difference connections in the 'S' reply
String gFlexHandleTx = "";//handle for active Tx
String gFlexClientID = ""; //get from sub client all
bool gFlexPan = false; //when true create a new panadapter when we create a new slice
int gFlexFreePanadapters = 0; //set from response from radio when ititializing - # of unused panadapters
uint8_t gFlexSlice = 0; //used to select active flex slice - display as A, B, C, etc
int gFlexFreeSlices = 0; //# of unused slices
uint8_t gFlexVol = 30; //audio level for selected slice
bool gFlexApiPTT = false; //default to use hardware PTT, turn this on to use API xmit command
bool gFlexApiKey = false; //default to use hardware Key, turn this on to use API to control Winkeyer in SmartSDR
uint16_t gFlexDelay = 0; //cw delay - use 30 to 200 when using network keying - max is 2000
uint8_t gFlexPO = 100; //power out - also put in bPwr for log
uint8_t gFlexTO = 10; //tune power
uint8_t gFlexWNB = 0; //msb is enabled flag on these
uint8_t gFlexNB = 0;
uint8_t gFlexNR = 0;
uint8_t gFlexAPF = 0; 
uint8_t gFlexANF = 0;
uint8_t gFlexAnt = 1;
uint8_t gFlexATU= 0xff;	//0=not installed, 1=bypassed, 2=turning, 3=tuned, 0xff=not installed
uint8_t gFlexATUPrev = 0xff;
bool gFlexInit = false; //goes true after we've completed initialization
uint16_t gFlexCmdNum = 0;//incrementing command # for flex commands
uint32_t gFlexLastReply = 0; //holds time of last flex reply - if >35 seconds, show radio failed off
uint32_t gFlexLastFreqUpdateTmr = 0; //holds the last time the frequency was updated - resend it 2 seconds later to make sure it takes

uint8_t gTxMode = RADIO_TX_OFF; //goes to 1 when btnA is pressed - enables tx output, set to 2 when txing
uint8_t gTxModePrev = RADIO_TX_OFF; //used for change detect

bool gTxModeRmt = false;
bool gTxModeRmtPrev = false;

//linear amp sequencing
bool gEnabAmpSeq = false; //when true use both PTT and Key for outputs as described below
uint16_t gSeqPre = 1; //amp sequence delay in millisec - time to wait between PTT and Key outputs on Radio I/O module if gEnabAmpSeq is true
uint16_t gSeqPost = 1; //amp sequence delay in millisec - time to wait before dropping PTT after Key up (or PTT released in voice mode) if gEnabAmpSeq is true
volatile uint32_t gSeqPreTmr = 0; //set this to millis() + gSeqPre-1 to set PTT to Key delay for key down sequencing
volatile uint32_t gSeqPostTmr = 0; //set this to millis() + gSeqPost to set Key to PTT delay for key up sequencing

bool gShareDB = false; //mirrors gShareDB - if true, don't index the freq, tx, or band databases for each radio, use a common db for all radios

bool gLock = false; // ignore freq changes when true - keeps radio on locked freq

uint8_t radioOnline = 0; //after 3 radio comm fails, post alert in tInfo1 and 2
uint8_t gIcmAdr = 0x0; //default adrs - allow user to set it after selecting Icom CAT
uint8_t radioRxBytes = 0; //holds # of bytes we're waiting for from radio
bool radioCommBusy = false; //goes true when any routine is tx or rx to radio - blocks ISR from overwriting comm

/* Keyer Mode Descriptions
 See https://ag6qr.net/index.php/keys-bugs-paddles-and-keyers-a-terminology-introduction/ for description of modes
 0= Stright: Connect manual key to tip of paddle (PDL_LEFT) - no keyer action, just pass PDL_LEFT to KEY_OUT
 1= Iambic A: Squeezing both paddles causes alternating dit-dah-dit-dah or dah-dit-dah-dit depending on which paddle was pressed first - sequence ends as soon as both paddles open
 2= Iambic B: Same as A but send one additional element after both paddles released
 3= Ultimatic: Second paddle pressed has priority - press dit first and dits are generated. Press dah while holding dit and dahs are generated
 4= Bug: Auto generate dits only - manual dahs
*/

uint8_t gCWSpeed = 20; //wpm from CW page
uint8_t gFarnsSpeed = gCWSpeed; //farnsworth spacing from CW page
bool speedToggle = false; //toggle speed with '*' chr,  
uint16_t keyerDitDur = 0; //use CalcDitDur() to calc
uint16_t keyerDitDurPrev = keyerDitDur; //holds original speed before speed toggle
uint16_t keyerDitFarnsDur = 0; //use CalcDitFarnsDur() to calculate
uint16_t keyerDitFarnsDurPrev = keyerDitFarnsDur; //holds original spacing before speed toggle

uint8_t gKeyType = KEYER_IAMBIC_B; //mode 0=none, 1=straight, 2=Passthru, 3=Iambic A, 4=Iambic B, 5=Ultimatic, 6=bug 
uint8_t gKeyTypePrev = gKeyType; // store current keyer mode here if we are using keyer to enter txtInput
bool gSideTone = true; //use internal buzzer
volatile bool gKeyDown = false; //follow KEY_OUT so we can use sidetone without tx being enabled
bool gKeyDownPrev = false; //holds last state of gKeyDown

bool gPaddle = false; //dit on left, dah on right. If true, dah on left, dit on right
uint8_t gRmtMode = MNU_RMT_NORM; //three modes, 0=normal PTT, 1=latching PTT, 2=straight key - if straight key use ISR otherwise poll from Loop()
bool gBtnPTT = false; //goes true when [C] button is used to control PTT 
bool gPTTState = false; //follows PTT state regardless of source (hardware ptt or flex api ptt)

volatile uint8_t gRmtFuncBtn = 0xff; //updated from ChkRmtPTT() ISR

//variables for Antenna Switch Controller (ASC)
bool gAntSwOnline = false; //goes true when we receive link packets from ASC
bool gAntSwActive = false;
bool gAntSwDual = false; //goes true when dual radio ports selected 
bool gAntSw8Port = false; //goes true when 8-ant port switch is selected
bool gAntSwLock = false; //goes true when ASC is locked
bool gAntSwSendBand = false; //if true, send band data to ASC on band change

int gAntSwA = 0; //switch assigned to radio A - ones based, 0=none, 1-8 depending on gAntSwDual
int gAntSwAPrev = 0;
int gAntSwB = 0; //radio b
int gAntSwBPrev = 0;

const uint8_t gAntSwCmdLength = 20; //# of bytes in asc command

//Timers *********************************************************************************************************************************

uint32_t radioPollTmrDefault = 250; //chk radio mode every 300 ms
uint32_t radioPollTmr = millis();

uint32_t radioMsgTimeoutDefault = 150; //timeout for rx msg from radio - adjust for each radio based on baud rate & # of chrs

const uint32_t hmiEncoderTimeoutDefault = 7000; //time out encoder select after 7 seconds of inactivity
uint32_t hmiEncoderTimeout = millis(); //set to millis() to reset timeout

const uint32_t pttWatchDogDefault = 150000; // 3 minute timeout or there about
uint32_t pttWatchDog = millis();

//tx timers
const uint32_t txHangTmrDefault = 500; //hang 500 mS after key/ptt released
uint32_t txHangTmr = millis();


//Arrays **********************************************************************************************************************************

const int txtInputMax = 75;
char txtInput[txtInputMax]; //array for GetTxt()
int txtInputPtr = 0; //points to char in txt to edit
int txtInputSrc = 0; //points to source of txtInput[]
int txtInputSrcMax = 0; //holds max# chrs to edit
int txtInputSrcStart = 0; //holds starting chr of txtInput if it is > 10 chrs
volatile char txtInputCW = 0; //holds chr from keyer when TxtInput is open

const int radioRxBufMax = 100;
char radioRx[radioRxBufMax]; //radio rx buffer

String mnuArray[25]; //holds the current menu array

//  ^ sends call, # sends contest sn, & sends contest exchange, < sends error
// using K1EL compatible prosigns from https://www.k1elsystems.com/files/WKUSB_Manual_v1.3.pdf
// code 39 = ', code 92 = \ //
// ^ is shortcut for call sign
// # is shortcut for contest SN
// % is shortcut for contest exchange
// ~ is shortcut for Error
const uint8_t chrMax = 89;
char chr[chrMax] = { 'a', 'b', 'c', 'd', 'e', 'f', 'g', 'h', 'i', 'j', 'k', 'l', 'm', 'n', 'o', 'p', 'q', 'r', 's', 't', 'u', 'v', 'w', 'x', 'y', 'z',
				'1', '2', '3', '4', '5', '6', '7', '8', '9', '0', '.', ',', ':',
				'A', 'B', 'C', 'D', 'E', 'F', 'G', 'H', 'I', 'J', 'K', 'L', 'M', 'N', 'O', 'P', 'Q', 'R', 'S', 'T', 'U', 'V', 'W', 'X', 'Y', 'Z',
				'"', '$', 39,  '(', ')', '+', '-', '/', ';', '<', '=', '>', '?', '@', '[',  92, ']', '^', '#', '%', '~', '*', ' ', 0};

const int mnuRadioCATMax = 14;
//these are enumerated above
String mnuRadioCAT[mnuRadioCATMax] = { "None", "Flex", "PCR1000", "Icom", "Kenwood1", "Kenwood2", "Yaesu FTdx", "Yaesu 100", "Yaesu 757", "Yaesu 8x7", "Yaesu 890", "Yaesu 1000", "NET rigctld", "IP Adrs"};
//baud rates for selected radio
int gRadioBaud[mnuRadioCATMax] = { 0, 0, 9600, 19200, 9600, 4800, 38400, 38400, 4800, 38400, 38400, 38400, 0, 0 };
int gRadioBaudDefault[mnuRadioCATMax] = { 0, 0, 9600, 19200, 9600, 4800, 38400, 38400, 4800, 38400, 38400, 38400, 0,0 }; //to restore on erase

//menu selections from baud rate
const int mnuRadioBaudMax = 8;
String mnuRadioBaud[mnuRadioBaudMax] = { "1200", "2400","4800","9600","19200","38400","57600","115200" };


const int mnuRadioTagMax = 21;
String mnuRadioTag[mnuRadioTagMax] = { "Port 1", "Port 2", "Port 3", "Port 4", "Port 5", "Port 6", "Port 7", "Port 8", "Port 9", "Port 10", "Port 11", "Port 12",
                                       "Port 13", "Port 14", "Port 15", "Port 16", " ", "Copy", "Rename", " ", "Erase"};

const int mnuTxMsgMax = 14;
String mnuTxMsg[mnuTxMsgMax]; //14 records fit on menu display in dual column, saved to TXMSG.TXT	
uint8_t mnuTxMsgSel = 0; //holds message to tx - this is gTxMem[] in CTR2
String gTxBuffer = ""; //load with tx buffer - pick off one chr at a time and run it through ascii/morse converter - save this to mnuTxMsg[0] so F1 can repeat it
uint8_t gPlay = KEYER_MODE_STOP; //0=keyer stopped, 1=keyer running, 2=keyer paused, press BtnA to cont
uint8_t gPlayPrev = gPlay; //detect change
																												 //default Frequency Memories - loaded into 80-89 and 96-100 when creating a new file
//NOTE: 60m frequencies are based on the ARRL Frequency Chart for the US - edit them below for other countries
const int vfoMemMax = 100;
String defaultVFOMemLabel[20] = { "60m1 CW","60m2 CW","60m3 CW","60m4 CW","60m5 CW","60m1 Voc","60m2 Voc","60m3 Voc","60m4 Voc","60m5 Voc","20m beacon","17m beacon","15m beacon","12m beacon","10m beacon" };//default label
uint32_t defaultVFOMemFreq[20] = { 5332000,5348000,5358500,5373000,5405000,5330500,5346500,5357000,5371500,5403500,14100000, 18110000,21150000,24930000,28200000 };//arrays to hold default freq & RMode and OMode 
uint8_t defaultVFOMemMode[20] = { 0,0,0,0,0,3,3,3,3,3,0,0,0,0,0,6,6,6,6,6 };
uint8_t	defaultVFOMemBW[20] = { 0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0 };

//arrays for file
String vfoMemLabel[vfoMemMax];
uint32_t vfoMemA_Freq[vfoMemMax]; //holds VFOA settings
uint8_t vfoMemA_Mode[vfoMemMax];
uint8_t vfoMemA_BW[vfoMemMax];
uint32_t vfoMemB_Freq[vfoMemMax]; //holds VFOB settings
uint8_t vfoMemB_Mode[vfoMemMax]; 
uint8_t vfoMemB_BW[vfoMemMax];

//initial seeting for BANDLIST.CSV
const int bandMemMax = 16; //memories for each band - load them into current VFO when changing bands
String bandMemLabel[bandMemMax] = { "160m","80m","60m","40m","30m","20m","17m","15m","12m","10m","6m","2m","1.25m","70cm","33cm","Gen" };
uint32_t defaultMemFreq[bandMemMax] = { 1850000,3550000,5330500,7100000, 10125000, 14100000,18110000,21150000,24930000,28200000,50100000,144500000,223000000,421000000,903000000, 9500000 };//arrays to hold saved freq & RMode and OMode 
uint32_t bandMemFreq[bandMemMax]; //loaded with values from BANDLIST.CSV file, or default if that file doesn't exist
uint8_t bandMemRMode[bandMemMax] = { 2,2,3,2,2,3,3,3,3,3,3,4,4,4,4,6 };
uint8_t bandMemBW[bandMemMax] = { 0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0 };

uint32_t freqChngTmrDefault = 10000; //freq must be stable for 10 sec before saving it to prevFreq[]
uint32_t freqChngTmr = millis();
uint32_t lastStableFreq = 0; //holds last stable freq of radio - must change and be stable for 10 seconds before it's added to prevFreq[]
uint8_t lastStableRMode = 255; //holds last stable RMode - same as lastStableFreq
uint8_t lastStableOMode = 255; //holds last stable OMode - same as lastStableFreq
uint32_t prevFreq[11]; //arrays to hold last 10 frequency/mode sets
uint8_t prevRMode[11];

const uint8_t pcCmdRxLen = 75;
char pcCmdRx[pcCmdRxLen]; //buffer for receiving commands on Serial (USB port) from PC - read/set freq and mode on HMI from PC program, keyboard control in terminal mode

int keyerQue[22] = { 0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0 }; //keyer sequence array - each element holds the time counter for that element - if positive, key down, if negative, key up (space) - dec until [0] gets to 0 then shift down one

//Menu arrays

volatile bool hmiMnuChng = false; //goes true if any menu changes - forces CreateMnu()
volatile bool gEncoderClicked = false; //true when encoder clicked in ISR - Loop updates selected option

//main menu options
//Change NextMnu() if changing order of menus
const int mnuMainMax = 13;
String mnuMain[mnuMainMax] = { "Frequency", "Band", "Mode", "VFO", "Tx Msg", "Edit Tx Msg", "Keyer", "WiFi", "Link", "Ant Sw", "Radio Port", "Config", ""}; //Settings menu visible for Flex & PCR1000 only

volatile int gMnu = 0; //selects main menu 
int gMnuPrev = -1; //used to detect when refresh is required
volatile int gMnuSel = 0; //tracks selected option in the selected menu
int gMnuSelB4Help = 0; //holds the selected menu item before help is opened so it can be restored when help is closed

int mnuMainSel = 0; //holds the last active main menu so we select it on first encoder click
volatile int gMnuSelMax = mnuMainMax; //holds total # of entries in the selected menu 
									  //this value changes when gMnu is changed

const int mnuBandMax = 16;
//use bandMemLabel[] for menu

//mode menu options - offset gRMode by + 1 because Toggle Tx Enab 
const int mnuModeMax = 10;
String mnuMode[mnuModeMax] = { "Toggle Tx Enab", "CW", "CWr", "LSB", "USB", "FMn", "FMw", "AM", "DigL", "DigH" }; //mode descriptions ARE 1's BASED!

//vfo menu options
const int mnuVFOMax = 13;
String mnuVFO[mnuVFOMax] = { "VFO-A", "VFO-B", "Split Mode", "Memory Tune", "Memory Mode", "Toggle Lock", " ", "Copy A>B", "Copy B>A", "Swap A/B", "Copy M>VFO", "Push", "Pull"};
int mnuVFOSel = MNU_VFO_A;

//keyer menu options
const int mnuKeyerMax = 10;
String mnuKeyer[mnuKeyerMax] = { "Speed", "Farns", "Sidetone", "Keyer Type", "Paddle", "Rmt PTT"," ", "Call: ", "SN: ", "Exch: " };
int mnuKeyerSel = MNU_KEYER_SPEED;
int mnuLastTxMsg = 0; //holds last TXMSG we edited
String mnuKeyerCall = ""; //hold user call
uint16_t mnuKeyerSN = 1; //holds contest sn
uint16_t mnuKeyerSNPrev = 1; //detect change so we can update Init file
String mnuKeyerExch = ""; //holds contest exchange

//keyer types
const int mnuCWKeyerMax = 7;
String mnuCWKeyer[mnuCWKeyerMax] = { "None","Straight","Passthru","Iambic-A","Iambic-B","Ultimatic","Bug" };
int mnuCWKeyerSel = KEYER_NONE;

const int mnuFreqMax = 7;
String mnuFreq[mnuFreqMax] = { "Zero Low Digits", "RIT ()", "Share Freq DB", "Load from Memory","Save to Memory", "Rename Memory", "Prev Freq/Mode" };
int mnuFreqSel = 0;

const int mnuFreqListMax = 7; //show 7 freqList entries at a time
String mnuFreqList[mnuFreqListMax];
int mnuFreqListSel = 0;
int mnuFreqListStart = 0; //tracks starting freqList entry on starting row

const int mnuConfigMax = 7;
String mnuConfig[mnuConfigMax] = { "[C] Btn", "Function Btns", "Theme", "Tag Line", "Radio CAT", "Radio Baud", "Amp Seq"};
int mnuConfigSel = MNU_CONFIG_C_BTN;

const int mnuCBtnMax = 17;
String mnuCBtn[mnuCBtnMax] = { "PTT", "RIT", "Speed", "Frequency", "Mode", "Band", "VFO", "Tx Msg","Edit Tx Msg", "Keyer", "WiFi", "Link", "Ant Sw", "Function Btn", "Push Info", "Pull Info",  "Radio" };
int mnuCBtnSel = MNU_C_BTN_TXMSG; //default to TxMsg

//arrays for Function Button menus
const int mnuFuncBtnMax = 6;
String mnuFuncBtn[mnuFuncBtnMax] = { "F1", "F2", "F3", "F4", "F5", "RESET KEYPAD LOCKOUT"};
int mnuFuncBtnSel = RMT_FUNC_F1-1;

String mnuFuncBtnLbl[mnuFuncBtnMax] = { "TxMsg 1", "TxMsg 2", "TxMsg 3", "TxMsg 4", "TxMsg 5"}; //labels to use on func btn menu come from mnuFuncBtnOption[]
uint8_t mnuFuncBtnIdx[mnuFuncBtnMax] = { 0,1,2,3,4 }; //points to selected function

const int mnuFuncBtnOptionMax = 15;
String mnuFuncBtnOption[mnuFuncBtnOptionMax] = { "TxMsg 1", "TxMsg 2", "TxMsg 3", "TxMsg 4", "TxMsg 5", "TxMsg 6", "TxMsg 7", "RIT", "Speed", "TxMsg", "Btn [A]", "Btn [B]", "Btn [C]", "Ant Sw"};
uint8_t mnuFuncBtnOptionSel = MNU_FUNC_OPTION_TXMSG1;

const int mnuLinkMax = 7;
String mnuLink[mnuLinkMax] = { "Connect", "Sync Mode", "Server/Client", "BLE/WIFI", "Auto-Connect", "UDP Adrs",  "Zero-Beat"};
uint8_t mnuLinkSel = MNU_LINK_CONNECT;

//Array for WiFi
const int mnuWiFiMax = 7;
String mnuWiFi[mnuWiFiMax] = { "Connect","SSID","PW" ,"Auto-Connect", "Sta Adrs", "Subnet", "Gateway"};
int mnuWiFiSel;

const int mnuThemeMax = 7; //# of themes available
String mnuTheme[mnuThemeMax] = {"Grey", "Violet", "Blue", "Maroon", "Cyan", "Green", "Olive" };
int mnuThemeSel = 2; //default to blue

const int mnuFlexMax = 17;
String mnuFlex[mnuFlexMax] = { "IP", "Port", "Panadapter", "Slice", "Ant", "ATU", "Vol", "PO", "TO", "WNB", "NB", "NR", "ANF", " ", "PTT", "Key", "Delay" };
int mnuFlexSel = 0;

String gFlexSliceLbl[8] = { "A","B","C","D","E","F","G","H" };
//use gFlexSlice as the index - declared above

const int mnuPCRMax = 11;
String mnuPCR[mnuPCRMax] = { "Vol: ", "Sq: ", "IF Shift: ", "Zero IF: ", "DSP:: ", "DSP NR: ", "Auto Notch: ", "BW: ", "AGC: ", "Atten: ", "NB: " };
int mnuPCRSel;

const int mnuAmpSeqMax = 3;
String mnuAmpSeq[mnuAmpSeqMax] = { "Enable Amp Seq", "PTT->Key Pre Delay", "Key->PTT Post Delay" };
int mnuAmpSeqSel;

const int mnuASCMax = 7;
String mnuASC[mnuASCMax] = { "Port-A", "Port-B", "Swap A/B","Lock Mode", "Switch Mode", "Edit Antenna Labels", "Send Band Data" };
int mnuASCSel = MNU_ANT_SW_PORT_A;

const int mnuAntLblMax = 9;
String mnuAntLbl[mnuAntLblMax] = { "OPEN", "Ant 1", "Ant 2", "[OFF]", "[OFF]", "[OFF]", "[OFF]", "[OFF]", "[OFF]" }; //ones based -- only show ants over 2 that != '[OFF]'
String mnuAntList[mnuAntLblMax]; //used to list ants for MNU_RADIO_A and _B selection
int mnuAntEditSel = 0;

//using gAntSwA and gAntSwB to point to selected menu option


//radio db menu options
//use RadioTag[] array for menu

//radio model options
//use RadioType[] array for menu


//PCR1000 Rx support ********************************************************************************************************************

//variables for PCR1000 radio support

//PCR uses gTxMode as power on/off flag
bool gPCRInit = false; // goes true after PCR1000 has been initialized the first time
bool gPCRAGC = false;
bool gPCRAtten = false;
bool gPCRNB = false;
bool gPCRDSP = false;
bool gPCRANch = false; //auto notch
bool gPCRHasDSP = false; //turn on if we get a GD01 from the GD? request
uint8_t gPCRBW = 0; //0=3k, 1=6k, 2=15k, 3=50k, 4=230k 
uint8_t gPCRVol = 70; //pcr volume - 0 to 0xff but scaled to 0-100 on Mini
uint8_t gPCRSq = 0; //pcr squelch, same scaling as pcrVol
uint8_t gPCRIF = 0x80; //pcr IF shift, 0x80 is center, 0 to 0xff
uint8_t gPCRDSPNR = 0; //pcr DSP noise reduction, 0 to 15

//SD card


//ISR **********************************************************************************************************************************

//IntervalTimer decoderUpdateTmr; //called every 2 mS to time demodulators

//variable passed from KeyboardPress_ISR
volatile int keyboardChr;

//SD Card ******************************************************************************************************************************

//File myFile;


//WiFi Server =============================================================
WiFiServer server(80);
//WiFiClient client; //make it global
WiFiClient tcp; //use for flex && rigctld
WiFiUDP udp; //udp port for Link data sharing

//crc16 ===================================================================
Crc16 crc;

//Callbacks for BLE


class MyClientCallback : public BLEClientCallbacks
{
	void onConnect(BLEClient* pClient)
	{
		Serial.println("Client connected to Server");
	}

	void onDisconnect(BLEClient* pClient)
	{
		gBLEConnected = false;
		Serial.println("Client Disconnected from Server");
	}
};

class MyServerCallbacks : public BLEServerCallbacks {
	void onConnect(BLEServer* pServer) 
	{
		Serial.println("Mini-A Server Connected ");
		gBLEConnected = true;
	}

	void onDisconnect(BLEServer* pServer) 
	{
		Serial.println("Server Disconnected");
		gBLEConnected = false;
	}
};

//BLE setups
/* Scan for BLE servers and find the first one that advertises the service we are looking for. */
class MyAdvertisedDeviceCallbacks : public BLEAdvertisedDeviceCallbacks
{
	/* Called for each advertising BLE server. */
	void onResult(BLEAdvertisedDevice advertisedDevice)
	{
		Serial.print("\n\rBLE Advertised Device found: ");
		Serial.println(advertisedDevice.toString().c_str());

		/* We have found a device, let us now see if it contains the service we are looking for. */
		//Looking for device name instead of services - current configuration doen't broadcast services ???
		//if (advertisedDevice.haveservUUID() && advertisedDevice.isAdvertisingService(servUUID))
		if (advertisedDevice.getName() == "CTR2-Mini-BLE-Server")
		{
			BLEDevice::getScan()->stop();
			pRmtDevice = new BLEAdvertisedDevice(advertisedDevice);// if you pass BLEAdvertisedDevice instead of address, it will be recognized type of peer device address (public or private)
			gBLEKillScan = true; //this will cause a call to GetLinkParams() in Loop()
		}
	}
};

//callback for server side 
class MyLocalCallback: public BLECharacteristicCallbacks {
	void onWrite(BLECharacteristic *pCharacteristic) 
	{
		//this is the update from the remote client
		pLocalParams = pCharacteristic;
		gLinkRxData = true; //loop calls GetBLEParam  
    }
};


//===================================================================================================
//this is the starting setup routine

void setup()
{
	DoSetupTasks();
}

//==========================================================================================================
//this is the main loop routine


void loop()
{
	DoLoopTasks();
}



