// =====================================================================================================================================================================================================
// (c) 2022 Lynn Hansen, KU7Q															                                                                                                               |
// This Source Code Form is subject to the terms of the GNU GENERAL PUBLIC LICENSE, Version 3, 29 June 2007. A copy of this license can be found here: https://choosealicense.com/licenses/gpl-3.0/|
// =====================================================================================================================================================================================================


//This module holds the file system routines

//FILE ROUTINES ********************************************************************************************

void Chk4Folders()
{
	//chk to make sure the required folders exist. If not, create them
	char dirName[15];
	String name = "/MISC"; //band registers and radio labels go here
	name.toCharArray(dirName, name.length() + 1); //add 1 to length for terminator	
	if (!SD.exists(dirName))
	{
		SerialOut("Creating " + name, true);
		PrtStatus("Creating " + name, false);
		SD.mkdir(dirName);
	}

	name = "/RADIO";
	name.toCharArray(dirName, name.length() + 1);	
	if (!SD.exists(dirName))
	{
		SerialOut("Creating " + name, true);
		PrtStatus("Creating " + name, false);
		SD.mkdir(dirName);
	}
	name = "/FREQ"; //vfoMem goes here
	name.toCharArray(dirName, name.length() + 1);	
	if (!SD.exists(dirName))
	{
		SerialOut("Creating " + name, true);
		PrtStatus("Creating " + name, false);
		SD.mkdir(dirName);
	}
	name = "/TXMSG"; //Tx Msg files go here	
	name.toCharArray(dirName, name.length() + 1);
	if (!SD.exists(dirName))
	{
		SerialOut("Createing " + name, true);
		PrtStatus("Creating " + name, false);
		SD.mkdir(dirName);
	}
}

void LoadInitFile()
{
	//load intial startup data
	char fName[15];
	String name = "Init.txt";
	name.toCharArray(fName, name.length() + 1);

	if (!SD.exists(fName))
	{
		//preset first bootup conditions		
		mnuThemeSel = HMI_THEME_BLUE;		
		SaveInitFile();
	}
	else
	{
		SerialOut("Opening Init.txt file", true);
		char chr = 0;
		//parse file and reload variables
		String txt = "";
		int ptr = 0;
		myFile = SD.open(fName, FILE_READ);
		while (myFile.available())
		{
			txt = "";
			while (myFile.available())
			{
				chr = myFile.read();
				if (chr != 10 && chr != 13)
				{
					txt += String(chr);
				}
				else if (chr == 13)
				{
					break;
				}
			}

			ptr = txt.indexOf("=") + 1;
			delay(20);
			if (txt.length() > ptr)
			{
				//SerialOut(txt, true);
				if (txt.substring(0, ptr) == "gRadioPort=")
				{
					gRadioPort = ExtractValue(txt);
					if (gRadioPort > 15)
					{
						gRadioPort = 0;
					}
				}
				else if (txt.substring(0, ptr) == "gUnit=")
				{
					gUnit = ExtractValue(txt);
					if (gUnit < 1 || gUnit > 100000000)
					{
						gUnit = 100; //default
					}
					//SerialOut("gUnit= " + String(gUnit) + ", txt= " + txt, true);
				}
				else if (txt.substring(0, ptr) == "gZeroDigits=")
				{
					gZeroDigits = ExtractValue(txt);
				}
				else if (txt.substring(0, ptr) == "mnuCBtnSel=")
				{
					mnuCBtnSel = ExtractValue(txt);
				}
				else if (txt.substring(0, ptr) == "mnuFuncBtnIdx0=")
				{
					mnuFuncBtnIdx[0] = ExtractValue(txt);
					mnuFuncBtnLbl[0] = mnuFuncBtnOption[mnuFuncBtnIdx[0]];
				}
				else if (txt.substring(0, ptr) == "mnuFuncBtnIdx1=")
				{
					mnuFuncBtnIdx[1] = ExtractValue(txt);
					mnuFuncBtnLbl[1] = mnuFuncBtnOption[mnuFuncBtnIdx[1]];
				}
				else if (txt.substring(0, ptr) == "mnuFuncBtnIdx2=")
				{
					mnuFuncBtnIdx[2] = ExtractValue(txt);
					mnuFuncBtnLbl[2] = mnuFuncBtnOption[mnuFuncBtnIdx[2]];
				}
				else if (txt.substring(0, ptr) == "mnuFuncBtnIdx3=")
				{
					mnuFuncBtnIdx[3] = ExtractValue(txt);
					mnuFuncBtnLbl[3] = mnuFuncBtnOption[mnuFuncBtnIdx[3]];
				}
				else if (txt.substring(0, ptr) == "mnuFuncBtnIdx4=")
				{
					mnuFuncBtnIdx[4] = ExtractValue(txt);
					mnuFuncBtnLbl[4] = mnuFuncBtnOption[mnuFuncBtnIdx[4]];
				}
				else if (txt.substring(0, ptr) == "mnuThemeSel=")
				{
					mnuThemeSel = ExtractValue(txt);
					SetThemeColors();
				}
				else if (txt.substring(0, ptr) == "hmiTagLine=")
				{
					hmiTagLine = txt.substring(ptr);
				}
				else if (txt.substring(0, ptr) == "mnuKeyerCall=")
				{
					mnuKeyerCall = txt.substring(ptr);
				}
				else if (txt.substring(0, ptr) == "mnuKeyerSN=")
				{
					mnuKeyerSN = ExtractValue(txt);
				}
				else if (txt.substring(0, ptr) == "mnuKeyerExch=")
				{
					mnuKeyerExch = txt.substring(ptr);
				}
				else if (txt.substring(0, ptr) == "gSSID=")
				{
					gSSID = txt.substring(ptr);
				}
				else if (txt.substring(0, ptr) == "gStaKey=")
				{
					gStaKey = txt.substring(ptr);
				}
				else if (txt.substring(0, ptr) == "gWifiAutoConnect=")
				{
					gWiFiAutoConnect = ExtractValue(txt);
				}
				else if (txt.substring(0, ptr) == "gLinkMiniA=")
				{
					gLinkMiniA = ExtractValue(txt);
				}
				else if (txt.substring(0, ptr) == "gLinkAutoConnect=")
				{
					gLinkAutoConnect = ExtractValue(txt);
				}
				else if (txt.substring(0, ptr) == "gSyncMode=")
				{
					if (gLinkMiniA)
					{
						gSyncMode = ExtractValue(txt); //client gets set remotely
					}
				}
				else if (txt.substring(0, ptr) == "gLinkTypeUDP=")
				{
					gLinkTypeUDP = ExtractValue(txt);
				}
				else if (txt.substring(0, ptr) == "gUDPAdrs=")
				{
					gUDPAdrs = txt.substring(ptr);
				}
				else if (txt.substring(0, ptr) == "gUDPAdrsASC=")
				{
					gUDPAdrsASC = txt.substring(ptr);
				}
				else if (txt.substring(0, ptr) == "gAntSwSendBand=")
				{
					if (gSyncMode == SYNC_MODE_BASIC)
					{
						//only set sendband if we're not using basic+freq mode - can't switch ants and  A & B on same band
						if (gSyncMode != SYNC_MODE_FREQ)
						{
							gAntSwSendBand = ExtractValue(txt);
						}						
					}
				}
			}
		}
		myFile.close();
	}

}

void SaveInitFile()
{
	//NOTE: Put Init.txt file in root of drive - File system crashes if it has to look into a subdirectory and the SD card is blank
	char fName[15];
	String name = "Init.txt";
	name.toCharArray(fName, name.length() + 1);
	if (SD.exists(fName))
	{
		//delete it and start a new one
		SD.remove(fName); 
		delay(100);
	}

	SerialOut("Saving Init.txt file", true);
	//create file using defaults
	blockISR = true;
	myFile = SD.open(fName, FILE_WRITE);
	if (myFile)
	{
		myFile.println("CTR2-Mini Initialization File");
		myFile.println(version);
		myFile.println(" ");
		myFile.println("gUnit=" + String(gUnit));
		myFile.println("gZeroDigits=" + String(gZeroDigits));
		myFile.println("gRadioPort=" + String(gRadioPort));
		delay(1);
		myFile.println("mnuCBtnSel=" + String(mnuCBtnSel));
		myFile.println("mnuFuncBtnIdx0=" + String(mnuFuncBtnIdx[0]));
		myFile.println("mnuFuncBtnIdx1=" + String(mnuFuncBtnIdx[1]));
		myFile.println("mnuFuncBtnIdx2=" + String(mnuFuncBtnIdx[2]));
		delay(1);
		myFile.println("mnuFuncBtnIdx3=" + String(mnuFuncBtnIdx[3]));
		myFile.println("mnuFuncBtnIdx4=" + String(mnuFuncBtnIdx[4]));
		myFile.println("mnuThemeSel=" + String(mnuThemeSel));
		myFile.println("hmiTagLine=" + hmiTagLine);
		delay(1);
		myFile.println("mnuKeyerCall=" + mnuKeyerCall);
		myFile.println("mnuKeyerSN=" + String(mnuKeyerSN));
		myFile.println("mnuKeyerExch=" + mnuKeyerExch);
		myFile.println("gSSID=" + gSSID);
		delay(1);
		myFile.println("gStaKey=" + gStaKey); //wifi passkey
		myFile.println("gWifiAutoConnect=" + String(gWiFiAutoConnect)); //auto connect on boot
		myFile.println("gLinkMiniA=" + String(gLinkMiniA));
		myFile.println("gLinkAutoConnect=" + String(gLinkAutoConnect));
		myFile.println("gSyncMode=" + String(gSyncMode)); //don't save tracking state - force user to enable it after they set both freqs on Minis
		delay(1);
		myFile.println("gLinkTypeUDP=" + String(gLinkTypeUDP));
		myFile.println("gUDPAdrs=" + gUDPAdrs);
		myFile.println("gUDPAdrsASC=" + gUDPAdrsASC);
		myFile.println("gAntSwSendBand=" + String(gAntSwSendBand));
	}
	myFile.close();
	blockISR = false;
}

void Chk4BandListFile(bool resetFile)
{
	//chk if freq list file exists - if it does, load it, if not, create it
	String fileName = "BNDLST.CSV"; //if using a common db
	if (gShareDB == false || resetFile == true)
	{
		fileName = "BNDLST_" + String(gRadioPort, HEX) + ".CSV";
	}
	fileName = "/MISC/" + fileName; //add dir

	char fName[25];
	fileName.toCharArray(fName, fileName.length() + 1);	
	if (!SD.exists(fName) || resetFile == true)
	{
		SerialOut(String(fName) + " not found, creating new one", true);
		PrtStatus("Saving Band List", false);
		//clear array		
		for (int i = 0; i < bandMemMax; i++)
		{			
			bandMemFreq[i] = defaultMemFreq[i];
			if (i < 4)
			{
				bandMemRMode[i] = 2; //lsb
			}
			else if (i < 11)
			{
				bandMemRMode[i] = 3; //usb
			}
			else
			{
				bandMemRMode[i] = 4; //fm narrow
			}
			bandMemBW[i] = 0;
		}
		SaveBandList(gRadioPort); //save new default file
	}

	LoadBandList();

}

void LoadBandList()
{

	//loaded on initial startup- then these are managed in memory and saved periodically (when band changes)
	String txt = "";
	char chr = 0;
	int rowCtr = 0;
	int colCtr = 0;
	bool commentLn = false; //goes true if line starts with #
	String fileName = "BNDLST.CSV"; //if using a common db
	if (gShareDB == false)
	{
		fileName = "BNDLST_" + String(gRadioPort, HEX) + ".CSV";
	}
	fileName = "/MISC/" + fileName;
	char fName[25];
	fileName.toCharArray(fName, fileName.length() + 1);
	PrtStatus("Loading Band List", false);	
	delay(500);

	//load bandMem arrays with defaults
	for (int i = 0; i < bandMemMax; i++)
	{		
		bandMemFreq[i] = defaultMemFreq[i];
		bandMemRMode[i] = defaultVFOMemMode[i];
		bandMemBW[i] = defaultVFOMemBW[i];
	}

	//SD.remove("/MISC/BANDLIST.CSV"); //use to clear list

	myFile = SD.open(fName, FILE_READ);
	if (myFile)
	{
		//SerialOut("Opened /MISC/BNDLST.CSV", true);
		while (myFile.available())
		{
			chr = (myFile.read());
			if (chr == '#' && txt.length() == 0)
			{
				commentLn = true;
			}
			if (commentLn == true)
			{
				//read until we hit the end of this line
				if (chr == 13)
				{
					//reached the linefeed, start new line
					commentLn = false;
					colCtr = 0;
					txt = "";
				}
			}
			else
			{
				if (chr == ',')
				{
					//end of field, save value
					ExtractBandInfo(colCtr, rowCtr, txt);
					colCtr++;
					txt = "";
				}
				else if (chr == 10)
				{
					//just ignore lf, use cr to inc 
					txt = "";
				}
				else if (chr == 13)
				{
					//we don't know which column we ended on, check colCtr					
					ExtractBandInfo(colCtr, rowCtr, txt);
					//new line					
					txt = "";
					colCtr = 0;
					rowCtr++;
					if (rowCtr == bandMemMax)
					{
						goto ExitHr;
					}
				}
				else
				{
					if (colCtr < 2 && txt.length() < 16)
					{
						txt += char(chr);
					}
					else if (isDigit(chr))
					{
						txt += char(chr);
					}
				}
			}
		}
	ExitHr:
		myFile.close();
		PrtStatus(hmiTagLine, false);
		return;
	}
	else
	{
		//failed to open
		PrtStatus("Fail: " + fileName, true);
		myFile.close();
		delay(1500);			
		SaveBandList(gRadioPort); //create a new default list
	}
	PrtStatus(hmiTagLine, false);
}

void ExtractBandInfo(uint8_t colCtr, uint8_t rowCtr, String txt)
{
	//move txt to bandMemxxxx[]
	int i = 0;

	//SerialOut("rowCtr= " + String(rowCtr) + ", colCtr= " + String(colCtr) + ", txt= " + txt, true);
	switch (colCtr)
	{
	case 0:
		//ignore rec# column
		txt = "";
		break;
	case 1:
		if (txt.length() == 0)
		{
			txt = " ";
		}
		//bandMemLabel[rowCtr] = txt; //don't reset this, it's used for the band mnu
		break;
	case 2:
		bandMemFreq[rowCtr] = txt.toInt();
		break;
	case 3:
		i = txt.toInt();
		if (i >= 0 && i < 9)
		{
			bandMemRMode[rowCtr] = i;
		}
		break;
	case 4:
		i = txt.toInt();
		if (i >= 0 && i < 5)
		{
			bandMemBW[rowCtr] = i;
		}
		break;
	}
}

void SaveBandList(uint8_t radioNum)
{
	//save bandMem to BANDLIST.CSV file on SD card	
	if (gTxMode > RADIO_TX_ENAB)
	{
		gSaveRadioCtr = 0; //reset counter - wait 30 more seconds before trying again
		return;
	}
	String fileName = "BNDLST.CSV"; //if using a common db
	if (gShareDB == false)
	{
		fileName = "BNDLST_" + String(radioNum, HEX) + ".CSV";
	}
	fileName = "/MISC/" + fileName;
	char fName[25];
	fileName.toCharArray(fName, fileName.length() + 1);
	//SerialOut("Saving " + String(fName), true);
	if (SD.exists(fName));
	{
		SD.remove(fName);
		delay(1000);
	}
	blockISR = true;
	myFile = SD.open(fName, FILE_WRITE);
	if (myFile)
	{
		//write header first
		myFile.println("### CSV Band Frequency & mode file for CTR2 Radio " + String(radioNum));
		myFile.println("### Enter the Label - Frequency (in hz) - radio mode - operation mode in the columns below");
		myFile.println("###");
		myFile.println("### Radio Mode: 0=CW 1=CWr 2=LSB 3=USB 4=FMn 5=FMw 6=AM 7=DigL =DigH");				
		myFile.println("###");
		myFile.println("###Record#,Label,Freq(hz),Radio Mode");
		for (int i = 0; i < bandMemMax; i++)
		{
			String txt = bandMemLabel[i];
			if (txt.length() == 0)
			{
				txt = " ";
			}
			txt += String(i + 1) + "," + txt + "," + String(bandMemFreq[i]) + "," + String(bandMemRMode[i]) + "," + String(bandMemBW[i]);
			myFile.println(txt);
			//SerialOut(txt, true);
		}
	}
	myFile.close();
	blockISR = false;
	ChkFreqModeChng(); //save current freq as last prev
}

void ChangeBaudRate(int curDir)
{
	int temp = gRadioBaud[gRadioCAT];
	if (curDir == 1)
	{
		//inc baud rate
		switch (temp)
		{
		case 1200:
			temp = 2400;
			break;
		case 2400:
			temp = 4800;
			break;
		case 4800:
			temp = 9600;
			break;
		case 9600:
			temp = 19200;
			break;
		case 19200:
			temp = 38400;
			break;
		case 38400:
			temp = 57600;
			break;
		case 57600:
			temp = 115200;
			break;
		case 115200:
			temp = 1200;
			break;
		}
	}
	else
	{
		//dec baud
		switch (temp)
		{
		case 1200:
			temp = 115200;
			break;
		case 2400:
			temp = 1200;
			break;
		case 4800:
			temp = 2400;
			break;
		case 9600:
			temp = 4800;
			break;
		case 19200:
			temp = 9600;
			break;
		case 38400:
			temp = 19200;
			break;
		case 57600:
			temp = 38400;
			break;
		case 115200:
			temp = 57600;
			break;
		}
	}
	gRadioBaud[gRadioCAT] = temp;
	SetRadioBaud();
	SaveRadioBaud(gRadioPort);
}

void LoadRadioBaud(uint8_t radioNum)
{
	//load radio baud list if it exists. It will only exist if the user has changed a default in gRadioBaud[]
	//NOTE: Each radio port can have its own baud list, reload when loading radio
	String fileName = "/MISC/BAUD_" + String(radioNum, HEX) + ".CSV";
	char fName[25];
	char chr = 0;
	String txt = "";
	bool commentLn = false;
	uint8_t colCtr = 0;
	fileName.toCharArray(fName, fileName.length() + 1);	
	if (SD.exists(fName))
	{
		myFile = SD.open(fName, FILE_READ);
		if (myFile)
		{
			while (myFile.available())
			{
				chr = (myFile.read());
				if (chr == '#' && txt.length() == 0)
				{
					commentLn = true;
				}
				if (commentLn == true)
				{
					//read until we hit the end of this line
					if (chr == 13)
					{
						//reached the linefeed, start new line
						commentLn = false;
						colCtr = 0;
						txt = "";
					}
				}
				else
				{
					if (chr == ',')
					{
						//end of field, save value
						gRadioBaud[colCtr] = txt.toInt();
						//SerialOut("Baud " + String(colCtr) + " = " + txt, true);
						colCtr++;
						txt = "";
					}
					else if (chr == 10)
					{
						//read in only one line
					}
					else if (chr == 13)
					{
						//we're done, exit
						myFile.close();
						return;
					}
					else if (isDigit(chr))
					{
						txt += char(chr);
					}
				}
			}
		}
		myFile.close();
	}
	else
	{
		//create a default file for this port
		SaveRadioBaud(radioNum);
	}
}


void SaveRadioBaud(uint8_t radioNum)
{
	//save gRadioBaud[] array to allow user to change default baude
	String fileName = "/MISC/BAUD_" + String(radioNum, HEX) + ".CSV";
	char fName[25];
	fileName.toCharArray(fName, fileName.length() + 1);
	if (!gComMode)
	{
		SerialOut("Saving " + String(fName), true);
	}
	if (SD.exists(fName));
	{
		SD.remove(fName);
		delay(1000);
	}
	blockISR = true;
	myFile = SD.open(fName, FILE_WRITE);
	if (myFile)
	{
		//write header first
		myFile.println("### CSV Baud List for CTR2-Mini");
		myFile.println("###");
		myFile.println("###        ,Flex, PCR1000, Icom, Kenwood1, Kenwood2, Yaesu FTdx, Yaesu 100, Yaesu 757, Yaesu 8x7, Yaesu 890, Yaesu 1000");
		myFile.println("###Default:,WIFi, 9600, 19200, 9600-1 stopbit, 4800-2 stopbit, 38400, 4800, 38400, 38400, 38400, 38400,");
		myFile.println("###");
		String txt = "";
		for (int i = 0; i < mnuRadioCATMax; i++)
		{
			txt += String(gRadioBaud[i]) + ",";
		}
		txt += "0"; //terminator
		//SerialOut("New baud rates= " + txt, true);
		myFile.println(txt);
	}
	myFile.close();
	blockISR = false;
}


void EraseRadio()
{
	//reset selected radio to defaults and erase support files
	gShareDB = false; //so we don't delete shared files
	String txt = "Port " + String(gRadioPort + 1);
	mnuRadioTag[gRadioPort] = txt;
	SaveRadioTags();
	//restore baud defaults
	for (int i = 0; i < mnuRadioCATMax; i++)
	{
		gRadioBaud[i] = gRadioBaudDefault[i];
	}
	SaveRadioBaud( gRadioPort);
	gRadioCAT = 0;
	gFreqRadio = 10000000;
	gBand = 15;
	for (int i = 0; i < 4; i++)
	{
		gVFO_Freq[i] = 10000000;
		gVFO_Mode[i] = 0;
		gVFO_BW[i] = 0;
	}
	gMSel = 0;
	gMSelMode = false; //use vfo-a for memory tuning
	gCWSpeed = 15;
	gFarnsSpeed = 15;
	CalcDitDur();
	CalcDitFarnsDur();
	gKeyType = 0;
	gRmtMode = MNU_RMT_KEY; //set this to Key mode - if we initialize it in Norm or Latch mode and the CTR2-Mini board is not plugged in, it's read as a LOW causing Tx Enable to cycle
	blockISR = true;
	pinMode(RMT_BTNS, OUTPUT); //toggle mode to reset it otherwise key will get stuck on
	delay(10);
	pinMode(RMT_BTNS, INPUT_PULLUP);
	blockISR = false;
	gSideTone = true;
	gPaddle = false;	
	gIPAdrs = gIPAdrsDefault;
	gFlexPort = "4992";
	gFlexSlice = 0;
	gFlexVol = 30;
	gFlexPO = 100;
	gFlexTO = 10;
	gFlexWNB = 0;
	gFlexNB = 0;
	gFlexNR = 0;
	gFlexAPF = 0;
	gFlexANF = 0;
	gFlexAnt = 1;
	gFlexApiPTT = 0;
	gFlexApiKey = 0;
	gFlexDelay = 0;

	gPCRAGC = false;
	gPCRANch = false;
	gPCRAtten = false;
	gPCRBW = 0; //3khz on PCR1000
	gPCRDSP = false;
	gPCRDSPNR = 0;
	gPCRIF = 0x80; //centered
	gPCRNB = false;
	gPCRVol = 70;
	gPCRSq = 0;

	//amp sequencing
	gEnabAmpSeq = false;
	gSeqPre = 0;
	gSeqPost = 0;

	Chk4BandListFile(true); //erase band and ant data

	txt = "/TXMSG/TXMSG_" + String(gRadioPort, HEX) + ".TXT";
	PrtStatus("Erasing Tx Messages", false);
	char fName[25];
	txt.toCharArray(fName, txt.length() + 1);
	if (SD.exists(fName))
	{
		SD.remove(fName);
	}
	Chk4TxMsgFile(); //create empty file

	txt = "/FREQ/VFOMEM_" + String(gRadioPort, HEX) + ".CSV";
	PrtStatus("Erasing Frequency List", false);
	txt.toCharArray(fName, txt.length() + 1);
	if (SD.exists(fName))
	{
		SD.remove(fName);
	}
	Chk4FreqListFile();

	txt = "/MISC/BNDLST_" + String(gRadioPort, HEX) + ".CSV";
	PrtStatus("Erasing Band List", false);
	txt.toCharArray(fName, txt.length() + 1);
	if (SD.exists(fName))
	{
		SD.remove(fName);
	}
	SaveBandList(gRadioPort);
	delay(1000);

}

void SaveRadio(uint8_t radioNum, bool force)
{
	//radioNum is 0-based
	//save the settings of the previously selected radio then update radio parmaeters to currently selected radio

	if (!force)
	{
		gSaveRadioCtr = 0; //reset counter 
		ChkFreqModeChng(); //save any changes to prev lists
		if (gTxMode > RADIO_TX_ENAB || hmiMode != HMI_HOME)
		{
			if (!gComMode)
			{
				SerialOut("Save radio file blocked for 30 sec - millis=" + String(millis()), true);
			}
			return;
		}
	}

	//if contest sn has changes, save init file
	if (mnuKeyerSN != mnuKeyerSNPrev)
	{
		SaveInitFile();
		mnuKeyerSNPrev = mnuKeyerSN;
	}

	//reset save flag
	//SerialOut("Saving radio file - millis= " + String(millis()), true);
	gSaveRadio = false;

	String fileName = "/RADIO/RADIO_" + String(radioNum, HEX) + ".INI";
	char fName[25];
	fileName.toCharArray(fName, fileName.length() + 1);

	if (SD.exists(fName));
	{
		SD.remove(fName);
Wait2Rmv:
		delay(200);
		if (SD.exists(fName))
		{
			SerialOut("Radio file still exists", true);
			goto Wait2Rmv;
		}
	}

	blockISR = true;
	myFile = SD.open(fName, FILE_WRITE);
	delay(100);
	if (myFile)
	{
		if (!gComMode)
		{
			SerialOut("Saving " + String(fName), true);
		}
		myFile.println("Radio setting file for Radio " + String(radioNum + 1));
		myFile.println(" ");
		myFile.println("label=" + mnuRadioTag[radioNum - 1]);
		myFile.println("gFreq=" + String(gFreqRadio));
		delay(1);
		for (int i = 0; i < 4; i++)
		{
			myFile.println("gVFO_Freq" + String(i) + "=" + String(gVFO_Freq[i]));
			myFile.println("gVFO_Mode" + String(i) + "=" + String(gVFO_Mode[i]));
			myFile.println("gVFO_BW" + String(i) + "=" + String(gVFO_BW[i]));
		}
		delay(1);
		myFile.println("gRMode=" + String(gRMode));
		myFile.println("gMSelMode=") + String(gMSelMode);
		myFile.println("gRadioCAT=" + String(gRadioCAT));
		// don't save this - it's manually selected by operator
		//myFile.println("gRadioPort=" + String(gRadioPort));

		myFile.println("gBand=" + String(gBand));
		delay(1);
		myFile.println("gShareDB=" + String(gShareDB));

		//CW page
		myFile.println("gCWSpeed.val=" + String(gCWSpeed));		
		myFile.println("gFarnsSpeed.val=" + String(gFarnsSpeed));
		myFile.println("gKeyType.val=" + String(gKeyType));
		delay(1);
		myFile.println("gSideTone.val=" + String(gSideTone));
		myFile.println("gPaddle.val=" + String(gPaddle));
		myFile.println("gRmtMode.val=" + String(gRmtMode));
		myFile.println("gLink_ZeroBeat=" + String(gLink_ZeroBeat));
		delay(1);
		//Flex or rigCtl adrs
		if (!gIPAdrs.length())
		{
			gIPAdrs = gIPAdrsDefault; //default label
		}
		myFile.println("gIPAdrs=" + gIPAdrs);		
		myFile.println("gFlexPort=" + gFlexPort);		
		myFile.println("gFlexSlc.val=" + String(gFlexSlice)); //use nextion variable for CTR2 compatability
		myFile.println("gFlexAnt.val=" + String(gFlexAnt));
		delay(1);
		myFile.println("gFlexVol.val=" + String(gFlexVol));
		myFile.println("gFlexApiPTT.val=" + String(gFlexApiPTT));
		myFile.println("gFlexApiKey.val=" + String(gFlexApiKey));
		myFile.println("gFlexDelay.val=" + String(gFlexDelay));
		delay(1);

		//PCR page		
		myFile.println("gPCRAGC=" + String(gPCRAGC));
		myFile.println("gPCRAtten=" + String(gPCRAtten));
		myFile.println("gPCRNB=" + String(gPCRNB));
		myFile.println("gPCRDSP=" + String(gPCRDSP));
		delay(1);
		myFile.println("gPCRANch=" + String(gPCRANch));
		myFile.println("gPCRBW=" + String(gPCRBW));
		myFile.println("gPCRVol=" + String(gPCRVol));
		myFile.println("gPCRSq=" + String(gPCRSq));
		delay(1);
		myFile.println("gPCRIF=" + String(gPCRIF));
				
		//Amp sequencing
		myFile.println("gEnabAmpSeq=" + String(gEnabAmpSeq)); 
		myFile.println("gSeqPre=" + String(gSeqPre));
		myFile.println("gSeqPost=" + String(gSeqPost));
	}
	else
	{
		SerialOut("Error creating Radio File", true);
	}
	myFile.close();
	blockISR = false;
	delay(10);
}

void LoadRadio()
{	
	if (gRadioPort > 15)
	{
		SerialOut("Invalid gRadioPort= " + (gRadioPort), true);
		return;
	}

	char chr = 0;
	char fName[25];

	String fileName = "/RADIO/RADIO_" + String(gRadioPort, HEX) + ".INI";

	fileName.toCharArray(fName, fileName.length() + 1);

	PrtStatus("Loading Radio Port " + String(gRadioPort +1), false);
	if (!SD.exists(fName))
	{
		//save current radio and exit				
		SerialOut("Creating " + fileName,true);
		EraseRadio(); //set to defaults and create new aux files
		SaveRadio(gRadioPort, true);		
	}

	//SerialOut("Start: " + String(millis()), true);
	uint8_t ptr = 0; //points to '='

	LoadRadioBaud(gRadioPort); //get baud list for this port

	//reset amp sequencing so it must be re-enabled by new setting
	gEnabAmpSeq = false;
	gSeqPre = 0;
	gSeqPost = 0;

	myFile = SD.open(fName, FILE_READ);
	if (myFile)
	{
		//parse file and reload variables
		String txt = "";		
		//String sel = String(gRadioPort);

		while (myFile.available())
		{
			txt = "";
			while (myFile.available())
			{
				chr = myFile.read();
				if (chr != 10 && chr != 13)
				{
					txt += String(chr);
				}
				else if (chr == 13)
				{
					break;
				}
			}
			//SerialOut(txt, true);
			ptr = txt.indexOf("=") + 1;
			delay(20);
			if (txt.length() > ptr)
			{
				//SerialOut(txt, true);
				if (txt.substring(0, ptr) == "label=")
				{
					//mnuRadioTag[gRadioPort] = ExtractValue(txt); //no need to do this
				}
				else if (txt.substring(0, ptr) == "gFreq=") //Home page variables
				{
					gFreqRadio = ExtractValue(txt);
				}
				else if (txt.substring(0, ptr) == "gVFO_Freq0=")
				{
					gVFO_Freq[0] = ExtractValue(txt);
				}
				else if (txt.substring(0, ptr) == "gVFO_Freq1=")
				{
					gVFO_Freq[1] = ExtractValue(txt);
				}
				else if (txt.substring(0, ptr) == "gVFO_Freq2=")
				{
					gVFO_Freq[2] = ExtractValue(txt);
				}
				else if (txt.substring(0, ptr) == "gVFO_Freq3=")
				{
					gVFO_Freq[3] = ExtractValue(txt);
				}
				else if (txt.substring(0, ptr) == "gVFO_Mode0=")
				{
					gVFO_Mode[0] = ExtractValue(txt);
				}
				else if (txt.substring(0, ptr) == "gVFO_Mode1=")
				{
					gVFO_Mode[1] = ExtractValue(txt);
				}
				else if (txt.substring(0, ptr) == "gVFO_Mode2=")
				{
					gVFO_Mode[2] = ExtractValue(txt);
				}
				else if (txt.substring(0, ptr) == "gVFO_Mode3=")
				{
					gVFO_Mode[3] = ExtractValue(txt);
				}
				else if (txt.substring(0, ptr) == "gVFO_BW0=")
				{
					gVFO_BW[0] = ExtractValue(txt);
				}
				else if (txt.substring(0, ptr) == "gVFO_BW1=")
				{
					gVFO_BW[1] = ExtractValue(txt);
				}
				else if (txt.substring(0, ptr) == "gVFO_BW2=")
				{
					gVFO_BW[2] = ExtractValue(txt);
				}
				else if (txt.substring(0, ptr) == "gVFO_BW3=")
				{
					gVFO_BW[3] = ExtractValue(txt);
				}
				else if (txt.substring(0, ptr) == "gRMode=")
				{
					gRMode = ExtractValue(txt);
				}
				else if (txt.substring(0, ptr) == "gMSelMode=")
				{
					gMSelMode = ExtractValue(txt);
				}
				else if (txt.substring(0, ptr) == "gRadioCAT=")
				{
					gRadioCAT = ExtractValue(txt);
					if (gRadioCAT >= RADIO_CAT_NUMRADIOS)
					{
						gRadioCAT = 0;
					}
				}
				else if (txt.substring(0, ptr) == "gShareDB=")
				{
					gShareDB = ExtractValue(txt);
				}

				else if (txt.substring(0, ptr) == "gCWSpeed.val=") // CW page
				{
					gCWSpeed = ExtractValue(txt);
					CalcDitDur();
					CalcDitFarnsDur(); //do them both 
				}
				else if (txt.substring(0, ptr) == "gFarnsSpeed.val=")
				{
					gFarnsSpeed = ExtractValue(txt);
					CalcDitFarnsDur();
				}
				else if (txt.substring(0, ptr) == "gKeyType.val=")
				{
					gKeyType = ExtractValue(txt);
					gKeyTypePrev = gKeyType; //used to restore previous keyer model if menu was opened					
					if (gKeyType > 6)
					{
						gKeyType = 0; //chk for bad data
					}
				}
				else if (txt.substring(0, ptr) == "gSideTone.val=")
				{
					gSideTone = ExtractValue(txt);
				}
				else if (txt.substring(0, ptr) == "gPaddle.val=")
				{
					gPaddle = ExtractValue(txt);
				}
				else if (txt.substring(0, ptr) == "gRmtMode.val=") //original label from CTR2 for compatibility
				{
					gRmtMode = ExtractValue(txt);
				}
				else if (txt.substring(0, ptr) == ("gLink_ZeroBeat="))
				{
					gLink_ZeroBeat = ExtractValue(txt);
				}
				else if (txt.substring(0, ptr) == "gIPAdrs=") //Flex page
				{
					gIPAdrs = txt.substring(ptr);
					gIPAdrs.trim();
					if (!gIPAdrs.length())
					{
						gIPAdrs = gIPAdrsDefault; //default label
					}
				}
				else if (txt.substring(0, ptr) == "gFlexPort=")
				{
					gFlexPort = txt.substring(ptr);
					gFlexPort.trim();
				}
				else if (txt.substring(0, ptr) == "gFlexSlc.val=")
				{
					gFlexSlice = ExtractValue(txt);
				}		
				else if (txt.substring(0, ptr) == "gFlexVol.val=")
				{
				gFlexVol = ExtractValue(txt);
				}
				else if (txt.substring(0, ptr) == "gFlexApiPTT.val=")
				{
					gFlexApiPTT = ExtractValue(txt);
				}
				else if (txt.substring(0, ptr) == "gFlexApiKey.val=")
				{
					gFlexApiKey = ExtractValue(txt);
				}
				else if (txt.substring(0, ptr) == "gFlexDelay.val=")
				{
					gFlexDelay = ExtractValue(txt);
				}
				/* don't save these, use what the radio has set on it
				else if (txt.substring(0, ptr) == "gFlexPO.val=")
				{
					gFlexPO = ExtractValue(txt);				
				}
				else if (txt.substring(0, ptr) == "gFlexTO.val=")
				{
					gFlexTO = ExtractValue(txt);			
				}
				else if (txt.substring(0, ptr) == "gFlexWNB.val=")
				{
					gFlexWNB = ExtractValue(txt);						
				}
				else if (txt.substring(0, ptr) == "gFlexNB.val=")
				{
					gFlexNB = ExtractValue(txt);
				}
				else if (txt.substring(0, ptr) == "gFlexNR.val=")
				{
					gFlexNR = ExtractValue(txt);
				}
				else if (txt.substring(0, ptr) == "gFlexAPF.val=")
				{
					gFlexAPF = ExtractValue(txt);
				}
				else if (txt.substring(0, ptr) == "gFlexANF.val=")
				{
					gFlexANF = ExtractValue(txt);
				}
				*/
				else if (txt.substring(0, ptr) == "gFlexAnt.val=")
				{
					gFlexAnt = ExtractValue(txt);
				}
				//get PCR1000 properties
				else if (txt.substring(0, ptr) == "gPCRAGC=")
				{
					gPCRAGC = ExtractValue(txt);
				}
				else if (txt.substring(0, ptr) == "gPCRAtten=")
				{
					gPCRAtten = ExtractValue(txt);
				}
				else if (txt.substring(0, ptr) == "gPCRNB=")
				{
					gPCRNB = ExtractValue(txt);
				}
				else if (txt.substring(0, ptr) == "gPCRDSP=")
				{
					gPCRDSP = ExtractValue(txt);
				}
				else if (txt.substring(0, ptr) == "gPCRANtch=")
				{
					gPCRANch = ExtractValue(txt);
				}
				else if (txt.substring(0, ptr) == "gPCRBW=")
				{
					gPCRBW = ExtractValue(txt);
					switch (gRMode)
					{
					case RADIO_FM_W:
						if (gPCRBW < 2 || gPCRBW > 4)
						{
							gPCRBW = 3; //force to 50khz if invalid
						}						
						break;
					case RADIO_CW:
					case RADIO_LSB:
					case RADIO_USB:
					case RADIO_DIG_L:
					case RADIO_DIG_H:
						//if > 6khz force to 3khz
						if (gPCRBW > 1)
						{
							gPCRBW = 0;
						}
						break;
					case RADIO_AM:
					case RADIO_FM_N:
						//if > 15khz force to 6 khz
						if (gPCRBW > 2)
						{
							gPCRBW = 1;
						}
						break;
					}					
				}
				else if (txt.substring(0, ptr) == "gPCRVol=")
				{
					gPCRVol = ExtractValue(txt);
				}
				else if (txt.substring(0, ptr) == "gPCRSq=")
				{
					gPCRSq = ExtractValue(txt);
				}
				else if (txt.substring(0, ptr) == "gPCRIF=")
				{
					gPCRIF = ExtractValue(txt);
				}
				else if (txt.substring(0, ptr) == "gEnabAmpSeq=")
				{
					gEnabAmpSeq = ExtractValue(txt);
				}
				else if (txt.substring(0, ptr) == "gSeqPre=")
				{
					gSeqPre = ExtractValue(txt);	
				}
				else if (txt.substring(0, ptr) == "gSeqPost=")
				{
					gSeqPost = ExtractValue(txt);
				}
			}
		}		
	}
	else
	{
		PrtStatus("Fail: " + fileName, true);
		delay(1500);	
		gRadioPort = 0; //load radio 1 files
	}
	myFile.close();
	PrtStatus(hmiTagLine, false);

	gTxMode = RADIO_TX_OFF;
	gMSel = 0;
	gMSelPrev = 0;
	gMSelPtr = 0;

	//reset save flag
	gSaveRadio = false;
	SetRadioBaud();

	if (gRadioCAT == RADIO_CAT_PCR1000)
	{		
		SerialOut("Pinging PCR1000", true);
		switch (PCR_Ping())
		{
		case 2:			
			gTxMode = RADIO_TX_ENAB;
			gTxModePrev = gTxMode;
			PCR_Initialize(); //initialize it
			break;
		}		
	}

	Chk4FreqListFile();
	Chk4BandListFile(false); //don't erase band data
	Chk4TxMsgFile();
	ChkFreqModeChng(); //save any changes to prev lists
	CalcBand(); //sets gBand
	SerialOut("Radio Port: [" + mnuRadioTag[gRadioPort] + "] has been loaded", true); //just report radio loaded
	if (gComMode == COM_MODE_TERM)
	{
		SerialOut("Enter a new Radio Port name now or press [Enter] to continue...", true);
	}
	gMemLabel = " "; //revert to default until we load a memory location
	gInit = true; //hmi is initialized 		
	SendLinkParams();

}

void LoadRadioTags()
{
	String txt = "";
	char chr = 0;
	uint8_t ptr = 0;
	int idx = 0;
	int jdx = 0;
	//if label file doesn't exist, create it

	//use this code to reset labels file

	/*
	if (SD.exists("/MISC/RAD-TAGS.TXT"))
	{
		SerialOut("============ Deleted RAD-TAGS", true);
		SD.remove("R-LABEL.TXT");
		delay(10000);
	}
	*/


	if (!SD.exists("/MISC/RAD-TAGS.TXT"))
	{
		SerialOut("Creating RAD-TAGS.TXT", true);
		myFile = SD.open("/MISC/RAD-TAGS.TXT", FILE_WRITE);
		for (int i = 0; i < 16; i++)
		{
			//first 16 are radio tags, last 8 are ant tags
			txt = "=-- " + String(i + 1) + " --";
			if (i == gRadioPort && gShareDB == true)
			{
				//add astrick to name if sharing dbase
				txt = '*' + txt;
			}
			else
			{
				//not shared - just space
				txt = ' ' + txt;
			}

			myFile.println(String(i + 1) + txt);
		}
		myFile.close();
	}
	delay(500);

	myFile = SD.open("/MISC/RAD-TAGS.TXT", FILE_READ);
	if (myFile)
	{
		//parse file and reload variables
		while (myFile.available())
		{
			txt = "";
			while (myFile.available())
			{
				chr = myFile.read();
				if (chr != 10 && chr != 13)
				{
					txt += String(chr);
				}
				else if (chr == 13)
				{
					break;
				}
			}
			jdx = txt.indexOf("=") + 1;
			if (jdx > 1)
			{
				//ignore unless it's a pair
				idx = txt.toInt() - 1; //labels saved in idx=txt format				
				mnuRadioTag[idx] = txt.substring(jdx); //these will be uploaded to HMI Radio page when it's selected		
				String temp = mnuRadioTag[idx];
				//SerialOut("Load: " + mnuRadioTag[idx] + ", Len= " + String(temp.length()), true);
				if (idx == gRadioPort)
				{
					//chk to see if "*" is in name, if so, we're sharing dbase
					txt = mnuRadioTag[idx];
					if (txt.charAt(0) == '*')
					{
						gShareDB = true;
					}
					else
					{
						gShareDB = false;
					}
				}
				ptr++;
				if (ptr > 15)
				{
					break;
				}
			}
		}
	}
	else
	{
		SerialOut("Failed to open /MISC/RAD-TAGS.TXT", true);
	}
	myFile.close();
	delay(100);
}

void SaveRadioTags()
{
	String txt = "";
	if (SD.exists("/MISC/RAD-TAGS.TXT"))
	{
		SD.remove("/MISC/RAD-TAGS.TXT");
		delay(1000);
	}
	//SerialOut("Saving Radio Labels", true);
	blockISR = false;
	myFile = SD.open("/MISC/RAD-TAGS.TXT", FILE_WRITE);
	if (myFile)
	{
		myFile.println("# This file contains the labels used for the 16 radio channels");
		myFile.println("# and the 8 antenna switches.");
		myFile.println("# To reset these labels to default, delete this file.");
		myFile.println("#");
		for (int i = 0; i < mnuRadioTagMax; i++)
		{
			//mark radio tag with backspace chr if shareDB is selected
			txt = mnuRadioTag[i];
			if (i < 16)
			{
				//radio tags 0 to 15, CTR2 uses antenna tags 16 to 23
				if (i == gRadioPort)
				{
					if (txt.charAt(0) != '*' && gShareDB == true)
					{
						mnuRadioTag[i] = '*' + txt.substring(1); //add shared flag to front of txt
					}
					else if (txt.charAt(0) == '*' && gShareDB == false)
					{
						//remove it if we're not sharing db
						mnuRadioTag[i] = ' ' + txt.substring(1);
					}
					if (txt.charAt(0) != ' ' && txt.charAt(0) != '*')
					{
						//pad with char(1) if not using it for shared 
						mnuRadioTag[i] = ' ' + mnuRadioTag[i];
					}
				}
			}
			myFile.println(String(i + 1) + "=" + mnuRadioTag[i]);
			//SerialOut("Saving " + mnuRadioTag[i], true);
		}
	}
	myFile.close();
	blockISR = false;
}

uint32_t ExtractValue(String txt )
{
	//NOTE: THIS ROUTINE EXPECTS AN INTEGER VALUE IN TXT (NO PERIODS)

	//extract the value from the pair and return it
	//count in digits one at a time - the .toInt() function only works up to 31 bits then overflows 

	uint8_t strt = txt.indexOf("=") + 1;
	uint32_t val = 0;
	txt = txt.substring(strt);

	//count digits in and multiply them by their digit position

	long multiplier = 1; //inc by 10 each interation
	for (int i = txt.length(); i > 0; i--)
	{
		String temp = txt.substring(i - 1, i);
		val += (temp.toInt() * multiplier);
		multiplier *= 10;
	}

	//SerialOut("Extracted val=" + String(val), true);

	return val;
}


void Chk4TxMsgFile()
{
	//chk for existing TxMsg file, if none, create it
	String temp = "";
	if (gShareDB == false)
	{
		temp = String(gRadioPort, HEX);
	}
	String fileName = "/TXMSG/TXMSG" + temp + ".TXT";

	char fName[25];
	fileName.toCharArray(fName, fileName.length() + 1);	
	if (!SD.exists(fName))
	{
		if (!gComMode)
		{
			SerialOut("Saving " + String(fName), true);
		}
		PrtStatus("Saving Tx Message File", false);
		myFile = SD.open(fName, FILE_WRITE);
		myFile.println("### Transmit Message Buffer file for CTR2-Mini");
		myFile.println("### You can edit this file in a text editor");
		myFile.println("### CTR2-Mini supports 16 buffers with a maximum length of 75 chrs each");
		myFile.println("### The first entry by default is the Contest Exchange");
		myFile.println("### The second entry always holds the last buffer sent");
		myFile.println("### ..................Entries start below this line .....................");
		//save current buffers to new radio port
		for (int i = 0; i < mnuTxMsgMax; i++)
		{						
			temp = mnuTxMsg[i];
			temp.trim();
			if (!i && !temp.length())
			{
				temp = "^#&"; //default to contest exchange if first entry is empty
				mnuTxMsg[i] = temp;
			}
			myFile.println(temp);
		}
		myFile.close();		
		delay(2000);
	}
	else
	{
		LoadTxMsg();
	}
}

void LoadTxMsg()
{
	String temp = "";
	if (gShareDB == false)
	{
		temp = String(gRadioPort, HEX);
	}
	String fileName = "/TXMSG/TXMSG" + temp + ".TXT";

	char fName[25];
	fileName.toCharArray(fName, fileName.length() + 1);
	//SerialOut("Loading " + String(fName), true);
	String txt;
	char chr = 0;
	uint8_t rowCtr = 0;
	bool commentLn = false; //goes true if line starts with #
	
	//clear array
	for (int i = 0; i < mnuTxMsgMax; i++)
	{
		mnuTxMsg[i] = " ";
	}

	myFile = SD.open(fName, FILE_READ);
	if (myFile)
	{
		while (myFile.available() && rowCtr < mnuTxMsgMax)
		{
			chr = (myFile.read());
			if (chr == '#' && txt.length() == 0)
			{
				commentLn = true;
			}
			if (commentLn == true)
			{
				//read until we hit the end of this line
				if (chr == 13)
				{
					//reached the linefeed, start new line
					commentLn = false;
					txt = "";
				}
			}
			else
			{
				if (chr == 10)
				{
					//just ignore lf, use cr to inc 
					//txt = "";					
				}
				else if (chr == 13)
				{
					//save to gTxMsg array
					txt.trim();
					if (!rowCtr && !txt.length())
					{
						txt = "^#%"; //default to contest exchange if #1 slot is empty
					}
					mnuTxMsg[rowCtr] = txt; 
					//SerialOut(String(rowCtr) + ": " + txt, true);
					//new line
					txt = "";
					rowCtr++;
				}
				else
				{
					txt += char(chr);
				}
			}
		}
	}
	else
	{
		//failed to open
		PrtStatus("Fail: " + fileName,true);		
		delay(500);
	}	
	myFile.close();
	PrtStatus(hmiTagLine, false);

	//send new strings to hmi

}



void UpdateTxMsgSave()
{
	SaveTxMsg(gRadioPort);
}


void SaveTxMsg(uint8_t radioNum)
{
	//save TxMsg strings to TXMSG.TXT	
	String temp = "";
	if (gShareDB == false)
	{
		temp = String(radioNum, HEX);
	}

	String fileName = "/TXMSG/TXMSG" + temp + ".TXT";

	char fName[25];
	fileName.toCharArray(fName, fileName.length() + 1);
	if (SD.exists(fName));
	{
		//SerialOut("Removing " + String(fName), true);
		SD.remove(fName);
		delay(1000);
	}
	//SerialOut("Saving " + String(fName), true);

	blockISR = true;
	myFile = SD.open(fName, FILE_WRITE);

	if (myFile)
	{
		//write header first
		myFile.println("### Transmit Message TXT file for CTR2-Mini Radio " + String(radioNum));
		myFile.println("### You can edit this file in a text editor");
		myFile.println("### Msg #2 is used to store the last message sent");
		myFile.println("### 14 msgs total with maximum # a chrs per message of 75");
		myFile.println("###");
		String txt = "";
		for (int i = 0; i < mnuTxMsgMax; i++)
		{
			txt = mnuTxMsg[i];
			if (txt.length() == 0)
			{
				txt = " ";
			}
			myFile.println(txt);			
		}
	}
	blockISR = false;
	myFile.close();
}


void Chk4FreqListFile()
{
	//chk if freq list file exists
	String fileName = "VFOMEM.CSV"; //if using a common db
	if (gShareDB == false)
	{
		fileName = "VFOMEM_" + String(gRadioPort, HEX) + ".CSV";
	}
	fileName = "/FREQ/" + fileName;

	char fName[25];
	fileName.toCharArray(fName, fileName.length() + 1);

	//--------------------------------------------------------------------------
	//enable this line to delete current freq list and start with a blank list
	//SD.remove(fName);
	//--------------------------------------------------------------------------

	if (!SD.exists(fName))
	{
		//move the default values from the first 5 locations to the last 5		
		//clear vfoMem arrays
		PrtStatus("Creating Frequeny List File", false);
		for (int i = 0; i < vfoMemMax; i++)
		{
			vfoMemLabel[i] = " ";
			vfoMemA_Freq[i] = 0;
			vfoMemA_Mode[i] = 0;
			vfoMemA_BW[i] = 0;
			vfoMemB_Freq[i] = 0;
			vfoMemB_Mode[i] = 0;
			vfoMemB_BW[i] = 0;	
		}
		for (int i = 0; i < 10; i++)
		{
			//save 60m cw & voice memorys in 80-89						
			vfoMemLabel[79 + i] = defaultVFOMemLabel[i];
			vfoMemA_Freq[79 + i] = defaultVFOMemFreq[i];			
			vfoMemA_Mode[79 + i] = defaultVFOMemMode[i];			
			vfoMemA_BW[i] = 0;
		}
		for (int i = 10; i < 15; i++)
		{
			//save beacon freqs in last 96-100
			vfoMemLabel[85 + i] = defaultVFOMemLabel[i];
			vfoMemA_Freq[85 + i] = defaultVFOMemFreq[i];			
			vfoMemA_Mode[85 + i] = defaultVFOMemMode[i];			
			vfoMemA_BW[i] = 0;
		}
		//now save the correct data to the new file
		SaveFreqList(gRadioPort); //save new default file
	}
	else
	{
		LoadFreqList();
	}
}

void LoadFreqList()
{
	String txt = "";
	String temp = "";
	char chr = 0;
	int rowCtr = 0;
	int colCtr = 0;
	int i = 0;
	bool commentLn = false; //goes true if line starts with #

	String fileName = "VFOMEM.CSV"; //if using a common db
	if (gShareDB == false)
	{
		fileName = "VFOMEM_" + String(gRadioPort, HEX) + ".CSV";
	}
	fileName = "/FREQ/" + fileName;
	char fName[25];
	fileName.toCharArray(fName, fileName.length() + 1);
	PrtStatus("Loading Frequency List File", false);
	
	//clear vfoMem arrays
	for (int i = 0; i < vfoMemMax; i++)
	{
		vfoMemLabel[i] = " ";
		vfoMemA_Freq[i] = 0;
		vfoMemA_Mode[i] = 0;
		vfoMemA_BW[i] = 0;
		vfoMemB_Freq[i] = 0;
		vfoMemB_Mode[i] = 0;
		vfoMemB_BW[i] = 0;		
	}

	int loc = 0;	

	myFile = SD.open(fName, FILE_READ);
	if (myFile)
	{
		while (myFile.available())
		{
			chr = (myFile.read());
			if (chr == '#' && txt.length() == 0)
			{
				commentLn = true;
			}
			if (commentLn == true)
			{
				//read until we hit the end of this line
				txt += chr;
				if (chr == 13)
				{
					//reached the linefeed, start new line				
					commentLn = false;
					colCtr = 0;
					txt = "";
				}
			}
			else
			{
				//look for column terminator
				if (chr == ',')
				{
					//SerialOut("Loading", true);					
					//end of field, save value
					switch (colCtr)
					{
					case 0:
						//ignore the record #, just inc column
						colCtr++;
						break;
					case 1:
						if (txt.length() == 0)
						{
							txt = " ";
						}
						vfoMemLabel[rowCtr] = txt;
						colCtr++;
						break;
					case 2:
						if (txt.indexOf(".") == -1)
						{
							//no period - freq in Hz (default)
							vfoMemA_Freq[rowCtr] = txt.toInt();
							if (vfoMemA_Freq[rowCtr] < 100000)
							{
								//convert to kHz
								vfoMemA_Freq[rowCtr] *= 1000;
							}
						}
						else
						{
							if (txt.toFloat() < 1000)
							{
								//convert it to MHz if the result is lower then 1 GHz
								vfoMemA_Freq[rowCtr] = uint32_t(txt.toFloat() * 1000000); //convert to MHz
							}
							else
							{
								vfoMemA_Freq[rowCtr] = uint32_t(txt.toFloat() * 1000); //convert to kHz
							}							
						}												
						colCtr++;
						break;
					case 3:
						i = txt.toInt();
						if (i >= 0 && i < 9)
						{
							vfoMemA_Mode[rowCtr] = i;
						}
						colCtr++;
						break;
					case 4:
						i = txt.toInt();
						if (i >= 0 && i < 5)
						{
							vfoMemA_BW[rowCtr] = i;
						}
						colCtr++;
						break;
					case 5:						
						if (txt.indexOf(".") == -1)
						{
							//no period - freq in Hz (default)
							vfoMemB_Freq[rowCtr] = txt.toInt();
							if (vfoMemB_Freq[rowCtr] < 100000)
							{
								//convert to kHz
								vfoMemB_Freq[rowCtr] *= 1000;
							}
						}
						else
						{
							if (txt.toFloat() < 1000)
							{
								//convert it to MHz if the result is lower then 1 GHz
								vfoMemB_Freq[rowCtr] = uint32_t(txt.toFloat() * 1000000); //convert to MHz
							}
							else
							{
								vfoMemB_Freq[rowCtr] = uint32_t(txt.toFloat() * 1000); //convert to kHz
							}
						}						
						colCtr++;
						break;
					case 6:
						i = txt.toInt();
						if (i >= 0 && i < 9)
						{
							vfoMemB_Mode[rowCtr] = i;
						}
						colCtr++;
						break;
					case 7:
						i = txt.toInt();
						if (i >= 0 && i < 5)
						{
							vfoMemB_BW[rowCtr] = i;
						}
						colCtr++;
						break;
					}
					txt = "";
				}
				else if (chr == 13)
				{
					//new line					
					txt = "";
					colCtr = 0;
					rowCtr++;
					if (rowCtr == vfoMemMax)
					{
						goto ExitHr;
					}
				}
				else if (chr == 10)
				{
					//just ignore lf, use cr to inc 
					txt = "";
				}
				else
				{
					if (colCtr < 2 && txt.length() < 16)
					{
						txt += char(chr);
					}
					else if (isDigit(chr) || chr == '.')
					{						
						//valid number
						txt += char(chr);
					}
				}
			}
		}

ExitHr:
		myFile.close();
		PrtStatus(hmiTagLine, false);
		return;
	}
	else
	{
		//failed to open
		PrtStatus("Fail: " + fileName, true);
		myFile.close();
		delay(1500);		
	}
	PrtStatus(hmiTagLine, false);

}


void SaveFreqList(uint8_t radioNum)
{
	//save vfoMem to VFOMEM_[radio#].CSV file on SD card	
	String fileName = "VFOMEM.CSV"; //if using a common db
	if (gShareDB == false)
	{
		fileName = "VFOMEM_" + String(radioNum, HEX) + ".CSV";
	}
	fileName = "/FREQ/" + fileName;
	char fName[25];
	fileName.toCharArray(fName, fileName.length() + 1);
	if (!gComMode)
	{
		SerialOut("Saving " + String(fName), true);
	}
	if (SD.exists(fName))
	{
		SD.remove(fName);
		delay(1000);
	}

	blockISR = true;
	myFile = SD.open(fName, FILE_WRITE);
	if (myFile)
	{
		//write header first		
		myFile.println("### CSV Frequency & mode file for CTR2 Radio " + String(radioNum));
		myFile.println("### Enter the Label - Frequency (in hz) - radio mode - operation mode in the columns below");
		myFile.println("###");
		myFile.println("### Radio Mode: 0=CW 1=CWr 2=LSB 3=USB 4=FMn 5=FMw 6=AM 7=DigL 8=DigH");
		myFile.println("### PCR1000 BW: 0=3khz 1=6kHz 2=15kHz 3=50kHz 4=230kHz (FMw only)");
		myFile.println("###");
		myFile.println("### NOTE 1: Frequencies below 100000 are converted to kHz");
		myFile.println("### NOTE 2: Frequencies with decimal points are converted to MHz");
		myFile.println("###");
		myFile.println("###Record#,Label,VFOA,VFOA Mode,VFOA BW, VFOB, VFOB Mode, VFOB BW");
				
		for (int i = 0; i < vfoMemMax; i++)
		{
			String txt = vfoMemLabel[i];
			if (txt.length() == 0)
			{
				txt = " ";
			}			
			txt = String(i + 1) + "," + txt + "," + String(vfoMemA_Freq[i]) + "," + String(vfoMemA_Mode[i]) + "," + String(vfoMemA_BW[i]) + "," +
				String(vfoMemB_Freq[i]) + "," + String(vfoMemB_Mode[i]) + "," + String(vfoMemB_BW[i]) + ", "; //terminate with a blank 				
			myFile.println(txt);
			//SerialOut(txt, true);
		}
	}
	blockISR = false;
	myFile.close();
}

