// =====================================================================================================================================================================================================
// (c) 2022 Lynn Hansen, KU7Q															                                                                                                               |
// This Source Code Form is subject to the terms of the GNU GENERAL PUBLIC LICENSE, Version 3, 29 June 2007. A copy of this license can be found here: https://choosealicense.com/licenses/gpl-3.0/|
// =====================================================================================================================================================================================================


//This module contains routines to service the HMI display

void ChkChr2HMI()
{
	//process keyboard input
	uint8_t newBand = 0;
	
	if (keyboardChr != 0)
	{
		//send chr from KeyboardPress_ISR()
		//Serial.println("keyboard chr=" + String(keyboardChr));//terminal mode blocks normal SerialOut()				 
		int curDir = 0;

		//functions common to all pages
		switch (keyboardChr)
		{
		case 9: //tab
			break; //already processed - just show menu
		case 13: //cr
			ProcessBtn(BTN_SW); //encoder click
			break;
		case 27: //esc - return home
			//already processed
			//PrtHomePg();
			break;
		case 194: //f1 - f1-f9= send buffers 0 to 8, f1 repeats last buffer
		case 195: //f2 		
		case 196: //f3
		case 197: //f4
		case 198: //f5
		case 199: //f6
		case 200: //f7
		case 201: //f8
		case 202: //f9				
			gTxBuffer = gTxBuffer + " " + mnuTxMsg[keyboardChr - 194]; //remove offset to get index
			SaveRepeatBuffer(mnuTxMsg[keyboardChr - 194]);//so F1 can repeat it			
SendIt:		//send to keyer			
			if (!gPlay)
			{
				TermOut("\n\n\r[Insert] to pause, [End] to stop ", true);
				TermOut("Send:" + gTxBuffer, false);
			}
			else
			{
				TermOut(mnuTxMsg[0], false); //just add new chrs it to existing line
			}
			gPlay = KEYER_MODE_RUN;
			break;		
		case 303: //shift f3 - buffer 9
		case 304: //shift f4 - 10
		case 305: //shift f5 - 11		
		case 306: //shift f6 - 12
		case 307: //shift f7 - 13	
			gTxBuffer = gTxBuffer + " " + mnuTxMsg[keyboardChr - 294];
			SaveRepeatBuffer(mnuTxMsg[keyboardChr - 294]);//so F1 can repeat it	
			goto SendIt;
			break;
		case 203: //f10 - send call sign
			gTxBuffer = gTxBuffer + "^"; // " " + mnuKeyerCall;
			mnuTxMsg[0] = mnuKeyerCall;
			goto SendIt;
			break;
		case 204: //f11 - send contest SN			
			gTxBuffer = gTxBuffer + "#"; //insert prosign so we can chk for duplicates - ChkPaddles() replaces this with actual SN	and increments mnuKeyerSN		
			mnuTxMsg[0] = String(mnuKeyerSN); //save current contest SN before # increments it again to repeat with F1
			goto SendIt;
			break;
		case 205: //f12 - send exhange
			gTxBuffer = gTxBuffer + "%"; // " " + mnuKeyerExch;
			mnuTxMsg[0] = mnuKeyerExch;
			goto SendIt;
			break;
		case 209: //[insert] - pause tx buffer
			if (gPlay == KEYER_MODE_PAUSE)
			{
				gPlay = KEYER_MODE_RUN;
				SerialOut("Keyer restarted... " + gTxBuffer, true);
			}
			else if (gPlay == KEYER_MODE_RUN)
			{
				gPlay = KEYER_MODE_PAUSE;
				SerialOut("Keyer paused...", true);
			}			
			break;	
		case 213: // [end] - stop buffer
			if (gPlay)
			{				
				gPlay = KEYER_MODE_STOP;
				ResetSpeedBoost();
				gTxBuffer = "";
				PrtHomePg();
				SerialOut("Keyer stopped", true);
			}					
			break;		

		case 210: //home - toggle tx off/enab
			if (gTxMode)
			{
				gTxMode = RADIO_TX_OFF;
				if (!mnuCBtnSel)
				{
					//[C] btn menu set to PTT, 
					gBtnPTT = false;
					ControlPTT(LOW);
					PrtBtnC();
				}
			}
			else
			{
				gTxMode = RADIO_TX_ENAB;
			}
			PrtBtnA(true);
			break;

		//cursor control keys work in all modes
		case 211: //pgUp = open Freq menu
			hmiMode = HMI_MENU;
			gMnu = MNU_FREQ;
			CreateMnu("", false);
			break;
		case 214: //pgDown = open user menu
			ProcessHomeBtn(BTN_C);
			break;
		case 215: //right arrow
			ProcessBtn(BTN_RIGHT);
			break;
		case 216: //left arrow
			ProcessBtn(BTN_LEFT);
			break;
		case 217: //down arrow - dec selected control value
			if (hmiMode == HMI_MENU)
			{
				curDir = 1; //menus are backwards
			}
			else
			{
				curDir = -1;
			}
			SetEncoderCtrl(curDir);			
			break;
		case 218: //up arrow = inc selected control value
			if (hmiMode == HMI_MENU)
			{
				curDir = -1; //menus are backwards
			}
			else
			{
				curDir = 1;
			}
			SetEncoderCtrl(curDir);
			break;
		case 308: //shift f8 - these are not used yet
		case 309: //shift f9 
		case 310: //shift f10
			break;
		default:
			//check modes if not control
			if (keyboardChr > 31 && keyboardChr < 128)
			{
				switch (hmiMode)
				{
				case HMI_HOME:
					TermOut(String(char(keyboardChr)), false); // just print it
					break;
				case HMI_MENU:
					break;
				case HMI_VALUE:
					break;
				case HMI_TXTINPUT:
					break;
				}
			}
		}		

	}
	keyboardChr = 0;
}

void PrtBtnA(bool refresh)
{
	//txt box below left btn
	static String lastTxt = " "; //only update if txt changes
	String txt = "";
	String term = ""; //holds string for terminal
	long x = 0;
	long y = 0;
	long w = 65;
	long h = 20;	
	uint16_t fGnd = hmiFGnd;
	uint16_t bGnd = hmiBase;	
	switch (gMnu)
	{
	case MNU_NONE:
		if (!gPlay)
		{
			//normal mode, we're not txing buffer 
			switch (gTxMode)
			{
			case RADIO_TX_OFF:
				fGnd = hmiFGnd;
				bGnd = hmiBase;
				if (gRadioCAT != RADIO_CAT_PCR1000)
				{
					txt = "Tx Off";
					if (gSyncOffset)
					{
						txt = "Offset";						
						bGnd = TFT_BLUE;
					}
					term = "Tx Mode: Disabled";
					if (!mnuCBtnSel)
					{
						PrtBtnC(); //update [C] btn if PTT is selected
					}
				} 
				else
				{
					txt = "PcrOff";
					term = "PCR1000: Power Off";
				}								
				break;
			case RADIO_TX_ENAB:				
				if (gRadioCAT != RADIO_CAT_PCR1000)
				{
					txt = "TxEnab";
					term = "Tx Mode: Enabled";
					if (!mnuCBtnSel)
					{
						PrtBtnC(); //update [C] btn if PTT is selected
					}
				}
				else
				{
					txt = "PcrOn";
					term = "PCR1000: Power On";
				}
				fGnd = TFT_BLACK;
				bGnd = TFT_GREEN;				
				break;
			case RADIO_TX_KEY:
				txt = "TX CW";
				term = "Tx Mode: " + txt;				
				if (gRadioCAT != RADIO_CAT_FLEX || !gFlexApiKey || gFlexHandleTx.length())
				{					
					fGnd = TFT_WHITE;
					bGnd = TFT_RED;
				}
				else
				{
					//post warning that SmartSDR must be online
					PrtStatus("SmartSDR Must Be Online", true);
				}
				break;
			case RADIO_TX_PTT:				
				txt = "TX PTT";
				term = "Tx Mode: " + txt;
				fGnd = TFT_WHITE;
				bGnd = TFT_RED;				
				break;
			case RADIO_TX_WDOG:
				//watchdog timer expired, display yellow warning
				txt = "WDog!";
				term = "Tx Mode: " + txt;
				fGnd = TFT_BLACK;
				bGnd = TFT_YELLOW;
				break;
			}
			term += " - [Home] to toggle";
		}
		else if (gPlay == KEYER_MODE_RUN || gPlay == KEYER_MODE_PAUSE)
		{
			txt = "Stop";
			fGnd = TFT_WHITE;
			bGnd = TFT_RED;
			PrtBtnB(); //add Paused label to btnB
		}
		break;
	default: //we're in a menu
		if (!gBtnALongPress)
		{
			txt = "Ok";
			fGnd = hmiFGnd;
			bGnd = hmiBGnd;
		}
		else 
		{
			txt = "Esc";
			fGnd = TFT_BLACK;
			bGnd = TFT_YELLOW;
		}
		break;
	}
	if (refresh || txt != lastTxt)
	{
		SetFont(0, 1, fGnd);
		tft.fillRoundRect(x, y, w, h, 5, bGnd);
		tft.drawCentreString(txt, x + (w / 2), y + 2, 1);
		if (txt != lastTxt && term.length())
		{
			//only if state actually changes
			TermOut(term, true);
		}
	}
		
	lastTxt = txt;

}

void PrtBtnB()
{
	//txt box below center btn
	long x = 80;
	long y = 0;
	long w = 60;
	long h = 20;
	String txt = "";
	uint16_t fGnd = TFT_GREEN;
	uint16_t bGnd = hmiBase;
	if (!gPlay || hmiMode != HMI_HOME)
	{
		//normal if keyer isn't active	
		switch (gMnu)
		{
		case MNU_NONE:
			//vfo functions
			if (!gLock)
			{
				switch (gVFO)
				{
				case MNU_VFO_A:
					txt = "VFO-A";
					break;
				case MNU_VFO_B:
					txt = "VFO-B";
					fGnd = TFT_CYAN;
					break;
				case MNU_VFO_SPLIT:
					txt = "SPLIT";
					fGnd = TFT_YELLOW;
					break;
				case MNU_VFO_MEM:
					if (!gMSelMode)
					{
						txt = "MEM-A";
					}
					else
					{
						txt = "MEM-B";
					}
					fGnd = TFT_RED;
					break;
				}
			}
			else
			{
				txt = "LOCK";
				fGnd = TFT_DARKGREY;
			}			
			break;
		default:
			//inc value
			txt = "^  <";
			fGnd = hmiFGnd;
			bGnd = hmiBGnd;
			break;
		}
	}
	else
	{		
		//keyer active
		if (gPlay == KEYER_MODE_RUN)
		{
			txt = "Pause";
			fGnd = TFT_BLACK;
			bGnd = TFT_YELLOW;
		}
		else
		{
			txt = "Start";
			fGnd = TFT_BLACK;
			bGnd = TFT_GREEN;
		}
		
	}
	SetFont(0, 1, fGnd);
	tft.fillRoundRect(x, y, w, h, 5, bGnd);
	tft.drawCentreString(txt, x + (w / 2), y + 2, 1);

}

void PrtBtnC()
{
	//txt box below right btn
	long x = 160;
	long y = 0;
	long w = 66;
	long h = 20;	
	String txt = "";
	uint16_t fGnd = hmiFGnd;
	uint16_t bGnd = hmiBase;
	switch (gMnu)
	{
	case MNU_NONE:		
		txt = mnuCBtn[mnuCBtnSel];
		//use abbriviated labels for btnC for these		
		switch (mnuCBtnSel)
		{
		case MNU_C_BTN_FREQ:
			txt = "Freq";
			break;
		case MNU_C_BTN_TXMSG:
			txt = "TxMsg";
			break;	
		case MNU_C_BTN_EDIT_TXMSG:
			txt = "Edit";
			break;
		case MNU_C_BTN_FUNC_BTN:
			txt = "Func";
			break;
		case MNU_C_BTN_RADIO:
			//don't update the radio button until after the first radio profile is loaded - otherwise, if Flex or PCR was selected the default will overwrite the selected c-btn option
			if (gInit)
			{
				switch (gRadioCAT)
				{
				case RADIO_CAT_FLEX:
					txt = "Flex";
					break;
				case RADIO_CAT_PCR1000:
					txt = "PCR";
					break;
				default:
					txt = "TxMsg"; //fall back to TxMsg if radio selection changed
					mnuCBtnSel = MNU_C_BTN_TXMSG;
					break;
				}
				break;
			}
		case MNU_C_BTN_PUSH:
			txt = "Push";
			break;
		case MNU_C_BTN_PULL:
			txt = "Pull";
			break;
		}
		break;
	default:
		txt = " v  ?";
		fGnd = hmiFGnd;
		bGnd = hmiBGnd;
		break;
	}
	
	//SerialOut("mnuCBtnSel= " + String(mnuCBtnSel) + ", gTxMode= " + String(gTxMode) + ", hmiMode= " + String(hmiMode), true);
	if (!mnuCBtnSel && gTxMode && hmiMode == HMI_HOME && gRmtMode != MNU_RMT_KEY)
	{
		//use color background if [C] Btn menu is set for PTT and Tx is enabled		
		if (!gPTTState)
		{
			SetFont(0, 1, bGnd);
			tft.fillRoundRect(x, y, w, h, 5, TFT_GREEN);
		}
		else
		{
			SetFont(0, 1, fGnd);
			tft.fillRoundRect(x, y, w, h, 5, TFT_RED);
		}
	}
	else
	{
		SetFont(0, 1, fGnd);
		tft.fillRoundRect(x, y, w, h, 5, bGnd);
	}
	tft.drawCentreString(txt, x+(w/2), y+2, 1);
}

void PrtSplash()
{
	//print splash screen
	tft.fillRoundRect(20, 20, 280, 200, 5, hmiBGnd);	
	SetFont(10, 2, hmiFGnd);
	tft.drawCentreString("CTR2-Mini", 160, 40, 1);
	SetFont(10, 1, hmiFGnd);
	tft.drawCentreString("Lynn Hansen, KU7Q", 160, 90, 1);
	tft.drawCentreString(version, 160, 120, 1);	
	tft.drawCentreString("GPL v3 License", 160, 150, 1);
	tft.drawCentreString("(c)2022", 160, 180, 1);
	delay(2000);	
	PrtHomePg();
}

void PrtHomePg()
{
	//this routine updates the entire home page
	static bool alreadyHr = false;

	if (alreadyHr)
	{
		return;
	}
	alreadyHr = true;

	if (!gPlay)
	{
		blockISR = true; //so we don't overwrite during ISR processing
	}	
	
	//clear terminal screen
	if (gComMode==COM_MODE_TERM)
	{
		TermOut(TERM_CLS + " " + TERM_SHOW_CURSOR, false);
	}
	lastStatusMsg = ""; //only used in menus
	hmiMode = HMI_HOME;	
	gMnu = 0;		

	gBtnCLongPress = false;  //clear long press flag now
	gKeyType = gKeyTypePrev; //restore previous keyer in case we changed it to enter txt input
	tft.fillScreen(hmiBase); // fills entire the screen with color black	
	if (!gPlay)
	{
		TermOut(" ", true); //add white space if we just stopped sending buffer
	}
	PrtFreq(true); //also calls PrtRMode()
	PrtRadioTag(true);
	//chk for matching memory
	if (gMSel > 0 && gFreqRadio == vfoMemA_Freq[gMSel - 1])
	{
		PrtMemLabel(vfoMemLabel[gMSel - 1]);
	}
	else
	{		
		PrtMemLabel(gMemLabel); //use existing label if any - may be copying between shared and normal database
	}
	PrtRMode(true);
	PrtRadioCAT(true);
	PrtBtnA(true);	
	PrtBtnB();
	PrtBtnC();
	PrtWiFi();	
	PrtOffVFO();
	PrtRIT();
	PrtOptions();
	PrtStatus(hmiTagLine, false);
	if (!gPlay)
	{
		TermOut("Type chrs then [Enter] to Tx, [Enter] for Menu", true);
		TermOut(" ", true);
	}
	blockISR = false;
	alreadyHr = false;
}

void PrtHelp(String topic, String txt)
{	
	long x = 0;
	long y = 0;
	long w = 320;
	long h = 28;
	
	hmiMode = HMI_HELP;
	tft.fillScreen(hmiBase); // fills entire the screen with theme base color
	SetFont(2, 1, hmiFGnd);
	tft.fillRect(x, y, w, h, hmiBGnd);
		
	tft.drawCentreString(topic, x + (w / 2), y + 3, 4);
	//starting cursor locations
	x = 10;
	y = 45;
	h = 23; //space between lines

	tft.cursor_x = x;
	tft.cursor_y = y;
	int ptr = 0; //points to next space
	char chr = 0;
	String temp = "";
	//print individual txt lines
	for (int i = 0; i < txt.length(); i++)
	{
		ptr = txt.indexOf(' ', i); //inc line if this word won't fit on this line
		chr = txt.charAt(i);
		//SerialOut("i= " + String(i) + ", ptr = " + String(ptr) + ", chr = " + String(chr) + ", " + String(chr, HEX), true);
		if (chr == 10 || chr == 13)
		{
			temp.trim();			
			if (temp.length())
			{
				tft.print(temp);
			}
			temp = "";
			y = y + h;
			tft.cursor_x = x;
			tft.cursor_y = y;
		}
		else
		{
			if (temp.length() < 28 && ((ptr - i + temp.length() < 28) || ptr == -1))
			{
				temp += String(chr);
			}
			else
			{
				temp.trim();
				tft.print(temp);
				//SerialOut(temp,true);
				temp = String(chr); //start with the last chr				
				y = y + h;
				tft.cursor_x = x;
				tft.cursor_y = y;
			}
		}
	}
	if (temp.length())
	{
		//print last line		
		temp.trim();
		tft.println(temp);
		//SerialOut(temp, true);
	}
	TermOut("\n\r===============================================================================", true);
	TermOut(topic + "\n\r", true);
	TermOut(txt, true);
	TermOut("\n\rPress [Esc] to exit Help", true);
	TermOut("===============================================================================", true);
}

void PrtRadioCAT(bool refresh)
{
	if (gMnu)
	{
		return;
	}
	//print memory label
	String txt = mnuRadioCAT[gRadioCAT];
	String lead = ""; //leading 0
	if (gRadioCAT == RADIO_CAT_FLEX)
	{
		txt += ": " + gFlexSliceLbl[gFlexSlice];
	}
	else if (gRadioCAT == RADIO_CAT_ICOM && gIcmAdr)
	{
		if (gIcmAdr < 0x10)
		{
			lead = "0";
		}
		txt += ": 0x" + lead + String(gIcmAdr, HEX);
	}
	else if (gRadioCAT == RADIO_CAT_RIGCTLD)
	{
		txt = "rigctld";
	}
	//override if we're Mini-A in remote link mode
	uint16_t bColr = hmiBGnd;
	if ((gBLEConnected || gUDPConnected) && (gSyncMode == SYNC_MODE_RMT && gLinkMiniA))
	{
		txt = "Link"; //print 'Remote' in the Tag label
		bColr = TFT_RED; //display red box
	}
	txt.trim();
	static String lastLabel = txt;
	long x = 162;
	long y = 30;
	long w = 148;
	long h = 33;
	long r = 5;
	uint16_t bGnd = hmiBGnd;
	if (txt.length() == 0)
	{
		bGnd = hmiBase; //blank out if nothing to display
	}
	SetFont(0, 1, hmiFGnd);
	if (txt != lastLabel || refresh == true)
	{
		tft.fillRoundRect(x, y, w, h, r, bColr); //erase old band		
		lastLabel = txt;
	}
	tft.drawCentreString(txt, x + (w / 2), y + 5, 4);
}

void PrtMemLabel(String txt)
{
	if (gMnu || !gInit)
	{
		return; //don't run until arrays are initialized - causes lockup
	}
		
	if (txt == " ")
	{
		//txt not included, chk for match				
		for (int i = 0; i < (vfoMemMax - 1); i++)
		{
			if (gFreqRadio == vfoMemA_Freq[i] && gRMode == vfoMemA_Mode[i]) // --add to chk for both freq& mode
			{				
				txt = vfoMemLabel[i];
				break;
			}
		}
	}
	
	//print memory label	
	txt.trim();
	long x = 162;
	long y = 123;
	long w = 148;
	long h = 33;
	long r = 5;
	uint16_t fGnd = hmiFGnd;
	uint16_t bGnd = hmiBGnd;
	if (!txt.length())
	{
		fGnd = hmiBase; //blank out if nothing to display
		bGnd = hmiBase;
	}
	SetFont(0, 1, fGnd);
	tft.fillRoundRect(x, y, w, h, r, bGnd); 		
	tft.drawCentreString(txt, x + (w / 2), y + 5, 4);	
}

void PrtOffVFO()
{
	//prints the offline vfo freq & mode
	String frq = "";
	uint8_t md = 0;
	String txt = "";
		
	switch (gVFO)
	{
	case MNU_VFO_A:
	case MNU_VFO_SPLIT:
		if (gVFO_Freq[MNU_VFO_B] >= 100000)
		{
			frq = String(gVFO_Freq[MNU_VFO_B]);
			md = gVFO_Mode[MNU_VFO_B];
		}
		break;
	case MNU_VFO_B:
		//vfo-A of offline
		frq = String(gVFO_Freq[MNU_VFO_A]);
		md = gVFO_Mode[MNU_VFO_A];
		break;
	case MNU_VFO_MEM:
		if (!gMSelMode)
		{
			if (gVFO_Freq[MNU_VFO_B+2] >= 100000)
			{
				frq = String(gVFO_Freq[MNU_VFO_B+2]);
				md = gVFO_Mode[MNU_VFO_B+2];
			}
		}
		else
		{
			frq = String(gVFO_Freq[MNU_VFO_A+2]);
			md = gVFO_Mode[MNU_VFO_A+2];
		}
		break;
	}
			
	if (frq.length())
	{
		//format frq
		int ctr = 0;
		for (int i = frq.length()-1; i > -1; i--)
		{
			txt = String(frq.charAt(i)) + txt;
			ctr++;
			if (ctr == 3 && i)
			{
				txt = "." + txt;
				ctr = 0;
			}
		}
		txt += " " + mnuMode[md + 1]; //offset for toggle in mnuMode[0]
	}

	long x = 15;
	long y = 160;
	long w = 155;
	long h = 25;
	SetFont(0, 1, hmiFGnd);
	tft.fillRect(x, y, w, h, hmiBase);
	tft.drawString(txt, x, y+3, 1);
}

void PrtOptions()
{
	//print options just above the status line
	if (gMnu)
	{
		return;
	}

	//load option string
	String txt = "";
	if (gLock)
	{
		txt = "FREQ & MODE ARE LOCKED";
	}
	else if (gRadioCAT == RADIO_CAT_FLEX)
	{
		switch(gTCPCon)
		{
		case TCP_CON_OFF:
			txt = "[Offline]";
			break;
		case FLEX_CON_GUI:
			txt = "[GUI]";
			break;
		case FLEX_CON_NON_GUI:
			txt = "[Non-GUI]";
			break;
		case FLEX_CON_CONNECTING:
			txt = "[Connecting]";
			break;
		}
		txt += "   Slice:" + gFlexSliceLbl[gFlexSlice] + "   Ant:" + String(gFlexAnt) + "   Vol:" + String(gFlexVol);
	}
	else if (gRadioCAT == RADIO_CAT_PCR1000)
	{
		txt = "Vol:" + String(uint8_t(gPCRVol / 2.56));
		txt += " Sq:" + String(uint8_t(gPCRSq / 2.56));
		txt += " BW:";
		switch (gPCRBW)
		{
		case 1:
			txt += "6";
			break;
		case 2:
			txt += "15";
			break;
		case 3:
			txt += "50";
			break;
		case 4:
			txt += "230";
			break;
		default:
			txt += "3";
			break;
		}			
		if (gPCRDSP)
		{
			txt += " DSP";
		}
		if (gPCRAGC)
		{
			txt += " AGC";
		}
		if (gPCRAtten)
		{
			txt += " Atn";
		}
		if (gPCRNB)
		{
			txt += " NB";
		}		
	}
	else
	{
		//everything else
		txt = "Spd:" + String(gCWSpeed) + "   STn:";
		if (gSideTone)
		{
			txt += "On   ";
		}
		else
		{
			txt += "Off   ";
		}
		txt += mnuCWKeyer[gKeyType] + "   ";
		if (gShareDB)
		{
			txt += "Shared";
		}
	}

	long x = 0;
	long y = 185;
	long w = 320;
	long h = 25;	
	SetFont(0, 1, hmiFGnd);
	tft.fillRect(x, y, w, h,  hmiBase);		
	tft.drawCentreString(txt, x + (w / 2), y+5, 1);
}

void PrtRadioTag(bool refresh)
{
	if (gMnu)
	{
		return;
	}	
	String txt = mnuRadioTag[gRadioPort];
	txt.trim();
	if (txt.charAt(0) == '*')
	{
		txt = txt.substring(1, txt.length()); //trim off shared db flag
	}
	//override if we're Mini-A in remote link mode	
	uint16_t bColr = hmiBGnd;
	if ((gBLEConnected || gUDPConnected) && (gSyncMode == SYNC_MODE_RMT && gLinkMiniA))
	{
		txt = "Remote"; //print 'Link' in the CAT label
		bColr = TFT_RED; //display red box
	}
	static String lastRadio = txt;
	long x = 10;
	long y = 30;
	long w = 147;
	long h = 33;
	long r = 5;
	SetFont(0, 1, hmiFGnd);	
	if (txt != lastRadio || refresh == true)
	{
		tft.fillRoundRect(x, y, w, h, r, bColr); //erase old band		
		lastRadio = txt;
	}
	tft.drawCentreString(txt, x + (w / 2), y + 5, 4);
}

void PrtBand(bool refresh)
{
	if (gMnu)
	{
		return;
	}	

	static String oldBnd = "";
	long x = 10;
	long y = 123;
	long w = 70;
	long h = 33;
	long r = 5;
	uint16_t bColr = hmiBGnd;
	uint16_t fColr = hmiFGnd;
	//switch band color to ASC color if sending band data
	if (gAntSwSendBand && gAntSwOnline)
	{
		if (gLinkMiniA)
		{
			bColr = TFT_RED;
			fColr = TFT_WHITE;
		}
		else
		{
			bColr = TFT_YELLOW;
			fColr = TFT_BLACK;
		}
	}
	SetFont(0, 1, fColr);
	String bnd= CalcBand();
	if (refresh || oldBnd != bnd)
	{
		tft.fillRoundRect(x, y, w, h, r, bColr); //erase old band		
		tft.drawCentreString(bnd, x + (w / 2), y + 5, 4);
		oldBnd = bnd;
	}
}

void PrtRIT()
{
	//prints the rx freq offset below the memory label
	if (gMnu)
	{
		if (gMnu == MNU_FREQ && mnuFreqSel == MNU_FREQ_RIT)
		{
			//update RIT window if it's open
			SetMnuParam();
		}
		return;
	}
	long x = 160;
	long y = 160;
	long w = 160;
	long h = 25;
	String txt = "";
	if (!gRIT)
	{
		txt = "                   ";
	}
	else
	{
		txt = "RIT: " + String(gRIT) + " Hz";
	}
	SetFont(0, 1, hmiFGnd);
	tft.fillRect(x, y, w, h, hmiBase);
	tft.drawCentreString(txt, x+80, y + 3, 1);

}

void PrtRMode(bool refresh)
{
	if (gMnu)
	{
		return;
	}	
	static String lastMode = mnuMode[gRMode + 1];
	long x = 86;
	long y = 123;
	long w = 70;
	long h = 33;
	long r = 5;
	
	SetFont(0, 1, hmiFGnd);
	
	if (mnuMode[gRMode + 1] != lastMode || refresh == true)
	{
		tft.fillRoundRect(x, y, w, h, r, hmiBGnd); //erase old mode		
		lastMode = mnuMode[gRMode + 1];			
	}
	//offset mnuMode by 1 for Toggle Tx Enab
	tft.drawCentreString(mnuMode[gRMode + 1], x+(w/2), y+5, 4);
}

void PrtStatus(String txt, bool useRedBGnd)
{
	//print txt on bottom status line
	long x = 0;
	long y = 210;
	long w = 320;
	long h = 30;	
	static String lastMsg = ""; //holds last alert msg while hmiTagLineRefresh is active

	SetFont(0, 1, hmiFGnd);
	
	if (useRedBGnd || hmiTagLineRefresh)
	{
		if (lastMsg.length() && hmiTagLineRefresh)
		{
			txt = lastMsg;
		}
		tft.fillRect(x, y, w, h, TFT_RED);
		hmiTagLineRefresh = 1; //wait hmiTagLineRefreshMax milliseconds then restore sticky msg
		lastMsg = txt;
	}
	else
	{
		tft.fillRect(x, y, w, h, hmiBGnd); //normal msg
		hmiTagLineRefresh = 0; //reset countup timer
		lastMsg = "";
	}

	tft.drawCentreString(txt, w/2, y+5, 4);	

	if (gComMode == COM_MODE_TERM && hmiMode == HMI_MENU && txt != hmiTagLine)
	{
		TermOut("\n\r   " + txt, true);
	}
}

void PrtTxtInput(bool refresh)
{
	//get text value for labels - returns with new string
	static char oldTxt[txtInputMax]; //holds the string we started with, only update changed values so display doesn't flicker	

	if (refresh)
	{
		tft.fillRoundRect(2, 29, 316, 175, 5, hmiBase);
		tft.drawRoundRect(2, 29, 316, 175, 5, hmiBGnd);
		SetFont(1, 1, hmiFGnd);
		tft.drawCentreString("Start with space to erase all", 160, 107, 1);
		tft.drawCentreString("Key: 6 Dits=erase, 6 Dahs=space", 160, 132, 1);
		tft.drawCentreString("Encoder: Push-Turn moves cursor", 160, 157, 1);
		tft.drawCentreString("Encoder: Push Saves changes", 160, 182, 1);		
		switch (gMnuSel)
		{
		case MNU_FREQ_SV:			
		case MNU_FREQ_REN:
			txtInputSrcMax = 10; //max# of chrs
			PrtStatus("Edit Memory [" + String(txtInputSrc + 1) + "] Label", false);
			break;
		case MNU_ANT_SW_LBL_EDIT:
			txtInputSrcMax = 10;
			PrtStatus("Edit Ant " + String(mnuAntEditSel + 1) + "'s Label", false);
			break;
		case MNU_RADIOPORT:
			txtInputSrcMax = 11; //max# of chrs
			PrtStatus("Edit Radio Port [" + String(txtInputSrc + 1) + "] Name", false);
			break;
		case MNU_TXMSG_EDIT:
			txtInputSrcMax = txtInputMax; //max# of chrs
			PrtStatus("Edit Tx Msg [" + String(mnuTxMsgSel + 1) + "]", false);
			break;
		case MNU_WIFI_SSID_EDIT:
			txtInputSrcMax = 40; //max# of chrs
			PrtStatus("Edit the WiFi SSID", false);
			break;
		case MNU_WIFI_STAKEY_EDIT:
			txtInputSrcMax = 40; //max# of chrs
			PrtStatus("Edit the WiFi Password", false);
			break;
		case MNU_UDP_ADRS_EDIT:
			txtInputSrcMax = 15;
			PrtStatus("Edit the Rmt Mini's IP Adrs", false);
			break;
		case MNU_LINK_OPTION6:
			txtInputSrcMax = 15;
			PrtStatus("Edit the Ant Switch IP Adrs", false);
			break;
		case MNU_RIGCTLD_IP_EDIT:
			txtInputSrcMax = 16;
			PrtStatus("Edit the rigctld IP Adrs", false);
			break;
		case MNU_FLEX_IP_EDIT:
			txtInputSrcMax = 16;
			PrtStatus("Edit the Flex IP Adrs", false);
			break;
		case MNU_FLEX_PORT_EDIT:
			txtInputSrcMax = 5;
			PrtStatus("Edit the Flex IP Port", false);
			break;
		case MNU_FLEX_SLICE_EDIT:
			txtInputSrcMax = 1;
			PrtStatus("Select the Active Slice", false);
			break;
		case MNU_KEYER_CALL_EDIT:
			txtInputSrcMax = 20;
			PrtStatus("Edit Your Call", false);
			break;
		case MNU_KEYER_EXCH_EDIT:
			txtInputSrcMax = 20;
			PrtStatus("Edit the Contest Exchange", false);
			break;
		case MNU_CONFIG_TAGLINE_EDIT:
			txtInputSrcMax = 30;
			PrtStatus("Edit the HMI Tag Line", false);
			break;
		}

		if (txtInputSrcMax > 10)
		{
			//draw progress bar showing location of current window
			int xx = 10; //coord for border
			int yy = 90;
			int ww = 300;
			int hh = 10;
			float scale = float(10) / float(txtInputSrcMax);
			int bar = ww * scale;
			int start = xx + ((float(txtInputSrcStart)/float(txtInputSrcMax)) * ww);
			if (start + bar > ww)
			{
				start = 310 - bar;
			}			
			tft.drawRoundRect(xx, yy, ww, hh, 5, hmiBGnd); 
			tft.fillRoundRect(start, yy, bar, hh, 5, hmiBGnd);
		}
	}

	int x = 5;
	int y = 40;
		
	SetFont(2, 2, TFT_GREEN); //use mono-bold 18pt

	for (int i = txtInputSrcStart; i < txtInputSrcStart + 14; i++)
	{
		if (txtInput[i] != oldTxt[i] || refresh)
		{				
			if (txtInputPtr == i)
			{
				tft.setTextColor(TFT_RED);
			}
			tft.fillRect(x, y, 24, 34, hmiBase); //blank old value
			DispNum(txtInput[i], x, y);
			if (txtInputPtr == i)
			{
				tft.drawRect(x, y, 22, 30, TFT_RED); 
			}
		}
		x += 22;

		tft.setTextColor(TFT_GREEN);
	}

	//save current setting
	for (int i = 0; i < txtInputMax; i++)
	{
		oldTxt[i] = txtInput[i];
	}

}

void PrtTxBuffer()
{
	//print txBuffer left justified on bottom status line
	//use red background if gTxMode is enabled
	long x = 0;
	long y = 210;
	long w = 320;
	long h = 30;
	SetFont(0, 1, hmiFGnd);
	if (!gTxMode)
	{
		tft.fillRect(x, y, w, h, hmiBGnd);
		hmiTagLineRefresh = 0;
	}
	else
	{
		tft.fillRect(x, y, w, h, TFT_RED);
		hmiTagLineRefresh = 1; //wait hmiTagLineRefreshMax milliseconds then restore sticky msg
	}
	tft.drawString(gTxBuffer.substring(0,30), x + 10, y + 5, 4);
}


void PrtFreq(bool refreshAll)
{
	static bool alreadyHr = false;

	if (gMnu || alreadyHr)
	{
		return;
	}
	alreadyHr = true;

	//print frequency display    
	int x = 22; //start on this column - each chr consumes 28 pixels
	int y = 80; //display on this row    
	char sFrq[12] = { '0','0','0','0','0','0','0','0','0','0','0','0' }; //array to hold individual digits, 100mhz digit starts at [2]
	static char oFrq[12] = { '0','0','0','0','0','0','0','0','0','0','0','0' }; //this array holds the last value printed, only update digit if it changes to eliminate flicker
	bool newFreq = false;
	static uint32_t oldUnit = gUnit;

	PrtBand(refreshAll);

	if (refreshAll)
	{
		for (int i = 0; i < 12; i++)
		{
			oFrq[i] = 0; //force all to update
		}
		tft.fillRoundRect(11, 71, 298, 45, 5, hmiBase); //blank old value
		if (gBLEConnected || gUDPConnected)
		{
			SendLinkParams(); //send update
		}
	}
	
	uint16_t fGnd = TFT_GREEN; //for vfoA
	switch (gVFO)
	{	
	case MNU_VFO_B:
		fGnd = TFT_CYAN;
		break;
	case MNU_VFO_SPLIT:
		fGnd = TFT_YELLOW;
		break;
	case MNU_VFO_MEM:
		fGnd = TFT_RED;
		break;
	}
	if (gSyncOffset)
	{
		fGnd = TFT_BLUE; //if we're offset because other Mini has Tx enabled in Sync mode
	}

	if (gLock)
	{
		//override w/grey text
		fGnd = TFT_DARKGREY;
	}
	SetFont(6, 2, fGnd);

	tft.drawRoundRect(10, 70, 300, 46, 5, hmiBGnd);

	//derive freq digit, display, and inc column

	//calculate step digit to highlight from gUnit	
	int step = String(gUnit).length();
			
	CalcKenwoodFreq(sFrq);
	uint8_t ctr = 9; //start from 100mhz digit
	bool lead = true; //display space until we hit the first digit above 0
	for (int i = 2; i < 11; i++)
	{
		if (lead == true && sFrq[i] == '0')
		{
			sFrq[i] = ' '; //replace 0 with space 
		}
		else
		{
			lead = false;
		}
		if (ctr == 6 || ctr == 3)
		{
			//insert '.'
			tft.drawString(".", x-1, y);
			x += 16;
		}
		if (sFrq[i] != oFrq[i])
		{
			newFreq = true;
			tft.fillRect(x, y, 28, 28, hmiBase); //blank old value
			if (step == ctr && !gLock)
			{
				tft.setTextColor(TFT_RED);
			}
			if (step > ctr && gZeroDigits)
			{
				//greyout low digits and show them as small 0s
				SetFont(1, 1, fGnd);
				tft.setTextColor(TFT_DARKGREY);
				tft.drawString("0", x, y+14);
				SetFont(6, 2, fGnd);
			}
			else
			{
				DispNum(sFrq[i], x, y);
			}
		}		
		ctr--;
		x += 28;

		tft.setTextColor(fGnd);
	}
	
	//save new freq to old freq array
	for (int i = 0; i < 12; i++)
	{
		oFrq[i] = sFrq[i];
	}

	//update terminal if we're not sending gTxBuffer
	if (gComMode == COM_MODE_TERM && hmiMode == HMI_HOME && !gPlay)
	{
		String f = String(gFreqRadio);
		String frq= "";
		frq = f.substring(0, f.length() - 3) + "." + f.substring(f.length() - 3, f.length()) + " kHz";
		if (oldUnit != gUnit)
		{
			String stp = String(float(gUnit) / 1000, 3) + " kHz";
			int z = frq.length() - stp.length(); //leading zeros so step lines up with freq
			for (int i = 0; i < z; i++)
			{
				stp = " " + stp;
			}			
			TermOut("   Step: " + stp, true);
			oldUnit = gUnit;
		}
		else if (newFreq || refreshAll)
		{
			//only display if gUnit didn't change	
			if (!gPlay)
			{
				if (gVFO != MNU_VFO_SPLIT)
				{
					TermOut("   Freq: " + frq + "  Mode: " + mnuMode[gRMode + 1], true);
				}
				else
				{					
					//display tx freq too					
					f = String(gVFO_Freq[MNU_VFO_B]);
					String frqB = f.substring(0, f.length() - 3) + "." + f.substring(f.length() - 3, f.length()) + " kHz";
					//make both the same length
					while (frq.length() < frqB.length())
					{
						frq = " " + frq;
					}
					while (frqB.length() < frq.length())
					{						
						frqB = " " + frqB;
					}
					TermOut("Rx Freq: " + frq + "  Mode: " + mnuMode[gRMode + 1], true);
					TermOut("Tx Freq: " + frqB + "  Mode: " + mnuMode[gVFO_Mode[MNU_VFO_B] + 1], true);
				}
				TermOut(" ", true); //white space
			}
		}
	}
	alreadyHr = false;
}

void PrtWiFi()
{
	//prints a WiFi icon in the top right if WiFi is connected	
	int offset = 0; //to center link letter if failed off

	uint16_t m = 300; //middle of ant icon
	if (gWiFi)
	{
		uint16_t colr = TFT_GREEN;
		tft.drawLine(m - 13, 5, m + 13, 5, colr);
		tft.drawLine(m - 13, 6, m + 13, 6, colr);
		tft.drawLine(m - 8, 9, m + 8, 9, colr);
		tft.drawLine(m - 8, 10, m + 8, 10, colr);
		tft.drawLine(m - 3, 13, m + 3, 13, colr);
		tft.drawLine(m - 3, 14, m + 3, 14, colr);
		tft.drawLine(m - 1, 17, m + 1, 17, colr);
		tft.drawLine(m - 1, 18, m + 1, 18, colr);
	}
	else
	{
		tft.fillRect(m - 13, 5, 27, 13, hmiBase);
	}

	uint16_t bColr = hmiBase;
	uint16_t fColr = hmiFGnd;
	if (gUDPAdrs.length() || gBLEConnected)
	{

		if (gBLEConnected)
		{
			bColr = TFT_BLUE;
		}
		else if (gUDPConnected)
		{
			bColr = TFT_GREEN;
			fColr = TFT_BLACK;
		}

		if ((gBLEConnected || gUDPConnected) && gLinkCommFail)
		{
			//link comms failed, grey out
			bColr = TFT_LIGHTGREY;
			fColr = TFT_BLACK;
			offset = 4; // move link letter over to center it
		}
	}

	tft.fillRoundRect(261, 3, 22, 20, 3, bColr);
	if (bColr != hmiBase)
	{
		String mode = "A"; //Mini A is the server
		String modifier = "";
		if (!gLinkMiniA)
		{
			mode = "B"; //Mini B is the client
		}			
		if ((gBLEConnected || gUDPConnected) && !gLinkCommFail)
		{
			//add function modifier

			switch (gSyncMode)
			{
			case SYNC_MODE_FREQ:
				if (gLink_ZeroBeat > 0)
				{
					modifier = ">";
				}
				else if (gLink_ZeroBeat < 0)
				{
					modifier = "<";
				}
				else
				{
					modifier = "f";
				}
				break;
			case SYNC_MODE_TRACK:
				modifier = "t";
				break;
			case SYNC_MODE_RMT:
				if (gLinkMiniA)
				{
					modifier = "r"; //we're the remote unit
				}
				else
				{
					modifier = "l"; //we're the local unit
				}
				break;
			default:
				modifier = "b";
				break;
			}
		}
		SetFont(1, 1, fColr); //ss-bold
		tft.drawString(mode, 262 + offset, 6);
		SetFont(2, 1, fColr); //mono-bold
		tft.drawString(modifier, 273, 2);		
	}
	
	//if ASC is connected we'll have a value in gAntSwA or B, print it
	String ant = "";	 
	tft.fillRect(242, 3, 14, 20, hmiBase); //erase previous value
	if (gAntSwOnline)
	{
		if (gAntSwActive)
		{
			if (gLinkMiniA || gSyncMode == SYNC_MODE_RMT)
			{
				bColr = TFT_RED;
				fColr = TFT_WHITE;
			}
			else
			{
				if (gAntSwDual)
				{
					bColr = TFT_YELLOW;
					fColr = TFT_BLACK;
				}
				else
				{
					bColr = hmiBase; //hide b if not dual port
					fColr = hmiBase;
				}
			}
		}
		else
		{
			bColr = TFT_LIGHTGREY;
			fColr = TFT_BLACK;
		}
		ant = "-";
		if (gLinkMiniA || gSyncMode == SYNC_MODE_RMT)
		{			
			if (gAntSwA && gAntSwOnline && gAntSwActive)
			{			
				ant = String(gAntSwA);
			}			
			tft.fillRoundRect(242, 3, 14, 20, 3, bColr);
			SetFont(0, 1, fColr); // color of A on ASC			
		}
		else 
		{	
			if (gAntSwB && gAntSwOnline && gAntSwActive)
			{			
				ant = String(gAntSwB);
			}			
			tft.fillRoundRect(242, 3, 14, 20, 3, bColr);
			SetFont(0, 1, fColr); //color of B on ASC				
		}
		tft.drawCentreString(ant, 248, 6, 1);
	}	
}

void PrtTxt(String txt, uint8_t fnt, uint8_t siz, uint16_t colr, long x, long y, long w)
{
	SetFont(fnt, siz, colr);	
	tft.drawCentreString(txt, x + (w/2), y, siz);
}


void SaveRepeatBuffer(String txt)
{
	//Save txt to mnuTxMsg[0] - if it has prosigns for contest exchanges, save full text instead of prosign
	if (!txt.indexOf("^") && !txt.indexOf("#") && !txt.indexOf("%"))
	{
		mnuTxMsg[0] = txt; //just save it
	}
	else
	{
		String msg = "";
		for (int i = 0; i < txt.length(); i++)
		{
			if (txt.charAt(i) == '^')
			{
				msg.trim();
				msg += " " + mnuKeyerCall + " ";
			}
			else if (txt.charAt(i) == '#')
			{
				msg.trim();
				msg += " " + String(mnuKeyerSN) + " ";
			}
			else if (txt.charAt(i) == '%')
			{
				msg.trim();
				msg += " " + mnuKeyerExch + " ";
			}
			else
			{
				msg += String(txt.charAt(i));
			}
		}
		mnuTxMsg[0] = msg;
	}
}


void DispNum(char num, int x, int y)
{
	//prints a chr at referenced location in specified foreground and background colors
	tft.drawString(String(char(num)), x, y);	
}

void SetFont(uint8_t fnt, uint8_t siz, uint16_t colr)
{
	//setup font - not sure how to pass the GFXfont code so we'll do it by numbers
	//add additional fonts as needed
	switch (fnt)
	{
	case 1:
		tft.setFreeFont(FSSB9); //san serif-bold 9pt
		break;
	case 2:
		tft.setFreeFont(FMB9); //mono-bold 9pt
		break;
	case 10:
		tft.setFreeFont(FMB12); //mono-bold 12pt
		break;
	default:
		//default, fnt==0
		tft.setFreeFont(FSS9); //regular 9pt san serif
		break;
	}
	tft.setTextSize(siz);
	tft.setTextColor(colr);

}

void SetThemeColors()
{
	switch (mnuThemeSel)
	{
	case HMI_THEME_GREY:
		hmiFGnd = TFT_WHITE;
		hmiBGnd = TFT_DARKGREY;
		hmiBase = TFT_BLACK;
		break;
	case HMI_THEME_VIOLET:
		hmiFGnd = TFT_WHITE;
		hmiBGnd = TFT_PURPLE;
		hmiBase = TFT_BLACK;
		break;
	case HMI_THEME_BLUE:
		hmiFGnd = TFT_WHITE;
		hmiBGnd = 0x2af0; //dark blue
		hmiBase = TFT_BLACK;
		break;
	case HMI_THEME_MAROON:
		hmiFGnd = TFT_WHITE;
		hmiBGnd = TFT_MAROON;
		hmiBase = TFT_BLACK;
		break;
	case HMI_THEME_CYAN:
		hmiFGnd = TFT_WHITE;
		hmiBGnd = TFT_DARKCYAN;
		hmiBase = TFT_BLACK;
		break;
	case HMI_THEME_GREEN:
		hmiFGnd = TFT_WHITE;
		hmiBGnd = TFT_DARKGREEN;
		hmiBase = TFT_BLACK;
		break;
	case HMI_THEME_OLIVE:
		hmiFGnd = TFT_WHITE;
		hmiBGnd = TFT_OLIVE;
		hmiBase = TFT_BLACK;
		break;
	}
}