// =====================================================================================================================================================================================================
// (c) 2022 Lynn Hansen, KU7Q															                                                                                                               |
// This Source Code Form is subject to the terms of the GNU GENERAL PUBLIC LICENSE, Version 3, 29 June 2007. A copy of this license can be found here: https://choosealicense.com/licenses/gpl-3.0/|
// =====================================================================================================================================================================================================



//This module contains the help pages
//Each page is referred to by the Menu # and gMnuSel that was selected with Long-press btnC

void GetHelp()
{
	String topic = "";
	String txt = "";

	//==================================================================
	//Help guidelines
	// 1. Keep it short - MNU_TXMSG in MNU_MAIN is about the max that can be displayed
	// 2. Use \n to break a line and start a new line
	// 3. Use \n\n to add white space between lines
	// 4. Leaving an extra space before \n and at the end helps with word wrap

	switch (gMnu)
	{
	case MNU_MAIN:
		//subtract 2 from the Main menu index because they start at 2
		mnuMainSel = gMnuSel;
		switch (gMnuSel)
		{
		case MNU_FREQ - 2:
			topic = "FREQUENCY MENU";
			txt = "This menu contains options for controlling the radio's frequency. ";
			break;
		case MNU_BAND - 2:
			topic = "BAND MENU";
			txt = "This menu allows you to select a band. \n\r";
			txt += "The last frequency and mode are saved for each band and will be restored. ";
			break;
		case MNU_MODE - 2:
			topic = "MODE MENU";
			txt = "This menu allows you to select a radio mode. \n\r";
			txt += "Radio modes selected for DigL and DigH are radio dependent. ";
			break;
		case MNU_VFO - 2:
			topic = "VFO TUNING MODE";
			txt = "Select a VFO tuning mode or Lock/Unlock the frequency from this mneu. \n\r";
			txt += "You can also step through these modes using the [B] button. \n\r";
			txt += "Long-press the [B] button to toggle Lock On and Off. ";
			break;
		case MNU_TXMSG - 2:
			topic = "TX MESSAGE MENU";
			txt = "This menu allows you to select a Transmit message to send. \n\r";
			txt += "Sending will commence immediately after the selection is made. \n\r";
			txt += "Press [A] to STOP and [B] to PAUSE. ";
			break;
		case MNU_TXMSG_EDIT - 2:
			topic = "EDIT TX MESSAGE";
			txt = "This menu allows you edit the Tx Message buffers. \n\r";
			txt += "The 1st buffer defaults to Contest Exch (^#%). The 2nd is always the last buffer sent. ";
			break;
		case MNU_KEYER - 2:
			topic = "KEYER MENU";
			txt = "This menu allows you to change Keyer setting and options. \n\r";
			txt += "Use the encoder to scroll and click to select. \n\r";
			txt += "Some options just toggle On/Off. ";
			break;
		case MNU_WIFI - 2:
			topic = "WIFI MENU";
			txt = "This menu allows you to change WiFi setting and options. \n\r";
			txt += "Use the encoder to scroll and click to select. \n\r";
			txt += "Click to Connect option to toggle WiFi On and Offline. ";
			break;
		case MNU_LINK - 2:
			topic = "LINK OPTIONS";
			txt = "This menu allows you to connect two Mini's and their radio's together using BLE and the CTR2-Mini SO2R Controller. \n\r";
			txt += "NOTE: Always connect the A radio's Mini first. ";
			break;
		case MNU_ANT_SW - 2:
			topic = "ANTENNA SWITCH";
			txt = "This menu allows you to remotely control a CTR2-Mini Antenna Switch Controller. \n\r";
			txt += "This is only available when WiFi linking is enabled. ";
			break;
		case MNU_RADIOPORT - 2:
			topic = "RADIO PORT MENU";
			txt = "This menu allows you to select a Radio Port. \n\r";
			txt += "Once the database has been loaded you can edit the Port Name or press [OK] to accept the current name. ";
			break;
		case MNU_CONFIG - 2:
			topic = "CONFIG MENU";
			txt = "This menu provides additional configuration and setting options. ";
			break;
		case MNU_RADIO - 2:
			if (gRadioCAT == RADIO_CAT_FLEX)
			{
				topic = "FLEX 6000 MENU";
				txt = "This menu provides additional controls and settings for Flex 6000 radios. ";
			}
			else if (gRadioCAT == RADIO_CAT_PCR1000)
			{
				topic = "PCR-1000 MENU";
				txt = "This menu provides additional controls and settings for the Icom PCR-1000 receiver. ";
			}
			break;
		}
		break;
	case MNU_FREQ:
		mnuFreqSel = gMnuSel; //so we return where we started
		switch (gMnuSel)
		{
		case MNU_FREQ_ZERO:
			topic = "ZERO LOW DIGITS";
			txt = "Select this option to zero the frequency digits below the Step Digit. ";
			break;
		case MNU_FREQ_RIT:
			topic = "RX FREQUENCY OFFSET";
			txt = "This option allows you to offset the Rx frequency while leaving the Tx frequency on the VFO setting. \n\r";
			txt += "The value can be (+) or (-). Set to 0 or tune VFO to turn RIT off. ";
			break;
		case MNU_FREQ_DB:
			topic = "DATABASE SELECTION";
			txt = "This option allows you to toggle between the Radio Port's database and the Shared Radio database. \n\r";
			txt += "The Shared database can be used on any Radio Port. ";
			break;
		case MNU_FREQ_LOAD:
			topic = "LOAD FROM MEMORY";
			txt = "This option LOADS the Mini's VFOs and their modes from memory. \n\r";
			txt += "Use the encoder to scroll and click to select a slot. \n\r";
			txt += "Select [Mem] on [B] to tune the radio using the memory list. ";
			break;
		case MNU_FREQ_SAVE:
			topic = "SAVE TO MEMORY";
			txt = "This option SAVES the current VFO settings to the selected memory slot. \n\r";
			txt += "Use the encoder to scroll and click to select a slot. \n\r";
			txt += "You can edit the memory's label after it's been saved. ";
			break;
		case MNU_FREQ_RENAME:
			topic = "RENAME MEMORY";
			txt = "This option allows you to RENAME the selected memory slot. \n\r";
			txt += "Use the encoder to scroll and click to select a slot.";
			break;
		case MNU_FREQ_PREV:
			topic = "PREVIOUS FREQ & MODE";
			txt = "This option loads the previous frequency and mode to the radio. \n\r";
			txt += "The Previous Frequency buffer holds up to 10 previous frequencies and their modes. ";
			break;
		}
		break;
	case MNU_BAND:
		topic = "BAND SELECTION";
		txt = "Select one of the bands from the list using the encoder to scroll then click to select. \n\r";
		txt += "The last frequency & mode on the selected band will be restored. ";
		break;
	case MNU_MODE:
		topic = "MODE SELECTION";
		txt = "Select a radio mode from the list using the encoder to scroll then click to select. ";
		break;
	case MNU_VFO:
		switch (gMnuSel)
		{
		case MNU_VFO_A:
			topic = "VFO-A";
			txt = "This option selects VFO-A as the operating VFO. \n\r";
			txt += "Button [B] can be used to step through the VFO tuning options. \n\r";
			txt += "Long-press [B] to lock the frequency and mode. ";
			break;
		case MNU_VFO_B:
			topic = "VFO-B";
			txt = "This option selects VFO-B as the operating VFO. \n\r";
			txt += "Button [B] can be used to step through the VFO tuning options. \n\r";
			txt += "Long-press [B] to lock the frequency and mode. ";
			break;
		case MNU_VFO_SPLIT:
			topic = "SPLIT MODE";
			txt = "In Split Mode VFO-A is used for receive and VFO-B is used for transmit. \n\r";
			txt += "Button [B] can be used to step through the VFO tuning options. \n\r";
			txt += "Long-press [B] to lock the frequency and mode. ";
			break;
		case MNU_VFO_MEM:
			topic = "MEMORY TUNE";
			txt = "In Memory Tune the encoder tunes the radio using the selected memory. \n\r";
			txt += "Button [B] can be used to step through the VFO tuning options. \n\r";
			txt += "Long-press [B] to copy memory to the VFOs. ";
			break;
		case MNU_VFO_MODE:
			topic = "TUNE MEMORY MODE";
			txt = "This option selects either A or B memory to be used in the active VFO when in Memory Tuning is enabled. \n\r";
			txt += "Long-press [B] to copy memory to the VFOs. ";
			break;
		case MNU_VFO_LOCK:
			topic = "TOGGLE LOCK";
			txt = "Click this option to toggle Frequency and Mode Lock. \n\r";
			txt += "Long-press [B] will also set the Lock, Short-press [B] will reset it. ";
			break;
		case MNU_VFO_COPYAB:
			topic = "COPY VFO-A > VFO-B";
			txt = "Click this option to Copy VFO-A's settings to VFO-B and return to the Home page in VFO-B mode. ";
			break;
		case MNU_VFO_COPYBA:
			topic = "COPY VFO-B > VFO-A";
			txt = "Click this option to Copy VFO-B's settings to VFO-A and return to the home page in VFO-A mode. \n\r";
			txt += "If VFO-B has not been set, VFO-A's settings will be copied to VFO-B. ";
			break;
		case MNU_VFO_SWAPAB:
			topic = "SWAP VFO-A <> VFO-B";
			txt = "Click this option to Swap VFO-A's and VFO-B's settings. \n\r";
			txt += "If VFO-B has not been set, VFO-A's settings will be copied to VFO-B. ";
			break;
		case MNU_VFO_COPY2A:
			topic = "COPY MEMORY TO VFOs";
			txt = "This option copies the selected memory frequencies and modes to the VFOs and returns to the Home page in VFO-A mode. \n\r";
			txt += "Long-press [B] while in Memory Tuning does the same thing. ";
			break;
		case MNU_VFO_PUSH:
			topic = "PUSH FREQ/MODE";
			txt = "This option will push the freq, mode, band, and RIT settings from this Mini to the other Mini when operating in Basic Link mode. \n\r";
			txt += "This menu is only available in Basic Link mode. ";
			break;
		case MNU_VFO_PULL:
			topic = "PULL FREQ/MODE";
			txt = "This option will pull the freq, mode, band, and RIT settings from the linked Mini to this Mini when operating in Basic Link mode. \n\r";
			txt += "This menu is only available in Basic Link mode. ";
			break;
		}
		break;
	case MNU_TXMSG:
	case MNU_TXMSG_EDIT:
		switch (gMnuSel)
		{
		case 0: //this is the repeat buffer
			topic = "REPEAT BUFFER";
			txt = "Buffer #1 always holds the last buffer transmitted. \n\r";
			txt += "Select this buffer to resend the prev msg. \n\r";
			txt += "When using a Terminal, press F1 to send the Repeat Buffer. ";
			break;
		case 1: //CQ buffer
			topic = "CQ BUFFER";
			txt = "Buffer #2 can be used as a CQ buffer. \n\r";
			txt += "Just enter   CQ CQ CQ de ^ k   to send CQ and your call. \n\r";
			txt += "When using a Terminal, press F2 to send this buffer. ";
			break;
		case 2: //contest exchange
			topic = "CONTEST EXCHANGE";
			txt = "Buffer #3 can be used to send a Contest Exchange. \n\r";
			txt += "'^' sends your call\n\r '#' sends the contest SN\n\r '%' sends the contest exch ";
			break;
		default:
			topic = "TRANSMIT MESSAGE";
			txt = "Scroll to a Tx message using the encoder then click to select. \n\r";
			txt += "- In Tx Msg, sending will commence immediately after a msg is selected. \n\r";
			txt += "- Use the EDIT TX MSG option to edit each msg. ";
			break;
		}
		break;
	case MNU_KEYER:
		mnuKeyerSel = gMnuSel;
		switch (gMnuSel)
		{
		case MNU_KEYER_SPEED:
			topic = "KEYER SPEED";
			txt = "Use the encoder change the Keyer Speed. \n\r";
			txt += "The Farnsworth Spacing automatically follows the Keyer Speed. You can adjust this after setting the Speed. ";
			break;
		case MNU_KEYER_FARNS:
			topic = "FARNSWORTH SPACING";
			txt = "Use the encoder change the Farnsworth Spacing. \n\r";
			txt += "The Farnsworth Spacing automatically follows the Keyer Speed when it's changed. Adjust this setting after setting the Speed. ";
			break;
		case MNU_KEYER_SIDETONE:
			topic = "SIDETONE ON/OFF";
			txt = "Toggle the keyer sidetone On and Off using this option. \n\r";
			txt += "To use the Mini for code practice, set TX ENABLE OFF and turn the Sidetone ON. ";
			break;
		case MNU_KEYER_TYPE:
			topic = "KEYER TYPE MENU";
			txt = "This menu allows you to select the type of Keyer you want to use. \n\r";
			break;
		case MNU_KEYER_PDL:
			topic = "KEYER PADDLES";
			txt = "This option allows you to change the paddle configuration. \n\r";
			txt += "- NORMAL: DIT is on the left and DAH is on the right. \n\r";
			txt += "- REVERSED: DAH is on the left and DIT is on the right. ";
			break;
		case MNU_KEYER_PTT:
			switch (gRmtMode)
			{
			case MNU_RMT_NORM:
				topic = "REMOTE PTT (NORMAL)";
				txt = "In this mode the Remote PTT Input controls the PTT Output on the Radio I/O. \n\r";
				txt += "Grounding the Remote PTT Input grounds the PTT Output. \n\r";
				txt += "PTT Output has a 3-min watchdog timer. ";
				break;
			case MNU_RMT_LATCH:
				topic = "REMOTE PTT (LATCH)";
				txt = "This mode latches PTT Output ON with the first Remote PTT input and turns it OFF ";
				txt += "with the next so there is no need to hold the PTT button continuously to transmit. \n\r";
				txt += "PTT Output has a 3-min watchdog timer. ";
				break;
			case MNU_RMT_KEY:
				topic = "REMOTE PTT (KEY)";
				txt = "In this mode a straight key on the Remote PTT Input will control the Key Output. \n\r";				
				txt += "NOTE: THE FUNCTION KEYPAD IS NOT AVAILABLE IN THIS MODE. ";
				break;
			}
			break;
		case MNU_KEYER_CALL:
			topic = "USER CALL SIGN";
			txt = "Enter your Call Sign here. \n\r";
			txt += "Your call will be sent automatically when you embed the '^' prosign in a Tx Message buffer. ";
			break;
		case MNU_KEYER_SN:
			topic = "CONTEST SN";
			txt = "Edit or reset the Contest Serial # here. \n\r";
			txt += "The Contest SN will be sent automatically when you embed the '#' prosign in a Tx Message buffer. ";
			break;
		case MNU_KEYER_EXCH:
			topic = "CONTEST EXCHANGE";
			txt = "Edit the Contest Exchange here. \n\r";
			txt += "The Contest Exchange will be sent automatically when you embed the '%' prosign in a Tx Message buffer. ";
			break;
		}
		break;
	case MNU_WIFI:
		mnuWiFiSel = gMnuSel;
		switch (gMnuSel)
		{
		case MNU_WIFI_CONNECT:
			topic = "WIFI CONNECT";
			txt = "Toggle this option to CONNECT or DISCONNECT from your WiFi network. \n\r";
			txt += "NOTE \nYou must set the Network SSID and Password before you can connect. ";
			break;
		case MNU_WIFI_SSID:
			topic = "WIFI SSID";
			txt = "Enter your WiFi networks' SSID here. \n\rThis is case sensitive. \n\r";
			txt += "The Adrs, Subnet, and Gateway fields will fill in once connected. ";
			break;
		case MNU_WIFI_STAKEY:
			topic = "WIFI PASSWORD";
			txt = "Enter your WiFi networks' PASSWORD here. \n\rThis is case sensitive. \n\r";
			txt += "The Adrs, Subnet, and Gateway fields will fill in once connected. ";
			break;
		case MNU_WIFI_AUTO:
			topic = "WIFI AUTO-CONNECT";
			txt = "Enabling this option allows the Mini to connect to your WiFi network once the initial boot up is complete. \n\r";
			txt += "Press and hold the encoder switch during boot to invert this option. ";
			break;
		default:
			topic = "STATION INFO";
			txt = "This field will be filled in when you connect to a WiFi network. ";
			break;
		}
		break;
	case MNU_RADIOPORT:
		switch (gMnuSel)
		{
		case MNU_PORT_COPY:
			topic = "RADIO PORT COPY";
			txt = "This option allows you to copy the current Radio Port settings to another Port. \n\r";
			txt += "Copied data includes Tx Msg buffers, Favorite Frequencies, and Band Stacking Registers. ";
			break;
		case MNU_PORT_RENAME:
			topic = "RENAME RADIO PORT";
			txt = "This option allows you to edit the name assigned to the current Radio Port. ";
			break;
		case MNU_PORT_ERASE:
			topic = "RADIO PORT ERASE";
			txt = "This option allows you reset the current Radio Port back to default settings. \n\r";
			txt += "Tx Msg buffers, Favorite Frequencies, and Band Stacking Registers are also reset. ";
			break;
		default:
			topic = "RADIO PORT";
			txt = "Select a Radio Port from the list using the encoder to scroll and click to select. \n\r";
			txt += "You can edit the Port Name after the Port database has been loaded. ";
			break;
		}
		break;
	case MNU_CONFIG:
		mnuConfigSel = gMnuSel;
		switch (gMnuSel)
		{
		case MNU_CONFIG_C_BTN:
			topic = "[C] BUTTON MENU";
			txt = "Use this menu to assign one of the options on the list to the [C] button. \n\r";
			txt += "Press [C] to execute the option. \n\r";
			txt += "Long-press [C] to open this menu to change this setting. ";
			break;
		case MNU_CONFIG_FUNC:
			topic = "FUNCTION BUTTONS";
			txt = "This menu allows you to define a function for each Function Button. \n\r";
			txt += "Function buttons are brought in on the Rmt PTT Input jack. ";
			break;
		case MNU_CONFIG_THEME:
			topic = "THEME MENU";
			txt = "This menu allows you to change the theme colors. \n\r";
			txt += "Use the encoder to scroll and click to select. ";
			break;
		case MNU_CONFIG_TAGLINE:
			topic = "TAG LINE EDIT";
			txt = "This menu allows you to change default Tag Line at the bottom of the Home page. \n\r";
			txt += "Suggested settings are your name and call but you can set it to anything you like. ";
			break;
		case MNU_CONFIG_CAT:
			topic = "RADIO CAT MENU";
			txt = "This menu allows you to change the CAT protocol used on the selected Radio Port. \n\r";
			txt += "Use the encoder to scroll and click to select. \n\r";
			txt += "Once you select the CAT protocol, check the Radio Baud settings. ";
			break;
		case MNU_CONFIG_BAUD:
			topic = "RADIO BAUD MENU";
			txt = "This menu allows you to change the baud rate for the current CAT protocol on the Radio Port. \n\r";
			txt += "Use the encoder to scroll and click to select. ";
			break;
		case MNU_CONFIG_AMPSEQ:
			topic = "AMP SEQUENCE MENU";
			txt = "This menu allows you to enable Amp Sequencing and change the PTT and Key pickup and release time. \n\r";
			txt += "Wire the Radio I/O PTT Out to your Amp's PTT and the KEY Out to your transmitter's PTT or Key. ";
			break;
		}
		break;
	case MNU_RADIO:
		//specialty radios		
		switch (gRadioCAT)
		{
		case RADIO_CAT_FLEX:
			mnuFlexSel = gMnuSel;
			switch (gMnuSel)
			{
			case MNU_FLEX_IP:
				topic = "FLEX IP ADDRESS";
				txt = "Enter the IP address for your Flex radio here. \n\r";
				txt += "You can find the radio's IP address in SmartSDR under \n\r";
				txt += "Settings->Radio Setup-> Network. \n\r";
				txt += "Use xxx.xxx.xxx.xxx format. ";
				break;
			case MNU_FLEX_PORT:
				topic = "FLEX IP PORT";
				txt = "Enter the IP Port for your Flex radio here. \n\r";
				txt += "The default value is 4992. \n\r";
				txt += "You can change this if you are using port forwarding for remote operation. ";
				break;
			case MNU_FLEX_PAN:
				topic = "PANADAPTER";
				txt = "Toggle this setting to create a new panadapter when you create a new slice. \n\r";
				txt += "This setting will automatically reset to use an Existing Panadapter after the connection.";
				break;
			case MNU_FLEX_SLICE:
				topic = "SLICE SELECTION";
				txt = "Select a receiver slice to connect to the Mini. \n\r";
				txt += "If the slice doesn't exist it will be created. \n\r";
				txt += "A new panadapter will also be created if you've chosen that option above. ";
				break;
			case MNU_FLEX_ANT:
				topic = "ANTENNA SELCTION";
				txt = "Select the antenna assigned when you create a new slice. \n\r";
				txt += "Both Tx and Rx will be assigned to this antenna. ";
				break;
			case MNU_FLEX_TOGGLE:
				topic = "ANT TUNER ACTIVATE";
				txt = "Click this option to START the Antenna Tuner. \n\r";
				txt += "The tuning status will be shown in the Status bar at the bottom of the display. ";
				break;
			case MNU_FLEX_VOL:
				topic = "VOLUME";
				txt = "This option adjusts the Rx Audio Level on the selected slice. ";
				break;
			case MNU_FLEX_PO:
				topic = "POWER OUTPUT";
				txt = "Use this control to set the desired transmit power output. ";
				break;
			case MNU_FLEX_TO:
				topic = "TUNE POWER OUTPUT";
				txt = "Use this control to set the desired tune power output. ";
				break;
			case MNU_FLEX_WNB:
				topic = "WIDE NOISE BLANKER";
				txt = "Adjust the Wide Noise Blanker level here. \n\r";
				txt += "  - To Enable WNB set this value above 0. \n\r";
				txt += "  - To Disable WNB set this value to 0. ";
				break;
			case MNU_FLEX_NB:
				topic = "NOISE BLANKER";
				txt = "Adjust the Noise Blanker level here. \n\r";
				txt += "  - To Enable NB set this value above 0. \n\r";
				txt += "  - To Disable NB set this value to 0. ";
				break;
			case MNU_FLEX_NR:
				topic = "NOISE REDUCTION";
				txt = "Adjust the Noise Reduction level here. \n\r";
				txt += "  - To Enable NR set this value above 0. \n\r";
				txt += "  - To Disable NR set this value to 0. ";
				break;
			case MNU_FLEX_AF:
				if (gRMode < RADIO_LSB)
				{
					//CW only
					topic = "AUTO-PEAKING FILTER";
					txt = "Adjust the Auto-Peaking Filter for CW reception here. \n\r";
					txt += "  - To Enable APF set this value above 0. \n\r";
					txt += "  - To Disable APF set this value to 0. ";
				}
				else
				{
					topic = "AUTO-NOTCH FILTER";
					txt = "Adjust the Auto-Notch Filter level here. \n\r";
					txt += "  - To Enable ANF set this value above 0. \n\r";
					txt += "  - To Disable ANF set this value to 0. ";
				}
				break;
			case MNU_FLEX_PTT:
				topic = "FLEX PTT OUTPUT CONTROL";
				txt = "PTT (Hw) uses the Radio I/O PTT Output to control PTT. \n\r";
				txt += "PTT (Net) uses the Flex API to control PTT via the WiFi network. \n\r";
				txt += "Use PTT (Net) for remote operation. ";
				break;
			case MNU_FLEX_KEYER:
				topic = "FLEX KEY OUTPUT CONTROL";
				txt = "Key (Hw) uses the Radio I/O Key Output to key the radio. \n\r";
				txt += "Key (Net) uses the Flex API to key the radio via the WiFi network. ";
				txt += "Adjust the CW Delay for network latency. ";
				break;
			case MNU_FLEX_DELAY:
				topic = "FLEX BREAK-IN DELAY";
				txt = "Adjust the CW break-in delay here. \n\r";
				txt += "When using Network CW Keying adjust this value to compensate for network latency. \n\r";
				txt += "Try delays over 100 mS. ";
				break;
			}
			break;

		case RADIO_CAT_PCR1000:
			mnuPCRSel = gMnuSel;
			switch (gMnuSel)
			{
			case MNU_PCR_VOL:
				topic = "VOLUMNE";
				txt = "Use this option to adjust the PCR-1000's Volume. \n\r";
				txt += "Range is 0 to 100. ";
				break;
			case MNU_PCR_SQ:
				topic = "SQUELCH";
				txt = "Use this option to adjust the PCR-1000's Squelch. \n\r";
				txt += "Range is 0 to 100. ";
				break;
			case MNU_PCR_IFSHIFT:
				topic = "IF SHIFT";
				txt = "Adjust the IF Shift here. This allows you to shift the receiver's passband away from an interfering signal. \n\r";
				txt += "Range 0 to 100 with 50 being the center of the IF (default). ";
				break;
			case MNU_PCR_ZEROIF:
				topic = "ZERO IF PASSBAND";
				txt = "Click this option to center the IF in the middle of its passband. \n\r";
				txt += "You can also set the IF Shift to 50 to center the passband. ";
				break;
			case MNU_PCR_DSP:
				topic = "DSP ENABLE/DISABLE";
				txt = "Click this option to ENABLE or DISABLE the Digital Signal Processor board in the radio. \n\r";
				txt += "If the radio doesn't have the DSP option, N/A will be displayed here. ";
				break;
			case MNU_PCR_DSPNR:
				topic = "NOISE REDUCTION";
				txt = "Adjust the amount of DSP Noise Reduction applied to the received signal. \n\r";
				txt += "Your radio must have the DSP option and it must be ENABLED to use this feature. \n\r";
				txt += "If DSP is not installed, N/A will be displayed here. ";
				break;
			case MNU_PCR_ANch:
				topic = "AUTO-NOTCH FILTER";
				txt = "ENABLE or DISABLE the DSP Auto-Notch filter. \n\r";
				txt += "The radio must have the DSP option and it must be ENABLED to use this feature. \n\r";
				txt += "If DSP is not installed, N/A will be displayed here. ";
				break;
			case MNU_PCR_BW:
				topic = "BANDWIDTH";
				txt = "Click this option to step through the available receive Bandwidth settings. \n\r";
				txt += "The radio's mode determines the bandwidth options available. \n\r";
				txt += "50 KHz and 230 KHz are only available in FM-Wide mode. ";
				break;
			case MNU_PCR_AGC:
				topic = "AUTO GAIN CONTROL";
				txt = "Use this option to ENABLE or DISABLE the AGC on the receiver. \n\r";
				txt += "AGC will normalize the received signals so they all sound about the same. ";
				break;
			case MNU_PCR_ATTEN:
				topic = "ATTENUATOR";
				txt = "Use this option to ENABLE or DISABLE the Attenuator on the receiver. \n\r";
				txt += "Use the attenuator to reduce the level of strong signals that swamp the frontend of the receiver. ";
				break;
			case MNU_PCR_NB:
				topic = "NOISE BLANKER";
				txt = "Use this option to ENABLE or DISABLE the Noise Blanker on the receiver. \n\r";
				txt += "The noise blanker will help reduce pulsing noise like ignition noise. ";
				break;
			}
		}
		break;
	case MNU_FREQ_LD:
		mnuFreqListSel = gMnuSel;
		topic = "LOAD MEMORY";
		txt = "Use the encoder to scroll through the memory channels then click to select the desired channel. \n\r";
		txt += "The A and B VFOs and their Modes will be loaded from memory and the radio will switch to VFO A. ";
		break;
	case MNU_FREQ_SV:
		mnuFreqListSel = gMnuSel;
		topic = "SAVE MEMORY";
		txt = "Use the encoder to scroll through the memory channels then click to select the channel to SAVE the current VFOs and Modes to. \n\r";
		txt += "You will be able to name the memory channel once it's saved. ";
		break;
	case MNU_FREQ_REN:
		mnuFreqListSel = gMnuSel;
		topic = "RENAME MEMORY";
		txt = "Use the encoder to scroll through the memory channels then click to select the channel you want to RENAME. ";
		break;
	case MNU_C_BTN:
		mnuCBtnSel = gMnuSel;
		switch (gMnuSel)	
		{
		case MNU_C_BTN_PTT:
			topic = "[C] BUTTON PTT MODE";
			txt = "This option enables latching PTT with the [C] button. \n\r";
			txt += "This option is not available if Rmt PTT Input is mapped to the Key Output or if Amplifier Sequencing is Enabled. ";		
			break;
		case MNU_C_BTN_RIT:
			topic = "[C] BUTTON RIT";
			txt = "This options assigns RIT (Rx offset) to the [C] button to allow quick access to this feature. \n\r";		
			txt += "This window remains open until you close it by clicking the encoder. ";
			break;
		case MNU_C_BTN_SPEED:
			topic = "[C] BUTTON SPEED";
			txt = "This option assigns Keyer Speed to the the [C] button to allow quick access to this feature. \n\r";			
			txt += "This window remains open until you close it by clicking the encoder. ";
			break;
		case MNU_C_BTN_ASC:
			topic = "[C] BUTTON ANT SW";
			txt = "This option assigns the Antenna Switch Controller to the [C] button. \n\r";
			txt += "NOTE: This option is only available when the Ant Switch is online. ";	
			break;
		case MNU_C_BTN_PUSH:
			topic = "[C] BUTTON PUSH INFO";
			txt = "This option pushes freq, mode, band, and RIT information to a linked Mini. \n\r";
			txt += "Two Minis must be linked together using the Basic Link mode to use this option. ";
			break;
		case MNU_C_BTN_PULL:
			topic = "[C] BUTTON PULL INFO";
			txt = "This option pulls freq, mode, band, and RIT information from a linked Mini. \n\r";
			txt += "Two Minis must be linked together using the Basic Link mode to use this option. ";
			break;
		default:
			topic = "[C] BUTTON MENU";
			txt = "Use the encoder to scroll through the available options that you can assign to the [C] (right) button. \n\r";
			txt += "Once assigned you can quickly open the selected option by clicking the [C] button. \n\r";
			txt += "Long-press the [C] button to open this menu to select another option. ";
			break;
		}
		break;
	case MNU_FUNC_BTN:		
		mnuFuncBtnSel = gMnuSel;		
		if (gMnuSel == mnuFuncBtnMax - 1)
		{
			topic = "RESET KEYPAD LOCKOUT";
			txt = "This option allows you to reset a Function Keypad Lockout. \n\r";
			txt += "The Function Keypad is locked out if you hold a function button down longer than 5 seconds, ";
			txt += "or a hardware failure occurs.";
		}
		else
		{
			topic = "FUNCTION BUTTONS";
			txt = "This option allows you to define the function of each button on the Function Button pad. \n\r";
			txt += "You can also define TxMsg to one button and buttons [A], [B], and [C] to other buttons for navigation. ";
		}
		break;
	case MNU_LINK:
		mnuLinkSel = gMnuSel;
		switch (gMnuSel)
		{
		case MNU_LINK_CONNECT:
			topic = "CONNECT/DISC LINK";
			txt = "Use this option to connect or disconnect the Link connecting two Mini's together. \n\r";
			txt += "When using WiFi UDP disconnecting the Link does not disconnect WiFi. ";
			break;
		case MNU_LINK_TYPE:
			topic = "CONNECTION TYPE";
			txt = "This option toggles between Bluetooth LE and WiFi UDP for the link interface. \n\r";
			txt += "Use WiFi UDP if WiFi is available, otherwise use BLE if WiFi is not available. ";
			break;
		case MNU_LINK_UDP_ADRS:
			topic = "LINK WIFI ADDRESS";
			txt = "Enter the WiFi address of the other Mini when using WiFi UDP to pass link data. ";
			break;
		case MNU_LINK_SVR_CLIENT:
			topic = "LINK MODE";
			txt = "Select which radio (A or B) this Mini controls through the wireless Link interface. \n\r";
			txt += "Radio A is the server and Radio B is the client. ";
			break;
		case MNU_LINK_MODE:
			switch (gSyncMode)
			{
			case SYNC_MODE_BASIC:
				topic = "SYNC: BASIC";
				txt = "When Link is connected this mode will sync Keyer Speed & Type, Farnsworth Speed, and Keyer SN ";
				txt += "between the two Minis. Frequency and RIT ARE NOT synced.\n\r Tx can only be enabled on one Mini. ";
				break;
			case SYNC_MODE_FREQ:
				topic = "SYNC: BASIC+FREQ";
				txt = "This option synchronizes the basic settings and syncs the same Frequency on both Minis. \n\r";
				txt += "Freq offsets by 10 MHz when the other Mini has Tx Enabled. ";
				break;
			case SYNC_MODE_TRACK:
				topic = "SYNC: BASIC+TRACKING";
				txt = "This option synchronizes the basic settings in addition to locking frequencies by the ";
				txt += "freq offset present when this mode is first enabled.\n";
				txt += "Tracking is disabled if Tx is Enabled on either Mini. ";
				break;
			case SYNC_MODE_RMT:
				if (gLinkMiniA)
				{
					topic = "SYNC: AS REMOTE";
					txt = "In this mode, Mini-A is used to remotly control Mini-B. \n\r";
					txt += "Mini-B will be set to [Sync: As Local] and must be connected to the radio. ";
				}
				else
				{
					topic = "SYNC: AS LOCAL";
					txt = "In this mode, Mini-B is connected to the radio and can be remotly controlled by Mini-A. \n\r";
					txt += "Setting Mini-A to [Sync: As Remote] enables this mode on Mini-B. ";
				}
				break;
			}
			break;
		case MNU_LINK_AUTO:
			topic = "BLE AUTO-CONNECT";
			txt = "Enabling this option starts a BLE connection to another Mini once the initial boot up is complete. \n\r";
			txt += "If WiFi UDP is used, select Auto-Connect in the WiFi menu. ";
			break;
		case MNU_LINK_OPTION6:
			if (!gLinkMiniA)
			{
				//Mini-B option
				topic = "RADIO-B ZERO-BEAT";
				txt = "This option allows you to adjust the freq of the radio on Mini-B to zero-beat with the radio on Mini-A. \n\r";
				txt += "This allows you to tune out minor frequency differences. ";
			}
			else
			{
				//Mini-A option
				topic = "ANT SWITCH IP ADRS";
				txt = "Enter the WiFi address of the CTR2-Antenna Switch Controller. \n\r";
				txt += "This allows this Mini to control the ASC when WiFi Link mode is enabled. ";				
			}
			break;
		}
		break;
	case MNU_ANT_SW:
		mnuASCSel = gMnuSel;
		switch (gMnuSel)
		{
		case MNU_ANT_SW_PORT_A:
		case MNU_ANT_SW_PORT_B:
			topic = "SELECT ANTENNA";
			txt = "This menu allows you to select an antenna for the selected port from the available antennas on your CTR2-Mini Antenna Switch Controller. ";
			break;
		case MNU_ANT_SW_SWAP:
			topic = "SWAP A & B ANTENNAS";
			txt = "Executing this option swaps the antennas on Radio Ports A and B on the antenna switch. \n\r";
			txt += "This option is only available on dual-port antenna switches. ";
			break;
		case MNU_ANT_SW_LOCK:
			topic = "OPERATE MODE";
			txt = "Executing this option toggles the Ready/Lock operate mode on the antenna switch controller. \n\r";
			txt += "Other options are not available when the switch is locked. ";
			break;
		case MNU_ANT_SW_ENABLE:
			topic = "SWITCH MODE";
			txt = "Executing this option toggles the mode of the antenna switch. \n\rWhen the switch is OFF all antenna ports are OPEN. \n\r";
			break;
		case MNU_ANT_SW_LABELS:
			topic = "EDIT LABELS";
			txt = "This menu allows you to edit the label used on each antenna port. ";
			break;
		case MNU_ANT_SW_SEND_BAND:
			topic = "SEND BAND DATA";
			txt = "Enable this option to send band data to the Antenna Switch. ";
			txt += "Assign bands in the Ant Sw's Config>Band Assignments menu. \n\r";
			txt += "NOT AVAILABLE in Basic+Freq sync mode. ";
			break;
		}
		break;
	case MNU_FUNC_BTN_OPTIONS:
		mnuFuncBtnOptionSel = gMnuSel;
		switch (gMnuSel)
		{
		case MNU_FUNC_OPTION_RIT:
			topic = "[RIT] FUNCTION BTN";
			txt = "This option assigns RIT (Rx offset) to a function btn to allow quick access to this feature. \n\r";
			txt += "Assign btns [A], [B], and [C] to other function btns for navigation. ";
			break;
		case MNU_FUNC_OPTION_SPEED:
			topic = "[SPEED] FUNCTION BTN";
			txt = "This option assigns Keyer Speed to the one of the function buttons to allow quick access to this feature. \n\r";
			txt += "Assign buttons [A], [B], and [C] to other function btns for navigation. ";
			break;
		case MNU_FUNC_OPTION_TXMSG:
			topic = "[TX MSG] FUNCTION BTN";
			txt = "This option assigns the Tx Msg menu to a function button. This allows quick access to this menu. \n\r";
			txt += "Assign buttons [A], [B], and [C] to other function btns for navigation. ";
			break;
		case MNU_FUNC_OPTION_BTNA:
			topic = "[BTN A] FUNCTION BTN";
			txt = "This option assigns button [A] (left btn) to a function button. \n\r";
			txt += "Btns [B] and [C] can also be assigned to function btns to allow full navigation from the btn pad. ";
			break;
		case MNU_FUNC_OPTION_BTNB:
			topic = "[BTN B] FUNCTION BTN";
			txt = "This option assigns button [B] (center btn) to a function button. \n\r";
			txt += "Btns [A] and [C] can also be assigned to function btns to allow full navigation from the btn pad. ";
			break;
		case MNU_FUNC_OPTION_BTNC:
			topic = "[BTN C] FUNCTION BTN";
			txt = "This option assigns button [C] (right btn) to a function button. \n\r";
			txt += "Btns [A] and [B] can also be assigned to function btns to allow full navigation from the btn pad. ";
			break;
		default:
			topic = "FUNCTION BTN TX MSGS";
			txt = "This option assigns this Tx Msg buffer to the selected function button. \n\r";
			txt += "You can also assign TxMsg to one btn and [A], [B], and [C] to other btns for navigation. ";
			break;
		}
		break;
	case MNU_THEME:
		mnuThemeSel = gMnuSel;
		topic = "SELECT THEME";
		txt = "Select a color theme for your Mini. ";
		break;
	case MNU_RADIOCAT:		
		switch (gMnuSel)
		{
		case RADIO_CAT_RIGCTLD:
			topic = "HAMLIB RIGCTLD";
			txt = "This option allows you to control your radio over an IP network. Your radio must have Hamlib rigctld built-in. \n\r";
			txt += "Enter the radio's IP address below when this option is selected. ";
			break;
		case RADIO_CAT_IPADRS:
			topic = "RIGCTLD IP ADRS";
			txt = "Enter the IP address of your Hamlib rigctld enabled rig here. \n\r";
			txt += "The default IP port for rigctld is 4532. ";
			break;
		default:
			topic = "SELECT RADIO CAT";
			txt = "Select the CAT protocol to use with the radio connected to the Radio I/O module. \n\r";
			txt += "NOTE:\nKenwood1 uses 1 stop bit. \nKenwood2 uses 2 stop bits. ";
			break;
		}
		break;
	case MNU_RADIOBAUD:
		topic = "SELECT RADIO BAUD";
		txt = "Select the Baud Rate that matches your radio's CAT port's Baud setting. ";		
		break;		
	case MNU_AMPSEQ:
		mnuAmpSeqSel = gMnuSel;
		switch (gMnuSel)
		{
		case 0: //enable/disable
			topic = "AMP SEQUENCE ENABLE";
			txt = "Click to ENABLE or DISABLE Amplifier Sequencing. \n\r";
			txt += "Enable this option if you are using an older linear amplifier that requires time to switch to Tx mode before power is applied. ";
			break;
		case 1: //pickup delay
			topic = "PICKUP DELAY";
			txt = "Set the time to wait after PTT is asserted to the Amp before keying the radio in milliseconds. \n\r";
			txt += "Wire Radio I/O PTT Out to Amp PTT In and KEY Out to radio's PTT or Key In. ";
			break;
		case 2: //release delay
			topic = "RELEASE DELAY";
			txt = "Set the time to wait after Radio I/O KEY Out is released before releasing PTT Out in milliseconds. \n\r";
			txt += "Wire Radio I/O PTT Out to Amp PTT In and KEY Out to radio's PTT or Key In. ";
			break;
		}
		break;	
	case MNU_CW_KEYERS:		
		mnuCWKeyerSel = gMnuSel;
		switch (gMnuSel)
		{
		case KEYER_NONE:
			topic = "NO KEYER MODE";
			txt = "Selecting [None] disables the keyer and the Radio I/O's Key Output. \n\r";
			txt += "No sidetone will be generated in this mode. ";
			break;
		case KEYER_STRAIGHT:
			topic = "STRAIGHT KEY MODE";
			txt = "Plug your straight key into the Paddles jack to key your radio manually. \n\r";
			txt += "Wire your key to the Tip and Sleeve of the 3.5mm phone plug. ";
			break;
		case KEYER_PASSTHRU:
			topic = "PASSTHRU MODE";
			txt = "Inputs from the paddles are sent directly to the Key and PTT outputs. \n\r";
			txt += "Use this mode when you want to use your radio's built-in keyer or an external keyer. \n\r";
			txt += "No sidetone will be generated. ";
			break;
		case KEYER_IAMBIC_A:
			topic = "IAMBIC-A MODE";
			txt = "In Iambic-A mode squeezing both paddles at the same time results in an alternating DIT-DAH-DIT-DAH pattern. \n\r";
			txt += "Extra elements are not inserted when you release the paddles. ";
			break;
		case KEYER_IAMBIC_B:
			topic = "IAMBIC-B MODE";
			txt = "In Iambic-B mode squeezing both paddles at the same time results in an alternating DIT-DAH-DIT-DAH pattern. \n\r";
			txt += "An additional element may be inserted when you release the paddles. ";
			break;
		case KEYER_ULTIMATIC:
			topic = "ULTIMATIC MODE";
			txt = "In Ultimatic mode when squeezing both paddles the keyer will continually send the element from the last paddle pressed. ";			
			break;
		case KEYER_BUG:
			topic = "BUG MODE";
			txt = "In Bug mode the keyer only generates a string of DITS when the Dit paddle is pressed. The DAH paddle acts like a straight key. ";
			break;
		default:
			topic = "KEYER TYPES";
			txt = "Select the Keyer type you want to use with the Mini. \n\r";
			txt = "If you want to use your radio's keyer, select PASSTHRU. \n\r";
			txt += "This will send the DIT paddle to the Radio I / O's KEY Output and the DAH paddle to the PTT Output. \n\r";
			txt += "Wire these outputs to your radio's Key jack. ";
			break;
		}
		break;
	}
	if (txt.length())
	{	
			PrtHelp(topic, txt);
	}
	else
	{
		PrtStatus("No Help Topic for this menu", true);
	}
}
