// =====================================================================================================================================================================================================
// (c) 2022 Lynn Hansen, KU7Q															                                                                                                               |
// This Source Code Form is subject to the terms of the GNU GENERAL PUBLIC LICENSE, Version 3, 29 June 2007. A copy of this license can be found here: https://choosealicense.com/licenses/gpl-3.0/|
// =====================================================================================================================================================================================================



//This module has I/O routines

void SerialOut(String txt, bool lfcr)
{
	//if enabSerialOut is defined, send txt to Serial.print/println - use to turn off serial output to free port for external programs to use
	if (gComMode == COM_MODE_CAT)
	{
		//Serial is being used as a virtual CAT comm port by another app - these messages will interfere with that traffic
		return;
	}
#ifdef enabSerialOut
	if (!lfcr)
	{
		Serial.print(txt);
	}
	else
	{
		Serial.println(txt);
	}
#endif
}

uint8_t ChkBtns()
{
	//check buttons for activity
	if (!digitalRead(WIO_KEY_C))
	{
		return BTN_A; //A and C are backwards to me, A should be on the left
	}
	else if (!digitalRead(WIO_KEY_B))
	{
		return BTN_B;
	}
	else if (!digitalRead(WIO_KEY_A))
	{
		return BTN_C;
	}
	else if (!digitalRead(WIO_5S_UP))
	{
		return BTN_UP;
	}
	else if (!digitalRead(WIO_5S_DOWN))
	{
		return BTN_DOWN;
	}
	else if (!digitalRead(WIO_5S_LEFT))
	{
		return BTN_LEFT;
	}
	else if (!digitalRead(WIO_5S_RIGHT))
	{
		return BTN_RIGHT;
	}
	else if (!digitalRead(WIO_5S_PRESS))
	{
		return BTN_SW;
	}
	return BTN_NONE;
}

void ProcessA_Btn()
{
	//process A button press if keyer is not running

	Flex_TxCW(true); //in case it's on
	//if linked, send new gTxMode to other Mini			
	if ((gBLEConnected || gUDPConnected) && gTxMode != gTxModePrev)
	{
		SendLinkParams();
	}
	if (!mnuCBtnSel)
	{
		//[C] btn menu set to PTT, 
		gBtnPTT = false;
		ControlPTT(LOW);
		PrtBtnC();
	}
	switch (gRadioCAT)
	{
	case RADIO_CAT_FLEX:
		//set radio key mode if using keyer
		if (gTCPCon && gRMode < RADIO_LSB && gKeyType != KEYER_NONE && gKeyType != KEYER_PASSTHRU)
		{
			if (!gTxMode)
			{
				//turn off key
				Tx2WiFi(WIFI_FLEX, "|cw break_in off");
			}
			else
			{
				Tx2WiFi(WIFI_FLEX, "|cw break_in on");
				Tx2WiFi(WIFI_FLEX, "|cw break_in_delay " + String(gFlexDelay));
			}
		}
		break;
	case RADIO_CAT_PCR1000:
		//toggle power on PCR		
		if (gTxMode)
		{
			switch (PCR_Ping())
			{
			case 0:
				PrtStatus("PCR-1000 not powered", true);
				gTxMode = RADIO_TX_OFF;
				break;
			case 1:
				if (!gPCRInit)
				{
					PCR_Initialize();
				}
				break;
			}
		}
		PCR_Update(0xff); //set power on/off controls
	}
}

void ProcessBtn(uint8_t btn)
{
	//button just pressed, process button
	uint16_t fgColor = hmiFGnd;
	uint16_t bgColor = hmiBase;
	String txt = "";
	int val = 0;

	hmiEncoderTimeout = millis();
		
	switch (hmiMode)
	{
	case HMI_HOME:
		ProcessHomeBtn(btn);
		break;
	case HMI_MENU:
		ProcessMenuBtn(btn);
		break;
	case HMI_HELP:
		ProcessHelpBtn(btn);
		break;
	case HMI_VALUE:
		ProcessValueBtn(btn);
		break;
	case HMI_TXTINPUT:
		ProcessTxInputBtn(btn);
		break;
	}

}

void ProcessHomeBtn(uint8_t btn)
{
	//process button presses while in MNU_HOME mode
	uint32_t val = 0;
	String txt = "";
	switch (btn)
	{
	case BTN_SW:
		//5-way btn pressed, open Tx Msg menu
		hmiMode = HMI_MENU;
		gMnu = MNU_MAIN;
		gMnuPrev = 0; //force refresh
		gMnuSel = mnuMainSel; //the last main menu selection			
		hmiMnuChng = true; //force CreateMnu() in Loop
		break;
	case BTN_A:
		//toggle Tx Mode
		switch (gPlay)
		{
		case KEYER_MODE_STOP:
			//normal mode						
			gTxMode++;
			if (gTxMode == RADIO_TX_ENAB && gRmtMode != MNU_RMT_KEY)
			{
				//we just enabled Tx, switch RMT_BTNS to output and send a 10 mS pulse to switch SO2R relays, then return to normal
				pinMode(RMT_BTNS, OUTPUT);
				digitalWrite(RMT_BTNS, HIGH);
				delay(10);
				digitalWrite(RMT_BTNS, LOW); //this will toggle SEL relays on SO2R
				delay(10);
				digitalWrite(RMT_BTNS, HIGH);
				delay(100);
				pinMode(RMT_BTNS, INPUT_PULLUP); //return to normal
			}
			if (gTxMode > RADIO_TX_ENAB)
			{
				gTxMode = RADIO_TX_OFF; //loop back to off
			}			
			ProcessA_Btn();
			SendASCCmd(ASC_CMD_TXMODE, 0, 0);
			break;
		case KEYER_MODE_RUN:
StopKeyer:	//turn off tx buffer
			gPlay = KEYER_MODE_STOP;
			gTxBuffer = "";
			delay(500); //let last chr clear the keyer
			ResetSpeedBoost();
			return;
			break;
		case KEYER_MODE_PAUSE:
			goto StopKeyer;
			break;
		}
		PrtBtnA(true);
		break;
	case BTN_B:
		if (gLock)
		{
			//unlock and return to VFO mode
			gLock = false;
			PrtBtnB();			
		}
		else if (!gPlay)
		{
			//keyer not active and we're not locked, normal functions
			if (!gLock && !gPlay && gTxMode < RADIO_TX_KEY)
			{
				gVFO++;
				if (gVFO > MNU_VFO_MEM)
				{
					gVFO = MNU_VFO_A;
				}

				switch (gVFO)
				{
				case MNU_VFO_A:
				case MNU_VFO_SPLIT:
					gFreqRadio = gVFO_Freq[MNU_VFO_A];
					gRMode = gVFO_Mode[MNU_VFO_A];
					gPCRBW = gVFO_BW[MNU_VFO_A];
					break;
				case MNU_VFO_B:
					if (gVFO_Freq[MNU_VFO_B] == 0)
					{
						gVFO_Freq[MNU_VFO_B] = gFreqRadio;
						gVFO_Mode[MNU_VFO_B] = gRMode;
						gVFO_BW[MNU_VFO_B] = gPCRBW;
					}
					gFreqRadio = gVFO_Freq[MNU_VFO_B]; //load b
					gRMode = gVFO_Mode[MNU_VFO_B];
					gPCRBW = gVFO_BW[MNU_VFO_B];
					break;				
				case MNU_VFO_MEM:
					//turning encoder will tune to next programmed memory slot
					int offset = MNU_VFO_A;
					if (gMSel)
					{						
						if (!gMSelMode || vfoMemB_Freq[gMSel - 1] < 100000)
						{
							offset = MNU_VFO_A + 2;
							gVFO_Freq[offset] = vfoMemA_Freq[gMSel - 1];
							gVFO_Mode[offset] = vfoMemA_Mode[gMSel - 1];
							gVFO_BW[offset] = vfoMemA_BW[gMSel - 1];							
							
						}
						else
						{
							offset = MNU_VFO_B + 2;
							gVFO_Freq[offset] = vfoMemB_Freq[gMSel - 1];
							gVFO_Mode[offset] = vfoMemB_Mode[gMSel - 1];
							gVFO_BW[offset] = vfoMemB_BW[gMSel - 1];
						}							
					}
					gFreqRadio = gVFO_Freq[offset];
					gRMode = gVFO_Mode[offset];
					gPCRBW = gVFO_BW[offset];
					break;
				}
				ChkFreqModeChng();
				PrtFreq(true); //reset digits to new vfo color
				PrtRMode(true);
				PrtOffVFO();
				PrtRIT();
				PrtOptions();
				PrtBtnB();
				return;
			}
		}
		else
		{
			if (gPlay == KEYER_MODE_RUN)
			{
				gPlay = KEYER_MODE_PAUSE; //pause keyer
			}
			else
			{
				gPlay = KEYER_MODE_RUN; //restart keyer
			}
		}
		break;
	case BTN_C:
		//execute selected [C] button option				
		hmiMode = HMI_MENU;
		switch (mnuCBtnSel)
		{
		case MNU_C_BTN_PTT:
			//toggle PTT if tx is enabled & stay on home page
			hmiMode = HMI_HOME;
			if (gTxMode)
			{				
				if (gRmtMode == MNU_RMT_KEY)
				{
					PrtStatus("PTT SET TO 'KEY OUT'", true);
					gRmtFuncBtn = RMT_FUNC_OFF;
				}
				else
				{
					gBtnPTT = !gBtnPTT;	//Loop() will call ChkPTT() to control ptt						
				}
			}
			else
			{
				PrtStatus("TX IS NOT ENABLED", true);
			}
			return;
			break;
		case MNU_C_BTN_RIT:
			//just open the RIT window
			gMnu = MNU_FREQ;
			gMnuSel = MNU_FREQ_RIT;
			mnuFreqSel = gMnuSel;
			CreateMnu("", false);			
			SetMnuParam();
			return;
			break;
		case MNU_C_BTN_SPEED:
			//just open the keyer speed window
			gMnu = MNU_KEYER;
			gMnuSel = MNU_KEYER_SPEED;
			mnuKeyerSel = gMnuSel;
			CreateMnu("", false);
			SetMnuParam();
			return;
			break;
		case MNU_C_BTN_FREQ:
			gMnu = MNU_FREQ;
			gMnuSel = mnuFreqSel;
			break;
		case MNU_C_BTN_MODE:
			gMnu = MNU_MODE;
			gMnuSel = gRMode;
			break;
		case MNU_C_BTN_BAND:
			gMnu = MNU_BAND;
			gMnuSel = gBand;
			break;
		case MNU_C_BTN_VFO:
			gMnu = MNU_VFO;
			gMnuSel = gVFO;
			break;
		case MNU_C_BTN_TXMSG:
			gMnu = MNU_TXMSG;
			gMnuSel = mnuTxMsgSel;
			break;
		case MNU_C_BTN_EDIT_TXMSG:
			gMnu = MNU_TXMSG_EDIT;
			gMnuSel = mnuLastTxMsg;
			break;
		case MNU_C_BTN_KEYER:
			gMnu = MNU_KEYER;
			gMnuSel = mnuKeyerSel;
			break;
		case MNU_C_BTN_WIFI:
			gMnu = MNU_WIFI;
			gMnuSel = mnuWiFiSel;
			break;
		case MNU_C_BTN_LINK:
			gMnu = MNU_LINK;
			gMnuSel = mnuLinkSel;
			break;
		case MNU_C_BTN_ASC:
			if (gAntSwOnline)
			{
				gMnu = MNU_ANT_SW;
				gMnuSel = mnuASCSel;
			}
			else
			{
				PrtStatus("ANT SWITCH OFFLINE", true);
				return;
			}
			break;
		case MNU_C_BTN_FUNC_BTN:
			gMnu = MNU_FUNC_BTN;
			gMnuSel = mnuFuncBtnSel;
			break;
		case MNU_C_BTN_RADIO:
			switch (gRadioCAT)
			{
			case RADIO_CAT_FLEX:
				gMnu = MNU_RADIO;
				gMnuSel = mnuFlexSel;
				break;
			case RADIO_CAT_PCR1000:
				gMnu = MNU_RADIO;
				gMnuSel = mnuPCRSel;
				break;
			default:
				//bad option, default to TxMsg
				gMnu = MNU_TXMSG;
				gMnuSel = mnuTxMsgSel;
				break;
			}
			break;
		case MNU_C_BTN_PUSH:
			gMnu = MNU_VFO;
			gMnuSel = MNU_VFO_PUSH;
			ExecMnuSel();
			PrtHomePg();
			return;
			break;
		case MNU_C_BTN_PULL:
			gMnu = MNU_VFO;
			gMnuSel = MNU_VFO_PULL;
			ExecMnuSel();
			PrtHomePg();
			return;
			break;
		}
		gMnuPrev = -1;
		CreateMnu("", false);
		PrtBtnC();
		return;
		break;
	case BTN_UP:
		//move freq if not locked or sending buffer
		if (!gLock && !gPlay)
		{			
			gFreqRadio += gUnit;
			if (gFreqRadio > 999999999)
			{
				gFreqRadio = 100000;
			}
		}
		break;
	case BTN_DOWN:
		//move freq down if not locked or sending buffer
		if (!gLock && !gPlay)
		{
			gFreqRadio -= gUnit;
			if (gFreqRadio > 999999999)
			{
				gFreqRadio = 100000;
			}
			else if (gFreqRadio < 100000)
			{
				gFreqRadio = 999999999;
			}
			while (String(gUnit).length() > String(gFreqRadio).length())
			{
				gUnit /= 10; //find the top of the freq
			}
		}
		break;
	case BTN_LEFT:
		//move gUnit left
		if (!gLock)
		{
			gUnit *= 10;
			if (String(gUnit).length() > String(gFreqRadio).length())
			{
				//roll over
				gUnit = 1;
			}			
		}		
		break;
	case BTN_RIGHT:
		//move gUnit right			
		if (!gLock)
		{
			if (gUnit > 1)
			{
				gUnit /= 10;
			}
			else
			{
				//move to top of freq
				gUnit = 10;
				while (String(gUnit).length() < String(gFreqRadio).length())
				{
					gUnit *= 10; //find the top of the freq
				}
			}			
		}		
		break;
	}
	PrtFreq(true);
}

void ProcessMenuBtn(uint8_t btn)
{
	//process button presses while in HMI_MENU mode

	switch (btn)
	{
	case BTN_SW:
		//same as encoder press
		if (gBtnSwLongPress)
		{
			//already did PrtHomePg();
		}
		else
		{
			gEncoderClicked = true;
		}		
		break;
	case BTN_A:
		if (gBtnALongPress == false)
		{
			//short press - select current menu option
			ExecuteMenuSelection();
		}
		else
		{
			//long press (>1 sec) = esc from menu
			PrtHomePg();
		}
		break;
	case BTN_B:
		//^ - move menu option up
		goto MnuUp;
		break;
	case BTN_C:
		goto MnuDown;
		break;
	case BTN_DOWN:
		//v - inc selected menu option
MnuDown:
		gMnuSel++;
		if (!mnuArray[gMnuSel].length() && gMnu != MNU_TXMSG && gMnu != MNU_TXMSG_EDIT)
		{
			gMnuSel++; //skip empty items
		}
		if (gMnu != MNU_FREQ_LD && gMnu != MNU_FREQ_SV && gMnu != MNU_FREQ_REN)
		{			
			if (gMnuSel >= gMnuSelMax)
			{
				gMnuSel = 0;
			}
			if (gMnu == MNU_MAIN)
			{
				//if we're on the main mnu, track last selection
				mnuMainSel = gMnuSel;
			}
		}
		else
		{
			//change start pointer if past # of displayed memory slots
			if (gMnuSel > mnuFreqListMax - 1)
			{
				gMnuSel = 0;
				gMnuPrev = -1; //refresh
				mnuFreqListStart = mnuFreqListStart + mnuFreqListMax;
				if (mnuFreqListStart > 99)
				{
					mnuFreqListStart = 0; //roll over
				}
			}
		}
		hmiMnuChng = true; //update in Loop
		break;
	case BTN_UP:
		//^ - dec selected mnu option
MnuUp:
		gMnuSel--;		
		if (gMnu != MNU_FREQ_LD && gMnu != MNU_FREQ_SV && gMnu != MNU_FREQ_REN)
		{
			if (gMnuSel < 0)
			{
				gMnuSel = gMnuSelMax - 1;
			}
			if (!mnuArray[gMnuSel].length() && gMnu != MNU_TXMSG && gMnu != MNU_TXMSG_EDIT)
			{
				gMnuSel--; //skip empty items
			}
			if (gMnu == MNU_MAIN)
			{
				//if we're on the main mnu, track last selection
				mnuMainSel = gMnuSel;
			}
		}
		else
		{
			//change start pointer if past # of displayed memory slots
			if (gMnuSel < 0)
			{
				if (mnuFreqListStart > 0 && mnuFreqListStart < 6)
				{
					mnuFreqListStart = 0;
				}
				else if (mnuFreqListStart - mnuFreqListMax + 1 > 0)
				{
					mnuFreqListStart = mnuFreqListStart - mnuFreqListMax;
				}
				else
				{
					mnuFreqListStart = vfoMemMax - mnuFreqListMax;
				}
				gMnuSel = mnuFreqListMax - 1;
				gMnuPrev = -1;
			}
		}
		hmiMnuChng = true; //update in Loop
		break;
	case BTN_LEFT:
MnuLeft:	
		PrevMnu();
		break;
	case BTN_RIGHT:
MnuRight:
		NextMnu();
		break;
	}
}

void ProcessHelpBtn(uint8_t btn)
{
	//return to calling menu when btn released	
	//Rewrite the entire screen since Help uses it all
	hmiMode = HMI_MENU;
	tft.fillScreen(hmiBase);	
	PrtBtnA(true);
	PrtBtnB();
	PrtBtnC();
	gMnuPrev = -1; //force refresh
	CreateMnu("", false);
}

void ProcessFunctionBtn()
{
	//process a remote PTT function button
	if (gRmtFuncBtn > RMT_FUNC_PTT && gRmtFuncBtn < RMT_FUNC_OFF)
	{
		//SerialOut("Function key " + String(gRmtFuncBtn) + ", func btn idx= " + String(mnuFuncBtnIdx[gRmtFuncBtn-1]), true);
		//get the function to execute with this button
		switch (mnuFuncBtnIdx[gRmtFuncBtn-1])
		{
		case MNU_FUNC_OPTION_TXMSG1:
		case MNU_FUNC_OPTION_TXMSG2:
		case MNU_FUNC_OPTION_TXMSG3:
		case MNU_FUNC_OPTION_TXMSG4:
		case MNU_FUNC_OPTION_TXMSG5:
		case MNU_FUNC_OPTION_TXMSG6:
		case MNU_FUNC_OPTION_TXMSG7:
			//send selected buffer
			if (gTxBuffer.length())
			{
				//add BT between buffers
				gTxBuffer = gTxBuffer += " = " + mnuTxMsg[gRmtFuncBtn - 1];
			}
			else
			{
				gTxBuffer = mnuTxMsg[gRmtFuncBtn - 1];
			}
			SaveRepeatBuffer(mnuTxMsg[gRmtFuncBtn-1]); //save for F1 to repeat it
			gPlay = KEYER_MODE_RUN; //start keyer
			//send buffer to terminal
			TermOut("\n\rSend: " + gTxBuffer, true);
			break;
		case MNU_FUNC_OPTION_RIT:		
			//just open the RIT window
			hmiMode = HMI_MENU;
			gMnu = MNU_FREQ;
			gMnuPrev = -1; //refresh
			gMnuSel = MNU_FREQ_RIT;
			mnuFreqSel = gMnuSel;
			CreateMnu("", false);
			SetMnuParam();			
			break;
		case MNU_FUNC_OPTION_SPEED:
			//just open the keyer speed window
			hmiMode = HMI_MENU;
			gMnu = MNU_KEYER;
			gMnuPrev = -1; //refresh
			gMnuSel = MNU_KEYER_SPEED;
			mnuKeyerSel = gMnuSel;
			CreateMnu("", false);
			SetMnuParam();			
			break;
		case MNU_FUNC_OPTION_TXMSG:
			hmiMode = HMI_MENU;
			gMnu = MNU_TXMSG;
			gMnuPrev = -1; //refresh screen
			CreateMnu("", false);
			break;
		case MNU_FUNC_OPTION_BTNA:
			ProcessBtn(BTN_A);			
			break;
		case MNU_FUNC_OPTION_BTNB:		
			ProcessBtn(BTN_B);			
			break;
		case MNU_FUNC_OPTION_BTNC:
			ProcessBtn(BTN_C);			
			break;
		case MNU_FUNC_OPTION_ANT_SW:
			if (gAntSwOnline)
			{
				hmiMode = HMI_MENU;
				gMnu = MNU_ANT_SW;
				gMnuPrev = -1;
				CreateMnu("", false);
			}
			else
			{
				PrtStatus("ANT SWITCH IS OFFLINE", true);
			}
			break;
		}

	}

}

void ProcessValueBtn(uint8_t btn)
{
	//process button presses while in HMI_VALUE mode
	static uint32_t lastCmd = 0;// used to detect fast turning

	gMnuValuePrev = 0; //reset if any movement
	switch (btn)
	{
	case BTN_SW:
		//same as btn A
		if (gBtnSwLongPress)
		{
			//already did PrtHomePg();
		}
		else
		{
			ExecuteMenuSelection();			
		}
		break;
	case BTN_A:
		//long press = esc, short= save selection & return to calling mnu
		if (gBtnALongPress)
		{
			PrtHomePg();
		}
		else
		{
			ExecuteMenuSelection();
		}
		break;
	case BTN_B:
		//^ - inc value
MnuInc:
		gMnuValue++;
		if (gMnu != MNU_KEYER)
		{
			if (millis() - lastCmd < 50)
			{				
				if (gMnu == MNU_FREQ && gMnuSel == MNU_FREQ_RIT)
				{
					gMnuValue += 99; //inc by 100 if spinning dial in RIT
					gMnuValue = int(gMnuValue / 10) * 10; //round to 10s
				}
				else
				{
					gMnuValue += 9; //inc by 10 otherwise
				}
			}
			else if (gMnu == MNU_FREQ && mnuFreqSel == MNU_FREQ_RIT)
			{
				gMnuValue += 9;
				gMnuValue = int(gMnuValue / 10) * 10; //round to 10s
			}
		}
		if (gMnuValue > gMnuValueMax)
		{
			gMnuValue = gMnuValueMax;
		}
		GetMnuValue(false, "", LONG_PRESS_LAST);
		UpdateParam(); //send updated parameter to radio
Send2Term:
		if (gComMode == COM_MODE_TERM)
		{			
			Serial.print(TERM_CURSOR_UP); //move cursor up one line up to overwrite last value
			TermOut(" Val: " + String(gMnuValue) + "                                                                   ", true); //remove any longer data from previous value
		}
		lastCmd = millis();
		if (gRadioCAT == RADIO_CAT_FLEX)
		{
			lastCmd += 20; //add extra time for update delay
		}
		return;
		break;
	case BTN_C:
		//v - dec value
MnuDec:		
		gMnuValue--;
		//no speed up in keyer speeds
		if (gMnu != MNU_KEYER)
		{			
			if (millis() - lastCmd < 50)
			{
				if (gMnu == MNU_FREQ && gMnuSel == MNU_FREQ_RIT) 	
				{
					gMnuValue -=99 ; //dec by 100 if spinning dial in RIT
					gMnuValue = int(gMnuValue / 10) * 10; //round to 10s
				}				
				else
				{
					gMnuValue -= 9; //otherwise dec by 10
				}
			} 
			else if (gMnu == MNU_FREQ && mnuFreqSel == MNU_FREQ_RIT)
			{
				gMnuValue -= 9;
				gMnuValue = int(gMnuValue / 10) * 10; //round to 10s
			}
		}		
		if (gMnuValue < gMnuValueMin || gMnuValue > gMnuValueMax)
		{
			gMnuValue = gMnuValueMin;
		}		
		GetMnuValue(false, "", LONG_PRESS_LAST);
		UpdateParam(); //send updated parameter to radio
		goto Send2Term;		
		break;
	case BTN_UP:
		goto MnuInc; //inc value
		break;
	case BTN_DOWN:
		goto MnuDec; //dec value
		break;
	case BTN_LEFT:
		goto MnuDec;
		break;
	case BTN_RIGHT:
		goto MnuInc;
		break;
	}	
}

void ProcessTxInputBtn(uint8_t btn)
{
	//process button presses while in HMI_TXTINPUT mode

	int val = 0;
	static bool newChr = false; //goes true if we move to the next chr in TxtInput. When true, use the last chr entered to start with when inc/dec chr

	switch (btn)
	{
	case BTN_SW:
		//same as btnA
		if (gBtnSwLongPress)
		{
			//already did PrtHomePg(); //esc
		}
		else
		{
			ExecuteMenuSelection(); //save txt
		}
	case BTN_A:
		//long press = esc, short= return to calling mnu
		if (gBtnALongPress)
		{
			PrtHomePg(); //esc
		}
		else
		{
			ExecuteMenuSelection(); //save txt
		}
		break;
	case BTN_B:
		//^ - inc char val			
ChrInc:
		if (!newChr || !txtInputPtr)
		{
			//not a new chr or we're just starting, use the current chr
			val = txtInput[txtInputPtr];
		}
		else
		{
			//start with the chr to the left
			val = txtInput[txtInputPtr - 1];
			newChr = false; //reset
		}
		//find chr in chr[]
		for (int i = 0; i < chrMax - 1; i++)
		{
			if (val == chr[i])
			{
				val = i; //use this chr
				goto IncAscii;
			}
		}
		val = chrMax - 1;//default to space if chr not in chr[]

IncAscii:
		val++;
		//ignore error prosign
		if ((val > chrMax - 2))
		{
			val = 0; //rollover
		}
		if (chr[val] == '~')
		{
			val++; //ignore error prosign
		}
		txtInput[txtInputPtr] = chr[val];
		PrtTxtInput(false);
		return;
		break;
	case BTN_C:
		//v - dec chr val			
ChrDec:
		if (!newChr || !txtInputPtr)
		{
			//not a new chr or we're just starting, use the current chr
			val = txtInput[txtInputPtr];
		}
		else
		{
			//start with the chr to the left
			val = txtInput[txtInputPtr - 1];
			newChr = false; //reset
		}
		//find chr in assci[]
		//SerialOut("txtInput val= " + String(val) + ", " + String(txtInput[txtInputPtr]), true);
		for (int i = 0; i < chrMax - 1; i++)
		{
			if (val == chr[i])
			{
				val = i; //use this chr
				goto DecAscii;
			}
		}
		val = 0;//default to A if chr not in chr[]

DecAscii:	
		val--;
		if (val < 0)
		{
			val = chrMax - 2;
		}
		if (chr[val] == '~')
		{
			//ignore error prosign
			val--;
		}
		txtInput[txtInputPtr] = chr[val];
		PrtTxtInput(false);
		return;
		break;
	case BTN_UP:
		goto ChrInc;
		break;
	case BTN_DOWN:
		goto ChrDec;
		break;
	case BTN_LEFT:
		//dec ptr			
		txtInputPtr--;
		if (txtInputPtr < txtInputSrcStart)
		{
			txtInputSrcStart--;
			if (txtInputSrcStart < 0)
			{
				txtInputSrcStart = 0;
			}
			txtInputPtr = txtInputSrcStart;
		}
		else if (txtInputPtr < 0)
		{
			txtInputSrcStart = 0;
			txtInputPtr = 0;
		}
		PrtTxtInput(true);
		if (txtInputPtr > 0)
		{
			//move cursur to the left
			TermOut(String(char(8)), false);
		}
		break;
	case BTN_RIGHT:
		//inc ptr			
		txtInputPtr++;
		//chk to see if we're scrolling off to the right
		if (txtInputPtr > txtInputSrcStart + 12)
		{
			//scroll to the right
			txtInputSrcStart++;
			if (txtInputSrcStart >= txtInputSrcMax)
			{
				txtInputSrcStart = txtInputSrcMax - 12;
			}
			txtInputPtr = txtInputSrcStart + 12;//point to last chr
		}
		if (txtInputPtr >= txtInputSrcMax)
		{
			//we're at the end, stop scrolling
			txtInputPtr = txtInputSrcMax - 1;
		}

		if (!txtInput[txtInputPtr] || txtInput[txtInputPtr] == ' ')
		{
			//if nothing, insert a space - we'll trim the final result
			txtInput[txtInputPtr] = ' ';
			newChr = true; //this is a new chr, start with chr to the left in ChrInc:
		}
		PrtTxtInput(true);		
		if (txtInputPtr < txtInputSrcMax && keyboardChr > 128)
		{
			//move cursur to the right if cursor keys
			TermOut(" ", false);
		}
		break;
	}
}

void ChkPaddles()
{
	//key sidetone on left paddle to test
	static bool ditPrev = false;
	static bool dahPrev = false;
	static uint8_t ditDebounce = 0;//must = 5 for valid dit
	static uint8_t dahDebounce = 0;
	static uint8_t lastSent = 0; //tracks last element sent - 1=dit, 2=dah	
	static uint8_t elementCtr = 0; //counts the # of active elements in keyerQue[]
	static uint32_t wdt = 0; //watch dog for stuck key
	static uint32_t keyDownStart = 0; //start of key down, used to time btTxMode indication back to Tx Enab after 1000 mS	
	static bool bothPaddles = false; //goes true when both paddles are pressed together	
	static bool alreadyHr = false; //blocks recussion
	
	bool dit = false;
	bool dah = false;

	uint8_t lastElementCtr = 0; //holds last ctr value so we can detect when a new element was added
	static String cwChr = ""; //holds cw string for txtInput
	static uint32_t lastElementTmr; //tracks when last element was sent to detect chr space

	if (alreadyHr == true)
	{
		return;
	}
	alreadyHr = true;

	//invert paddle states for readability - paddle inputs are normally high when idle
	if (gPaddle == false)
	{
		//normal paddle input
		dit = !digitalRead(PDL_LEFT); //tip
		dah = !digitalRead(PDL_RIGHT); //ring
	}
	else
	{
		//reverse paddle input
		dit = !digitalRead(PDL_RIGHT); //ring
		dah = !digitalRead(PDL_LEFT); //tip
	}
	bool keyDown = false;
	bool rmtPTTIn = false; //active LOW - this converts it to active HIGH
	
	if (gRmtMode == MNU_RMT_KEY)
	{		
		if (gRmtFuncBtn == RMT_FUNC_PTT)
		{
			rmtPTTIn = true;
		}
	}
	

	//debounce for 4 mSec
	if (dit != ditPrev)
	{
		if (ditDebounce < 4)
		{
			ditDebounce++; //wait 4 mS before declaring it valid
			dit = !dit;
		}
		else
		{
			ditDebounce = 0;//reset debounce and allow dit to go through			
		}
	}
	if (dah != dahPrev)
	{
		if (dahDebounce < 4)
		{
			dahDebounce++;
			dah = !dah;
		}
		else
		{
			dahDebounce = 0;
		}
	}


	lastElementCtr = elementCtr; //if this changes we added an element, check to see if we're in MNU_TXTINPUT & decode new chr if we added a space

	if (!gPlay && gKeyType == KEYER_PASSTHRU)
	{
		//if not playing buffer, just pass dits and dahs to key and PTT toward radio
		goto Paddles;
	}

	//que processor

	if (elementCtr > 0)
	{
		//dec or inc keyerQue[0] until it reaches 0 - if positive, we're key down, if negative we're spacing
		if (keyerQue[0] > 0)
		{
			//counting down key down time
			keyerQue[0]--;
			//SerialOut(String(keyerQue[0]), true);
			keyDown = true;
		}
		else if (keyerQue[0] < 0)
		{
			//counting up space time
			keyerQue[0]++;
			keyDown = false;
		}
		if (keyerQue[0] == 0)
		{
			//we've reached the end of this period, rotate keyerQue[] to see if there's more to send
			for (int i = 1; i < 20; i++)
			{
				keyerQue[i - 1] = keyerQue[i];
				keyerQue[i] = 0;
			}
			elementCtr--;
			if (keyerQue[0] != 0)
			{
				//que is not empty, sent next element
				keyDown = true;
				if (keyerQue[0] >= keyerDitDur * 2)
				{
					lastSent = 2; //this element is a dah						
				}
				else if (keyerQue[0] < 0)
				{
					bothPaddles = false; //only look for both paddles after space
				}
				else if (keyerQue[0] < keyerDitDur * 2)
				{
					lastSent = 1; //this element is a dit										
				}
			}
			else
			{
				if (gKeyType == KEYER_IAMBIC_B && bothPaddles && !dit && !dah)
				{
					//if we're done sending see if we squeezed during input - send opposite chr as lastSent
					//SerialOut("Squeeze released, dit = " + String(dit) + ", dah = " + String(dah) + ", lastSent = " + String(lastSent), true);
					bothPaddles = false;
					//release squeezed, add opposite chr
					if (lastSent == 1)
					{
						//SerialOut("adding extra dah", true);
						keyerQue[elementCtr] = keyerDitDur * 3; //add a dah after ending space						
					}
					else
					{
						ditPrev = dit;
						//SerialOut("adding extra dit", true);
						keyerQue[elementCtr] = keyerDitDur; //add a dit						
					}
					elementCtr++;
					keyerQue[elementCtr] = -keyerDitDur; //add ending space
					elementCtr++;
				}
			}
		}
	}
	else
	{
		//this is where we send chrs from gTxBuffer
		
		//que is empty, check for tx buffer data
		//chk to see if we have a chr to send from gTxBuffer
		if (gTxBuffer.length() == 0 && gPlay != KEYER_MODE_STOP)
		{
			//we just finished sending last chr, reset mode
			gPlay = KEYER_MODE_STOP; //we're done, we'll upodate the Home pg in Loop			
			keyDown = false;
			alreadyHr = false;
			ResetSpeedBoost();
			return;
		}
		else if (gTxBuffer.length() > 0 && gPlay == KEYER_MODE_RUN)
		{
			//load chr elements from first chr in gTxBuffer then remove it from string & send string back to display						
			keyDown = false;
			String txChar = gTxBuffer.substring(0, 1);			

			gTxBuffer.remove(0, 1);

			//toggle speed 		
			if (txChar == "*")
			{
				if (!speedToggle)
				{
					//increase speed 50%
					keyerDitDurPrev = keyerDitDur;//save current speed
					keyerDitDur = keyerDitDur / 1.5; //50% speed increase
					keyerDitFarnsDurPrev = keyerDitFarnsDur;
					keyerDitFarnsDur = keyerDitFarnsDur / 2; //less space between chrs
				}
				else
				{
					keyerDitDur = keyerDitDurPrev;
					keyerDitFarnsDur = keyerDitFarnsDurPrev;
				}
				speedToggle = !speedToggle;
			}
			else if (txChar == "^")
			{
				//insert call sign in buffer
				gTxBuffer = " " + mnuKeyerCall + gTxBuffer;
				alreadyHr = false;
				return;
			}
			else if (txChar == "#")
			{
				//insert contest SN into buffer in place of #
				gTxBuffer = " " + String(mnuKeyerSN) + gTxBuffer;
				if (gTxBuffer.indexOf("#") == -1)
				{
					mnuKeyerSN++; //inc it for next exchange if it's not repeated in this buffer
				}
				alreadyHr = false;
				return;
			}
			else if (txChar == "%")
			{
				//insert contest exchange into buffer
				gTxBuffer = " " + mnuKeyerExch + gTxBuffer;
				alreadyHr = false;
				return;
			}
			if (txChar == " ")
			{
				//add word space 
				keyerQue[elementCtr] = -(keyerDitFarnsDur * 4); //only add four dit times instead of 7 because we already have 3 from the last chr space
				elementCtr++;
			}
			else
			{								
				String txt = TxChar2Morse(txChar); //convert chr to morse symbols then stuff the elements here and clear the buffer					
				txt.trim();
				if (txt.length())
				{
					for (uint i = 0; i < txt.length(); i++)
					{
						if (txt.charAt(i) == '.')
						{
							keyerQue[elementCtr] = keyerDitDur; //dit
							elementCtr++;
							//add element space
							keyerQue[elementCtr] = -keyerDitDur;
							elementCtr++;
						}
						else if (txt.charAt(i) == '-')
						{
							keyerQue[elementCtr] = keyerDitDur * 3; //dah
							elementCtr++;
							//add element space
							keyerQue[elementCtr] = -keyerDitDur;
							elementCtr++;
						}
					}
					//add chr space at the end of letter
					if (elementCtr)
					{
						keyerQue[elementCtr - 1] = -(keyerDitFarnsDur * 3); //replace last interdigit space with chr space							
						alreadyHr = false;
						return;
					}
				}
			}
		}

		//que completed, see if we have paddles still pressed 					
		if (dit && dah)
		{
			//both squeezed
			if (gKeyType == KEYER_IAMBIC_A || gKeyType == KEYER_IAMBIC_B)
			{
				if (lastSent == 2)
				{
					//last element sent was a dah, add a dit
					lastSent = 1; //toggle for next pass
					keyerQue[elementCtr] = keyerDitDur;
				}
				else
				{
					//add a dah 
					lastSent = 2;
					keyerQue[elementCtr] = keyerDitDur * 3;
				}
				elementCtr++;
			}
			else if (gKeyType == KEYER_ULTIMATIC)
			{
				//lastSent holds first element of squeezed chr
				bothPaddles = true;
				if (lastSent == 2)
				{
					//if dit was first, send continous dahs
					keyerQue[elementCtr] = keyerDitDur * 3;
				}
				else
				{
					//else if dah was first, send continous dits
					keyerQue[elementCtr] = keyerDitDur;
				}
				elementCtr++;
			}
			else if (gKeyType == KEYER_BUG)
			{
				//just send dits
				keyerQue[elementCtr] = keyerDitDur;
				elementCtr++;
			}
			//add a space
			keyerQue[elementCtr] = -(keyerDitDur);
			elementCtr++;
		}
		else if (gKeyType == KEYER_BUG && dit == false)
		{
			if (dah == true && dahPrev == true && !elementCtr)
			{
				//hold dah as long as wanted after last space has timed out
				keyDown = true;
			}
			else if (dah == false && dahPrev == true)
			{
				//add a space at end
				keyerQue[elementCtr] = -(keyerDitDur);
				elementCtr++;
				dahPrev = dah;
			}
		}
		else if (dit == true && ditPrev == true)
		{
			//add another dit						
			lastSent = 1;
			keyerQue[elementCtr] = keyerDitDur;
			elementCtr++;
			//add a space
			keyerQue[elementCtr] = -(keyerDitDur);
			elementCtr++;
		}
		else if (dah == true && dahPrev == true)
		{
			//add another dah			
			lastSent = 2;
			keyerQue[elementCtr] = keyerDitDur * 3;
			elementCtr++;
			//add a space
			keyerQue[elementCtr] = -(keyerDitDur);
			elementCtr++;
		}
	}


	//chk paddle positions
	//If gRmtMode is MNU_RMT_KEY, use remote PTT as a straight key

Paddles:	
	
	if (gRmtMode == MNU_RMT_KEY && rmtPTTIn == true)
	{
		//rmtPTT used for straight key. This allows a straight key to be plugged into the Rmt PTT jack so the op can switch between the keyer and straight key easily		
		dit = true;
		//SerialOut("Rmt PTT -> Straight Key", true);
		goto StghtKey;
	}

	if (gKeyType == KEYER_STRAIGHT)
	{
StghtKey:
		if (dit == true)
		{
			wdt++;
			keyDown = true;
		}
		else
		{
			wdt = 0;
		}
		ditPrev = dit;
	}
	else if (gKeyType == KEYER_PASSTHRU && !gEnabAmpSeq)
	{
		//just send dit to Key Out and dah to PTT toward radio so we can use radio's keyer
		if (gTxMode)
		{
			//control key_out for dit output
			if (dit == true)
			{
				if (gTxMode)
				{
					//key radio
					digitalWrite(KEY_OUT, HIGH);
					gTxMode = RADIO_TX_KEY;
					pttWatchDog = millis(); //start watchdog
					txHangTmr = millis(); 
				}				
			}
			else
			{
				digitalWrite(KEY_OUT, LOW);	//unkey radio			
			}
			ditPrev = dit;
			//control PTT_out for dah output
			if (dah == true)
			{
				if (gTxMode)
				{
					digitalWrite(PTT_OUT, HIGH); //key ptt
					gTxMode = RADIO_TX_KEY;
					pttWatchDog = millis(); //start watchdog
					txHangTmr = millis();
					//SerialOut("Dah keying radio", true);				}
				}				
			}
			else
			{
				digitalWrite(PTT_OUT, LOW); //release ptt				
			}
			dahPrev = dah;
		}
		alreadyHr = false;
		return; //nothing left to do in this mode
	}
	else if ((gKeyType == KEYER_IAMBIC_A) || (gKeyType == KEYER_IAMBIC_B) || (gKeyType == KEYER_ULTIMATIC))
	{		
		if (dit != ditPrev)
		{
BugMode:
			ditPrev = dit;
			if (dit == true)
			{
				//new dit - add a dit and space to keyerQue
				keyerQue[elementCtr] = keyerDitDur;
				elementCtr++;
				keyerQue[elementCtr] = -(keyerDitDur);
				elementCtr++;
				if (gKeyType == KEYER_ULTIMATIC && bothPaddles == false)
				{
					//save beginning element
					lastSent = 1;
				}
			}
		}
		else if (dah != dahPrev)
		{
			dahPrev = dah;
			if (dah == true)
			{
				//new dah - add a dah and space to keyerQue
				keyerQue[elementCtr] = keyerDitDur * 3; //dah timer
				elementCtr++;
				keyerQue[elementCtr] = -(keyerDitDur);
				elementCtr++;
				if (gKeyType == KEYER_ULTIMATIC && bothPaddles == false)
				{
					//save beginning element
					lastSent = 2;
				}
			}
		}
		if (dit == true && dah == true)
		{
			bothPaddles = true; //this sequence involved both paddles being pressed at the same time - treat special if Iambic-B or Ultimatic
		}
	}
	else if (gKeyType == KEYER_BUG)
	{
		if (dit != ditPrev)
		{
			goto BugMode;
		}
		else if (dah != dahPrev)
		{
			dahPrev = dah;
			if (dah == true)
			{
				wdt++;
				//new dah - add a dah to keyerQue - user can hold it longer if wanted
				keyerQue[elementCtr] = keyerDitDur * 2; //partial dah timer
				elementCtr++;
			}
			else
			{
				wdt = 0;
			}
		}
	}

	//chk to see if we have the txt editor open. If so, add the new element to cwChr
	if (hmiMode == HMI_TXTINPUT)
	{
		//there's always a space element added after a chr element		
		if (elementCtr > lastElementCtr + 1)
		{
			if (keyerQue[lastElementCtr] == keyerDitDur)
			{
				cwChr += ".";
				lastElementTmr = millis();
			}
			else if (keyerQue[lastElementCtr] == keyerDitDur * 3)
			{
				cwChr += "-";
				lastElementTmr = millis();
			}			
			lastElementCtr = elementCtr;			
		}
		if ((cwChr.length() > 0) && ((millis() - lastElementTmr) > (keyerDitDur * 5)))
		{
			//we're done, decode							
			txtInputCW = Morse2Char(cwChr); //we'll clear this once we add it to txtInput				
			cwChr = "";
		}
	}


	//control output based on options - turn off after 10 sec of key down	
	if (keyDown == true && wdt < 10000)
	{		
		//request to key radio
		if (gTxMode || gSideTone || hmiMode == HMI_TXTINPUT)
		{			
			gKeyDown = true; //for sidetone & Flex_TxCW()
			if (gTxMode && (hmiMode != HMI_TXTINPUT))
			{
				//key the radio
				//don't change [A] btn state if we're on a Flex and not in CW mode
				if (gRadioCAT != RADIO_CAT_FLEX || gRMode == RADIO_CW)
				{
					gTxMode = RADIO_TX_KEY; //tx indicator will be reset after KEY_OUT goes low for >500 mS
				}				
				keyDownStart = millis();				
				txHangTmr = millis();
				if (!gEnabAmpSeq)
				{
					//sequencing not enabled, just key radio
					if (gFlexApiKey && gRadioCAT == RADIO_CAT_FLEX)
					{
						//we'll control Flex from Loop() once we exit ISR
						//Flex_TxCW(false); //uses gKeyDown
					}
					else
					{
						//normal hardware keying
						digitalWrite(KEY_OUT, HIGH);						
					}
				}
				else if (digitalRead(PTT_OUT) == LOW)
				{
					//timer hasn't been started, key up PTT to enable amp Tx then start pre sequence timer
					digitalWrite(PTT_OUT, HIGH); //key amp
					if (gSeqPre)
					{
						gSeqPreTmr = millis() + gSeqPre - 1; //subtract the current 1ms cycle
					}
					else
					{
						//no pre time delay, just key the radio
						gSeqPreTmr = 0;
						digitalWrite(KEY_OUT, HIGH); //key radio
					}
				}
				else if (gSeqPreTmr < millis())
				{
					//pre timer has timed out, set key out to xcvr
					gSeqPostTmr = millis() + gSeqPost; //start post timer
					digitalWrite(KEY_OUT, HIGH);
				}
			}			
		}
		pttWatchDog = millis(); //start watchdog						
	}
	else if (!rmtPTTIn)
	{
		//only if we're not keying PTT with RMT PTT
		if (gTxMode == RADIO_TX_KEY && (millis() - keyDownStart > 10000))
		{
			gTxMode = RADIO_TX_ENAB; //reset red TX indicator back to green TX Enab
		}
		if (gEnabAmpSeq)
		{
			if (digitalRead(KEY_OUT) == HIGH)
			{
				//we're still keyed but just opened the key, start post timer				
				if (gSeqPost)
				{
					gSeqPostTmr = millis() + gSeqPost;
				}
				else
				{
					//post is 0, just release PTT
					digitalWrite(PTT_OUT, LOW);
				}
			}
			else
			{
				//key input is open, we are waiting for timer to expire before opening PTT to desable amp Tx, or clock has rolled over
				if (millis() >= gSeqPostTmr || millis() < 1000)
				{
					digitalWrite(PTT_OUT, LOW);
				}
			}
		}
		digitalWrite(KEY_OUT, LOW ); //release key out
		gKeyDown = false;		
		wdt = 0;		
	}

	//failsafe
	if (elementCtr > 20)
	{
		elementCtr = 20;
	}
	alreadyHr = false;

}

void ChkPTT()
{
	static bool alreadyHr = false; //blocks recurssion	
	static uint8_t keySeq = 0; //0=not keyed by any source, 1=keyed by ptt, 2=ptt latched, wait for next PTT or pttWatchdog timeout
	static bool watchDogFired = false; //goes true if pttWatchDog expired - this hold PTT inactive
	bool ptt = false;
	bool cPTT = false;
	static bool lastPTT = false; //remember last states
	static bool lastCPTT = false;
	static uint8_t lastRmtFuncBtn = 0xff;
		
	if (gRmtMode != MNU_RMT_KEY)
	{
		if (gRmtFuncBtn == RMT_FUNC_PTT)
		{
			//active Rmt PTT returns 0 
			ptt = true;
		}
	}
	if (gRmtFuncBtn != lastRmtFuncBtn)
	{
		if (gRmtFuncBtn > RMT_FUNC_PTT && gRmtFuncBtn < RMT_FUNC_OFF)
		{
			ProcessFunctionBtn();
		}
		lastRmtFuncBtn = gRmtFuncBtn; //wait for next COS
	}

	if (!mnuCBtnSel && gTxMode)
	{
		cPTT = gBtnPTT; //use current state
	}

	if (!gTxMode)
	{
		keySeq = 0; //reset if we're not enabled, in case we exited Tx Enable while sequencing
		watchDogFired = false;
		lastPTT = false;
		lastCPTT = false;	
	}

	if (gRmtMode == MNU_RMT_KEY)
	{
		//we're in straight key mode and [C] btn is not set to PTT, ignore PTT input, ChkPaddles() will output to Key when RmtPTT is active
		gBtnPTT = false; 
		alreadyHr = false; //just in case 
		return;
	}
	if (alreadyHr || !gTxMode)
	{		
		return;
	}
	alreadyHr = true;
	
	//SerialOut("ChkPTT: ptt= " + String(ptt), true);

	//check current state
	switch (keySeq)
	{
	case 0:
		//ptt is not currently active
		if (ptt || cPTT || watchDogFired)
		{
			//ptt is active now - switch to tx mode						
			//if watchDog has fired, ignore ptt until it's released
			if (!watchDogFired)
			{
				keySeq = 1;	//wait for ptt release
				lastPTT = ptt;
				lastCPTT = cPTT;
				delay(10); //debounce
				ControlPTT(HIGH); //set PTT active
				gTxMode = RADIO_TX_PTT;
				gTxModePrev = gTxMode; //so we don't reset it			
				PrtBtnA(true);
				pttWatchDog = millis(); //start watchdog
				//SerialOut("PTT active", true);
			}
			else if (!ptt && !cPTT)
			{
				//ptt is inactive now, reset watchdog latch if it was set
				watchDogFired = false;
				gTxMode = RADIO_TX_ENAB;
				gBtnPTT = false;
				//SerialOut("WD timer reset", true);
			}
		}		
		break;
	case 1:
		//ptt is active
		if (((!ptt && lastPTT) || (!cPTT && lastCPTT)) && gRmtMode != MNU_RMT_LATCH)
		{
			//ptt just released, unkey and reset gTxMode			
			//SerialOut("gRmtMode= " + String(gRmtMode) + ", PTT released", true);
			gTxMode = RADIO_TX_ENAB;
			delay(10);
PTTOff:		keySeq = 0;	//reset and wait for the next ptt down				
			ControlPTT(LOW);	
			gBtnPTT = false;
			lastPTT = false;
			lastCPTT = false;
			gTxModePrev = gTxMode; //so we don't reset it
			PrtBtnA(true);	
			if (gTxMode == RADIO_TX_WDOG)
			{				
				PrtStatus("WATCHDOG TIMEOUT", true);
				delay(3000);
			}
		}		
		else if (pttWatchDogDefault < millis() - pttWatchDog)
		{
FireWD:		//keyed for > 3 minutes - release and wait for RmtPTT to open			
			watchDogFired = true; 
			ptt = false;
			gBtnPTT = false;
			gTxMode = RADIO_TX_WDOG; //display yellow warning for Btn A
			//SerialOut("WD timer fired 1", true);
			goto PTTOff; //release PTT
		}
		else if ((!ptt && gRmtMode == MNU_RMT_LATCH) || (cPTT && !mnuCBtnSel))
		{
			//ptt released, wait for next PTT active
			lastPTT = ptt;
			lastCPTT = cPTT;
			delay(10);
			keySeq = 2;
			//SerialOut("first PTT released", true);
		}
		break;
	case 2:
		//we're in latch mode, wait for ptt to go active again to turn it off when released
		if (ptt && !lastPTT)
		{
			//ptt just went active again, hold ptt out active until we release ptt
			delay(10); //debounce
			keySeq = 3;
			//SerialOut("second PTT active", true);
		}
		else if (!cPTT && lastCPTT)
		{
			gTxMode = RADIO_TX_ENAB;
			//SerialOut("seq 2 - cbtn went to 0", true);
			goto PTTOff;
		}		
		else if (pttWatchDogDefault < millis() - pttWatchDog)
		{
			//keyed for > 3 minutes - release and wait for RmtPTT to open
			//NOTE: We can't goto FireWD here because the PTT is not active - FireWD displays WDog! label on [A] until ptt is released
			watchDogFired = true;
			ptt = false;
			gBtnPTT = false;
			lastPTT = false;
			lastCPTT = false;
			gTxMode = RADIO_TX_WDOG; //display yellow warning for Btn A
			//SerialOut("WD timer fired 2", true);
			keySeq = 0;	//reset and wait for the next ptt down				
			ControlPTT(LOW);
			gTxModePrev = gTxMode; //so we don't reset it
			PrtBtnA(true);
			PrtStatus("WATCHDOG TIMEOUT", true);
			delay(3000); //display WDog label on [A] for 3 seconds then go back to scanning for changes
		}
		break;
	case 3:
		//chk for PTT to be released or WD timer to time out
		if (!ptt)
		{
			delay(10);
			//SerialOut("second PTT released", true);
			gTxMode = RADIO_TX_ENAB;
			goto PTTOff;
		}
		else if (pttWatchDogDefault < millis() - pttWatchDog)
		{
			goto FireWD;
		}
		break;
	}

	alreadyHr = false;
}

char Morse2Char(String code)
{
	//decode the char represented by the code char array	
	if (code == ".-")		return chr[0]; // A
	if (code == "-...")		return chr[1]; // B
	if (code == "-.-.")		return chr[2]; // C
	if (code == "-..")		return chr[3]; // D
	if (code == ".")		return chr[4]; // E
	if (code == "..-.")		return chr[5]; // F
	if (code == "--.")		return chr[6]; // G
	if (code == "....")		return chr[7]; // H
	if (code == "..")		return chr[8]; // I
	if (code == ".---")		return chr[9]; // J
	if (code == "-.-")		return chr[10]; // K
	if (code == ".-..")		return chr[11]; // L
	if (code == "--")		return chr[12]; // M
	if (code == "-.")		return chr[13]; // N
	if (code == "---")		return chr[14]; // O
	if (code == ".--.")		return chr[15]; // P
	if (code == "--.-")		return chr[16]; // Q
	if (code == ".-.")		return chr[17]; // R
	if (code == "...")		return chr[18]; // S
	if (code == "-")		return chr[19]; // T
	if (code == "..-")		return chr[20]; // U
	if (code == "...-")		return chr[21]; // V
	if (code == ".--")		return chr[22]; // W
	if (code == "-..-")		return chr[23]; // X
	if (code == "-.--")		return chr[24]; // Y
	if (code == "--..")		return chr[25]; // Z

	if (code == ".----")	return chr[26]; // 1
	if (code == "..---")	return chr[27]; // 2
	if (code == "...--")	return chr[28]; // 3
	if (code == "....-")	return chr[29]; // 4
	if (code == ".....")	return chr[30]; // 5
	if (code == "-....")	return chr[31]; // 6
	if (code == "--...")	return chr[32]; // 7
	if (code == "---..")	return chr[33]; // 8
	if (code == "----.")	return chr[34]; // 9
	if (code == "-----")	return chr[35]; // 0
	
	//prosigns based of of K1EL's WinKeyer
	if (code == ".-.-.-")	return chr[36]; // .
	if (code == "--..--")	return chr[37]; // ,
	if (code == "-.--.")	return chr[38]; // : also KN

	if (code == ".-..-.")	return chr[65];   // " is RR
	if (code == "...-..-")	return chr[66];  // $ is SX
	if (code == ".----.")	return chr[67];  // ' is WG
	if (code == "-.--.")	return chr[68];  // ( is KN
	if (code == "-.--.-")	return chr[69];  // ) is KK
	if (code == ".-.-.")	return chr[70];  // + is AR
	if (code == "-....-")	return chr[71];  // - is DU
	if (code == "-..-.")	return chr[72];  // / is DN
	if (code == ".-.-")		return chr[73];  // ; is AA
	if (code == ".-.-.")	return chr[74];  // < is AR
	if (code == "-...-")	return chr[75];  // = is BT
	if (code == "...-.-")	return chr[76];  // > is SK
	if (code == "..--..")	return chr[77];  // ?
	if (code == ".--.-.")	return chr[78];  // @ is AC
	if (code == ".-...")	return chr[79];  // [ is AS

	//These are dups of chrs alreadly listed above and will never be returned
	if (code == "-..-.")	return chr[80];  // \ is DN (same as /) 
	if (code == "-.--.")	return chr[81];  // ] is KN (same as :)	

	//82, 83, and 84 are &, #, and % - these are shortcuts to send call, contest SN, and contest exhange when inserted in code

	if (code == "......" || code == "......." || code == "........" || code == ".........")	return chr[85]; // ~ = use multiple err codes for back space chr
	if (code == "------" || code == "-------" || code == "--------" || code == "---------") return chr[86]; // for txtInputCW - returns ' ' to inc to next chr
	
	return 0;
}

String TxChar2Morse(String txChar)
{
	//chr[] to morse symbols
	txChar.toLowerCase();
	char i = txChar.charAt(0);

	if (i == chr[0]) return ".-"; // a
	if (i == chr[1]) return "-..."; // b
	if (i == chr[2]) return "-.-."; // c
	if (i == chr[3]) return "-.."; // d
	if (i == chr[4]) return ".";  // e
	if (i == chr[5]) return "..-."; // f
	if (i == chr[6]) return "--.";  // g
	if (i == chr[7]) return "...."; //h
	if (i == chr[8]) return ".."; // i
	if (i == chr[9]) return ".---"; // j
	if (i == chr[10]) return "-.-"; // k
	if (i == chr[11]) return ".-.."; // l
	if (i == chr[12]) return "--"; // m
	if (i == chr[13]) return "-."; // n
	if (i == chr[14]) return "---"; // o
	if (i == chr[15]) return ".--."; // p
	if (i == chr[16]) return "--.-"; // q
	if (i == chr[17]) return ".-."; // r
	if (i == chr[18]) return "..."; // s
	if (i == chr[19]) return "-"; // t
	if (i == chr[20]) return "..-"; // u
	if (i == chr[21]) return "...-"; // v
	if (i == chr[22]) return ".--"; // w
	if (i == chr[23]) return "-..-"; // x
	if (i == chr[24]) return "-.--"; // y
	if (i == chr[25]) return "--.."; // z

	if (i == chr[26]) return ".----"; // 1
	if (i == chr[27]) return "..---"; // 2
	if (i == chr[28]) return "...--"; // 3
	if (i == chr[29]) return "....-"; // 4
	if (i == chr[30]) return "....."; // 5
	if (i == chr[31]) return "-...."; // 6
	if (i == chr[32]) return "--..."; // 7
	if (i == chr[33]) return "---.."; // 8
	if (i == chr[34]) return "----."; // 9
	if (i == chr[35]) return "-----"; // 0
	if (i == chr[36]) return ".-.-.-";  // . is period
	if (i == chr[37]) return "--..--";  // , is comma
	if (i == chr[38]) return "-.--.";   // : is KN (same as ])

	//prosigns based on W1EL WinKey - https://www.k1elsystems.com/files/WKUSB_Manual_v1.3.pdf
	if (i == chr[65]) return ".-..-.";  // " is RR
	if (i == chr[66]) return "...-..-"; // $ is SX
	if (i == chr[67]) return ".----.";  // ' is WG
	if (i == chr[68]) return "-.--.";   // ( is KN
	if (i == chr[69]) return "-.--.-";  // ) is KK	
	if (i == chr[70]) return ".-.-.";   // + is AR
	if (i == chr[71]) return "-....-";  // - is DU	
	if (i == chr[72]) return "-..-.";   // / is DN	
	if (i == chr[73]) return ".-.-";    // ; is AA
	if (i == chr[74]) return ".-.-.";   // < is AR
	if (i == chr[75]) return "-...-";   // = is BT
	if (i == chr[76]) return "...-.-";  // > is SK
	if (i == chr[77]) return "..--..";  // ? is question
	if (i == chr[78]) return ".--.-.";  // @ is AC
	if (i == chr[79]) return ".-...";   // [ is AS
	if (i == chr[80]) return "-..-.";   // \ is DN
	if (i == chr[81]) return "-.--.";   // ] is KN (same as :)
	// special prosigns
	if (i == chr[82]) return "CALL";   //^ = send call
	if (i == chr[83]) return "SN";     //# = send contest sn
	if (i == chr[84]) return "EX";     //% = send contest exchange	
	return " ";
}

void CalcDitDur()
{
	keyerDitDur = (1200 / (HW_TIMER_INTERVAL_1MS / 1000)) / gCWSpeed;
}

void CalcDitFarnsDur()
{
	keyerDitFarnsDur = (1200 / (HW_TIMER_INTERVAL_1MS / 1000)) / gFarnsSpeed;
}

void ChkEncoder()
{
	//called every 1 mS from ISR -
	// DON'T CALL ANY ROUTINES THAT WRITE TO COMM PORTS OR USE DELAY() FUNCTIONS - causes uProc to lock up

	//NOTE: Don't debounce these inputs - we're looking for the edge to fall on a or b to determine direction

	//roll polarities for easier code reading

	bool a = !digitalRead(ROTARY_A);
	bool b = !digitalRead(ROTARY_B);
	bool sw = !digitalRead(ROTARY_SW); // !=0 when encoder pressed

	static int curDir = 0; //indicates which way to move - a=low first, dec, b=low first, inc	
	static bool alreadyHr = false;

	static bool lastA = a; //track a input change
	static bool lastB = b; //track b input change
	static bool lastSw = sw; //track switch change
	static uint32_t rotaryTimer = 0; //rotary change timer
	static uint32_t swTimer = 0; //detects switch down time
	
	if (alreadyHr == true)
	{
		//nothing to do
		return;
	}

	alreadyHr = true;

	//detect switch press & release
	if (sw && !lastSw)
	{
		swTimer = millis(); //switch just pressed, start switch down timer and wait for swTimer to time out			
	}
	else if (!sw && lastSw && swTimer > 0)
	{
		//switch just released, inc gMnu if a & b not changed		
		//if we've waited 50 mS inc gMnu and reset, else wait		
		uint32_t tmr = millis() - swTimer;
		if (tmr > 50)
		{
			//SerialOut("Encoder Switch Released", true);
			switch (hmiMode)
			{
			case HMI_HOME:
				//start menu up - use < > keys to change
				hmiMode = HMI_MENU;
				if (tmr < 500)
				{
					//short press, open main menu
					gMnu = MNU_MAIN;
					gMnuSel = mnuMainSel; //the last main menu selection
				}
				else
				{
					//long press, open freq mnu					
					gMnu = MNU_FREQ;
					gMnuSel = mnuFreqSel;
				}
				gMnuPrev = -1; //force refresh				
				hmiMnuChng = true; //force CreateMnu() in Loop
				break;			
			case HMI_VALUE:
				if (tmr > 500 && gLongPress == LONG_PRESS_YES)
				{
					//long-click toggle value/0					
					if (gMnuValuePrev)
					{
						gMnuValue = gMnuValuePrev;
						gMnuValuePrev = 0;
					}
					else
					{
						gMnuValuePrev = gMnuValue;
						gMnuValue = 0;
					}					
					gLongPressToggle = true; //Loop() will update the parameter
				}
				else
				{
					//short click or long-press not allowed, close window					
					gMnuValuePrev = 0; //reset short-press toggle
					gEncoderClicked = true;
				}								
				break;
			case HMI_MENU:	//flow through
			case HMI_TXTINPUT:				
				if (tmr < 500)
				{
					gEncoderClicked = true;
				}
				else
				{
					//long press, go back to main menu
					gMnu = MNU_MAIN;
					gMnuPrev = -1;
					gMnuSel = mnuMainSel; //the last main menu selection	
					hmiMnuChng = true;
				}				
				break;
			}
			hmiEncoderTimeout = millis(); //restart timeout	
			swTimer = 0; //reset time so we can start again
			curDir = 0;
			rotaryTimer = 0;
		}
	}
	


	//detect encoder COS on trailing edge
	//if both inputs are true the rotaryTimer hasn't started, check for trailing edge to trigger
	if (a && b && rotaryTimer == 0)
	{
		if (!lastA && lastB)
		{
			curDir = 1;
			swTimer = 0;
			rotaryTimer = millis(); //start rotary timer			
		}
		else if (lastA && !lastB)
		{
			curDir = -1;
			swTimer = 0;
			rotaryTimer = millis();
		}
	}

	//execute rotary change if rotary timer times out	
	if (curDir != 0 && rotaryTimer != 0)
	{
		uint32_t tmr = millis() - rotaryTimer;
		//wait 30 mS to process change then reset
		if (tmr > 30)
		{
			if (!sw)
			{
				//sw not pressed, process knob turn
				gMnuValuePrev = 0; //incase it was active
				SetEncoderCtrl(curDir);
			}
			else
			{		
				//sw pressed, change step freq or gMnu
				switch (hmiMode)					
				{
				case HMI_HOME:
					//change step if switch pressed	and encoder turned								
					//Loop() will update freq display when it detects change in gUnit
					if (gVFO != MNU_VFO_MEM)
					{
						if (curDir == -1)
						{
							gUnit *= 10;
							if (gUnit > 100000000)
							{
								gUnit = 1;
							}
						}
						else if (curDir == 1)
						{
							gUnit /= 10;
							if (gUnit < 1)
							{
								gUnit = 100000000;
							}
						}							
					}
					break;
				case HMI_MENU:
					//drop through
					//no action for HMI_VALUE
				case HMI_TXTINPUT:
					if (curDir < 0)
					{
						gBtnPress = BTN_LEFT; //process it in Loop
					}
					else
					{
						gBtnPress = BTN_RIGHT; //process it in Loop
					}
					break;
				}
			}
			curDir = 0;
			rotaryTimer = 0;
			swTimer = 0;
		}
	}
	lastA = a;
	lastB = b;
	lastSw = sw;
	alreadyHr = false;

}


void SetEncoderCtrl(int curDir)
{
	//WARNING!!!
	//This routine is called during the ISR
	//Don't call any routines that can't operate during ISR here!!!
	// NOTE: Tx2HMI() and Tx2ESP() have been fixed to put cmd in hmiTxQue or espTxQue if it's busy in ISR digitalRead(KEY_OUT)
	//
	//change the current gMnu selected control by the value in curDir
	int temp = 0; //temp var for calcs
	uint8_t val = 0; //temp byte value for calcs
	String cmd = ""; //for hmi cmd
	float g = 0;
	static uint32_t lastCmd = 0; //speed up inc/dec of freq if spinning fast
	
	//SerialOut("curDir=" + String(curDir), true);
	
	if (hmiMode == HMI_HOME)
	{
		//change freq		
		//only tx every 30 mS to allow radio to process last transmission
		//change freq by selected unit if we're not locked or PTT is active
		temp = digitalRead(PTT_OUT) + digitalRead(KEY_OUT);
		//SerialOut("temp= " + String(temp) + ", Lock= " + String(gLock), true);
		if (!gLock)
		{
			if (gVFO != MNU_VFO_MEM)
			{
				//ignore if locked, playing gTxBuffer, or PTT or Key is active
				//gUnit change will be detected in Loop() and it will update the freq display
				//don't change freq if gUnit != gUnitPrev because we're in the middle of changing gUnit
				if (!gPlay && !temp && gUnit == gUnitPrev)
				{
					if (millis() - lastCmd < 40)
					{
						curDir *= 10; //speed up if spinning
					}
					lastCmd = millis();
					int32_t newFreq = gFreqRadio + (curDir * gUnit);
					if (newFreq < 100000)
					{
						//we rolled below 100k, move gUnit
						if (gUnit > 100000)
						{
							gUnit = gUnit / 10;
							if (gUnit < 100000)
							{
								gUnit = 100000;
							}
						}
						gSaveRadio = true;
						return;
					}
					//SerialOut(">> New Encoder Freq= " + String(newFreq), true);
					gFreqRadio = uint(newFreq);
					gSaveRadio = true;
				}
			}
			else
			{
				//find next programmed memory			
				FindNextMem(curDir);
			}
		}
	}
	else if (hmiMode == HMI_MENU)
	{		
		hmiMnuChng = true; //force CreateMnu() in Loop
		if (curDir < 0)
		{
			gBtnPress = BTN_UP; //process it in Loop
		}
		else
		{
			gBtnPress = BTN_DOWN; //process it in Loop
		}
	}
	else if (hmiMode == HMI_VALUE || hmiMode == HMI_TXTINPUT)
	{
		hmiMnuChng = false;
		if (curDir < 0)
		{
			gBtnPress = BTN_DOWN; //process it in Loop
		}
		else
		{
			gBtnPress = BTN_UP; //process it in Loop
		}
	}	
	hmiEncoderTimeout = millis(); //restart timeout
}

void FindNextMem(int dir)
{
	//inc (1) or dec (-1) to next valid memory slot
	//find first matching freq in list
	//gMSel range is 1 to 100, memory data stored in 0 to 99

	uint8_t i = 0;
	gMSelPrev = gMSel; //to detect gMSel change

	if (dir == -1)
	{
		//dec memory ptr to next valid memory slot				
		if (gMSel > 1)
		{
			for (i = gMSel-1; i > 0; i--)
			{
				if (vfoMemA_Freq[i - 1] >= 100000)
				{					
					gMSel = i; 
					return;
				}
			}
		}
		//we didn't find one above, try starting from end
		for (i = vfoMemMax; i > gMSel; i--)
		{
			if (vfoMemA_Freq[i - 1] >= 100000)
			{
				gMSel = i;
				return;
			}
		}		
		return; //didn't find anything
	}
	else
	{
		if (gMSel < vfoMemMax + 2)
		{
			for (i = gMSel + 1; i < vfoMemMax + 1; i++)
			{
				if (vfoMemA_Freq[i - 1] >= 100000)
				{
					gMSel = i; //load this rec on exit					
					return;
				}
			}
		}
		//we didn't find one above, try starting from first entry
		for (i = 1; i < gMSel; i++)
		{
			if (vfoMemA_Freq[i - 1] >= 100000)
			{
				//SerialOut("+ from start 1, gMsel= " + String(i), true);
				gMSel = i;
				return;
			}
		}
	}	
}

void ChkPCCmd()
{
	/*
	PC command uses Serial conole to Rx and Tx messages to a PC onnected to the Wio Terminal's micro-USB port
	* Make sure Serial.print calls are minimized to reduce bad replies
	* Any baud can be set on the host program,
	* This enulates the Kenwood TS-2000 xcvr because it will reply with ID019; to ID; requests
	* gComMode determines what external app is connected
	  - once it's triggered it can only toggle between CAT and TERM mode, debug data (SerialOut()) is turned off until a power cylce
	*/

	static String temp = "";
	static uint8_t pcCmdRxCtr = 0; //counts # of rx bytes - must see cmd start with F or M and end with ;
	char rxByte = 0;
	String txt = "";


ChkAgn:
	if (!Serial.available())
	{
		return;
	}

	//data is available on Serial 
	
	delay(1); //wait for other chrs of esc sequence
	uint8_t newChrs = pcCmdRxCtr; //how many new chrs just arrived

	while (Serial.available())
	{
		//add all rx chr to buffer
		rxByte = Serial.read();
		if (rxByte != 10)
		{
			pcCmdRx[pcCmdRxCtr] = rxByte;
			pcCmdRxCtr++;
			if (rxByte == ';' && gComMode == COM_MODE_CAT)
			{
				break; //process each command as they come it
			}
		}		
	}
	newChrs = pcCmdRxCtr - newChrs;
	
	// debug
	/*
	for (int i = 0; i < pcCmdRxCtr; i++)
	{
		SerialOut(String(pcCmdRx[i], HEX) + ".", false);
	}
	SerialOut(", rxByte= " + String(rxByte,HEX) + ", newChrs= " + String(newChrs), true);
	*/
	

	if (newChrs == 1)
	{		
		if (!gComMode && rxByte == 13)
		{
			PrtStatus("Entering Terminal Mode", true);
			gComMode = COM_MODE_TERM;
			pcCmdRxCtr = 0;
			SendTermMenu(); //send start menu on cr
			return;
		}
		else if (gComMode == COM_MODE_TERM)
		{			
			hmiEncoderTimeout = millis(); //refresh timer if we're in terminal mode
			if (rxByte == 9 || rxByte == 13)
			{
				//send cr/lf to terminal
				//TermOut("", true);
			}
			// chk for '/' or '?' - if so, print summary info to terminal
			if ((rxByte == '?' || rxByte == '/') && hmiMode == HMI_MENU)
			{
				//Serial.print(char(8)); //backspace to remove last chr from display
				//SendStatus2Term();
				GetHelp();
				pcCmdRxCtr = 0;
				return;
			}			
			//chk for backspace key
			if (rxByte == 8 && hmiMode == HMI_HOME)
			{
				pcCmdRxCtr--; //delete the delete chr and the last chr
				pcCmdRx[pcCmdRxCtr] = 0;
				if (pcCmdRxCtr)
				{
					pcCmdRxCtr--;
					pcCmdRx[pcCmdRxCtr] = 0;
				}
				Serial.print(char(rxByte)); //just backspace
				if (!pcCmdRxCtr)
				{
					//if keyboard buffer is empty and gTxBuffer has something, delete the last chr
					blockISR = true;
					if (gTxBuffer.length())
					{
						gTxBuffer.remove(gTxBuffer.length() - 1);
					}
					blockISR = false;
				}
				goto ChkAgn;
			}
		}		
	}

	//enable this to capture chr codes from terminal
	//SerialOut(String(int(rxByte)) + ", ctr: " + String(pcCmdRxCtr), true); //prints value of chr rx
	//always accept Enter as terminator.
	if (gComMode == COM_MODE_TERM)
	{
		//normal typing
		if (rxByte == 27 && newChrs == 1)
		{
			//esc pressed
			pcCmdRxCtr = 0;
			if (hmiMode == HMI_HOME && gPlay)
			{
				//stop keyer
				gTxBuffer = "";
				gPlay = KEYER_MODE_STOP;
				ResetSpeedBoost(); 
				TermOut(" ", true);
			}
			else if (hmiMode != HMI_HOME)
			{
				PrtHomePg();
			}
			else
			{
				TermOut(" ", true);
			}
			return;
		}
		if (newChrs == 1)
		{
			//process normal keys
			switch (hmiMode)
			{
			case HMI_HOME:
				//send string to tx buffer
				if (rxByte == 9)
				{
					SendTermMenu(); //tab to bring up menu
					pcCmdRxCtr = 0;
				}					
				else if (rxByte == 13 && pcCmdRxCtr > 1)
				{
					pcCmdRx[pcCmdRxCtr] = 0; //terminate string
					txt = String(pcCmdRx);
					blockISR = true;					
					if (gPlay == KEYER_MODE_PAUSE)
					{
						//insert new string at front of buffer if paused
						gTxBuffer = txt + " " + gTxBuffer;
					}
					else
					{
						//just tack it to the end
						gTxBuffer = gTxBuffer + " " + txt;
					}					
					//don't display new buffer - it corrupts the terminal display
					blockISR = false;
					SaveRepeatBuffer(txt); //so F1 can repeat what we just entered
					if (!gPlay)
					{
						TermOut("\n\n\r[Insert] to pause, [End] to stop ", true);
						TermOut("Send:" + gTxBuffer, true);
						TermOut("    ", false); //space for additional keyboard text to be displayed
					}					
					gPlay = KEYER_MODE_RUN;					
					pcCmdRxCtr = 0; //reset buffer
					return;
				}
				else
				{
					//Send to terminal 			
					if (rxByte == 13)
					{
						pcCmdRxCtr = 0; //reset counter
					}
					keyboardChr = rxByte; 	
				}
				break;
			case HMI_MENU:
				//just wait for cr
				if (rxByte == 13)
				{
					//select current menu option
					ProcessBtn(BTN_A);										
				}				
				pcCmdRxCtr = 0;			
				return;
				break;
			case HMI_VALUE:
				if (rxByte == 13)
				{
					//UpdateParam();
					ProcessBtn(BTN_A); //accept new value														
				}	
				pcCmdRxCtr = 0;
				return;
				break;
			case HMI_TXTINPUT:
				if (rxByte == 13)
				{									
					TermOut(" ", true); //white space		
					ProcessBtn(BTN_A); //accept new txt												
				}
				else
				{					
					//update txtInput array					
					if (rxByte == 8)
					{
						//del current chr
						txtInput[txtInputPtr] = ' ';
						ProcessTxInputBtn(BTN_LEFT); //dec pointer
					}
					else
					{
						txtInput[txtInputPtr] = char(rxByte);
						ProcessTxInputBtn(BTN_RIGHT); //inc pointer
						TermOut(String(char(rxByte)), false);
					}										
				}
				pcCmdRxCtr = 0; //only enter on chr at a time
				return;
				break;
			}
			//single chr, just send it to ChkChr2HMI()		
			keyboardChr = rxByte;
	
		}
		else
		{			
			//look for multiple function key codes with preamble 27 91 
			//SerialOut("# of chrs= " + String(pcCmdRxCtr), true);
			if (pcCmdRx[0] == 27)
			{				
				if (((pcCmdRxCtr == 1 || pcCmdRxCtr == 2) && !Serial.available()) || (pcCmdRx[1] == 91 && pcCmdRxCtr > 2))
				{					
					ProcessTerminalCmd(pcCmdRxCtr); //process function keys
				}
				pcCmdRxCtr = 0;
				return;
			}			
			//just accumulate chrs in pcCmdRx array
		}
		goto ChkAgn;
	}

	//if we get to here, chk for cat commands
	
	if (rxByte >= 'a' && rxByte <= 'z')
	{
		rxByte &= 0xDF; //convert to upper case
	}
	
	/*
	if (pcCmdRxCtr == 1)
	{
		if (rxByte != 'A' && rxByte != 'F' && rxByte != 'I' && rxByte != 'M' && rxByte != 'P' && rxByte != 'R' && rxByte != 'T')
		{
			//ignore, allow any chrs through
			//goto ChkAgn;
		}
	}
	*/

	//ignore lf/cr
	if (rxByte == char(13))
	{
		if (!gComMode)
		{
			if (pcCmdRxCtr)
			{
				pcCmdRxCtr--; //dec ctr to ignore it
			}
			goto ChkAgn;
		}			
	}
		
	if (pcCmdRxCtr > pcCmdRxLen)
	{
		pcCmdRxCtr = 0; //start over if cmd is too long
		goto ChkAgn;
	}

	if (rxByte == ';')
	{		
		pcCmdRx[pcCmdRxCtr] = '\0'; //terminate string

		if (gComMode != COM_MODE_CAT)
		{
			PrtStatus("CAT Mode started", true); //notify user 
			gComMode = COM_MODE_CAT; //block diagnostic reports on Serial() - they interfer with external apps trying to talk to this device
		}
		
		// debug code to see what is coming in
		//code to display incoming traffic in bRx.txt since we can't use serial monitor
		/*
		if (pcCmdRx[0] != 'F')
		{
			for (int i = 0; i < pcCmdRxCtr; i++)
			{
				temp += char(pcCmdRx[i]);
			}
			PrtStatus(temp, true);			
			temp = temp + " ";
			if (temp.length() > 30)
			{
				temp.substring(temp.length() - 30, temp.length());
			}
		}
		*/
		

		//we have received a full command - process it
		char sFrq[12] = { '0', '0', '0', '0', '0', '0', '0', '0', '0', '0', '0', 0x00 }; //put frq digits here
		//PrtStatus("Cmd: " + String(pcCmdRx) + " - " + String(pcCmdRxCtr), true);
		if (pcCmdRx[0] == 'F')
		{			
			if (pcCmdRx[1] == 'A' || pcCmdRx[1] == 'B' || pcCmdRx[1] == 'C')
			{
				//vfo A or b, or subVFO freq query or set
				if (pcCmdRxCtr == 3)
				{
					//freq request, return gFreq
RtnFrq:				CalcKenwoodFreq(sFrq);					
					char buf[15] = {'F', pcCmdRx[1], '0', '0', '0', '0',  '0',  '0',  '0',  '0',  '0',  '0',  '0', ';',  0 };
					for (int i = 0; i < 11; i++)
					{
						buf[i + 2] = sFrq[i];
					}
					Serial.write(buf, 14); //reply to selected cmd							
					//PrtStatus("Cmd: " + String(buf), true);							
				}
				else if (pcCmdRxCtr == 14)
				{
					//set gFreq					
					if (!gLock && !gPlay)
					{
						gFreqRadio = ((pcCmdRx[4] & 0x0f) * 100000000) + ((pcCmdRx[5] & 0x0f) * 10000000) + ((pcCmdRx[6] & 0x0f) * 1000000) + ((pcCmdRx[7] & 0x0f) * 100000) + ((pcCmdRx[8] & 0x0f) * 10000)
							+ ((pcCmdRx[9] & 0x0f) * 1000) + ((pcCmdRx[10] & 0x0f) * 100) + ((pcCmdRx[11] & 0x0f) * 10) + (pcCmdRx[12] & 0x0f);
						ChkFreqModeChng();
						//PrtStatus("Set Freq= " + String(gFreqRadio), true);
						goto RtnFrq;
					}
				}
			}
		}
		else if (pcCmdRx[0] == 'M' && pcCmdRx[1] == 'D')
		{
			//mode query or set
			if (pcCmdRxCtr == 3)
			{
				//calc mode digit and return mode
				char mdDigit = 0;
				switch (gRMode) 
				{
				case 0: //cw
					mdDigit = '3';
					break;
				case 1: //cwr
					mdDigit = '7';
					break;
				case 2: //lsb
					mdDigit = '1';
					break;
				case 3: //usb
					mdDigit = '2';
					break;
				case 4: //fmn
					mdDigit = '4';
					break;
				case 5: //fmw - force to fmn
					mdDigit = '4';
					break;
				case 6: //am
					mdDigit = '5';
					break;
				case 7: //digL - force to fsk reverse
					mdDigit = '6';
					break;
				case 8: //digH - force to fsk
					mdDigit = '9';
					break;
				}
				char buf[5] = { 'M', 'D', mdDigit, ';', 0 };
				Serial.write(buf, 4); //reply to selected cmd	
				//PrtStatus("Cmd: " + String(buf), true);
			}
			else if (pcCmdRxCtr == 4)
			{
				//translate kenwood mode to CTR2 mode and set gRMode
				int mdDigit = pcCmdRx[2] - 0x30; //convert ascii to number
				switch (mdDigit)
				{
				case 3:
					gRMode = RADIO_CW;
					break;
				case 7:
					gRMode = RADIO_CW_R;
					break;
				case 1:
					gRMode = RADIO_LSB;
					break;
				case 2:
					gRMode = RADIO_USB;
					break;
				case 4:
					gRMode = RADIO_FM_N;
					break;
				case 5:
					gRMode = RADIO_AM;
					break;
				case 6:
					gRMode = RADIO_DIG_L; //NOTE: if radio doesn't support these digL and digR modes, it will revert to previous mode
					break;
				case 9:
					gRMode = RADIO_DIG_H;
					break;
				}
				char buf[5] = { 'M', 'D', char(mdDigit + 0x30), ';', 0 }; //just reply what we were told, MD; will get actual radio mode
				Serial.write(buf, 4); //reply to selected cmd					
				//PrtStatus("Cmd: " + String(buf), true);
				ChkFreqModeChng(); //send mode change to radio immediately so radio mode doesn't revert hmi back to previous mode
				PrtRMode(true);
			}
		}
		else if (pcCmdRx[0] == 'R' && pcCmdRx[1] == 'X')
		{
			if (gTxMode > RADIO_TX_OFF)
			{
				//set rx mode
				gTxMode = RADIO_TX_ENAB; //make sure they're different
				ControlPTT(LOW); //release PTT			
				char buf[5] = { 'R', 'X', '0', ';', 0 };
				Serial.write(buf, 4); //reply to selected cmd					
				//PrtStatus("Cmd: " + String(pcCmdRx) + "  Reply: " + String(buf), true);
				PrtBtnA(true);
			}
		}
		else if (pcCmdRx[0] == 'T' && pcCmdRx[1] == 'X')
		{
			//if Tx is enabled and Use CAT PTT is enabled
			if (gTxMode > RADIO_TX_OFF)
			{
				gTxMode = RADIO_TX_PTT;							
				ControlPTT(HIGH); //key ptt
				//wsjt-x doesn't send TX#;, just TX;
				char buf[5] = { 'T', 'X', '0', ';', 0 };
				if (pcCmdRxCtr == 4)
				{
					buf[2] = pcCmdRx[2]; // add the tx # if request included it
				}					
				Serial.write(buf, 4); //reply to selected cmd					
				//PrtStatus("Cmd: " + String(pcCmdRx) + "  Reply: " + String(buf), true);
				PrtBtnA(true);
			}
		}
		else if (pcCmdRx[0] == 'I' && pcCmdRx[1] == 'D')
		{
			//reply with TS2000 ID 019#
			//ts-2000 is 019, 440 is 004, 480 is 020
			char buf[7] = { 'I', 'D', '0', '1', '9', ';', 0 };
			Serial.write(buf, 6);	
			//PrtStatus("Cmd: " + String(buf), true);
		}
		else if (pcCmdRx[0] == 'P' && pcCmdRx[1] == 'S')
		{
			//reply power switch is on - wsjt-x asks for this		
			char buf[5] = { 'P', 'S', '1', ';', 0 };
			Serial.write(buf, 4);			
			//PrtStatus("Cmd: " + String(buf), true);
		}
		else if (pcCmdRx[0] == 'A' && pcCmdRx[1] == 'I')
		{
			//auto information is off - wsjt-x asks for this	
			char buf[5] = { 'A', 'I', '0', ';', 0 };
			Serial.write(buf, 4);			
			//PrtStatus("Cmd: " + String(buf), true);
		}
		else if (pcCmdRx[0] == 'T' && pcCmdRx[1] == 'O')
		{
			//tone is off - wsjt-x asks for this	
			char buf[5] = { 'T', 'O', '0', ';', 0 };
			Serial.write(buf, 4);			
			//PrtStatus("Cmd: " + String(buf), true);
		}
		else if (pcCmdRx[0] == 'I' && pcCmdRx[1] == 'F')
		{
			//reply if info packet
			char buf[40];
			for (int i = 0; i < 40; i++)
			{
				buf[i] = 0x30; //preload with ascii 0s
			}
			buf[0] = 'I';
			buf[1] = 'F';
			CalcKenwoodFreq(sFrq);
			for (int i = 0; i < 11; i++)
			{
				buf[i + 2] = sFrq[i]; //freq
			}
			uint8_t ptt = !(digitalRead(PTT_OUT)); //invert
			buf[28] = ptt + 0x30; //ptt convert to ascii
			buf[29] = gRMode + 0x30;
			buf[37] = ';';
			buf[38] = 0; //make sure we're terminated
			Serial.write(buf, 38);				
			//PrtStatus("Cmd: " + String(buf), true);
		}
		//Serial.flush();
		pcCmdRxCtr = 0; //reset for next command
	}
	delay(1);
	goto ChkAgn;
}

void ControlPTT(bool state)
{
	//ptt and key are active when HIGH
	static bool lastState = false;
	if (gTxMode)
	{
		if (gEnabAmpSeq)
		{
			//using Amp/Radio sequencing
			if (state)
			{
				// we just keyed ptt
				digitalWrite(PTT_OUT, state); //pick up PTT first to enable amp
				delay(gSeqPre); //use pre-seq timer
				digitalWrite(KEY_OUT, state); //then key radio PTT send RF to amp
			}
			else
			{
				// we just unkeyed ptt
				digitalWrite(KEY_OUT, state); //drop radio PTT first to drop RF to amp
				delay(gSeqPost); //use pre-seq timer
				digitalWrite(PTT_OUT, state); //then unkey amp Tx Enable
			}
		}
		else
		{
			if (gRadioCAT != RADIO_CAT_FLEX || !gFlexApiPTT)
			{
				//if not Flex or we're not using API for PTT, control hardware PTT
				digitalWrite(PTT_OUT, state); //just control PTT to radio
			}
			else
			{
				//must be online
				if (gTCPCon)
				{
					Tx2WiFi(WIFI_FLEX, "|xmit " + String(state));
				}
				else
				{
					gBtnPTT = false;
					state = false;
					lastState = false;
					gPTTState = false;
					PrtStatus("Flex radio Offline", true);
				}
			}
		}

		if (state != lastState)
		{
			if (gRmtMode != MNU_RMT_KEY)
			{
				if (state)
				{
					PrtStatus("PTT ACTIVE", true);
				}
				else
				{
					PrtStatus("PTT Released", true);
				}
			}
			gPTTState = state;
			if (!mnuCBtnSel)
			{
				PrtBtnC();
			}
			lastState = state;
		}
	}
	else
	{
		gPTTState = false;
		lastState = false;
		digitalWrite(PTT_OUT, LOW);
	}
}

void ChkRmtFuncBtns()
{
	//read RmtPTT ADC and update gRmtFuncBtn with results
	/*
	Rmt PTT can have a resistive ladder for different function keys
	Each value has about a 125 count zone to be valid

	NOTE: Ranges in the code represent the generic resistance values - change if we go with precise values on the board build

											 Precise		Generic
	Volts	Counts	Function		func    Resistance	   Resistance
	0.0v 	<50		PTT				n/a
	0.4v =	124		Reset TxEnab	0		  1.38K			1.38K
	0.8v =	248		F1 btn			1         3.2K          2.2K + 1K
	1.6v =	372		F2 btn			2         5.715K        4.7K + 1K
	2.0v =	496		F3 btn			3         9.41K         10K
	2.4v =  620		F4 btn			4         15.385K       10K + 4.7K
	2.8v =	744		F5 btn			5         26.67K		22K + 4.7K
	*/
		
	uint16_t adc = 0;
	uint16_t btnPressed = 870; //anything below this is valid
	uint16_t lastADC = 0;
	String txt = "";

	//chk for stuck button on startup - this will catch a bad ADC port including PTT levels
	if (!gStuckBtn && gRmtMode != MNU_RMT_KEY)
	{
		for (int i = 0; i < 300; i++)
		{
			adc = analogRead(RMT_BTNS); //current value
			if (adc > btnPressed)
			{
				gStuckBtn = -1;
				goto Start;
			}
			delay(10);
		}
		gStuckBtn = 1; //disable func buttons
		hmiTagLineRefresh = 0;
		PrtStatus("FUNCTION KEY DISABLED", true);
		delay(1500);
		hmiTagLineRefresh = 0;
		PrtStatus("FUNCTION KEY LOCKOUT", true);
		SerialOut("Function key stuck - disabling Rmt Function Keypad", true);
		return;
	}

Start:
	if (gRmtMode == MNU_RMT_KEY || gStuckBtn> 0)
	{
		//service analogRead(RMT_BTNS) in ISR or button stuck > 5 sec - this can happen on Flex-Only unit
		return; 
	}
	
	int ctr = 0;
	
	adc = analogRead(RMT_BTNS); //current value
	//SerialOut(String(adc) + "  ", false);
	if (adc < btnPressed)
	{
		lastADC = 0;
		for (int i = 0; i < 15; i++)
		{
			delay(2); //debounce and average for 30 mS			
			adc = analogRead(RMT_BTNS); //get it again
			lastADC = (lastADC + adc) / 2;
			if (lastADC > btnPressed)
			{
				return;
			}
		}
		//SerialOut("lastADC= " + String(lastADC) + ", adc= " + String(adc) + " = " + String(lastADC - adc), true);
		//must be stable within a 10 count average window
		if ((adc > lastADC - 5) && (adc < lastADC + 5) && lastADC < btnPressed)
		{
			if (adc < 50)
			{
				gRmtFuncBtn = RMT_FUNC_PTT;
			}
			if (gRmtFuncBtn == RMT_FUNC_OFF)
			{
				//give the first level more room - it's switched through diodes so light pressure on the switch can cause higher than normal values
				if (adc < 220)
				{
					gTxMode = !gTxMode;
					SendASCCmd(ASC_CMD_TXMODE, 0, 0);
					gRmtFuncBtn = RMT_FUNC_OFF; //disable Tx Enable (the other Mini was just enabled)	
					txt = "Tx Enable Btn Pressed";
					SerialOut("adc= " + String(adc), true);
				}
				else if (adc < 341)
				{
					gRmtFuncBtn = RMT_FUNC_F1; //f1							
					txt = "F1 Pressed";
				}
				else if (adc < 496)
				{
					gRmtFuncBtn = RMT_FUNC_F2; //f2						
					txt = "F2 Pressed";
				}
				else if (adc < 620)
				{
					gRmtFuncBtn = RMT_FUNC_F3; //f3						
					txt = "F3 Pressed";
				}
				else if (adc < 744)
				{
					gRmtFuncBtn = RMT_FUNC_F4; //f4
					txt = "F4 Pressed";
				}
				else if (adc < btnPressed)
				{
					gRmtFuncBtn = RMT_FUNC_F5; //f5
					txt = "F5 Pressed";
				}
				else
				{
					return;
				}
			}
			if (adc > 50)
			{
				//exit immediately on PTT, otherwise wait for btn release
				//keep chking for stuck function buttons above PTT				
				hmiTagLineRefresh = 0; //forces update of txt in alert
				PrtStatus(txt, true);
				lastADC = adc;
				while (adc < (lastADC + 30))
				{
					//wait for btn to be released or time out					
					adc = analogRead(RMT_BTNS);
					delay(10);
					ctr++;
					if (ctr > 500)
					{
						//button stuck or we're on a Flex-Only unit - lock out function keypad
						gStuckBtn = 1;
						hmiTagLineRefresh = 0;
						PrtStatus("FUNCTION KEY TIMEOUT", true); //reset current alert if any
						delay(1500);
						hmiTagLineRefresh = 0;
						PrtStatus("FUNCTION KEY LOCKOUT", true);
						SerialOut("Function key stuck - disabling Rmt Function Keypad", true);
						gRmtFuncBtn = RMT_FUNC_OFF;
						gTxMode = false; //disable Tx Enable if it was forced on by a stuck button
						SendASCCmd(ASC_CMD_TXMODE, 0, 0);
						return;
					}
				}
			}
			delay(15);
		}
	}
	else
	{
		gRmtFuncBtn = RMT_FUNC_OFF;
	}
}

void ResetSpeedBoost()
{
	//reset speed when we stop
	if (speedToggle)
	{
		speedToggle = false;
		keyerDitDur = keyerDitDurPrev;
		keyerDitFarnsDur = keyerDitFarnsDurPrev;
	}
}

void SwapAnt()
{
	//Swap A/B antenna assignments if we're active an not in 1x8 mode
	if (gAntSwOnline && gAntSwActive && gAntSwDual)
	{
		if (gAntSwA && gAntSwB)
		{			
			SendASCCmd(ASC_CMD_ANT_SWAP, 0,0);
			PrtHomePg();
			PrtStatus("Swapping A and B Ants", true);
		}
		else
		{
			if (!gAntSwA)
			{
				PrtStatus("PORT-A SWITCH IS OPEN", true);
			}
			else
			{
				PrtStatus("PORT-B SWITCH IS OPEN", true);
			}
		}
	}
	else
	{
		PrtStatus("ANT SWITCH IS OFFLINE", true);
	}
}