// =====================================================================================================================================================================================================
// (c) 2022 Lynn Hansen, KU7Q															                                                                                                               |
// This Source Code Form is subject to the terms of the GNU GENERAL PUBLIC LICENSE, Version 3, 29 June 2007. A copy of this license can be found here: https://choosealicense.com/licenses/gpl-3.0/|
// =====================================================================================================================================================================================================




//This module contains code to support Bluetooth BLE

void OpenBLEServer()
{
	uint32_t zero = 0;

	PrtStatus("Starting BLE Server", true);

	BLEDevice::init("CTR2-Mini-BLE-Server");
	pServer = BLEDevice::createServer();
	pServer->setCallbacks(new MyServerCallbacks());

	pService = pServer->createService(servUUID);
	pLocalParams = pService->createCharacteristic
	(
		charUUID,
		BLECharacteristic::PROPERTY_READ |
		BLECharacteristic::PROPERTY_WRITE |
		BLECharacteristic::PROPERTY_NOTIFY

	);
	pLocalParams->setAccessPermissions(GATT_PERM_READ | GATT_PERM_WRITE);
	pLocalParams->setCallbacks(new MyLocalCallback());

	/* example of setting a label - it displays in Hex
	BLEDescriptor* pCharFrqLbl = pLocalParams->createDescriptor(
		charUUID,
		ATTRIB_FLAG_VOID | ATTRIB_FLAG_ASCII_Z,
		GATT_PERM_READ | GATT_PERM_WRITE, 4
	);
	pCharFrqLbl->setValue("Freq");
	*/

	//how to set a string=value pair
	//std::string str = "gTxMode=";// +String(gTxMode);
	//str += gTxMode + 0x30;		

	pService->start();

	gBLEHandle = pService->getHandle();

	//BLEAdvertising *pAdvertising = pServer->getAdvertising();
	pAdvertising = BLEDevice::getAdvertising();
	pAdvertising->addServiceUUID(servUUID);

	pAdvertising->setScanResponse(true);
	pAdvertising->setMinPreferred(0x06);  // functions that help with iPhone connections issue
	pAdvertising->setMinPreferred(0x12);
	BLEDevice::startAdvertising();
	pAdvertising->start();	
	
	gBLEConnected = true;
	PrtStatus("BLE Server Online", true);

}

void CloseBLEServer()
{	
	//pServer->disconnect(gBLEHandle);
	pService->stop();		
	ResetLink();
	gBLEConnected = false;
	PrtStatus("BLE Server Offline", true);	
}

void SendLinkParams()
{
	//sets pLocalParam values
	//val = __builtin_bswap32(gFreqRadio); //force to big endian	

	//Byte 1=flags, 2-5=freq, 6-9=offset, 10-13=RIT, 14=cw speed, 15=farns speed, 16=key type, 17-18=contest sn
	//receiving end will determine actions based on flags
	static bool alreadyHr = false;

	if (gLinkChngTimerDefault > millis() - gLinkRxChngTimer)
	{	
		return; //just received new freq, don't repeat it back to other Mini
	}

	if ((!gBLEConnected && !gUDPConnected) || alreadyHr || !gUDPAdrs.length())
	{
		return;
	}
	alreadyHr = true;

	uint8_t ary[gNumLinkBytes];
	for (int i = 0; i < gNumLinkBytes; i++)
	{
		ary[i] = 0;
	}

	uint32_t newFreq = gFreqRadio;

	if (!gLinkMiniA)
	{
		ary[0] = LINK_PREAMBLE_CLIENT;
		ary[gNumLinkBytes - 1] = LINK_POSTAMBLE_CLIENT; //mini b - client
	}
	else
	{
		ary[0] = LINK_PREAMBLE_SVR;
		ary[gNumLinkBytes - 1] = LINK_POSTAMBLE_SVR; //mini a - server
	}

	if (gSyncMode == SYNC_MODE_FREQ)
	{
		if (gTxMode && gRadioCAT != RADIO_CAT_PCR1000)
		{
			//Tx is enabled on this unit, send offset freq to remote end so we don't listen to other radio's Tx
			if (newFreq > 15000000)
			{
				newFreq -= 10000000;
			}
			else
			{
				newFreq += 10000000;
			}
			if (gSyncOffset)
			{
				gFreqRadio = newFreq;
				TxFreq2Radio(gFreqRadio);
				gFreqRadioPrev = gFreqRadio;
				//SerialOut("offset removed from freq 1", true);
			}
			gSyncOffset = false;
		}
		else if (!gTxMode && gTxMode != gTxModePrev)
		{
			//just went off
			if (gSyncOffset)
			{
				if (newFreq > 15000000)
				{
					newFreq -= 10000000;
				}
				else
				{
					newFreq += 10000000;
				}
			}
			gSyncOffset = false;
			//SerialOut("tx disabled, reset offset", true);
		}
	}

	//NOTE: The server and the client operate with the same set of data - delineated by the sync bytes

	uint8_t pcrFlags = (gPCRHasDSP << 6) + (gPCRDSP << 5) + (gPCRANch << 4) + (gPCRAGC << 3) + (gPCRAtten << 2) + (gPCRNB << 1);
	uint8_t flexFlags = (gFlexApiPTT << 8) + (gFlexApiKey << 7); //add binaries here
	
	//bits 5 & 4 are push/pull flag
	//bits 3 & 2 are syncmode
	ary[LINK_FLAGS] = (gVFO_PushPull << 4) + (gSyncMode << 2) + (gSideTone << 1);
	gVFO_PushPull == SYNC_MODE_BASIC_NONE; //reset it so we only send it once

	if ((gSyncMode == SYNC_MODE_RMT || gRadioCAT != RADIO_CAT_PCR1000) && gTxMode)
	{
		ary[LINK_FLAGS] |= 1; //or in gTxMode - far end will disable Tx mode if this end is active unless we're in SYNC_MODE_RMT
	}
	ary[LINK_FREQ] = newFreq & 0xff;
	ary[LINK_FREQ + 1] = (newFreq & 0xff00) >> 8;
	ary[LINK_FREQ + 2] = (newFreq & 0xff0000) >> 16;
	ary[LINK_FREQ + 3] = (newFreq & 0xff000000) >> 24;
	ary[LINK_BAND] = gBand & 0xff;
	ary[LINK_RMODE] = gRMode & 0xff;
	ary[LINK_RIT] = gRIT & 0xff;
	ary[LINK_RIT + 1] = (gRIT & 0xff00) >> 8;
	ary[LINK_RIT + 2] = (gRIT & 0xff0000) >> 16;
	ary[LINK_RIT + 3] = (gRIT & 0xff000000) >> 24;
	ary[LINK_CW_SPEED] = gCWSpeed & 0xff;
	ary[LINK_FARNS_SPEED] = gFarnsSpeed & 0xff;
	ary[LINK_KEY_TYPE] = gKeyType & 0xff;
	ary[LINK_KEYER_SN] = mnuKeyerSN & 0xff;
	ary[LINK_KEYER_SN + 1] = (mnuKeyerSN & 0xff00) >> 8;
	//ary[LINK_FUTURE1] = something we forgot to add :)
	
	if (gRadioCAT == RADIO_CAT_PCR1000)
	{
		ary[LINK_AUX1] = pcrFlags;
		ary[LINK_AUX2] = gPCRVol & 0xff;
		ary[LINK_AUX3] = gPCRSq & 0xff;
		ary[LINK_AUX4] = gPCRIF & 0xff;
		ary[LINK_AUX5] = gPCRDSPNR & 0xff;
		ary[LINK_AUX6] = gPCRBW & 0xff;
	}
	else if (gRadioCAT == RADIO_CAT_FLEX)
	{
		ary[LINK_AUX1] = flexFlags;
		ary[LINK_AUX2] = gFlexSlice & 0xff;
		ary[LINK_AUX3] = gFlexVol & 0xff;
		ary[LINK_AUX4] = (gFlexDelay & 0xff00)>>8; //16 bit msb
		ary[LINK_AUX5] = gFlexDelay & 0xff; //lsb
		ary[LINK_AUX6] = gFlexPO & 0xff;
		ary[LINK_AUX7] = gFlexTO & 0xff;
		ary[LINK_AUX8] = gFlexWNB & 0xff;
		ary[LINK_AUX9] = gFlexNB & 0xff;
		ary[LINK_AUX10] = gFlexNR & 0xff;
		ary[LINK_AUX11] = gFlexAPF & 0xff;
		ary[LINK_AUX12] = gFlexANF & 0xff;
		ary[LINK_AUX13] = gFlexATU & 0xff;
	}

	unsigned short crcVal = crc.XModemCrc(ary, 0, gNumLinkBytes - 4); //calc CRC on all bytes up to CRC MSB
	ary[gNumLinkBytes - 3] = (crcVal & 0xff00) >> 8;
	ary[gNumLinkBytes - 2] = crcVal & 0xff;
	
	/*
	SerialOut("ary[LINK_FLAGS]= " + String(ary[1], HEX), true);
	for (int i = 0; i < gNumLinkBytes; i++)
	{
		SerialOut(String(ary[i], HEX) + ":", false);
	}
	SerialOut(" ", true);
	*/

	if (gBLEConnected)
	{		
		if (gLinkMiniA)
		{
			pLocalParams->setValue(ary, gNumLinkBytes);
		}
		else
		{
			pRmtParams->writeValue(ary, gNumLinkBytes); //always send it, server will ignore it if it's not in sync or track mode
		}
	}
	else if (gUDPConnected)
	{
		udp.beginPacket(udpAdrs, udpPort);
		udp.write(ary, gNumLinkBytes);		
		udp.endPacket();
		udp.flush();				
	}

ExitHr:
	alreadyHr = false;
}

void GetLinkParams()
{
	if (!gBLEConnected && !gUDPConnected && !gBLEKillScan)
	{
		return;
	}
	int numBytes = gNumLinkBytes;

	uint8_t ary[gNumLinkBytes]; //use 0xff as validation flag

	for (int i=0; i<gNumLinkBytes; i++)
	{
		ary[i] = 0;
	}
	unsigned short crcRx = 0;
	unsigned short crcVal = 0;
	
	std::string msg = "";	

	static bool alreadyHr = false;

	if (alreadyHr)
	{
		return;
	}
	alreadyHr = true;

	gLinkRxData = false; //reset the flag

	if (!gBLEKillScan)
	{
		if (gBLEConnected)
		{
			if (gLinkMiniA)
			{
				msg = pLocalParams->getValue();
				if (msg.length() == gNumLinkBytes)
				{
					if (msg.at(LINK_PREAMBLE) == LINK_PREAMBLE_CLIENT && msg.at(gNumLinkBytes - 1) == LINK_POSTAMBLE_CLIENT)
					{
						//only accept client updates
						goto GotOne;
					}
				}
			}
			else
			{
				if (pRmtParams->canRead())
				{
					pRmtParams = pRmtService->getCharacteristic(charUUID);
					msg = pRmtParams->readValue();
					//only accept server updates
					if (msg.length() == gNumLinkBytes)
					{
						if (msg.at(LINK_PREAMBLE) == LINK_PREAMBLE_SVR && msg.at(gNumLinkBytes - 1) == LINK_POSTAMBLE_SVR)
						{							
							goto GotOne;
						}
					}
				}
			}
		}
		else if (gUDPConnected && udp.available())
		{			
			int ctr = 0;
			msg = "";
			while (udp.available())
			{
				msg += udp.read();
				//wait until we get a vaild start preamble 
				if (!ctr && msg.at(LINK_PREAMBLE) != LINK_PREAMBLE_ASC && msg.at(LINK_PREAMBLE) != LINK_PREAMBLE_SVR && msg.at(LINK_PREAMBLE) != LINK_PREAMBLE_CLIENT)
				{
					msg = "";					
				}
				else
				{
					ctr++;
				}
			}
			if ((ctr == gNumLinkBytes) && ((gLinkMiniA && msg.at(LINK_PREAMBLE) == LINK_PREAMBLE_CLIENT && msg.at(gNumLinkBytes - 1) == LINK_POSTAMBLE_CLIENT) || \
				(!gLinkMiniA && msg.at(LINK_PREAMBLE) == LINK_PREAMBLE_SVR && msg.at(gNumLinkBytes - 1) == LINK_POSTAMBLE_SVR)))
			{	
				goto GotOne;
			}
			else if ((ctr == gNumASCLinkBytes) && (msg.at(LINK_PREAMBLE) == LINK_PREAMBLE_ASC && msg.at(gNumASCLinkBytes - 1) == LINK_POSTAMBLE_ASC))
			{				
				goto GotOne;
			}
		}
	}
	else
	{
		//just connected, kill the scan and connect as client
		//SerialOut("Killing BLE scanner", true);
		BLEDevice::getScan()->start(0); //scan completed, kill it
		gBLEKillScan = false;
		gBLEConnected = ConnectToBLEServer();
		if (!gBLEConnected)
		{
			gBLEConnect = false;
		}		
	}
	goto ExitHr;

// Process valid data ===========================================================

GotOne:
	//ignore this packet if we just sent a change, just reset timeout
	if (gLinkChngTimerDefault > millis() - gLinkTxChngTimer)
	{
		if (msg.at(LINK_PREAMBLE) != LINK_PREAMBLE_ASC)
		{
			gLinkCommTimeout = millis(); //reset comm fail timeout
			if (gLinkCommFail)
			{
				gLinkCommFail = false;				
				PrtWiFi();
			}			
			goto ExitHr; 
		}		
	}
	//determine # of bytes based on message
	if (msg.at(LINK_PREAMBLE) == LINK_PREAMBLE_ASC)
	{
		numBytes = gNumASCLinkBytes;
	}

	crcRx = (msg.at(numBytes - 3) << 8) + msg.at(numBytes - 2);	//sent CRC

	for (int i = 0; i < numBytes - 4; i++)
	{
		ary[i] = msg.at(i);
	}
	crcVal = crc.XModemCrc(ary, 0, numBytes - 4); //calc CRC on all bytes up to CRC MSB
	//SerialOut("crcRx= " + String(crcRx, HEX) + ", crcVal= " + String(crcVal, HEX), true);

	if (crcRx == crcVal)
	{			
		/* debug
		switch (msg.at(LINK_PREAMBLE))
		{
		case LINK_PREAMBLE_ASC:
			SerialOut("Rx ASC update " + String(millis()), true);
			break;
		case LINK_PREAMBLE_SVR:
			SerialOut("Rx Server update " + String(millis()), true);
			break;
		case LINK_PREAMBLE_CLIENT:
			SerialOut("Rx Client update " + String(millis()), true);
			break;
		}
		*/

		if (msg.at(LINK_PREAMBLE) != LINK_PREAMBLE_ASC)
		{
			//normal link message
			if (gLinkCommFail)
			{
				gLinkCommFail = false;
				gLinkMsgPrev = ""; //so next block works
				PrtWiFi();
			}
			gLinkCommTimeout = millis(); //reset comm fail timeout
			if (msg != gLinkMsgPrev)
			{
				DecLinkMsg(msg);
				gLinkMsgPrev = msg;
			}
		}
		else if (msg.at(LINK_PREAMBLE) == LINK_PREAMBLE_ASC)
		{
			if (!gAntSwOnline)
			{
				gAntSwOnline = true;
				gLinkASCMsgPrev = ""; //so next block works
				PrtWiFi();
			}
			gLinkASCCommTimeout = millis();
			if (msg != gLinkASCMsgPrev)
			{
				DecLinkMsgASC(msg);
				gLinkASCMsgPrev = msg;
			}
		}
	}
	else
	{
		SerialOut("Invalid CRC, rx=" + String(crcRx, HEX) + ", chk=" + String(crcVal, HEX), true);
	}
		
ExitHr:
	alreadyHr = false;
	
}

void DecLinkMsg(std::string msg)
{

	/*
	for (int i = 0; i < gNumLinkBytes; i++)
	{
		SerialOut(String(msg.at(i), HEX) + ":", false);
	}
	SerialOut(" ", true);
	*/
	bool newSideTone = false;
	uint32_t newFreq = 0;
	uint8_t newBand = 0;
	uint8_t newRMode = 0;
	int newRIT = 0; //only used in Rmt Sync mode
	uint8_t newCWSpeed = 0;
	uint8_t newFarnsSpeed = 0;
	uint8_t newKeyType = 0;
	uint16_t newKeyerSN = 0;

	//if we just went online, reset all offsets and acs to start fresh
	if (!gLinkOnline)
	{
		ResetLink();
		gLinkOnline = true;
	}

	String txt = "";
	uint16_t newVal = 0; //used to decode aux values
	static uint16_t oldVal = 0;

	newSideTone = ((msg.at(LINK_FLAGS) & 2) >> 1);
	if (gSideTone != newSideTone)
	{
		gSideTone = newSideTone;
		PrtOptions();
	}

	if (!gLinkMiniA)
	{
		gSyncMode = ((msg.at(LINK_FLAGS) & 8) >> 2) + ((msg.at(LINK_FLAGS) & 4) >> 2);
		
		//SerialOut("msg.at(1)= " + String(msg.at(LINK_FLAGS), HEX) + ", gSyncMode= " + String(gSyncMode), true);
		if (gSyncMode != SYNC_MODE_TRACK)
		{
			gSyncOffset = false;
			gLinkTrackOffset = 0;
		}
		PrtWiFi();
	}
	
	//decode bytes

	for (int i = LINK_FREQ; i < LINK_FREQ + 4; i++)
	{
		newFreq += msg.at(i) << ((i - LINK_FREQ) * 8);
	}	

	newBand = msg.at(LINK_BAND);
	newRMode = msg.at(LINK_RMODE);

	for (int i = LINK_RIT; i < LINK_RIT + 4; i++)
	{
		newRIT += msg.at(i) << ((i - LINK_RIT) * 8);
	}	
	newCWSpeed = msg.at(LINK_CW_SPEED);
	newFarnsSpeed = msg.at(LINK_FARNS_SPEED);
	newKeyType = msg.at(LINK_KEY_TYPE);
	newKeyerSN = msg.at(LINK_KEYER_SN) + (msg.at(LINK_KEYER_SN + 1) << 8);
		
	//don't set gTxMode if we're running a PCR1000 - it's used for power
	if (gRadioCAT != RADIO_CAT_PCR1000 || gSyncMode == SYNC_MODE_RMT)
	{
		gTxModeRmt = (msg.at(LINK_FLAGS) & 1);
	}
	else
	{
		gTxModeRmt = false;
	}

	//disable tx on this radio if the other radio is enabled	

	if (gTxModeRmt != gTxModeRmtPrev)
	{
		gTxModeRmtPrev = gTxModeRmt;
		if (gSyncMode != SYNC_MODE_RMT)
		{
			if (gTxModeRmt)
			{
				//our Tx was enabled, turn it off 
				gTxMode = RADIO_TX_OFF;				
				if (gSyncMode == SYNC_MODE_FREQ)
				{
					gFreqRadio = newFreq;
					if (gFreqRadio < 100000 || gFreqRadio > 999999999)
					{
						gFreqRadio = gFreqRadioPrev;
					}
					TxFreq2Radio(gFreqRadio);
					gFreqRadioPrev = gFreqRadio;
					gSyncOffset = true;
				}
				PrtBtnA(true);
				PrtFreq(true);
				SendASCCmd(ASC_CMD_TXMODE, 0, 0);
				goto ExitHr;
			}
			else
			{
				gSyncOffset = false;
			}
			PrtBtnA(true);
			PrtFreq(true);
			SendASCCmd(ASC_CMD_TXMODE, 0, 0);
		}
		else
		{
			//tx mode follows rmt in sync all
			gTxMode = gTxModeRmt;						
			ProcessA_Btn();
			gSyncOffset = false;			
		}
	}

	if (gSyncMode == SYNC_MODE_TRACK && gLinkTrackOffset==0)
	{
		//just went to tracking mode, calc offset
		// newFreq is what the other Mini is set to
		gLinkTrackOffset = gFreqRadio - newFreq;
		SerialOut("offset freq= " + String(gLinkTrackOffset), true);
	}
	else if (gSyncMode != SYNC_MODE_TRACK)
	{
		gSyncOffset = false;
		gLinkTrackOffset = 0; //not tracking		
	}
	gSyncModePrev = gSyncMode;

	//SerialOut("\n\rgFreqRadio= " + String(gFreqRadio) + ", freq= " + String(newFreq) + ", offsetFreq= " + String(offsetFreq), true);
	
	//get vfo push/pull flag if we're in basic link mode
	if (gSyncMode == SYNC_MODE_BASIC && (msg.at(LINK_FLAGS) & 0x30))
	{		
		gVFO_PushPull = (msg.at(LINK_FLAGS) & 0x30) >> 4; //get push/pull flag from other mini		
	}
	else
	{
		gVFO_PushPull = SYNC_MODE_BASIC_NONE; //keep it turned off
	}
	
	if ((gSyncMode == SYNC_MODE_FREQ || gSyncMode == SYNC_MODE_RMT) && \
	   ((!gTxMode && !gTxModeRmt) || gSyncMode == SYNC_MODE_RMT || gRadioCAT == RADIO_CAT_PCR1000) && \
	   gFreqRadio != newFreq && newFreq >= 100000 && newFreq <= 999999999)
	{
		gFreqRadio = newFreq; //remote end updated us, keep in sync		
	}
	else if (gSyncMode == SYNC_MODE_TRACK && gLinkTrackOffset != 0 && !gTxMode && !gTxModeRmt)
	{
		//only track if both Minis have Tx disabled - this allows you to tune the offline Mini while making a contact on the other
		//ignore this if the local unit was disabled - send our freq to other Mini to bring it to this freq+offset
		//SerialOut("newFreq= " + String(newFreq) + ", offset= " + String(gLinkTrackOffset) + ", old gFreqRadio= " + String(gFreqRadio), false);
		gFreqRadio = newFreq + gLinkTrackOffset;
		//SerialOut(", new gFreqRadio= " + String(gFreqRadio), true);
		if (gFreqRadio < 100000 || gFreqRadio >999999999)
		{
			gFreqRadio = gFreqRadioPrev;//reset if out of range
		}		
	}
	else if (gSyncMode == SYNC_MODE_BASIC && gVFO_PushPull)
	{
		//we're in basic sync mode and the other mini asserted a push/pull flag
		if (gVFO_PushPull == SYNC_MODE_BASIC_PUSH)
		{
			//they want us to use their info
			gFreqRadio = newFreq;
			gRMode = newRMode;
			gBand = newBand;
			gRIT = newRIT;
			PrtRMode(true);
			PrtBand(true);
			PrtRIT();
			hmiTagLineRefresh = 0;
			PrtStatus("Freq/Mode Updated", true);
		}
		else if (gVFO_PushPull == SYNC_MODE_BASIC_PULL)
		{
			//they want our info
			gVFO_PushPull = SYNC_MODE_BASIC_PUSH; //push it back
			SendLinkParams();
			hmiTagLineRefresh = 0;
			PrtStatus("Pushing Freq/Mode", true);
		}		
		gVFO_PushPull = SYNC_MODE_BASIC_NONE;		
	}

	if (gFreqRadio != gFreqRadioPrev)
	{	
		ChkFreqModeChng();
		gLinkRxChngTimer = millis(); //block Link Tx for chng timer default so we don't oscilate
	}
	
	if (gRIT != newRIT && gSyncMode == SYNC_MODE_RMT)
	{
		gRIT = newRIT;
		PrtRIT();
	}

	if (newCWSpeed >= 5 && newCWSpeed <= 50)
	{
		if (gCWSpeed != newCWSpeed)
		{
			gCWSpeed = newCWSpeed;
			gFarnsSpeed = gCWSpeed;
			CalcDitDur();
			CalcDitFarnsDur();
			PrtOptions();
		}
	}

	if (newFarnsSpeed >= 5 && newFarnsSpeed <= 50)
	{
		if (gFarnsSpeed != newFarnsSpeed)
		{
			gFarnsSpeed = newFarnsSpeed;
			CalcDitFarnsDur();
		}
	}

	if (newKeyType <= KEYER_BUG)
	{
		if (gKeyType != newKeyType)
		{
			gKeyType = newKeyType;
			PrtOptions();
		}
	}

	if (newKeyerSN != mnuKeyerSN)
	{
		mnuKeyerSN = newKeyerSN;
	}

	if (gSyncMode == SYNC_MODE_RMT)
	{
		if (gBand != newBand)
		{
			gBand = newBand;
		}
		if (gRMode != newRMode)
		{
			gRMode = newRMode;
		}

		//PCR1000 settings
		if (gRadioCAT == RADIO_CAT_PCR1000)
		{			
			newVal = msg.at(LINK_AUX1); //(gPCRHasDSP << 6) + (gPCRDSP << 5) + (gPCRANch << 4) + (gPCRAGC << 3) + (gPCRAtten << 2) + (gPCRNB << 1);						

			if (gLinkMiniA || gSyncMode != SYNC_MODE_RMT)
			{
				//only update if coming from the client (Mini-B) connected to radio or not in sync all mode
				gPCRHasDSP = (newVal & 0x40) >> 6;
			}
			if (newVal != oldVal)
			{
				gPCRDSP = (newVal & 0x20) >> 5;
				gPCRANch = (newVal & 0x10) >> 4;
				gPCRAGC = (newVal & 0x08) >> 3;
				gPCRAtten = (newVal & 0x04) >> 2;
				gPCRNB = (newVal & 0x02) >> 1;
				PCR_Update(0xf0); //update binary values if anything changes
				oldVal = newVal;
			}

			newVal = msg.at(LINK_AUX2);
			if (gPCRVol != newVal)
			{
				gPCRVol = newVal;
				PCR_Update(MNU_PCR_VOL);
			}
			
			newVal = msg.at(LINK_AUX3);			
			if (gPCRSq != newVal)
			{
				gPCRSq = newVal;
				PCR_Update(MNU_PCR_SQ);
			}
			
			newVal = msg.at(LINK_AUX4);
			if (gPCRIF != newVal)
			{
				gPCRIF = newVal;
				PCR_Update(MNU_PCR_IFSHIFT);
			}
			
			newVal = msg.at(LINK_AUX5);
			if (gPCRDSPNR != newVal)
			{
				gPCRDSPNR = newVal;
				PCR_Update(MNU_PCR_DSPNR);
			}
			
			newVal = msg.at(LINK_AUX6);
			if (gPCRBW != newVal)
			{
				gPCRBW = newVal;
				PCR_Update(MNU_PCR_BW);
			}
						
			PrtOptions();			
		}
		else if (gRadioCAT == RADIO_CAT_FLEX)
		{
			//flex options
			newVal = msg.at(LINK_AUX1);
			gFlexApiPTT = (newVal | 0x80) >> 8;
			gFlexApiKey = (newVal | 0x40) >> 7;

			newVal = msg.at(LINK_AUX2);
			if (gFlexSlice != newVal)
			{				
				Flex_Update(MNU_FLEX_SLICE, newVal, txt);
			}

			newVal = msg.at(LINK_AUX3);
			if (gFlexVol != newVal)
			{				
				Flex_Update(MNU_FLEX_VOL, newVal, txt);
			}
			
			newVal = (msg.at(LINK_AUX4)<<8) + msg.at(LINK_AUX5); //16-bit
			if (gFlexDelay != newVal)
			{
				Flex_Update(MNU_FLEX_DELAY, newVal, txt);
			}
			
			newVal = msg.at(LINK_AUX6);
			if (gFlexPO != newVal)
			{
				Flex_Update(MNU_FLEX_PO, newVal, txt);
			}

			newVal = msg.at(LINK_AUX7);
			if (gFlexTO != newVal)
			{
				Flex_Update(MNU_FLEX_TO, newVal, txt);
			}

			newVal = msg.at(LINK_AUX8);
			if (gFlexWNB != newVal)
			{
				Flex_Update(MNU_FLEX_WNB, newVal, txt);
			}

			newVal = msg.at(LINK_AUX9);
			if (gFlexNB != newVal)
			{
				Flex_Update(MNU_FLEX_NB, newVal, txt);
			}

			newVal = msg.at(LINK_AUX10);
			if (gFlexNR != newVal)
			{
				Flex_Update(MNU_FLEX_NR, newVal, txt);
			}

			newVal = msg.at(LINK_AUX11);
			if (gFlexAPF != newVal)
			{
				Flex_Update(MNU_FLEX_AF, newVal, txt);
			}

			newVal = msg.at(LINK_AUX12);
			if (gFlexANF != newVal)
			{
				Flex_Update(MNU_FLEX_AF, newVal, txt);
			}

			newVal = msg.at(LINK_AUX13);
			if (gFlexATU != newVal)
			{
				Flex_Update(MNU_FLEX_ANT, newVal, txt);
			}
			PrtOptions();
		}
		
	}

ExitHr: 
	return;
}


void DecLinkMsgASC(std::string msg)
{
	//ASC msg is rcvd from ASC if this is Mini-A and from Mini-A if this is Mini-B
	bool changed = false; 

	gAntSwOnline = true;
	bool newAntSwActive = msg.at(LINK_ASC_SWACTIVE);
	gAntSwDual = msg.at(LINK_ASC_SWDUAL);
	if (!gAntSwDual && !gLinkMiniA)
	{
		gAntSwSendBand = false; //force off if this is Mini-B and we're not in dual port mode
		PrtBand(true); //reset color
	}
	gAntSw8Port = msg.at(LINK_ASC_SW8PORT);
	bool newAntSwLock = msg.at(LINK_ASC_SWLOCK);

	if (newAntSwActive != gAntSwActive)
	{
		gAntSwActive = newAntSwActive;
		hmiTagLineRefresh = 0;
		if (gAntSwActive)
		{			
			PrtStatus("Ant Switch is ACTIVE", true);
		}
		else
		{
			PrtStatus("Ant Switch is OFF", true);
		}
		PrtBand(true);
		changed = true;
	}
	if (newAntSwLock != gAntSwLock)
	{
		gAntSwLock = newAntSwLock;
		hmiTagLineRefresh = 0;
		if (gAntSwLock)
		{
			PrtStatus("Ant Switch is LOCKED", true);
		}
		else
		{
			PrtStatus("Ant Switch Unlocked", true);
		}
		changed = true;
	}

	gAntSwA = int(msg.at(LINK_ASC_SWA));
	gAntSwB = int(msg.at(LINK_ASC_SWB));
	
	if (gAntSwA != gAntSwAPrev || gAntSwB != gAntSwBPrev || changed)
	{
		changed = true;
		PrtWiFi(); //update ant #s on display
		if (gAntSwA != gAntSwAPrev && gLinkMiniA)
		{			
			hmiTagLineRefresh = 0;
			PrtStatus("Port-A set to " + mnuAntLbl[gAntSwA], true);
		}
		else if (gAntSwB != gAntSwBPrev && !gLinkMiniA)
		{		
			hmiTagLineRefresh = 0;
			PrtStatus("Port-B set to " + mnuAntLbl[gAntSwB], true);
		}
		gAntSwAPrev = gAntSwA;
		gAntSwBPrev = gAntSwB;	
	}
	if (changed && gMnu == MNU_ANT_SW)
	{
		gMnuPrev = -1; //force update
		CreateMnu("", false);
	}
	uint8_t startLbl = int(msg.at(LINK_ASC_STARTLBL)); //two labels sent - inc this counter by 2 after each msg
	String newAntSwLbl1 = "";
	String newAntSwLbl2 = "";
	for (int i = 0; i < 10; i++)
	{
		if (msg.at(LINK_ASC_STARTLBL + 1 + i) != 0)
		{
			newAntSwLbl1 += msg.at(LINK_ASC_STARTLBL + 1 + i);
		}
		if (msg.at(LINK_ASC_STARTLBL + 11 + i) != 0)
		{
			newAntSwLbl2 += msg.at(LINK_ASC_STARTLBL + 11 + i);
		}
	}
	newAntSwLbl1.trim();
	newAntSwLbl2.trim();
	mnuAntLbl[startLbl] = newAntSwLbl1;
	mnuAntLbl[startLbl + 1] = newAntSwLbl2;

	//SerialOut("A= " + String(gAntSwA) + ", B= " + String(gAntSwB) + ", Lbl A= " + gAntSwLblA + ", Lbl B= " + gAntSwLblB, true);
	
	//send updated status to Mini-B
	SendASCLinkParams();	

ExitHr:
	return;
}

//================================================================================================================
// BLE Client services

void OpenBLEClient()
{
	BLEDevice::init("CTR2-Mini-BLE-Client");

	/* Retrieve a Scanner and set the callback we want to use to be informed when we
	   have detected a new device.  Specify that we want active scanning and start the
	   scan to run for 30 seconds. */
	
	pBLEScan = BLEDevice::getScan();
	pBLEScan->setAdvertisedDeviceCallbacks(new MyAdvertisedDeviceCallbacks());
	pBLEScan->setInterval(1349);
	pBLEScan->setWindow(449);
	pBLEScan->setActiveScan(true);
	pBLEScan->start(30, false);
	PrtStatus("Scanning for Mini-A", true);
	gSyncMode = SYNC_MODE_BASIC; //server will update this
	gSyncModePrev = gSyncMode;
	delay(1000);	
}

void CloseBLEClient()
{
	pClient->disconnect();		
	pService->stop();	
	PrtStatus("BLE Client Offline", true);
	ResetLink();
	gBLEConnected = false;		
}

/* Start connection to the BLE Server */
bool ConnectToBLEServer()
{
	Serial.print("Forming a connection to ");
	Serial.println(pRmtDevice->getAddress().toString().c_str());

	pClient = BLEDevice::createClient();
	PrtStatus("Created BLE Client", true);

	pClient->setClientCallbacks(new MyClientCallback());

	/* Connect to the remote BLE Server */
	pClient->connect(pRmtDevice);  // if you pass BLEAdvertisedDevice instead of address, it will be recognized type of peer device address (public or private)
	PrtStatus("Connecting to Mini-A", true);
	
	/* Obtain a reference to the service we are after in the remote BLE server */
	pRmtService = pClient->getService(servUUID);		
	if (pRmtService == nullptr)
	{
		PrtStatus("Mini-A Connection Failed", true);
		delay(10);
		pClient->disconnect();
		return false;
	}

	/* Obtain a reference to the characteristic in the service of the remote BLE server */	
	pRmtParams = pRmtService->getCharacteristic(charUUID);
	if (pRmtParams == nullptr)
	{
		PrtStatus("Mini-A Get Data Failed", true);
		pClient->disconnect();
		return false;
	}	

	/* Read the value of the characteristic */
	/*if (pRmtParams->canRead())
	{
		uint32_t freq = 0;
		std::string frq = pRmtParams->readValue();
		for (int i = 0; i < frq.length(); i++)
		{
			freq += frq.at(i) << (i * 8);
		}
		SerialOut("Freq of Rmt is " + String(freq) + " Hz", true);
		
		if (gFreqRadio != freq && freq>=100000)
		{
			gFreqRadio = freq; //remote end updated us
		}
	}
	*/
		

	PrtStatus("Connected to Mini-A", true);	

	return true;
}

void ResetLink()
{
	//we're going offline, close link if it's open
	if (gSyncOffset)
	{
		//we're offset, restore original freq		
		gFreqRadio = gFreqRadio + gLinkTrackOffset;
		ChkFreqModeChng();
	}	
	if (gLinkMiniA)
	{
		gLink_ZeroBeat = 0;
	}
	gSyncOffset = false;	
	gLinkTrackOffset = 0;
	gLinkMsgPrev = "";
	gLinkOnline = false;
	ResetASC();
}

void SendASCLinkParams()
{
	//only Mini-A sends status to ASC
	static bool alreadyHr = false;
	static uint8_t startLbl = 1; //send two labels each poll - inc this by 2
		
	if ((!gBLEConnected && !gUDPConnected) || !gLinkMiniA || !gUDPAdrs.length()  || alreadyHr)
	{
		return;
	}
	alreadyHr = true;

	uint8_t ary[gNumASCLinkBytes];
	for (int i = 0; i < gNumASCLinkBytes; i++)
	{
		ary[i] = 0;
	}

	//sync bytes for ASC
	ary[LINK_ASC_PREAMBLE] = LINK_PREAMBLE_ASC;
	ary[gNumASCLinkBytes - 1] = LINK_POSTAMBLE_ASC;

	//NOTE: send ASC properties to Mini-A - it will respond with switching commands

	ary[LINK_ASC_SWACTIVE] = gAntSwActive;
	ary[LINK_ASC_SWDUAL] = gAntSwDual; //goes true when dual radio ports selected 
	ary[LINK_ASC_SW8PORT] = gAntSw8Port; //goes true when 8-ant port switch is selected
	ary[LINK_ASC_SWLOCK] = gAntSwLock; //if true, controls are locked
	ary[LINK_ASC_SWA] = gAntSwA; //switch assigned to Port-A - ones based, 0=none, 1-8 depending on gAntSwDual
	ary[LINK_ASC_SWB] = gAntSwB; //Port-B
	ary[LINK_ASC_STARTLBL] = startLbl;
	//send selected antenna labels
	for (int i = 0; i < mnuAntLbl[startLbl].length(); i++)
	{
		ary[LINK_ASC_STARTLBL + 1 + i] = mnuAntLbl[startLbl].charAt(i);
	}
	for (int i = 0; i < mnuAntLbl[startLbl + 1].length(); i++)
	{
		ary[LINK_ASC_STARTLBL + 11 + i] = mnuAntLbl[startLbl + 1].charAt(i);
	}
	startLbl += 2;
	if (startLbl > mnuAntLblMax)
	{
		startLbl = 1; //start over
	}
	unsigned short crcVal = crc.XModemCrc(ary, 0, gNumASCLinkBytes - 4); //calc CRC on all bytes up to CRC MSB
	ary[gNumASCLinkBytes - 3] = (crcVal & 0xff00) >> 8;
	ary[gNumASCLinkBytes - 2] = crcVal & 0xff;

	/*
		SerialOut("ary[LINK_FLAGS]= " + String(ary[1], HEX), true);
		for (int i = 0; i < gNumASCLinkBytes; i++)
		{
			SerialOut(String(ary[i], HEX) + ":", false);
		}
		SerialOut(" ", true);
		*/

	if (gBLEConnected)
	{
		if (gLinkMiniA)
		{
			pLocalParams->setValue(ary, gNumASCLinkBytes);
		}
		else
		{
			pRmtParams->writeValue(ary, gNumASCLinkBytes); //always send it, server will ignore it if it's not in sync or track mode
		}
	}
	else if (gUDPConnected)
	{
		//SerialOut("starting udp", true);
		udp.beginPacket(udpAdrs, udpPort);
		//SerialOut("writing to udp", true);
		udp.write(ary, gNumASCLinkBytes);
		//SerialOut("ending udp", true);
		udp.endPacket();
		udp.flush();
	}
	alreadyHr = false;
}


void SendASCCmd(uint8_t cmd, uint8_t miniSel, uint8_t antSel)
{
	//send control command to ASC from Mini-A
	
	if (!gAntSwOnline)
	{
		//don't send anything if we've failed off or this is Mini-B
		return;
	}
	
	uint8_t ary[gAntSwCmdLength];
	//clear buffer
	for (int i = 0; i < gAntSwCmdLength; i++)
	{
		ary[i] = 0;
	}
	String txt = "";

	//turn off sendband if we're in basic+freq mode - can't switch ants and  A & B on same band
	if (gSyncMode == SYNC_MODE_FREQ && gAntSwSendBand)
	{
		gAntSwSendBand = false;
		PrtBand(true);
		return;
	}

	//set band data
	if (cmd == ASC_CMD_BAND && gAntSwSendBand)
	{		
		if (gLinkMiniA)
		{
			miniSel = 1;
		}
		else
		{
			miniSel = 2;
		}
		antSel = gBand;		
	}
	
	ary[LINK_PREAMBLE] = LINK_PREAMBLE_ASC;
	ary[LINK_ASC_CMD] = cmd;
	if (cmd == ASC_CMD_A_PORT || cmd == ASC_CMD_B_PORT || cmd == ASC_CMD_BAND)
	{
		ary[LINK_ASC_AUX1] = miniSel;
		ary[LINK_ASC_AUX2] = antSel;
	}
	else if (cmd == ASC_CMD_LABEL)
	{
		//add new ant label to msg - just send antSel + label
		txt = mnuAntLbl[antSel];		
		ary[LINK_ASC_AUX1] = antSel;
		for (int i = 0; i < txt.length(); i++)
		{
			ary[i + 3] = txt.charAt(i);			
		}		
	}
	else if (cmd == ASC_CMD_TXMODE && gLinkMiniA)
	{
		//only send this command if this is Mini-A
		ary[LINK_ASC_AUX1] = gTxMode; //Mini-A's tx mode
		ary[LINK_ASC_AUX2] = gTxModeRmt; //Mini-B's tx mode
		//SerialOut("ary[2]= " + String(ary[2]) + ", ary[3]= " + String(ary[3]), true);
	}
	else
	{
		//just send command - no aux parameters
	}
	unsigned short crcVal = crc.XModemCrc(ary, 0, gAntSwCmdLength-3); //calc CRC on all bytes up to CRC MSB
	ary[gAntSwCmdLength-3] = (crcVal & 0xff00) >> 8;
	ary[gAntSwCmdLength-2] = crcVal & 0xff;
	ary[gAntSwCmdLength-1] = LINK_POSTAMBLE_ASC;

	uint8_t ptr = gUDPAdrsASC.indexOf('.');
	udpAdrsASC[0] = gUDPAdrsASC.substring(0, ptr).toInt();
	ptr++;
	udpAdrsASC[1] = gUDPAdrsASC.substring(ptr, gUDPAdrsASC.indexOf(ptr, '.')).toInt();
	ptr = gUDPAdrsASC.indexOf('.', ptr + 1);
	ptr++;
	udpAdrsASC[2] = gUDPAdrsASC.substring(ptr, gUDPAdrsASC.indexOf(ptr, '.')).toInt();
	ptr = gUDPAdrsASC.indexOf('.', ptr + 1);
	ptr++;
	udpAdrsASC[3] = gUDPAdrsASC.substring(ptr).toInt();	

	if (gUDPConnected)
	{
		udp.beginPacket(udpAdrsASC, udpPort);
		udp.write(ary, gAntSwCmdLength);
		udp.endPacket();
		udp.flush();
	}

	/* debug
	for (int i = 0; i < gAntSwCmdLength; i++)
	{
		SerialOut(String(ary[i]) + ":", false);
	}
	SerialOut(" ", true);
	for (int i = 0; i < gAntSwCmdLength; i++)
	{
		SerialOut(String(char(ary[i])), false);
	}
	SerialOut(" ", true);
	*/
}

void ResetASC()
{
	////reset ASC properties
	gLinkASCMsgPrev = "";
	gAntSwOnline = false;
	gAntSwA = 0; //blank Ant# on display
	gAntSwB = 0;
	gAntSwAPrev = -1; //force display update on first reply from asc
	gAntSwBPrev = -1;
	PrtWiFi();
}
