// =====================================================================================================================================================================================================
// (c) 2022 Lynn Hansen, KU7Q															                                                                                                               |
// This Source Code Form is subject to the terms of the GNU GENERAL PUBLIC LICENSE, Version 3, 29 June 2007. A copy of this license can be found here: https://choosealicense.com/licenses/gpl-3.0/|
// =====================================================================================================================================================================================================



//This module contains tasks done in Loop - so it can be called from WiFiLoop

void DoLoopTasks()
{
	//chk radio for changes
	static uint32_t funcBtnTmr = millis(); //chk functio btns every 100 mS
	static uint8_t txStateUpdate = 0; //refresh btn a every 1/2 second
	static uint8_t flexPingTmr = 0; //ping flex every 10 seconds (40 polls) when gTCPCon
	static uint8_t btnSwTmr = 0; //track how long 5-way sw btn is pressed before release
	static uint8_t btnATmr = 0; //[A] btn timer
	static uint8_t btnBTmr = 0; //[B] btn timer
	static uint8_t btnCTmr = 0; //[C] btn timer

	uint32_t freq = 0; //to read BLE

	if (100 < millis() - funcBtnTmr)
	{
		//chk for buttons every 100 mS
		ChkRmtFuncBtns();
		funcBtnTmr = millis();
	}

	//chk for gFreqRadio change
	if (gFreqRadio != gFreqRadioPrev || gRMode != gRModePrev)
	{
		ChkFreqModeChng();
		gLinkTxChngTimer = millis(); //block rx of other Mini so we don't get old freq
	}

	//chk for gUnit change in SetEncoderCtrl()
	if (gUnit != gUnitPrev)
	{
		PrtFreq(true);
		SaveInitFile();
		gUnitPrev = gUnit;
	}

	if (gBtnPress)
	{
		//btn press from ISR, process it
		ProcessBtn(gBtnPress);
		gBtnPress = 0; //reset it
	}

	//chk to see if we're in txt input mode and a new chr just showed up from the keyer

	Flex_TxCW(false); //chk for Flex Api keying

	if (hmiMode == HMI_TXTINPUT && txtInputCW)
	{
		//add it to the string
		if (txtInputCW == '~')
		{
			//error, backspace
			txtInput[txtInputPtr] = ' ';
			txtInputPtr--;
			if (txtInputPtr < txtInputSrcStart)
			{
				txtInputSrcStart--;
				if (txtInputSrcStart < 0)
				{
					txtInputSrcStart = 0;
				}
				txtInputPtr = txtInputSrcStart;
			}
			else if (txtInputPtr < 0)
			{
				txtInputSrcStart = 0;
				txtInputPtr = 0;
			}
		}
		else
		{
			txtInput[txtInputPtr] = txtInputCW;
			txtInputPtr++;
			if (txtInputPtr > txtInputSrcStart + 12)
			{
				txtInputSrcStart++;
				if (txtInputSrcStart >= txtInputSrcMax)
				{
					txtInputSrcStart = txtInputSrcMax - 12;
				}
				txtInputPtr = txtInputSrcStart + 12;//point to last chr
			}
			if (txtInputPtr >= txtInputSrcMax)
			{
				txtInputPtr = txtInputSrcMax - 1;
			}
		}		
		txtInputCW = 0;
		hmiEncoderTimeout = millis(); //reset timer
		PrtTxtInput(true);
	}

	if (gTCPCon)
	{
		if (tcp.available())
		{
			Flex_ChkRx();			
		}
		Flex_TxCW(false); //chk for Flex Api keying
	}
	
	//if link is active, chk for incoming packets
	if ((gBLEConnected || gBLEKillScan) && (gLinkRxData || !gLinkMiniA) || \
		(gUDPConnected && udp.parsePacket()))
	{
		GetLinkParams();
	}
	
	if (radioPollTmrDefault < millis() - radioPollTmr)
	{	
		//chk for link fail
		if ((gBLEConnected || gUDPConnected) && (gLinkCommTimeoutDefault < millis() - gLinkCommTimeout) && !gLinkCommFail)
		{
			gLinkCommFail = true;
			gLinkOnline = false; //restore when we reconnect
			PrtWiFi(); //update link status
		}	
		if (gUDPConnected && (gLinkCommTimeoutDefault < millis() - gLinkASCCommTimeout) && gAntSwOnline)
		{
			//monitors online status of asc link
			gAntSwOnline = false;
			PrtWiFi();
		}
		SendLinkParams(); //update every 250 mS
		//SerialOut("Loop " + String(millis()), true);
		if (gFlexInit && gTCPCon)
		{				
			//resent freq 1/2 second after it was sent by encoder change
			if (gFlexLastFreqUpdateTmr && (1000 < millis() - gFlexLastFreqUpdateTmr))
			{
				Flex_TxFreq(gFreqRadio);
				gFlexLastFreqUpdateTmr = 0; //reset timer
				flexPingTmr = 0;
			}
			
			if (flexPingTmr > 20)
			{
				//ping flex every 5 seconds to keep connection alive
				//flex reports changes asynchronously
				Tx2WiFi(WIFI_FLEX, "|ping"); //keep it alive. DON'T send freq change here because if two or more Minis are connected to the flex they will alternate 
				flexPingTmr = 0;

				//if no reply after 30 seconds, Flex has disconnected us - drop WiFi and reconnect
				if (30000 < millis() - gFlexLastReply)
				{
					//re-initialize flex connection
					SerialOut("Reconnecting to Flex Slice " + gFlexSliceLbl[gFlexSlice], true);
					gTCPCon = TCP_CON_OFF;
					PrtOptions();
					gFlexInit = false;
					gFlexHandle = "";
					gFlexHandleTx = "";
					gFlexClientID = "";
					tcp.stop();					
					WiFi.disconnect(true);
					gWiFi = false;
					if (gUDPConnected)
					{
						udp.stop();
						gUDPConnected = false;
					}					
					delay(1500);

					gFlexLastReply = 0;
					ConnectToWiFi();
					if (gWiFi)
					{
						gTCPCon = FLEX_CON_CONNECTING; //connecting, not bound 
						Flex_Initialize();
					}
				}
			}
		}
		else
		{
			//all others must be asked
			GetFreqMode(); //poll radio for changes		
			ChkFreqModeChng(); //chk for freq or mode change - save to prev arrays
		}

		
		radioPollTmr = millis();
		
		//inc every 250mS - times btn press length
		btnSwTmr++;
		btnATmr++; 
		btnBTmr++;
		btnCTmr++;
		flexPingTmr++;

		//refresh btn A (tx state) every 2 seconds
		//update Mini-B w/ASC status if link is active
		txStateUpdate++;
		if (txStateUpdate > 8 || gTxMode != gTxModePrev)
		{
			PrtBtnA(false);
			txStateUpdate = 0;
		}

		if (hmiTagLineRefresh)
		{
			//red alert posted, count up to 1500+ and reset it
			hmiTagLineRefresh += radioPollTmrDefault;
			if (hmiTagLineRefresh > hmiTagLineRefreshMax)
			{
				hmiTagLineRefresh = 0;
				PrtStatus(hmiTagLine, false);				
			}			
		}

		//if gSaveRadio is true, inc counter then save after 30 seconds
		if (gSaveRadio == true)
		{
			gSaveRadioCtr++; //inc every 250ms (radioPollTmrDefault)
			if (gSaveRadioCtr > 30000 / radioPollTmrDefault)
			{
				Flex_TxCW(false); //chk for Flex Api keying
				SaveRadio(gRadioPort, false);				
			}
		}
	}

	//chk for menu timeout
	if (gMnu)
	{
		//keep value window open until user closes it - suggested by VA3AR
		if ((hmiEncoderTimeoutDefault < millis() - hmiEncoderTimeout) && hmiMode != HMI_HELP)
		{
			if (hmiMode == HMI_VALUE)
			{				
				//just leave the value window up until the user closes it 
				hmiEncoderTimeout = millis(); //reset timer
			}
			else
			{
				PrtHomePg();
			}
		}

		if (gLongPressToggle)
		{
			GetMnuValue(false, "", LONG_PRESS_LAST); //update window
			UpdateParam(); //long press clicked
			gLongPressToggle = false;
		}
		
		if (hmiMnuChng)
		{
			//chk for changes			
			hmiMnuChng = false;
			CreateMnu(lastStatusMsg, false); //update menu			
		}
		else if (gEncoderClicked)
		{
			gEncoderClicked = false;			
			ExecuteMenuSelection(); //set selected option on click				
		}
	}

	//scan for button press
	static uint8_t btn = BTN_NONE; //holds current active button if any
	static bool btnSwBlock = false; //goes true when btnSw is first pressed, clear it after TxMsg menu opens
	static bool btnCBlock = false; //goes true when btnC is first pressed, clear it after right menu opens

	uint8_t val = ChkBtns();
	if (val)
	{
		if (!btn)
		{
			//btn just pressed
			btn = val;
			delay(20); //debounce
			val = ChkBtns(); //chk again
			if (val)
			{
				//restart timer if its btnA, B, or C
				switch (val)
				{
				case BTN_SW:
					btnSwTmr = 0;
					return;
				case BTN_A:
					gBtnALongPress = false;
					btnATmr = 0;
					return;
				case BTN_B:					
					btnBTmr = 0;
					return;
				case BTN_C:					
					btnCTmr = 0;					
					return;
				}
			}
			else
			{
				btn = BTN_NONE; //not stable reset btn				
			}
		}
		else if (btn == BTN_SW && btnSwTmr == 3)
		{
			//open TxMsg menu
			gBtnSwLongPress = true; //execute on release
			if (hmiMode == HMI_HOME)
			{
				hmiMode = HMI_MENU;
				gMnu = MNU_TXMSG;
				gMnuPrev = -1; //force refresh
				CreateMnu("", false);
				btnSwBlock = true; //ignore btnSw on long press release so we don't change menu selection
			}
			else
			{
				PrtHomePg();//esc
			}
			btnSwTmr++;
		}
		else if (btn == BTN_A && btnATmr == 3)
		{
			//update btnA label to 'Esc'
			gBtnALongPress = true;
			PrtBtnA(true);
			btnATmr++; //so we dont' hit this again
		}
		else if (btn == BTN_B)
		{
			if (btnBTmr == 3)
			{
				gBtnBLongPress = true; 			
				if (hmiMode == HMI_HOME)
				{
					if (gVFO != MNU_VFO_MEM || !gMSel)
					{
						//if not on Memory, invert Lock
						gLock = !gLock;						
					}
					else
					{
						//we're on Memory mode, copy current settings to VFO-A and set to VFO-A
						CopyMem2VFO();
					}					
					PrtBtnB();
					PrtFreq(true);
				}
				else
				{
					PrevMnu();
				}
				btnBTmr++;
			}
		}
		else if (btn == BTN_C)
		{
			if (btnCTmr == 3)
			{				
				//open [C] Btn menu so user can change default selection
				gBtnCLongPress = true;
				if (hmiMode != HMI_MENU)
				{
					hmiMode = HMI_MENU;
					gMnu = MNU_C_BTN;
					gMnuPrev = -1; //force refresh
					CreateMnu("", false);
					btnCBlock = true; //ignore btnC on long press release so we don't change menu selection
				}
				else
				{
					gMnuSelB4Help = gMnuSel; //save so we can restore it when exiting help - this is only needed when the selected option is a parameter like gBand, not a 'selected' option like mnuCBtnSel
					GetHelp();
				}
				btnCTmr++;
			}
		}			
	}
	else
	{
		if (btn)
		{
			//button released			
			if (!gBtnSwLongPress && !gBtnBLongPress && !btnSwBlock && !btnCBlock)
			{
				//process short button presses
				ProcessBtn(btn);
			}
			btn = BTN_NONE;
			btnSwBlock = false;//reset blocks so we can process btns as normal			
			gBtnSwLongPress = false;
			gBtnALongPress = false;
			gBtnBLongPress = false;	
			gBtnCLongPress = false;
			btnCBlock = false; 			
		}
	}

	Flex_TxCW(false); //chk for Flex Api keying

	//chk txBuffer status
	static int buffLen = 0;	
	if (gPlay != gPlayPrev)
	{
		gPlayPrev = gPlay;
		if (gPlay == KEYER_MODE_STOP)
		{
			//we just completed sending buffer, refresh home page
			PrtHomePg();
		}
	}
	if (gPlay)
	{
		//update Status
		if (gTxBuffer.length() != buffLen)
		{
			//only print if changed
			PrtTxBuffer(); //print txBuffer left justified, background red if gTxMode				
		}
		buffLen = gTxBuffer.length();
	}

	//check for new gMSel if memory tuning
	if (gMSel != gMSelPrev)
	{		
		gFreqRadioPrev = 0; //force update
		gRModePrev = 0xff;
		gRIT = 0; //turn it off			
		if (gMSel > 0)
		{
			gVFO_Freq[MNU_VFO_A+2] = vfoMemA_Freq[gMSel - 1];
			gVFO_Mode[MNU_VFO_A+2] = vfoMemA_Mode[gMSel - 1];
			gVFO_BW[MNU_VFO_A+2] = vfoMemA_BW[gMSel - 1];
			gVFO_Freq[MNU_VFO_B+2] = vfoMemB_Freq[gMSel - 1];
			gVFO_Mode[MNU_VFO_B+2] = vfoMemB_Mode[gMSel - 1];
			gVFO_BW[MNU_VFO_B+2] = vfoMemB_BW[gMSel - 1];
		}
		if (!gMSelMode)
		{
			gFreqRadio = gVFO_Freq[MNU_VFO_A+2];
			gRMode = gVFO_Mode[MNU_VFO_A+2];
			gPCRBW = gVFO_BW[MNU_VFO_A+2];
		}
		else
		{
			gFreqRadio = gVFO_Freq[MNU_VFO_B+2];
			gRMode = gVFO_Mode[MNU_VFO_B+2];
			gPCRBW = gVFO_BW[MNU_VFO_B+2];
		}
		gMSelPrev = gMSel;
		ChkFreqModeChng();
		PrtFreq(true);
		PrtRMode(true);
		PrtOffVFO();
		PrtRIT();
		PrtOptions();
		PrtMemLabel(vfoMemLabel[gMSel-1]);
		gMemLabel = vfoMemLabel[gMSel-1]; //use this if copying between shared and normal memory
	}

	if (gRIT != gRITPrev)
	{
		SetRIT(); //must do this outside of ISR
		gRITPrev = gRIT;
	}

	static bool lastgWiFi = false;

	//chk WiFi status
	if (gWiFi != lastgWiFi)
	{
		blockISR = true;
		if (gWiFi)
		{
			if (gRadioCAT != RADIO_CAT_PCR1000)
			{
				gTxMode = RADIO_TX_OFF;
			}			
			ConnectToWiFi();
			if (gWiFi && gRadioCAT == RADIO_CAT_FLEX) 
			{
				delay(100);
				gTCPCon = FLEX_CON_CONNECTING;
				Flex_Initialize();
			}
		}
		else
		{
			if (gTCPCon)
			{
				gTCPCon = TCP_CON_OFF;
				PrtOptions();
				gFlexInit = false;
				gFlexHandle = "";
				gFlexHandleTx = "";
				gFlexClientID = "";
				tcp.stop();				
			}
			if (gUDPConnected)
			{
				udp.stop();
				gUDPConnected = false;
			}
			PrtStatus("Disconnecting WiFi", true);			
			WiFi.disconnect(true);
			gWiFi = false;
			ResetLink();
			delay(1500);
		}
		PrtWiFi();
		blockISR = false;
		lastgWiFi = gWiFi;
	}

	ChkPTT(); //chk status of Rmt PTT input
	Flex_TxCW(false); //chk for Flex Api keying

	//if we're in tx mode and nothing is active, and txHangTmr expires, drop tx indicaton
	if (gTxMode == RADIO_TX_KEY && !digitalRead(PTT_OUT) && !digitalRead(KEY_OUT) && (txHangTmrDefault < millis() - txHangTmr))
	{
		gTxMode = RADIO_TX_ENAB; //turn off tx indication		
	}

	//watch for mode COS
	if (gTxMode != gTxModePrev)
	{
		SendLinkParams();
		Flex_TxCW(false); //chk for Flex Api keying
		PrtBtnA(true);		

		//if we're on the Flex and changed to Tx Enable, set selected slice to active so it's the tx slice
		if (gRadioCAT == RADIO_CAT_FLEX && gFlexInit && gTCPCon && gTxMode == RADIO_TX_ENAB && gTxModePrev == RADIO_TX_OFF)
		{
			Tx2WiFi(WIFI_FLEX, "|slice s " + String(gFlexSlice) + " active=1 tx=1 txant=ANT" + String(gFlexAnt));
			Flex_TxFreq(gFreqRadio); //move panadapter to this freq
			PrtStatus("Slice " + gFlexSliceLbl[gFlexSlice] + " is Active", true);
		}
		gTxModePrev = gTxMode;

		//if we're in split mode, set freq
		if (gVFO == MNU_VFO_SPLIT && gTxMode)
		{
			if (gTxMode > RADIO_TX_ENAB)
			{
				//set Tx freq
				gFreqRadio = gVFO_Freq[MNU_VFO_B];
				gRMode = gVFO_Mode[MNU_VFO_B];
				gPCRBW = gVFO_BW[MNU_VFO_B];
			}
			else
			{
				gFreqRadio = gVFO_Freq[MNU_VFO_A];
				gRMode = gVFO_Mode[MNU_VFO_A];
				gPCRBW = gVFO_BW[MNU_VFO_A];
			}
			ChkFreqModeChng();
			PrtFreq(true);
			PrtRMode(true);
			PrtOffVFO();
			PrtRIT();
			PrtOptions();
		}		
	}

	//chk for incoming data on micro-USB Serial
	ChkPCCmd();
	ChkChr2HMI(); //chk to see if chrs came in from PC

	Flex_TxCW(false); //chk for Flex Api keying

	//BLE services
	if (gBLEConnected != gBLEConnectedPrev)
	{
		//update the icon if state changes
		PrtWiFi();
		gBLEConnectedPrev = gBLEConnected;
	}

	if (gBLEConnect != gBLEConnectPrev)
	{		
		if (gLinkMiniA)
		{
			if (gBLEConnect)
			{
				OpenBLEServer();
			}
			else
			{
				CloseBLEServer();
			}
		}
		else
		{
			if (gBLEConnect)
			{
				OpenBLEClient();
			}
			else
			{
				CloseBLEClient();
			}
		}
		gBLEConnectPrev = gBLEConnect;

		//if we're in the Link Options menu when this happens, update menu
		if (gMnu)
		{
			gMnu = MNU_LINK;
			gMnuPrev = -1;
			CreateMnu("", false);
		}		
	}	


}

