// =====================================================================================================================================================================================================
// (c) 2022 Lynn Hansen, KU7Q															                                                                                                               |
// This Source Code Form is subject to the terms of the GNU GENERAL PUBLIC LICENSE, Version 3, 29 June 2007. A copy of this license can be found here: https://choosealicense.com/licenses/gpl-3.0/|
// =====================================================================================================================================================================================================


//This module contains the routines to manage the Menus
//Routines to display (Prt) the menus are in the HMI module


void NextMnu()
{
	//move to next sub menu based on selected gMnu
	//SerialOut("gMnu=" + String(gMnu) + ", gMnuSel= " + String(gMnuSel), true);
	uint8_t val = 0;

	switch (gMnu)
	{
	case MNU_MAIN: //top menu
		gMnu = gMnuSel + 2; //offset from MNU_NONE and MNU_MAIN
		if (gMnu == MNU_RADIO && (gRadioCAT != RADIO_CAT_FLEX && gRadioCAT != RADIO_CAT_PCR1000))
		{
			gMnu= MNU_MAIN; //ignore if not on a special radio
		}
		gMnuSel = 0;
		break;

	//these are submenus

	case MNU_FREQ:
		//SerialOut("gMnuSel=" + String(gMnuSel), true);
		mnuFreqSel = gMnuSel;
		switch (gMnuSel)
		{
		case MNU_FREQ_ZERO:
			//toggle gZeroDigits - if just turning ON, do it now
			gZeroDigits = !gZeroDigits;
			SaveInitFile();
			ChkFreqModeChng();						
			hmiMode = HMI_HOME; //goto home page				
			hmiMnuChng = true; 
			break;
		case MNU_FREQ_RIT:			
			SetMnuParam();
			break;
		case MNU_FREQ_DB:
			//toggle gShareDB
			gShareDB = !gShareDB;
			gMnuPrev = -1; //refresh
			gSaveRadio = true;
			Chk4FreqListFile(); //create new files if needed
			Chk4TxMsgFile();
			Chk4BandListFile(false);
			break;
		case MNU_FREQ_LOAD:
			gMnu = MNU_FREQ_LD; //load freq list
			//find current mem selected and set gMnuSel to point to it
			gMnuSel = 0;
			mnuFreqListStart = gMSel - 1;
			if (mnuFreqListStart < 0)
			{
				mnuFreqListStart = 0;
			}
			break;
		case MNU_FREQ_SAVE:
			//freq save
			//select from list of slots to save current freq & mode
			gMnu = MNU_FREQ_SV;
			gMnuSel = 0;
			mnuFreqListStart = gMSel - 1;
			if (mnuFreqListStart < 0)
			{
				mnuFreqListStart = 0;
			}
			break;
		case MNU_FREQ_RENAME:
			//rename memory slot
			//let user select from memory slots			
			gMnu = MNU_FREQ_REN;
			gMnuSel = 0;
			mnuFreqListStart = gMSel - 1;
			if (mnuFreqListStart < 0)
			{
				mnuFreqListStart = 0;
			}
			break;
		case MNU_FREQ_PREV:
			//load previous freq
			PopPrevFreqMode(); //get last freq & mode
			hmiMnuChng = true; //go to home page
			hmiMode = HMI_HOME;
			break;
		}
		break;
	
	case MNU_FREQ_REN:
		// open txtInput editor				
		txtInputSrc = mnuFreqListStart + mnuFreqListSel;
		//SerialOut("txtInputSrc= " + String(txtInputSrc) + ", freqListStart=" + String(mnuFreqListStart) + ", mnuFreqListSel=" + String(mnuFreqListSel), true);				
		GetTxtInput(vfoMemLabel[txtInputSrc]);
		return;
		break;	
	case MNU_KEYER:
		//we're on the Keyer menu, open keyer type menu
		mnuKeyerSel = gMnuSel;
		switch (gMnuSel)
		{
		case MNU_KEYER_TYPE:		
			gMnu = MNU_CW_KEYERS;
			gMnuSel = gKeyType;	
			mnuCWKeyerSel = gKeyType;			
			break;
		}
		break;	
	case MNU_ANT_SW:
		mnuASCSel = gMnuSel;
		gMnu = MNU_ANT_SW;
		gMnuPrev = -1;
		switch (gMnuSel)
		{
		case MNU_ANT_SW_PORT_A:
			gMnuSel = gAntSwA;
			break;
		case MNU_ANT_SW_PORT_B:
			gMnuSel = gAntSwB;
			break;
		case MNU_ANT_SW_LABELS:
			gMnuSel = mnuAntEditSel;
			break;
		}		
		break;
	case MNU_CONFIG:
		mnuConfigSel = gMnuSel;
		switch (gMnuSel)
		{
		case MNU_CONFIG_C_BTN:
			gMnu = MNU_C_BTN;
			break;
		case MNU_CONFIG_FUNC:
			gMnu = MNU_FUNC_BTN;
			break;
		case MNU_CONFIG_THEME:
			gMnu = MNU_THEME;
			break;
		case MNU_CONFIG_TAGLINE:			
			gMnuSel = MNU_CONFIG_TAGLINE_EDIT; //so ExecTxtInput() knows what to do with the value
			GetTxtInput(hmiTagLine); //just get the new text for the bottom tag line
			break;
		case MNU_CONFIG_CAT:
			gMnu = MNU_RADIOCAT;			
			break;
		case MNU_CONFIG_BAUD:
			gMnu = MNU_RADIOBAUD;			
			break;
		case MNU_CONFIG_AMPSEQ:
			gMnu = MNU_AMPSEQ;
			break;
		}
		break;
	case MNU_FUNC_BTN:
		mnuFuncBtnSel = gMnuSel;
		gMnu = MNU_FUNC_BTN_OPTIONS;
		break;
	}
	gEncoderClicked = false; //reset click event that got us here
	hmiMnuChng = true; //force CreateMnu() in Loop
}

void PrevMnu()
{
	//dec menu from this menu to the previous menu
	switch (gMnu)
	{
	case MNU_MAIN:
		//turn off menu		
		hmiMnuChng = true; //go to home page
		hmiMode = HMI_HOME;
		break; 
	case MNU_FREQ: //fall through
	case MNU_MODE:
	case MNU_BAND:
	case MNU_VFO:
	case MNU_TXMSG:	
	case MNU_TXMSG_EDIT:
	case MNU_KEYER:
	case MNU_WIFI:
	case MNU_LINK:
	case MNU_ANT_SW:
	case MNU_RADIOPORT:
	case MNU_CONFIG:
	case MNU_RADIO:		
		gMnuSel = gMnu - 2; //subtract for MNU_NONE and MNU_MAIN
		mnuMainSel = gMnuSel;
		gMnu = MNU_MAIN;
		break;

	//sub menus return here, go back to calling main menu with this menu selected
	//menus in Freq menu
	case MNU_FREQ_LD:
		gMnuSel = MNU_FREQ_LD;
Rtn2Freq:
		gMnu = MNU_FREQ; //menu to return to
		break;
	case MNU_FREQ_SV:
		gMnuSel = MNU_FREQ_SV;
		goto Rtn2Freq;
		break;
	case MNU_FREQ_REN:
		gMnuSel = MNU_FREQ_REN;
		goto Rtn2Freq;
		break;	
	//menus in Config menu
	case MNU_C_BTN:				
		if (gBtnCLongPress)
		{
			//we're here from a C btn long press in Loop(), just go back to home page			
			PrtHomePg();
			return;
		}
		gMnuSel = MNU_CONFIG_C_BTN;
Rtn2Config:
		gMnu = MNU_CONFIG;
		break;
	case MNU_FUNC_BTN:
		gMnuSel = MNU_CONFIG_FUNC;
		goto Rtn2Config;
		break;
	case MNU_FUNC_BTN_OPTIONS:
		gMnuSel = MNU_FUNC_BTN;
		gMnu = MNU_FUNC_BTN;
		break;
	case MNU_THEME:
		gMnuSel = MNU_CONFIG_THEME;
		goto Rtn2Config;
		break;	
	case MNU_TAGLINE:
		gMnuSel = MNU_CONFIG_TAGLINE;
		goto Rtn2Config;
		break;
	case MNU_RADIOCAT:		
		gMnuSel = MNU_CONFIG_CAT;
		goto Rtn2Config;
		break;
	case MNU_RADIOBAUD:
		gMnuSel = MNU_CONFIG_BAUD;
		goto Rtn2Config;
		break;
	case MNU_AMPSEQ:
		gMnuSel = MNU_CONFIG_AMPSEQ;
		goto Rtn2Config;
		break;
	//Keyer menu
	case MNU_CW_KEYERS:
		gMnuSel = 4;
		gMnu = MNU_KEYER;				
		break;
	case MNU_ANT_SW_ANT_A_EDIT:
		gMnuSel = mnuASCSel;
		gMnu = MNU_ANT_SW;
		break;
	case MNU_ANT_SW_ANT_B_EDIT:
		gMnuSel = mnuASCSel;
		gMnu = MNU_ANT_SW;
		break;
	}
	hmiMnuChng = true; //force CreateMnu() in Loop
}


void CreateMnu(String statusMsg, bool statusRed)
{
	//menu dispatcher
	//chk to see if we've changed menus, if not, don't overwrite the Status bar	
	bool refresh = true;
	int ptr = 0;
	String txt = "";
	if (gMnu == gMnuPrev)
	{
		refresh = false;
	}
	//SerialOut("refresh= " + String(refresh) + ", gMnu= " + String(gMnu) + ", gMnuSel= " + String(gMnuSel), true);
	
	switch (hmiMode)
	{
	case HMI_HOME:
		//just refresh home page
		gMnu = MNU_NONE;
		PrtHomePg();
		return;
		break;
	case HMI_MENU:
		hmiEncoderTimeout = millis(); //reset timer
		switch (gMnu)
		{
		case MNU_MAIN:
			for (int i = 2; i < mnuMainMax; i++)
			{
				switch (i)
				{
				case MNU_BAND:
					mnuMain[i - 2] = "Band: " + bandMemLabel[gBand];
					break;
				case MNU_MODE:
					//offset mnuMode by 1 for Toggle Tx Enab
					mnuMain[i-2] = "Mode: " + mnuMode[gRMode + 1];
					break;
				case MNU_VFO:
					if (!gLock)
					{
						switch (gVFO)
						{
						case MNU_VFO_A:
							txt = "VFO-A";
							break;
						case MNU_VFO_B:
							txt = "VFO-B";
							break;
						case MNU_VFO_SPLIT:
							txt = "Split";
							break;
						case MNU_VFO_MEM:
							if (!gMSelMode)
							{
								txt = "Mem-A";
							}
							else
							{
								txt = "Mem-B";
							}
							break;
						}
					}
					else
					{
						txt = "Locked";
					}
					mnuMain[i - 2] = "VFO: " + txt;
					break;
				case MNU_WIFI:
					if (gWiFi)
					{
						txt = "ON";
					}
					else
					{
						txt = "off";
					}
					mnuMain[i - 2] = "WiFi: " + txt;
					break;
				case MNU_LINK:
					if (gBLEConnected || gUDPConnected)
					{
						txt = "ON";
					}
					else
					{
						txt = "off";
					}
					mnuMain[i - 2] = "Link: " + txt;
					break;
				case MNU_ANT_SW:					
					//show if we're online, and Mini-A, or Mini-B if we're in dual mode
					if (gAntSwOnline && (gLinkMiniA || gAntSwDual || gSyncMode == SYNC_MODE_RMT))
					{						
						if (gAntSwLock)
						{
							txt = "LOCKED";
						}
						else if (!gAntSwActive)
						{
							txt = "OFF";
						}
						else
						{
							txt = "Active";
						}
						mnuMain[i - 2] = "Ant Sw: " + txt;
					}				
					else
					{
						mnuMain[i - 2] = "";
					}
				}
			}
			switch (gRadioCAT)
			{			
			case RADIO_CAT_FLEX:
				mnuMain[mnuMainMax - 1] = "Flex 6000";
				break;
			case RADIO_CAT_PCR1000:
				mnuMain[mnuMainMax - 1] = "PCR-1000";
				break;
			default:
				mnuMain[mnuMainMax - 1] = "";
			}
			gMnuSelMax = mnuMainMax;
			PrintMenu(mnuMain);
			if (refresh)
			{
				statusMsg="Select Option";
			}
			break;
		case MNU_FREQ:
			if (refresh)
			{
				gMnuSel = mnuFreqSel;		
				if (gZeroDigits)
				{
					txt = "ON";
				}
				else
				{
					txt = "off";
				}
				mnuFreq[MNU_FREQ_ZERO] = "Zero Low Digits: " + txt;					
				
				if (!gRIT)
				{
					mnuFreq[MNU_FREQ_RIT] = "RIT: off";
				}
				else
				{
					mnuFreq[MNU_FREQ_RIT] = "RIT: " + String(gRIT) + " Hz";
				}
				if (gShareDB)
				{
					mnuFreq[MNU_FREQ_DB] = "SHARING COMMON DATABASE";
				}
				else
				{
					mnuFreq[MNU_FREQ_DB] = "Use Common Database";
				}
				statusMsg= "Select Frequency Option";
			}
			gMnuSelMax = mnuFreqMax;
			PrintMenu(mnuFreq);
			break;
		case MNU_FREQ_LD:
			if (refresh)
			{
				if (!gMSelMode)
				{
					statusMsg = "Load from Memory A";
				}
				else
				{
					statusMsg = "Load from Memory B";
				}
			}
			goto LoadMemList;
		case MNU_FREQ_SV:
			if (refresh)
			{
				statusMsg = "Select Memory to Save";
			}
			goto LoadMemList;
		case MNU_FREQ_REN:
			if (refresh)
			{
				statusMsg = "Select Memory to Rename";
			}
LoadMemList:
			//load mem list
			ptr = mnuFreqListStart;
			//limit to 7 below top of array
			if (ptr > 93)
			{
				//put record 94 at top
				ptr = 93;
				mnuFreqListStart = ptr;
				//and adjust mnu pointer to point to correct record
				gMnuSel = gMSel - 94;
			}
			for (int i = 0; i < mnuFreqListMax; i++)
			{
				//load 7 freqs & modes from freqLists
				if (gVFO == MNU_VFO_B ||  gMSelMode)
				{
					if (vfoMemB_Freq[ptr + i] >= 100000)
					{
						mnuFreqList[i] = String(ptr + i + 1) + ": " + vfoMemLabel[ptr + i] + " : " + 
						vfoMemB_Freq[ptr + i] + " - " + mnuMode[vfoMemB_Mode[ptr + i] + 1];
					}
					else
					{
						mnuFreqList[i] = String(ptr + i + 1) + ": " + vfoMemLabel[ptr + i] + " :     ---- VFO B  Open  ----";
					}
				}	
				else
				{
					if (vfoMemA_Freq[ptr + i] >= 100000)
					{
						mnuFreqList[i] = String(ptr + i + 1) + ": " + vfoMemLabel[ptr + i] + " :  " +
						 vfoMemA_Freq[ptr + i] + " - " + mnuMode[vfoMemA_Mode[ptr + i] + 1];
					}
					else
					{
						mnuFreqList[i] = String(ptr + i + 1) + ": " + vfoMemLabel[ptr + i] + " :     ---- VFO A  Open  ----";
					}
				}				
			}
			gMnuSelMax = mnuFreqListMax;
			PrintMenu(mnuFreqList);
			break;
		case MNU_MODE:
			if (refresh)
			{
				if (gMnuSelB4Help)
				{
					//restore gMnuSel when closing help
					gMnuSel = gMnuSelB4Help;					
				}
				else
				{
					gMnuSel = gRMode + 1; //compensate for Toggle Tx Enab option
				}
				if (!gLock)
				{
					statusMsg = "Select Radio Mode";
				}
				else
				{
					statusMsg = "Freq & Mode are LOCKED";
				}
			}
			gMnuSelMax = mnuModeMax;
			PrintMenu(mnuMode);
			break;
		case MNU_BAND:
			if (refresh)
			{
				if (gMnuSelB4Help)
				{
					//restore gMnuSel when closing help
					gMnuSel = gMnuSelB4Help;					
				}
				else
				{
					gMnuSel = gBand;
				}
				statusMsg = "Select Band";
			}
			gMnuSelMax = mnuBandMax;
			PrintMenu(bandMemLabel);
			break;
		case MNU_VFO:
			if (refresh)
			{	
				mnuVFO[MNU_VFO_PUSH] = ""; //used only in Basic Link mode
				mnuVFO[MNU_VFO_PULL] = "";
				if (!gLock)
				{
					if (gMnuSelB4Help)
					{
						//restore gMnuSel when closing help
						gMnuSel = gMnuSelB4Help;						
					}
					else
					{
						gMnuSel = mnuVFOSel;
					}
					if (gVFO == MNU_VFO_MEM)
					{
						mnuVFO[MNU_VFO_COPY2A] = "Copy M>VFO";
					}
					else
					{
						mnuVFO[MNU_VFO_COPY2A] = " ";
					}	
					if (gLinkOnline && gSyncMode == SYNC_MODE_BASIC)
					{
						if (gLinkMiniA)
						{
							txt = "B";
						}
						else
						{
							txt = "A";
						}
						mnuVFO[MNU_VFO_PUSH] = "Push to " + txt;
						mnuVFO[MNU_VFO_PULL] = "Pull from " + txt;
					}
				}
				else
				{
					gMnuSel = MNU_VFO_LOCK;
				}
				if (!gMSelMode)
				{
					mnuVFO[MNU_VFO_MODE] = "Tune Mem A";
				}
				else
				{
					mnuVFO[MNU_VFO_MODE] = "Tune Mem B";
				}
				mnuVFOSel = gMnuSel;
				statusMsg = "Select a VFO Mode";
			}			
			gMnuSelMax = mnuVFOMax;
			PrintMenu(mnuVFO);
			break;
		case MNU_TXMSG:
			if (refresh)
			{
				gMnuSel = mnuTxMsgSel;
				statusMsg = "Select & Send a Message";
			}
			gMnuSelMax = mnuTxMsgMax;
			PrintMenu(mnuTxMsg);
			break;
		case MNU_RADIOPORT:
			if (refresh)
			{
				if (gMnuSelB4Help)
				{
					//restore gMnuSel when closing help
					gMnuSel = gMnuSelB4Help;					
				}
				else
				{
					gMnuSel = gRadioPort;
				}
				statusMsg = "Select Radio Database";
			}
			gMnuSelMax = mnuRadioTagMax;
			PrintMenu(mnuRadioTag);
			break;
		case MNU_CONFIG:
			if (refresh)
			{
				for (int i = 0; i < mnuConfigMax; i++)
				{
					switch (i)
					{
					case MNU_CONFIG_AMPSEQ:
						if (gEnabAmpSeq)
						{
							mnuConfig[i] = "Amp Seq: ON";
						}
						else
						{
							mnuConfig[i] = "Amp Seq: off";
						}						
						break;					
					}
				}
				
				//SerialOut("gMnuSel= " + String(mnuConfigSel), true);
				gMnuSel = mnuConfigSel;				
				statusMsg = "Select a Config Option";
			}
			gMnuSelMax = mnuConfigMax;
			PrintMenu(mnuConfig);
			break;
		case MNU_RADIO:
			//select flex or pcr1000 menu
			switch (gRadioCAT)
			{
			case RADIO_CAT_FLEX:			
				if (refresh)
				{
					gMnuSel = mnuFlexSel;
					if (gTCPCon)
					{
						statusMsg = "Select a Flex Option";
					}
					else
					{
						statusMsg = "The Flex is Offline";
					}					
				}
				for (int i = 0; i < mnuFlexMax; i++)
				{
					switch (i)
					{
					case MNU_FLEX_IP:
						mnuFlex[i] = gIPAdrs;
						break;
					case MNU_FLEX_PORT:
						mnuFlex[i] = "Port " + gFlexPort;
						break;
					case MNU_FLEX_PAN:
						if (!gFlexPan)
						{
							mnuFlex[i] = "Panadptr";
						}
						else
						{
							mnuFlex[i] = "NEW PAN";
						}
						break;
					case MNU_FLEX_SLICE:
						mnuFlex[i] = "Slice: " + gFlexSliceLbl[gFlexSlice];
						break;
					case MNU_FLEX_ANT:
						mnuFlex[i] = "Ant: " + String(gFlexAnt);
						break;
					case MNU_FLEX_TOGGLE:
						if (gFlexATU != 0xff && gTCPCon)
						{
							mnuFlex[i] = "Start ATU";
						}
						else
						{
							mnuFlex[i] = "ATU N/A"; //not available
						}
						break;
					case MNU_FLEX_VOL:
						mnuFlex[i] = "Vol: " + String(gFlexVol);
						break;
					case MNU_FLEX_PO:
						mnuFlex[i] = "PO: " + String(gFlexPO) + "w";
						break;
					case MNU_FLEX_TO:
						mnuFlex[i] = "TO: " + String(gFlexTO) + "w";
						break;
					case MNU_FLEX_WNB:
						mnuFlex[i] = "WNB: " + String(gFlexWNB);
						break;
					case MNU_FLEX_NB:
						mnuFlex[i] = "NB: " + String(gFlexNB);
						break;
					case MNU_FLEX_NR:
						mnuFlex[i] = "NR: " + String(gFlexNR);
						break;
					case MNU_FLEX_AF:
						if (gRMode < RADIO_LSB)
						{
							mnuFlex[i] = "APF: " + String(gFlexAPF); //auto peaking filter for CW
						}
						else
						{
							mnuFlex[i] = "ANF: " + String(gFlexANF); //auto notch filter for voice modes
						}
						break;
					case MNU_FLEX_PTT:
						mnuFlex[i] = "PTT: ";
						if (gFlexApiPTT)
						{
							mnuFlex[i] += "Net";
						}
						else
						{
							mnuFlex[i] += "Hw";
						}
						break;
					case MNU_FLEX_KEYER:
						mnuFlex[i] = "Key: ";
						if (gFlexApiKey)
						{
							mnuFlex[i] += "Net";
						}
						else
						{
							mnuFlex[i] += "Hw";
						}
						break;
					case MNU_FLEX_DELAY:
						mnuFlex[i] = "CW Delay";
						break;
					}
				}
				gMnuSelMax = mnuFlexMax;
				PrintMenu(mnuFlex);
				break;
			case RADIO_CAT_PCR1000:			
				if (refresh)
				{
					gMnuSel = mnuPCRSel;
					statusMsg = "Select a PCR-1000 Option";
				}
				gMnuSelMax = mnuPCRMax;
				for (int i = 0; i < mnuPCRMax; i++)
				{
					switch (i)
					{
					case MNU_PCR_VOL:
						mnuPCR[i] = "Vol: " + String(uint8_t(gPCRVol / 2.56)); //using 0-100 scale, actual is 0-255
						break;
					case MNU_PCR_SQ:
						mnuPCR[i] = " Sq: " + String(uint8_t(gPCRSq / 2.56));
						break;
					case MNU_PCR_IFSHIFT:
						mnuPCR[i] = "IF Shift: " + String(uint8_t(gPCRIF / 2.56));
						break;
					case MNU_PCR_ZEROIF:
						mnuPCR[i] = "Zero IF";
						break;
					case MNU_PCR_DSP:
						if (gPCRHasDSP)
						{
							if (gPCRDSP)
							{
								txt = "ON";
							}
							else
							{
								txt = "off";
							}
						}
						else
						{
							txt = "n/a";
						}
						mnuPCR[i] = "DSP: " + txt;
						break;
					case MNU_PCR_DSPNR:
						if (gPCRHasDSP)
						{
							txt = String(uint8_t(gPCRDSPNR));
						}
						else
						{
							txt = "n/a";
						}
						mnuPCR[i] = "DSP NR: " + txt;
						break;
					case MNU_PCR_ANch:
						if (gPCRHasDSP)
						{
							if (gPCRDSP)
							{
								if (gPCRANch)
								{
									txt = "ON";
								}
								else
								{
									txt = "off";
								}
							}
							else
							{
								txt = "off";
							}
						}
						else
						{
							txt = "n/a";
						}
						mnuPCR[i] = "AutoNtch: " + txt;
						break;
					case MNU_PCR_BW:
						switch (gPCRBW)
						{
						case 1:
							txt = "6 kHz";
							break;
						case 2:
							txt = "15 kHz";
							break;
						case 3:
							txt = "50 kHz";
							break;
						case 4:
							txt = "230 kHz";
							break;
						default:
							txt = "3 kHz";
							break;
						}
						mnuPCR[i] = "BW: " + txt;
						break;
					case MNU_PCR_AGC:
						if (gPCRAGC)
						{
							txt = "ON";
						}
						else
						{
							txt = "off";
						}
						mnuPCR[i] = "AGC: " + txt;
						break;
					case MNU_PCR_ATTEN:
						if (gPCRAtten)
						{
							txt = "ON";
						}
						else
						{
							txt = "off";
						}
						mnuPCR[i] = "Atten: " + txt;
						break;
					case MNU_PCR_NB:
						if (gPCRNB)
						{
							txt = "ON";
						}
						else
						{
							txt = "off";
						}
						mnuPCR[i] = "NB: " + txt;
						break;
					}
				}
				PrintMenu(mnuPCR);
				break;
			default:
				return; //no menu
			}
			break;
		case MNU_KEYER:
			for (int i = 0; i < mnuKeyerMax; i++)
			{
				//add params to mnu items
				switch (i)
				{
				case MNU_KEYER_SPEED:
					mnuKeyer[i] = "Speed: " + String(gCWSpeed);
					break;
				case MNU_KEYER_FARNS:
					mnuKeyer[i] = "FarnSpd: " + String(gFarnsSpeed);
					break;
				case MNU_KEYER_SIDETONE:
					if (!gSideTone)
					{
						mnuKeyer[i] = "Sidetone: off";
					}
					else
					{
						mnuKeyer[i] = "Sidetone: ON";
					}
					break;
				case MNU_KEYER_TYPE:
					mnuKeyer[i] = "Type: " + mnuCWKeyer[gKeyType];
					break;
				case MNU_KEYER_PDL:
					if (!gPaddle)
					{
						mnuKeyer[i] = "Pdls: Norm";
					}
					else
					{
						mnuKeyer[i] = "Pdls: REV";
					}
					break;
				case MNU_KEYER_PTT:
					switch (gRmtMode)
					{					
					case MNU_RMT_LATCH:
						mnuKeyer[i] = "Rmt: PTT Latch";
						break;
					case MNU_RMT_KEY:
						mnuKeyer[i] = "Rmt: KEY";
						break;
					default:
						mnuKeyer[i] = "Rmt: PTT";
						gRmtMode = MNU_RMT_NORM;
						break;
					}
					break;
				case MNU_KEYER_CALL:
					mnuKeyer[i] = "Call: " + mnuKeyerCall;
					break;
				case MNU_KEYER_SN:
					mnuKeyer[i] = "SN: " + String(mnuKeyerSN);
					break;
				case MNU_KEYER_EXCH:
					mnuKeyer[i] = "Exch: " + mnuKeyerExch;
					break;
				}
			}
			if (refresh)
			{
				statusMsg = "Select a Keyer Option";
				gMnuSel = mnuKeyerSel;
			}	
			gMnuSelMax = mnuKeyerMax;
			PrintMenu(mnuKeyer);
			break;
		case MNU_WIFI:
			if (refresh)
			{
				gMnuSel = mnuWiFiSel;
				statusMsg = "Select a WiFi Option";
			}
			//Update Wifi settings
			for (int i = 0; i < mnuWiFiMax; i++)
			{
				switch (i)
				{
				case MNU_WIFI_CONNECT:
					if (gWiFi)
					{
						mnuWiFi[i] = "Disconnect WiFi";
					}
					else
					{
						mnuWiFi[i] = "Connect to WiFi";
					}
					break;
				case MNU_WIFI_SSID:
					mnuWiFi[i] = "SSID: " + gSSID;
					break;
				case MNU_WIFI_STAKEY:
					mnuWiFi[i] = "PW: ***********"; // +gStaKey;
					break;
				case MNU_WIFI_AUTO:
					if (gWiFiAutoConnect)
					{
						mnuWiFi[i] = "Auto Connect: ON";
					}
					else
					{
						mnuWiFi[i] = "Auto Connect: off";
					}
					break;
				case 4: //these are not saved
					if (gWiFi)
					{
						mnuWiFi[i] = "Sta Adrs: " + gStaAdrs;
					}
					else
					{
						mnuWiFi[i] = "Sta Adrs:                           ";
					}
					break;
				case 5:
					if (gWiFi)
					{
						mnuWiFi[i] = "Subnet: " + gSubnet;
					}
					else
					{
						mnuWiFi[i] = "Subnet:                         ";
					}
					break;
				case 6:
					if (gWiFi)
					{
						mnuWiFi[i] = "Gateway: " + gGW;
					}
					else
					{
						mnuWiFi[i] = "Gateway:                           ";
					}
					break;
				}
			}
			gMnuSelMax = mnuWiFiMax;
			PrintMenu(mnuWiFi);
			break;
		case MNU_C_BTN:
			if (refresh)
			{
				gMnuSel = mnuCBtnSel;
				statusMsg = "Set [C] Button Function";
			}
			//hide Ant Switch if not online
			if (gAntSwOnline)
			{
				txt = "Ant Sw";
			}
			else
			{
				txt = "";
			}
			mnuCBtn[MNU_C_BTN_ASC] = txt;

			switch (gRadioCAT)
			{
			case RADIO_CAT_FLEX:
				mnuCBtn[mnuCBtnMax - 1] = "Flex 6000";
				break;
			case RADIO_CAT_PCR1000:
				mnuCBtn[mnuCBtnMax - 1] = "PCR-1000";
				break;
			default:
				//nothing
				mnuCBtn[mnuCBtnMax - 1] = "";
				break;
			}
			gMnuSelMax = mnuCBtnMax;
			PrintMenu(mnuCBtn);
			break;
		case MNU_FUNC_BTN:
			if (refresh)
			{
				gMnuSel = mnuFuncBtnSel;
				statusMsg = "Select a Function Button";
			}
			for (int i = 0; i < mnuFuncBtnMax - 1; i++)
			{
				mnuFuncBtn[i] = "F" + String(i + 1) + ": " + mnuFuncBtnLbl[i];
			}
			if (gStuckBtn > 0)
			{
				txt = "RESET KEYPAD LOCKOUT";
			}
			else
			{
				txt = "";
			}
			mnuFuncBtn[mnuFuncBtnMax - 1] = txt;
			gMnuSelMax = mnuFuncBtnMax;
			PrintMenu(mnuFuncBtn);
			break;
		case MNU_FUNC_BTN_OPTIONS:
			if (refresh)
			{
				gMnuSel = mnuFuncBtnOptionSel;
				statusMsg = "Select a Function for F" + String(gMnuSel + 1);
			}
			if (gAntSwOnline)
			{
				txt = "Ant Sw";
			}
			else
			{
				txt = "";
			}
			mnuFuncBtnOption[MNU_FUNC_OPTION_ANT_SW] = txt;
			gMnuSelMax = mnuFuncBtnOptionMax;
			PrintMenu(mnuFuncBtnOption);
			break;
		case MNU_LINK:
			if (refresh)
			{
				for (int i = 0; i < mnuLinkMax; i++)
				{
					switch (i)
					{
					case MNU_LINK_CONNECT:
						if (gBLEConnected || gUDPConnected)
						{
							mnuLink[i] = "Disconnect Link";
						}
						else
						{
							mnuLink[i] = "Connect Link";
						}
						break;
					case MNU_LINK_SVR_CLIENT:
						if (gLinkMiniA)
						{
							mnuLink[i] = "This is Mini-A (server)";
						}
						else
						{
							mnuLink[i] = "This is Mini-B (client)";
						}
						break;
					case MNU_LINK_MODE:
						switch (gSyncMode)
						{
						case SYNC_MODE_BASIC:
							mnuLink[i] = "Sync: Basic";
							break;
						case SYNC_MODE_FREQ:
							mnuLink[i] = "Sync: Basic + Freq";
							break;
						case SYNC_MODE_TRACK:
							mnuLink[i] = "Sync: Basic + Track";
							break;
						case SYNC_MODE_RMT:
							if (gLinkMiniA)
							{
								mnuLink[i] = "Sync: Remote Controller";
							}
							else
							{
								mnuLink[i] = "Sync: Local Controller";
							}
							break;
						}
						if (!gLinkMiniA)
						{
							mnuLink[i] = "[" + mnuLink[i] + "]"; //bracket Mini-B options since they can't be changed
						}
						break;
					case MNU_LINK_TYPE:
						if (!gLinkTypeUDP)
						{
							mnuLink[i] = "Link with Bluetooth LE";
						}
						else
						{
							mnuLink[i] = "Link with WiFi UDP";
						}
						break;
					case MNU_LINK_AUTO:
						if (gLinkAutoConnect)
						{
							mnuLink[i] = "Auto Connect: ON";
						}
						else
						{
							mnuLink[i] = "Auto Connect: off";
						}
						break;
					case MNU_LINK_UDP_ADRS:
						if (gLinkTypeUDP)
						{
							String txt = "Mini-A's ";
							if (gLinkMiniA)
							{
								txt = "Mini-B's ";
							}
							mnuLink[i] = txt + "IP: " + gUDPAdrs;
						}
						else
						{
							mnuLink[i] = ""; //hide if using BLE
						}
						break;					
					case MNU_LINK_OPTION6:
						if (!gLinkMiniA)
						{
							//only on Mini-B
							if (gLinkMiniA || gSyncMode != SYNC_MODE_FREQ)
							{
								mnuLink[i] = ""; //not used on Mini-A or if not in freq sync mode
							}
							else
							{
								mnuLink[i] = "Rx-B Zero Beat: " + String(gLink_ZeroBeat) + " Hz";
							}
						}
						else
						{
							//only on Mini-A
							if (gLinkTypeUDP)
							{
								mnuLink[i] = "Ant Sw IP: " + gUDPAdrsASC;
							}
							else
							{
								mnuLink[i] = "";
							}
						}						
						break;
					}
				}
				gMnuSel = mnuLinkSel;
				statusMsg = "Setup Link Functions";
			}
			gMnuSelMax = mnuLinkMax;
			PrintMenu(mnuLink);
			break;
		case MNU_TXMSG_EDIT:
			//open TxMsg edit menu
			if (refresh)
			{
				gMnuSel = mnuLastTxMsg;				
				statusMsg = "Select Tx Msg to Edit";
			}
			gMnuSelMax = mnuTxMsgMax;
			PrintMenu(mnuTxMsg);
			break;
		case MNU_RADIOCAT:			
			if (refresh)
			{
				if (gMnuSelB4Help)
				{
					//restore gMnuSel when closing help
					gMnuSel = gMnuSelB4Help;					
				}
				else
				{
					gMnuSel = gRadioCAT;
				}
				statusMsg = "Select Radio CAT";
			}
			if (gRadioCAT == RADIO_CAT_RIGCTLD)
			{
				mnuRadioCAT[RADIO_CAT_IPADRS] = gIPAdrs;
			}
			else
			{
				mnuRadioCAT [RADIO_CAT_IPADRS]= ""; //hide it
			}
			gMnuSelMax = mnuRadioCATMax;
			PrintMenu(mnuRadioCAT);
			break;
		case MNU_RADIOBAUD:
			if (refresh)
			{
				//find matching baud
				if (gMnuSelB4Help)
				{
					//restore gMnuSel when closing help
					gMnuSel = gMnuSelB4Help;					
				}
				else
				{
					gMnuSel = 3; //default to 9600
					for (int i = 0; i < mnuRadioBaudMax; i++)
					{
						if (gRadioBaud[gRadioCAT] == mnuRadioBaud[i].toInt())
						{
							gMnuSel = i;
							break;
						}
					}
				}
				statusMsg = "Select Radio Baud Rate";
			}
			gMnuSelMax = mnuRadioBaudMax;
			PrintMenu(mnuRadioBaud);
			break;
		case MNU_ANT_SW:
			if (refresh)
			{
				gMnuSel = mnuASCSel;
				statusMsg = "Select Ant Sw Option";
			}
			for (int i = 0; i < mnuASCMax; i++)
			{
				switch (i)
				{
				case MNU_ANT_SW_PORT_A:
					if (!gAntSwLock)
					{
						if (gAntSwActive)
						{
							mnuASC[i] = "Port-A: " + mnuAntLbl[gAntSwA] + " (" + String(gAntSwA) + ")";
						}
						else
						{
							mnuASC[i] = "Port-A: " + mnuAntLbl[0]; //open
						}
					}
					else
					{
						mnuASC[i] = ""; //blocked
					}
					break;
				case MNU_ANT_SW_PORT_B:
					if (!gAntSwLock)
					{
						if (gAntSwDual)
						{
							if (gAntSwActive)
							{
								mnuASC[i] = "Port-B: " + mnuAntLbl[gAntSwB] + " (" + String(gAntSwB) + ")";
							}
							else
							{
								mnuASC[i] = "Port-B: " + mnuAntLbl[0]; //open
							}
						}
						else
						{
							mnuASC[i] = ""; //blank
						}
					}
					else
					{
						mnuASC[i] = ""; //blocked
					}
					break;
				case MNU_ANT_SW_SWAP:
					if (gAntSwDual && !gAntSwLock && gAntSwActive)
					{
						mnuASC[i] = "Swap A/B Ants";
					}
					else
					{
						mnuASC[i] = ""; 
					}
					break;
				case MNU_ANT_SW_LOCK:
					if (gAntSwLock)
					{
						mnuASC[i] = "OP MODE: LOCKED";
					}
					else
					{
						mnuASC[i] = "Op Mode: Ready";
					}
					break;
				case MNU_ANT_SW_ENABLE:
					if (!gAntSwLock)
					{
						if (gAntSwActive)
						{
							mnuASC[i] = "Switch Mode: Active";
						}
						else
						{
							mnuASC[i] = "SWITCH MODE: OFF";
						}
					}
					else
					{
						mnuASC[i] = "";
					}
					break;	
				case MNU_ANT_SW_SEND_BAND:
					if (gSyncMode != SYNC_MODE_FREQ)
					{
						if (!gAntSwSendBand)
						{
							txt = "Off";
						}
						else
						{
							txt = "ENABLED";
						}
						mnuASC[i] = "Send Band Data: " + txt;
					}
					else
					{
						mnuASC[i] = ""; //hide
					}
					break;
				}
			}			
			gMnuSelMax = mnuASCMax;
			PrintMenu(mnuASC);
			break;
		case MNU_ANT_SW_ANT_A_EDIT:
			if (refresh)
			{
				gMnuSel = gAntSwA;
				statusMsg = "Select Ant for Port-A";
			}
			if (gAntSw8Port)
			{
				ptr = 9;
			}
			else
			{
				ptr = 5;
			}
			for (int i = 0; i < ptr; i++)
			{
				if (!i || gAntSwB != i)
				{
					mnuAntList[i] = mnuAntLbl[i];
				}
				else
				{
					mnuAntList[i] = "";//block b-radio ant 
				}
			}
			gMnuSelMax = ptr;
			PrintMenu(mnuAntList);
			break;
		case MNU_ANT_SW_ANT_B_EDIT:
			if (refresh)
			{
				gMnuSel = gAntSwB;
				statusMsg = "Select Ant for Port-B";
			}
			if (gAntSw8Port)
			{
				ptr = 9;
			}
			else
			{
				ptr = 5;
			}
			for (int i = 0; i < ptr; i++)
			{
				if (!i || gAntSwA != i)
				{
					mnuAntList[i] = mnuAntLbl[i];
				}
				else
				{
					mnuAntList[i] = "";//block b-radio ant 
				}
			}
			gMnuSelMax = ptr;
			PrintMenu(mnuAntList);
			break;
		case MNU_ANT_SW_LBL_EDIT:
			if (refresh)
			{
				gMnuSel = mnuAntEditSel;
				statusMsg = "Select Antenna to RENAME";
			}
			//don't print OPEN label
			if (gAntSw8Port)
			{
				ptr = 8;
			}
			else
			{
				ptr = 4;
			}
			for (int i = 0; i < ptr; i++)
			{
				mnuAntList[i] = String(i+1) + ": " + mnuAntLbl[i + 1];//ant labels			
			}
			gMnuSelMax = ptr;
			PrintMenu(mnuAntList);
			break;
		case MNU_THEME:
			if (refresh)
			{
				gMnuSel = mnuThemeSel;
				statusMsg = "Select a Theme";
			}
			gMnuSelMax = mnuThemeMax;
			PrintMenu(mnuTheme);
			break;
		case MNU_AMPSEQ:
			if (refresh)
			{
				gMnuSel = mnuAmpSeqSel;
				statusMsg = "Amp Sequencing Options";
			}
			for (int i = 0; i < gMnuSelMax; i++)
			{
				switch (i)
				{
				case 0: //enable-disable
					if (gEnabAmpSeq)
					{
						mnuAmpSeq[i] = "Amp Sequencing: ON";
					}
					else
					{
						mnuAmpSeq[i] = "Amp Sequencing: off";
					}
					break;
				case 1: //pre delay
					mnuAmpSeq[i] = "PTT->Key Pickup delay: " + String(gSeqPre) + " mS";
					break;
				case 2: //post delay
					mnuAmpSeq[i] = "Key->PTT Release delay: " + String(gSeqPost) + " mS";
					break;
				}
			}
			gMnuSelMax = mnuAmpSeqMax;
			PrintMenu(mnuAmpSeq);
			break;
		case MNU_CW_KEYERS:
			//open keyer select menu
			if (refresh)
			{
				gMnuSel = mnuCWKeyerSel;
				statusMsg = "Select Keyer Type";
			}
			gMnuSelMax = mnuCWKeyerMax;
			PrintMenu(mnuCWKeyer);
			break;
		}
	}

	if (statusMsg.length())
	{
		PrtStatus(statusMsg, statusRed);		
	}
	lastStatusMsg = statusMsg; //save msg so we can redisplay it in terminal when navigating menu with up/down keys
	TermOut(TERM_SHOW_CURSOR, false); //turn on cursor

}

void PrintMenu(String ary[])
{
	//draw  menu on first encoder click
	//default to 3 columns	
	static int lastgMnuSel = -1;
	bool refreshAll = true;
	String str = ""; //trimmed ary[] to fit box

	//store ary[] in mnuArray[] so we can see it outside of this routine
	for (int i = 0; i < gMnuSelMax; i++)
	{
		mnuArray[i] = ary[i];
		mnuArray[i].trim();
	}

	gMnuSelB4Help = 0; //reset this if it was active

	long x = 0;
	long y = 24;
	long w = 320;
	long h = 210 - y;

	if (gMnu == gMnuPrev)
	{
		//if we're on the same mnu, don't blank screen
		refreshAll = false;
	}
	else
	{
		refreshAll = true;
		lastgMnuSel = -1;
		tft.fillRect(x, y, w, h, hmiBase);
	}
	gMnuPrev = gMnu;

	//start expecting 3 columns
	x += 5;
	y += 10;
	w = 120;
	h = 24;

	if (gMnuSelMax < 8)
	{
		//no columns
		x = 10;
		w = 320;
	}
	else if (gMnuSelMax < 15)
	{
		//use 2 columns
		x = 15;
		w = 157;
	}
	
	if (gComMode == COM_MODE_TERM)
	{
		//clear terminal and move up # of lines we're going to print
		TermOut(TERM_CLS + TERM_HIDE_CURSOR, false);
		for (int i = 0; i < gMnuSelMax + 2; i++)
		{
			TermOut(TERM_CURSOR_UP, false);
		}
	}

	for (int i = 0; i < gMnuSelMax; i++)
	{
		if (i == gMnuSel)
		{
			//highlight
			tft.fillRoundRect(x - 5, y - 5, w - 10, h, 5, hmiBGnd);
		}
		else if (i == lastgMnuSel)
		{
			//blank out last selected item
			tft.fillRoundRect(x - 5, y - 5, w - 10, h, 5, hmiBase);
		}
		//trim to fit
		if (gMnuSelMax > 14)
		{
			//three columns
			str = ary[i].substring(0, 9);
		}
		else if (gMnuSelMax > 7)
		{
			str = ary[i].substring(0, 16);
		}
		else
		{
			str = ary[i];
		}
		PrtTxt(str, 1, 1, hmiFGnd, x-10, y, w);
		
		//send menu to terminal if enabled
		if (gComMode == COM_MODE_TERM)
		{
			String txtAttrib = TERM_NORM;
			if (i == gMnuSel)
			{
				txtAttrib = TERM_INV; //highlight selected option
			}			
			str.trim();
			if (!str.length())
			{
				str = "          "; //so we print an empty menu bar
			}
			if (gMnu == MNU_TXMSG || gMnu == MNU_TXMSG_EDIT)
			{
				//add record # to string
				String pad = "";
				if (i < 9)
				{
					pad = " ";
				}
				str = pad + "[" + String(i + 1) + "] " + str;
			}
			TermOut("   " + txtAttrib + str + TERM_NORM, true);
		}

		//next row
		y += h;
		//second and third columns if needed
		if (i > 12 && x < 200)
		{
			//column 3
			x = 220;
			y = 34;
		}
		else if (i > 5 && x < 100 && gMnuSelMax>7)
		{
			//column 2
			if (gMnuSelMax < 15)
			{
				x = 165;
			}
			else
			{
				//center of three columns
				x = 115;
			}
			y = 34;
		}
	}
	lastgMnuSel = gMnuSel;
	//top line menus
	PrtBtnA(true);	
	PrtBtnB();
	PrtBtnC();
	PrtWiFi();
}


void ExecuteMenuSelection()
{
	//encoder was clicked on gMnu, set the selected option	
	//SerialOut("gMnu= " + String(gMnu) + ", gMnuSel= " + String(gMnuSel), true);
	String txt = "";

	switch (hmiMode)
	{
	case HMI_MENU:		
		ExecMnuSel();
		break;
	case HMI_VALUE:
		ExecValueSel();
		break;
	case HMI_TXTINPUT:		
		ExecTxtInputSel();
		break;		
	}
}

void CopyMem2VFO()
{
	if (gVFO == MNU_VFO_MEM)
	{
		gVFO = MNU_VFO_A;
		gVFO_Freq[MNU_VFO_A] = gVFO_Freq[MNU_VFO_A + 2];
		gVFO_Mode[MNU_VFO_A] = gVFO_Mode[MNU_VFO_A + 2];
		gVFO_BW[MNU_VFO_A] = gVFO_BW[MNU_VFO_A + 2];
		gVFO_Freq[MNU_VFO_B] = gVFO_Freq[MNU_VFO_B + 2];
		gVFO_Mode[MNU_VFO_B] = gVFO_Mode[MNU_VFO_B + 2];
		gVFO_BW[MNU_VFO_B] = gVFO_BW[MNU_VFO_B + 2];
		if (!gMSelMode)
		{
			gFreqRadio = gVFO_Freq[MNU_VFO_A];
			gRMode = gVFO_Mode[MNU_VFO_A];
			gPCRBW = gVFO_BW[MNU_VFO_A];
		}
		else
		{
			gFreqRadio = gVFO_Freq[MNU_VFO_B];
			gRMode = gVFO_Mode[MNU_VFO_B];
			gPCRBW = gVFO_BW[MNU_VFO_B];
		}
	}
}

void ExecMnuSel()
{
	// update the menu in HMI_MENU mode
	String txt = "";
	uint32_t tempFreq = 0; //for vfo swap operation
	uint8_t tempMode = 0;
	uint8_t tempBW = 0;
	uint8_t offset = 0;
	uint8_t temp = 0;

	if (gVFO == MNU_VFO_MEM)
	{
		offset = 2; //to work with a & b memories instead of a & b vfos
	}

	//SerialOut("gMnu=  " + String(gMnu),true);

	switch (gMnu)
	{
	case MNU_MAIN:
	case MNU_FREQ:
	case MNU_CONFIG:
		//select sub menu from main menu
		gBtnPress = BTN_RIGHT; //Loop will send us to NextMnu()
		break;
	case MNU_BAND:
		ChngBand(gMnuSel);
		hmiEncoderTimeout = 0;
		break;
	case MNU_MODE:
		//select mode
		if (!gLock)
		{
			if (!gMnuSel)
			{
				gTxMode = !gTxMode;
				SendASCCmd(ASC_CMD_TXMODE, 0, 0);
			}
			else
			{
				gRMode = gMnuSel - 1;
				TxMode2Radio();				
				gSaveRadio = true;
				SendLinkParams();				
			}								
		}
		hmiEncoderTimeout = 0; //force back to home page in Loop
		break;
	case MNU_VFO:
		//select vfo mode
		mnuVFOSel = gMnuSel;
		switch (gMnuSel)
		{
		case MNU_VFO_A:
		case MNU_VFO_B:		
			gVFO = gMnuSel;
			gFreqRadio = gVFO_Freq[gVFO];
			if (gFreqRadio < 100000)
			{
				gFreqRadio = 10000000;
			}
			gRMode = gVFO_Mode[gVFO];
			gPCRBW = gVFO_BW[gVFO];
			gLock = false;
			break;		
		case MNU_VFO_SPLIT:
			gVFO = gMnuSel;
			gFreqRadio = gVFO_Freq[MNU_VFO_A];
			gRMode = gVFO_Mode[MNU_VFO_A];
			gPCRBW = gVFO_BW[MNU_VFO_A];
			gLock = false;
			break;
		case MNU_VFO_MEM:
			gVFO = gMnuSel;
			if (gMSel > 0)
			{
				gVFO_Freq[MNU_VFO_A + offset] = vfoMemA_Freq[gMSel - 1];
				gVFO_Mode[MNU_VFO_A + offset] = vfoMemA_Mode[gMSel - 1];
				gVFO_BW[MNU_VFO_A + offset] = vfoMemA_BW[gMSel - 1];
				gVFO_Freq[MNU_VFO_B + offset] = vfoMemB_Freq[gMSel - 1];
				gVFO_Mode[MNU_VFO_B + offset] = vfoMemB_Mode[gMSel - 1];
				gVFO_BW[MNU_VFO_B + offset] = vfoMemB_BW[gMSel - 1];
			}			
			if (!gMSelMode)
			{
				gFreqRadio = gVFO_Freq[MNU_VFO_A + offset];
				gRMode = gVFO_Mode[MNU_VFO_A + offset];
				gPCRBW = gVFO_BW[MNU_VFO_A + offset];
			}
			else
			{
				gFreqRadio = gVFO_Freq[MNU_VFO_B + offset];
				gRMode = gVFO_Mode[MNU_VFO_B + offset];
				gPCRBW = gVFO_BW[MNU_VFO_B + offset];
			}			
			gLock = false;
			break;
		case MNU_VFO_MODE:
			gMSelMode = !gMSelMode; //toggle mem tuning mode
			gVFO = MNU_VFO_MEM; //force to memory tuning mode
			gSaveRadio = true;
			break;
		case MNU_VFO_LOCK:
			gLock = !gLock;
			break;
		case MNU_VFO_FUTURE:
			//future use
			break;
		case MNU_VFO_COPYAB:			
			gVFO_Freq[MNU_VFO_B + offset] = gVFO_Freq[MNU_VFO_A + offset];
			gVFO_Mode[MNU_VFO_B + offset] = gVFO_Mode[MNU_VFO_A + offset];
			gVFO_BW[MNU_VFO_B + offset] = gVFO_BW[MNU_VFO_A + offset];
			if (!offset)
			{
				gVFO = MNU_VFO_B; //switch to VFO-B if not in memory mode
				gFreqRadio = gVFO_Freq[MNU_VFO_B + offset];
				gRMode = gVFO_Mode[MNU_VFO_B + offset];
				gPCRBW = gVFO_BW[MNU_VFO_B + offset];
			}			
			break;
		case MNU_VFO_COPYBA:
			//copy b to a only if b is valid			
			if (gVFO_Freq[MNU_VFO_B + offset] >= 100000)
			{
				gVFO_Freq[MNU_VFO_A + offset] = gVFO_Freq[MNU_VFO_B + offset];
				gVFO_Mode[MNU_VFO_A + offset] = gVFO_Mode[MNU_VFO_B + offset];
				gVFO_BW[MNU_VFO_A + offset] = gVFO_BW[MNU_VFO_B + offset];
				gFreqRadio = gVFO_Freq[MNU_VFO_A + offset];
				gRMode = gVFO_Mode[MNU_VFO_A + offset];
				gPCRBW = gVFO_BW[MNU_VFO_A + offset];
			}
			else
			{
				//copy a to b
				gVFO_Freq[MNU_VFO_B + offset] = gVFO_Freq[MNU_VFO_A + offset];
				gVFO_Mode[MNU_VFO_B + offset] = gVFO_Mode[MNU_VFO_A + offset];
				gVFO_BW[MNU_VFO_B + offset] = gVFO_BW[MNU_VFO_A + offset];
				gFreqRadio = gVFO_Freq[MNU_VFO_B + offset];
				gRMode = gVFO_Mode[MNU_VFO_B + offset];
				gPCRBW = gVFO_BW[MNU_VFO_B + offset];
			}
			if (gVFO != MNU_VFO_MEM)
			{
				gVFO = MNU_VFO_A; //switch to VFO-A if not in memory mode
			}
			break;
		case MNU_VFO_SWAPAB:
			tempFreq = gVFO_Freq[MNU_VFO_A + offset];
			tempMode = gVFO_Mode[MNU_VFO_A + offset];
			tempBW = gVFO_BW[MNU_VFO_A + offset];
			if (gVFO_Freq[MNU_VFO_B + offset] >= 100000)
			{
				//only if valid
				gVFO_Freq[MNU_VFO_A + offset] = gVFO_Freq[MNU_VFO_B + offset];
				gVFO_Mode[MNU_VFO_A + offset] = gVFO_Mode[MNU_VFO_B + offset];
				gVFO_BW[MNU_VFO_A + offset] = gVFO_BW[MNU_VFO_B + offset];
			}
			gVFO_Freq[MNU_VFO_B + offset] = tempFreq;
			gVFO_Mode[MNU_VFO_B + offset] = tempMode;
			gVFO_BW[MNU_VFO_B + offset] = tempBW;
			switch (gVFO)
			{
			case MNU_VFO_A:
			case MNU_VFO_SPLIT:
				gFreqRadio = gVFO_Freq[MNU_VFO_A];
				gRMode = gVFO_Mode[MNU_VFO_A];
				gPCRBW = gVFO_BW[MNU_VFO_A];
				break;
			case MNU_VFO_B:
				gFreqRadio = gVFO_Freq[MNU_VFO_B];
				gRMode = gVFO_Mode[MNU_VFO_B];
				gPCRBW = gVFO_BW[MNU_VFO_B];
				break;
			case MNU_VFO_MEM:
				if (!gMSelMode)
				{
					gFreqRadio = gVFO_Freq[MNU_VFO_A + offset];
					gRMode = gVFO_Mode[MNU_VFO_A + offset];
					gPCRBW = gVFO_BW[MNU_VFO_A + offset];
				}
				else
				{
					gFreqRadio = gVFO_Freq[MNU_VFO_B + offset];
					gRMode = gVFO_Mode[MNU_VFO_B + offset];
					gPCRBW = gVFO_BW[MNU_VFO_B + offset];
				}
				break;
			}
			break;
		case MNU_VFO_COPY2A:
			//copy mem to both vfos if in MEM mode
			CopyMem2VFO();
			break;
		case MNU_VFO_PUSH:
			if (gLinkOnline && gSyncMode == SYNC_MODE_BASIC)
			{
				gVFO_PushPull = SYNC_MODE_BASIC_PUSH; //push freq info to other mini			
				if (gLinkMiniA)
				{
					txt = "B";
				}
				else
				{
					txt = "A";
				}
				txt = "Push Freq/Mode to Mini-" + txt;
			}
			else
			{
				txt = "Basic Link Mode Not Active";
			}
			hmiTagLineRefresh = 0;
			PrtStatus(txt, true);
			SendLinkParams();
			break;
		case MNU_VFO_PULL:
			if (gLinkOnline && gSyncMode == SYNC_MODE_BASIC)
			{
				gVFO_PushPull = SYNC_MODE_BASIC_PULL; //pull freq info from other mini			
				if (gLinkMiniA)
				{
					txt = "B";
				}
				else
				{
					txt = "A";
				}
				txt = "Pull Freq/Mode from Mini-" + txt;
			}
			else
			{
				txt = "Basic Link Mode Not Active";
			}
			hmiTagLineRefresh = 0;
			PrtStatus(txt, true);
			SendLinkParams();
			break;
		}		
		gMnuPrev = -1;
		PrtHomePg();
		break;
	case MNU_TXMSG:
		PrtHomePg();
		mnuTxMsgSel = gMnuSel;
		if (gTxBuffer.length())
		{
			//add BT between buffers
			gTxBuffer = gTxBuffer += " = " + mnuTxMsg[mnuTxMsgSel];
		}
		else
		{
			gTxBuffer = mnuTxMsg[mnuTxMsgSel];
		}
		SaveRepeatBuffer(mnuTxMsg[mnuTxMsgSel]); //save for F1 to repeat it
		gPlay = KEYER_MODE_RUN; //start keyer
		//send buffer to terminal
		TermOut("\n\rSend: " + gTxBuffer, true);		
		return;
		break;	
	case MNU_TXMSG_EDIT:
		//open TxMsg to select Tx Msg to edit		
		mnuLastTxMsg = gMnuSel;
		txtInputSrc = gMnuSel;
		//SerialOut("ExecuteMenuSelection: txtInputSrc= " + String(gMnuSel), true);
		SetMnuParam(); //open paramter setting window
		break;		
	case MNU_KEYER:		
		SetMnuParam(); //open paramter setting window
		return;
		break;
	case MNU_WIFI:		
		mnuWiFiSel = gMnuSel;
		SetMnuParam(); //open paramter setting window
		return;
		break;
	case MNU_LINK:
		mnuLinkSel = gMnuSel;
		if (gLinkMiniA || gMnuSel != MNU_LINK_MODE)
		{
			SetMnuParam();
		}
		else
		{
			PrtStatus("Change Mode on Mini-A", true);
			delay(1500);
		}
		return;
		break;
	case MNU_ANT_SW:
		mnuASCSel = gMnuSel;
		switch (gMnuSel)
		{
		case MNU_ANT_SW_PORT_A:
			gMnu = MNU_ANT_SW_ANT_A_EDIT;
			gMnuSel = gAntSwA;
			break;
		case MNU_ANT_SW_PORT_B:
			gMnu = MNU_ANT_SW_ANT_B_EDIT;
			gMnuSel = gAntSwB;
			break;
		case MNU_ANT_SW_LABELS:
			gMnu = MNU_ANT_SW_LBL_EDIT;
			gMnuSel = mnuAntEditSel;
			break;
		case MNU_ANT_SW_SWAP:
			SwapAnt();
			break;
		case MNU_ANT_SW_LOCK:
			SendASCCmd(ASC_CMD_LOCK, 0, 0); //toggle lock
			break;
		case MNU_ANT_SW_ENABLE:
			SendASCCmd(ASC_CMD_ENABLE, 0, 0); //toggle active
			break;
		case MNU_ANT_SW_SEND_BAND:
			gAntSwSendBand = !gAntSwSendBand;
			SaveInitFile();
			if (gAntSwSendBand)
			{
				SendASCCmd(ASC_CMD_BAND, 0, 0);				
			}			
			PrtHomePg();
			return;
			break;
		}
		gMnuPrev = -1;
		CreateMnu("", false);
		break;
	case MNU_ANT_SW_ANT_A_EDIT:
		SendASCCmd(ASC_CMD_A_PORT, 1, gMnuSel);
		gMnu = MNU_ANT_SW;
		gMnuSel = mnuASCSel;
		gMnuPrev = -1;
		CreateMnu("", false);
		break;
	case MNU_ANT_SW_ANT_B_EDIT:
		SendASCCmd(ASC_CMD_B_PORT, 2, gMnuSel);
		gMnu = MNU_ANT_SW;
		gMnuSel = mnuASCSel;
		gMnuPrev = -1;
		CreateMnu("", false);
		break;
	case MNU_ANT_SW_LBL_EDIT:		
		txtInputSrc = gMnuSel;
		gMnuSel = gMnu;
		txt = mnuAntLbl[txtInputSrc+1];
		txt.trim();
		GetTxtInput(txt);
		break;
	case MNU_RADIOPORT:
		switch (gMnuSel)
		{
		case MNU_PORT_COPY:
		case MNU_PORT_ERASE:
			SetMnuParam();
			break;
		case MNU_PORT_RENAME:
			gMnuSel = gMnu; //so txt input knows what to do 
			txtInputSrc = gRadioPort;
			txt = mnuRadioTag[txtInputSrc];
			txt.trim();
			if (txt.charAt(0) == '*')
			{
				txt = txt.substring(1, txt.length()); //trim off shared db flag
			}
			GetTxtInput(txt);			
			break;
		default:
			SaveRadio(gRadioPort, true);
			gRadioPort = gMnuSel;
			SaveInitFile(); //so we start with this radio
			LoadRadio();
			PrtHomePg();
			break;
		}
		break;

		//sub menu options begin here
	case MNU_FREQ_LD:
		//load selected memory slot
		mnuFreqListSel = gMnuSel;
		gMSel = gMnuSel + mnuFreqListStart + 1; //gMSel is 1s based, Loop() will update freq
		gMSelPrev = 0xff;//so change occurs in Loop()
		gMnuPrev = -1;
		gSaveRadio = true;
		PrtHomePg();
		break;
	case MNU_FREQ_SV:
		//save freq & mode to this slot
		mnuFreqListSel = gMnuSel;
		gMSel = gMnuSel + mnuFreqListStart + 1; //gMSel is 1s based
		gMnuPrev = -1;		
		if (gMSelMode)
		{
			offset = 2; //use memory info
		}
		//save all current settings to this memory slot			
		vfoMemA_Freq[gMSel - 1] = gVFO_Freq[MNU_VFO_A + offset];
		vfoMemA_Mode[gMSel - 1] = gVFO_Mode[MNU_VFO_A + offset];
		vfoMemA_BW[gMSel - 1] = gVFO_BW[MNU_VFO_A + offset];

		vfoMemB_Freq[gMSel - 1] = gVFO_Freq[MNU_VFO_B + offset];
		vfoMemB_Mode[gMSel - 1] = gVFO_Mode[MNU_VFO_B + offset];
		vfoMemB_BW[gMSel - 1] = gVFO_BW[MNU_VFO_B + offset];
	
		SaveFreqList(gRadioPort);
		//open MNU_FREQ_RENAME to edit label
		gMnuSel = gMnu; //so txt input knows what to do 
		txtInputSrc = gMSel - 1;
		if (!txtInputSrc)
		{
			GetTxtInput(gMemLabel); //used mem label from last load in case we're copying shared or normal
		}
		else
		{
			GetTxtInput(vfoMemLabel[txtInputSrc]); //use this memory's label
		}		
		break;
	case MNU_FREQ_REN:
		//open txtInput edit
		mnuFreqListSel = gMnuSel;
		gMnuSel = gMnu; //so txt input knows what to do 
		gBtnPress = BTN_RIGHT; //Loop calls ProcessBtn() to respond to freq menu change
		break;
	case MNU_RADIO:
		//these menus only appear when the radio CAT is selected
		switch (gRadioCAT)
		{
		case RADIO_CAT_FLEX:
			SetMnuParam();
			return;
			break;
		case RADIO_CAT_PCR1000:
			SetMnuParam();
			return;
			break;
		default:
			return;//do nothing
			break;
		}
		break;
	//Config menu selections
	case MNU_C_BTN:		
		mnuConfigSel = MNU_CONFIG_C_BTN;
		mnuCBtnSel = gMnuSel;		
		SaveInitFile();
		gMnuPrev = -1; //refresh		
		PrtHomePg();
		break;	
	case MNU_FUNC_BTN:
		if (gMnuSel < mnuFuncBtnMax - 1)
		{
			mnuConfigSel = MNU_CONFIG_FUNC;
			mnuFuncBtnSel = gMnuSel;
			mnuFuncBtnOptionSel = mnuFuncBtnIdx[gMnuSel];
			gMnu = MNU_FUNC_BTN_OPTIONS;
			gMnuPrev = -1;
			CreateMnu("", false);
		}
		else
		{
			gStuckBtn = 0; //reset stuck btn and let ChkRmtFuncBtns() test it again
			PrtHomePg();
			PrtStatus("FUNC KEY LOCKOUT RESET", true);
		}
		break;	
	case MNU_FUNC_BTN_OPTIONS:				
		mnuFuncBtnOptionSel = mnuFuncBtnIdx[gMnuSel];
		mnuFuncBtnIdx[mnuFuncBtnSel] = gMnuSel;
		mnuFuncBtnLbl[mnuFuncBtnSel] = mnuFuncBtnOption[gMnuSel];
		SaveInitFile();
		gMnu = MNU_FUNC_BTN;
		gMnuPrev = -1;
		CreateMnu("", false);
		break;
	case MNU_THEME:
		mnuConfigSel = MNU_CONFIG_THEME;
		mnuThemeSel = gMnuSel;
		SetThemeColors();
		SaveInitFile();
		PrtHomePg();
		break;	
	case MNU_TAGLINE:
		mnuConfigSel = MNU_CONFIG_TAGLINE;
		//SerialOut("ExecuteMenuSelection - MNU-TAGLINE", true);
		return;
		break;

	case MNU_RADIOCAT:
		if (gMnuSel == RADIO_CAT_IPADRS)
		{
			SetMnuParam();			
			return;
		}
		temp = gRadioCAT; //temp hold
		gRadioCAT = gMnuSel;
		gSaveRadio = true; //save when we can
		mnuConfigSel = MNU_CONFIG_CAT;
		SetRadioBaud();
		if (gRadioCAT == RADIO_CAT_ICOM)
		{
			//allow user to edit icomAdr on Icom selection
			SetMnuParam(); //open paramter setting window
			return;
		}
		else if (gRadioCAT == RADIO_CAT_RIGCTLD && temp != RADIO_CAT_RIGCTLD)
		{
			//stay on CAT menu and show IP adrs if we just enabled it
			gMnuPrev = -1;
			CreateMnu("", false);
			return;
		}
		else
		{
			PrtHomePg();
		}
		break;
	case MNU_RADIOBAUD:
		//set selected baud
		gRadioBaud[gRadioCAT] = mnuRadioBaud[gMnuSel].toInt();
		gSaveRadio = true; //save when we can
		mnuConfigSel = MNU_CONFIG_BAUD;
		SetRadioBaud();
		SaveRadioBaud(gRadioPort);
		PrtHomePg();
		break;
	case MNU_AMPSEQ:
		mnuAmpSeqSel = gMnuSel;		
		switch (gMnuSel)
		{
		case 0: //enable/disable sequencing
			if (!gFlexApiPTT && !gFlexApiKey)
			{
				gEnabAmpSeq = !gEnabAmpSeq;	
				CreateMnu("", false);
			}
			else
			{
				gEnabAmpSeq = false;
				PrtStatus("Turn OFF Flex PTT/Key w/API", true);
			}			
			break;
		case 1: //pre or post key timers
		case 2:
			SetMnuParam();			
			break;
		}
		break;
	case MNU_CW_KEYERS:
		//select the desired keyer
		gKeyType = gMnuSel;
		gKeyTypePrev = gKeyType;
		mnuCWKeyerSel = gKeyType;
		SendLinkParams();
		gSaveRadio = true;
		PrevMnu(); //move back to CW menu
		break;
	}
}

void ExecValueSel()
{
	// update selected value from HMI_VALUE mode

	String txt = "";
		
	hmiMode = HMI_MENU;
	switch (gMnu)
	{
	case MNU_FREQ:
		if (gMnuSel == MNU_FREQ_RIT)
		{
			gRIT = gMnuValue;
			SendLinkParams();
			gMnuPrev = -1;
			PrtHomePg();	
		}
		break;
	case MNU_KEYER:
		switch (gMnuSel)
		{
		case MNU_KEYER_SPEED:
			gCWSpeed = gMnuValue;
			gFarnsSpeed = gMnuValue; //follow speed
			CalcDitDur();
			CalcDitFarnsDur();
			SendLinkParams();
			gMnuPrev = -1; //refresh
			gSaveRadio = true;
			PrtHomePg();
			break;
		case MNU_KEYER_FARNS:
			gFarnsSpeed = gMnuValue;
			CalcDitFarnsDur();
			SendLinkParams();
			gMnuPrev = -1; //refresh
			gSaveRadio = true;
			PrtHomePg();
			break;
		case MNU_KEYER_SN:
			mnuKeyerSN = gMnuValue;
			SendLinkParams();
			gMnuPrev = -1;
			CreateMnu("", false);
			break;
		}
		break;
	case MNU_LINK:
		gMnuPrev = -1; //refresh
		gSaveRadio = true;
		PrtHomePg();
		break;
	case MNU_RADIOPORT:
		switch (gMnuSel)
		{
		case MNU_PORT_COPY:
			if (gMnuValue != gRadioPort + 1)
			{
				//copy current settings to new port
				PrtStatus("Copying Port [" + String(gRadioPort+1) + "] to [" + String(gMnuValue) + "]", true);
				delay(500);
				mnuRadioTag[gMnuValue - 1] = mnuRadioTag[gRadioPort] + "~";
				SaveRadioTags();
				SaveBandList(gMnuValue - 1);
				SaveFreqList(gMnuValue - 1);
				SaveRadio(gMnuValue - 1, true);
				SaveRadioBaud(gMnuValue - 1);
				gRadioPort = gMnuValue - 1;
				SaveInitFile();				
			}
			PrtHomePg();
			return;
			break;
		case MNU_PORT_ERASE:
			if (gMnuValue == 1)
			{
				PrtStatus("Erasing Radio Port [" + String(gRadioPort + 1) + "]", true);
				delay(500);
				EraseRadio();				
			}
			PrtHomePg();
			return;
			break;
		}
		break;
	case MNU_RADIOCAT:
		//set Icom adrs is the only value on this mnu
		switch (gMnuSel)
		{
		case MNU_ICOM_ADRS_EDIT:
			if (gMnuValue == 0xE0)
			{
				PrtHomePg();
				PrtStatus("0xE0 is an Invalid Address", true);
				return;
			}
			gIcmAdr = gMnuValue;
			gMnuPrev = -1;
			gSaveRadio = true;
			PrtHomePg();			
			break;
		}
		break;
	case MNU_RADIO:
		switch (gRadioCAT)
		{
		case RADIO_CAT_FLEX:
			switch (gMnuSel)
			{
			//case MNU_FLEX_PAN:	
			//	gFlexPan = !gFlexPan;
			//	break;
			case MNU_FLEX_ANT:
				if (!gTCPCon)
				{
					//only change if offline
					gFlexAnt = gMnuValue;
				}
				break;
			case MNU_FLEX_VOL:
			case MNU_FLEX_PO:
			case MNU_FLEX_TO:
			case MNU_FLEX_WNB:
			case MNU_FLEX_NB:
			case MNU_FLEX_NR:
			case MNU_FLEX_AF:
				UpdateParam();
				break;
			}
			gMnuPrev = -1;
			gSaveRadio = true;
			CreateMnu("", false);
			break;
		case RADIO_CAT_PCR1000:
			//save new values to PCR
			UpdateParam();
			gMnuPrev = -1; //force refresh
			gSaveRadio = true;
			CreateMnu("", false);			
			break;
		default:
			//nothing to do			
			break;
		}
	case MNU_AMPSEQ:
		switch (gMnuSel)
		{
		case 1:
			gSeqPre = gMnuValue;
			gMnuPrev = -1;
			CreateMnu("", false);
			break;
		case 2:
			gSeqPost = gMnuValue;
			gMnuPrev = -1;
			CreateMnu("", false);
			break;
		}
		break;
	}	
}

void ExecTxtInputSel()
{
	//update selected text with new txt and save in HMI_TXTINPUT mode

	uint8_t oldSlice = gFlexSlice; //to detect change

	hmiMode = HMI_MENU;
	gMnuPrev = -1; //so we refresh on exit
	txtInput[txtInputMax - 1] = 0; //make sure we're terminated

	String txt = String(txtInput);
	if (!txt.length())
	{
		return; //nothing entered so ignore - to erase, enter ' '
	}
	if (txt.charAt(0) == ' ')
	{
		//delete string
		txt = "";
		//set default for this entry
		switch (gMnuSel)
		{
		case MNU_ANT_SW_LBL_EDIT:
			txt = "[OFF]";
			break;
		case MNU_RADIOPORT:
			txt = "Port " + String(gRadioPort + 1);
			break;
		case MNU_CONFIG_TAGLINE_EDIT:
			txt = "CTR2-Mini by KU7Q";
			break;
		case MNU_RIGCTLD_IP_EDIT:
		case MNU_FLEX_IP_EDIT:
			txt = gIPAdrsDefault;
			break;
		default:
			txt = " ";
			break;
		}
	}
	else
	{
		txt.trim();
	}
	//SerialOut("gMnuSel= " + String(gMnuSel), true);
	switch (gMnuSel)
	{
	case MNU_FREQ_SV: //fall through
	case MNU_FREQ_REN:
		//we just accepted the tag name for this memory slot
		vfoMemLabel[txtInputSrc] = txt;
		if (!txt.length())
		{
			//delete freq & mode if label is blank
			vfoMemA_Freq[txtInputSrc] = 0;
			vfoMemA_Mode[txtInputSrc] = 0;
			vfoMemA_BW[txtInputSrc] = 0;
			vfoMemB_Freq[txtInputSrc] = 0;
			vfoMemB_Mode[txtInputSrc] = 0;
			vfoMemB_BW[txtInputSrc] = 0;
		}
		SaveFreqList(gRadioPort);	
		PrtHomePg();
		return;
		break;
	case MNU_ANT_SW_LBL_EDIT:
		mnuAntLbl[txtInputSrc+1] = txt;
		SendASCCmd(ASC_CMD_LABEL, 0, txtInputSrc+1);
		mnuAntEditSel = txtInputSrc;
		gMnu = MNU_ANT_SW;
		gMnuPrev = -1;
		gMnuSel = MNU_ANT_SW_LABELS;
		break;
	case MNU_RADIOPORT:
		//we just accepted the tag name for this radio port
		mnuRadioTag[gRadioPort] = " " + txt; //make allowance for shared db flag
		SaveRadioTags();
		PrtHomePg();
		return;
		break;
	case MNU_TXMSG_EDIT:
		//update cw tx buffer	
		//SerialOut("Saving tx buffer: " + txt, true);
		mnuTxMsg[mnuTxMsgSel] = txt;
		SaveTxMsg(gRadioPort);
		gMnu = MNU_TXMSG_EDIT; //go back to calling menu		
		break;
	case MNU_WIFI_SSID_EDIT:
		gSSID = txt;
		SaveInitFile();
		gMnu = MNU_WIFI;
		gMnuSel = MNU_WIFI_SSID;
		break;
	case MNU_WIFI_STAKEY_EDIT:
		gStaKey = txt;
		SaveInitFile();
		gMnu = MNU_WIFI;
		gMnuSel = MNU_WIFI_STAKEY;
		break;
	case MNU_UDP_ADRS_EDIT:
		if (txt != gStaAdrs)
		{
			//only if the adrs they entered isn't this unit's station adrs
			gUDPAdrs = txt;
			SaveInitFile();
			if (gUDPConnected)
			{
				//turn UDP off
				udp.stop();
				gUDPConnected = false;
				ResetASC();
			}
		}
		else
		{
			txt = "A";
			if (!gLinkMiniA)
			{
				txt = "B";
			}
			PrtStatus("Use Mini " + txt + "'s Station Adrs", true);
			delay(1500);
		}		
		gMnu = MNU_LINK;
		gMnuSel = MNU_LINK_UDP_ADRS;
		break;
	case MNU_LINK_OPTION6:
		if (txt != gStaAdrs)
		{
			//only if the adrs they entered isn't this unit's station adrs
			gWiFi = false; //force offline if changing addresses
			gAntSwOnline = false;
			gUDPAdrsASC = txt;
			SaveInitFile();
		}
		else
		{
			PrtStatus("Use Ant Sw Station Adrs", true);
			delay(1500);
		}
		gMnu = MNU_ANT_SW;
		gMnuSel = MNU_LINK_OPTION6;
		break;
	case MNU_RIGCTLD_IP_EDIT:
		gIPAdrs = txt;
		gSaveRadio = true;
		SaveRadio(gRadioPort, true);
		gMnu = MNU_RADIOCAT;
		gMnuSel = RADIO_CAT_RIGCTLD;
		SerialOut("rigctld ip=" + gIPAdrs, true);
		break;
	case MNU_FLEX_IP_EDIT:		
		gIPAdrs = txt;
		gSaveRadio = true;		
		gMnu = MNU_RADIO;
		gMnuSel = MNU_FLEX_IP;
		//SerialOut("flex ip=" + gIPAdrs, true);
		break;
	case MNU_FLEX_PORT_EDIT:
		gFlexPort = txt;
		gSaveRadio = true;
		gMnu = MNU_RADIO;
		gMnuSel = MNU_FLEX_PORT;
		break;
	case MNU_FLEX_SLICE_EDIT:
		txt.toUpperCase();
		if (txt == "B")
		{
			gFlexSlice = 1;
		}else if (txt == "C")
		{
			gFlexSlice = 2;
		}else if (txt == "D")
		{
			gFlexSlice = 3;
		}
		else if (txt == "E")
		{
			gFlexSlice = 4;
		}else if (txt == "F")
		{
			gFlexSlice = 5;
		}else if (txt == "G")
		{
			gFlexSlice = 6;
		}else if (txt == "H")
		{
			gFlexSlice = 7;
		}
		else
		{
			gFlexSlice = 0;
		}
		gSaveRadio = true;
		gMnu = MNU_RADIO;
		gMnuSel = MNU_FLEX_SLICE;
		SerialOut("flex slice=" + gFlexSliceLbl[gFlexSlice], true);
		if (gTCPCon)
		{
			//reinitialize if swithing slices
			//if (oldSlice != gFlexSlice)
			{
				Flex_Initialize();				
				PrtHomePg();
				return;
			}
		}
		break;
	case MNU_KEYER_CALL_EDIT:
		mnuKeyerCall = txt;
		SaveInitFile();
		gMnu = MNU_KEYER;
		gMnuSel = MNU_KEYER_CALL;		
		break;
	case MNU_KEYER_EXCH_EDIT:
		mnuKeyerExch = txt;
		SaveInitFile();
		gMnu = MNU_KEYER;
		gMnuSel = MNU_KEYER_EXCH;
		break;
	case MNU_CONFIG_TAGLINE_EDIT:
		hmiTagLine = txt;		
		SaveInitFile();
		PrtHomePg();
		return;
		break;
	}
	CreateMnu("", false);
}

void SetMnuParam()
{
	//set the parameter from the selected menu option
	//SerialOut("SetMnuParam: gMnu= " + String(gMnu) + ", mnuPCRSel= " + String(mnuPCRSel), true);
	
	String txt = "";

	switch (gMnu)
	{
	case MNU_FREQ:
		if (gMnuSel == MNU_FREQ_RIT)
		{			
			PrtStatus("Set the Rx Freq Offset", false);
			gMnuValue = gRIT;
			gMnuValueMin = -9950;
			gMnuValueMax = 9950;
			hmiMode = HMI_VALUE;
			GetMnuValue(true, "RX FREQ OFFSET", LONG_PRESS_YES);
			return;
		}
		break;
	case MNU_RADIOCAT:
		if (gMnuSel == RADIO_CAT_ICOM)
		{
			//set icom adrs				
			gMnuSel = MNU_ICOM_ADRS_EDIT;
			PrtStatus("Set the Icom Address", false);
			gMnuValue = gIcmAdr;
			gMnuValueMin = 0;
			gMnuValueMax = 0xF9; //max allowed icom address - ignore 0xE0 (controller) if user tries to set it too
			hmiMode = HMI_VALUE;
			GetMnuValue(true, "ICOM ADDRESS", LONG_PRESS_YES); //display in hex
			return;
		}
		else if (gMnuSel == RADIO_CAT_IPADRS)
		{
			if (!gTCPCon)
			{
				gMnuSel = MNU_RIGCTLD_IP_EDIT;
				PrtStatus("Set the rigctld IP Adrs", false);
				txtInputSrcStart = 0;				
				if (gIPAdrs == gIPAdrsDefault)
				{
					GetTxtInput(""); //don't make user delete the default label
				}
				else
				{
					GetTxtInput(gIPAdrs);
				}
			}
			else
			{				
				gMnuSel = RADIO_CAT_RIGCTLD; //reverto to rigctld
				hmiTagLineRefresh = 0;
				PrtStatus("Disconnect WiFi First", true);
			}
		}
		return;
		break;
	case MNU_TXMSG_EDIT:
		//edit Tx Msg entry		
		mnuTxMsgSel = gMnuSel;
		mnuLastTxMsg = gMnuSel;
		gMnuSel = gMnu;
		mnuMainSel = gMnu - 2; //subtract for MNU_NONE and MNU_MAIN		
		//send buffer to terminal
		TermOut(" ", true);
		TermOut("Edit: " + mnuTxMsg[mnuTxMsgSel], true);		
		GetTxtInput(mnuTxMsg[mnuTxMsgSel]);
		return;
		break;
	case MNU_TAGLINE:
		//get new tagline			
		mnuConfigSel = gMnuSel;
		GetTxtInput(hmiTagLine);
		return;
		break;
	case MNU_KEYER:
		mnuKeyerSel = gMnuSel;
		mnuMainSel = gMnu - 2; //subtract for MNU_NONE and MNU_MAIN
		switch (gMnuSel)
		{
		case MNU_KEYER_SPEED:			
			PrtStatus("Set Keyer Speed", false);
			gMnuValue = gCWSpeed;
			gMnuValueMin = 5;
			gMnuValueMax = 50;
			hmiMode = HMI_VALUE;
			GetMnuValue(true, "KEYER SPEED", LONG_PRESS_NO);
			return;
			break;
		case MNU_KEYER_FARNS:
			PrtStatus("Set Farnsworth Spacing", false);
			gMnuValue = gFarnsSpeed;
			gMnuValueMin = 5;
			gMnuValueMax = 50;
			hmiMode = HMI_VALUE;
			GetMnuValue(true, "FARNSWORTH SPACING", LONG_PRESS_NO);
			return;
			break;
		case MNU_KEYER_SIDETONE:
			gSideTone = !gSideTone;
			SendLinkParams();
			CreateMnu("Toggled Sidetone function", true);			
			break;
		case MNU_KEYER_TYPE:
			gBtnPress = BTN_RIGHT; //process it in Loop
			break;
		case MNU_KEYER_PDL:
			//toggle option and return to menu
			gPaddle = !gPaddle;
			CreateMnu("Toggled Paddle function", true);			
			break;
		case MNU_KEYER_PTT:
			//step through normal, latching, and straight key modes
			gRmtMode++;
			if (gRmtMode > MNU_RMT_KEY)
			{
				gRmtMode = MNU_RMT_NORM;
			}
			blockISR = true;
			pinMode(RMT_BTNS, OUTPUT); //toggle mode to reset it otherwise key will get stuck on
			delay(10);
			pinMode(RMT_BTNS, INPUT_PULLUP);			
			blockISR = false;
			CreateMnu("Changed Rmt PTT function", true);
			break;
		case MNU_KEYER_CALL:
			//get new call sign		
			gMnuSel = MNU_KEYER_CALL_EDIT;
			GetTxtInput(mnuKeyerCall);
			return;
			break;
		case MNU_KEYER_SN:
			//get new sn
			PrtStatus("Set the Contest Serial #", false);
			gMnuValue = mnuKeyerSN;
			gMnuValueMin = 1;
			gMnuValueMax = 0xffff;
			hmiMode = HMI_VALUE;
			GetMnuValue(true, "CONTEST SERIAL #", LONG_PRESS_YES);
			return;
			break;
		case MNU_KEYER_EXCH:
			//get new exch					
			gMnuSel = MNU_KEYER_EXCH_EDIT;
			GetTxtInput(mnuKeyerExch);
			return;
			break;
		}
		break;
	case MNU_WIFI:
		mnuWiFiSel = gMnuSel;
		mnuMainSel = gMnu - 2; //subtract for MNU_NONE and MNU_MAIN
		switch (mnuWiFiSel)
		{
		case MNU_WIFI_CONNECT:
			//we'll connect in Loop 
			if (!gWiFi && gBLEConnected)
			{				
				PrtStatus("Turn off BLE first", true);
				delay(1500);				
				gMnuPrev = -1;
				CreateMnu("", true);
				return;
				break;
			}
			gWiFi = !gWiFi;			
			PrtHomePg();
			return; //don't set save flag
			break;
		case MNU_WIFI_SSID:
			//get SSID
			gMnuSel = MNU_WIFI_SSID_EDIT;
			txtInputSrcStart = 0;
			GetTxtInput(gSSID);
			break;
		case MNU_WIFI_STAKEY:
			gMnuSel = MNU_WIFI_STAKEY_EDIT;
			txtInputSrcStart = 0;
			GetTxtInput(gStaKey);
			break;
		case MNU_WIFI_AUTO:
			//toggle auto connect
			gWiFiAutoConnect = !gWiFiAutoConnect;
			if (!gWiFiAutoConnect && gLinkTypeUDP)
			{
				gLinkAutoConnect = false; //follow wifi autoconnect if using UDP
			}
			SaveInitFile();
			gMnuPrev = -1;			
			CreateMnu("", false);
			break;
		}
		break;
	case MNU_LINK:
		switch (mnuLinkSel)
		{
		case MNU_LINK_CONNECT:						
			if (!gBLEConnect && !gLinkTypeUDP && gWiFi)
			{
				PrtStatus("Turn off WiFi first", true);
ErrNotice:		delay(1500);
				gMnuPrev = -1;
				CreateMnu("", false);
				return;
				break;
			}
			if (!gLinkTypeUDP)
			{
				gBLEConnect = !gBLEConnect;
			}
			else
			{
				if (!gWiFi)
				{
					//just turn it on, it will automatically connect to router and start UDP if one of the UDP IP addresses are set
					gWiFi = true;
				}
				else
				{
					if ((gUDPAdrs.indexOf('.') == -1) && (gUDPAdrsASC.indexOf('.') == -1))
					{
						PrtStatus("Set an IP Adrs First", true);
						goto ErrNotice;
					}
					if (gUDPConnected)
					{
						//turn UDP off
						udp.stop();
						gUDPConnected = false;						
					}
					else
					{
						//turn it on						
						ConnectUDP();						
						PrtHomePg();
						return;
					}
				}				
			}	
			ResetLink();
			PrtHomePg();
			break;
		case MNU_LINK_MODE:
			if (gLinkMiniA)
			{
				ResetLink();
				gSyncMode++;				
				if (gSyncMode > SYNC_MODE_RMT)
				{
					gSyncMode = SYNC_MODE_BASIC;
				}
				//disconnect when changing modes
				if (gUDPConnected)
				{
					//turn UDP off					
					udp.stop();
					gUDPConnected = false;
				}
				else if (gBLEConnected)
				{
					gBLEConnect = false;
				}	
				gWiFi = false; //turn wifi mode off - if we're connected to a Flex we get out of sync
				SaveInitFile();
			}
			break;
		case MNU_LINK_TYPE:
DiscFirst:	if (gUDPConnected)
			{
				PrtStatus("Disconnect UDP first", true);
				goto ErrNotice;
			}
			else if (gBLEConnected)
			{
				PrtStatus("Disconnect BLE first", true);
				goto ErrNotice;
			}
			gLinkTypeUDP = !gLinkTypeUDP;
			gLinkAutoConnect = false; //turn this off when changing connection types
			break;
		case MNU_LINK_UDP_ADRS:
			gMnuSel = MNU_UDP_ADRS_EDIT;
			txtInputSrcStart = 0;
			txt = gUDPAdrs; //chk for change
			GetTxtInput(gUDPAdrs); //just opens edit page and returns - processed in ExecTxtInputSel()
			return;
			break;
		case MNU_LINK_SVR_CLIENT:
			if (!gBLEConnected && !gUDPConnected)
			{
				gLinkMiniA = !gLinkMiniA;
			}
			else
			{
				goto DiscFirst;
			}			
			break;
		case MNU_LINK_AUTO:
			gLinkAutoConnect = !gLinkAutoConnect;
			if (gLinkTypeUDP && gLinkAutoConnect)
			{
				gWiFiAutoConnect = true; //make sure this is on if link is in autoconnect for UDP
			}			
			break;
		case MNU_LINK_OPTION6:
			if (!gLinkMiniA)
			{
				PrtStatus("Zero-Beat with Rx-A", false);
				gMnuValue = gLink_ZeroBeat;
				gMnuValueMin = -9999;
				gMnuValueMax = 9999;
				hmiMode = HMI_VALUE;
				GetMnuValue(true, "RADIO-B ZERO-BEAT", LONG_PRESS_YES);
			}
			else
			{
				gMnuSel = MNU_UDP_ADRS_ASC_EDIT;
				txtInputSrcStart = 0;
				txt = gUDPAdrsASC;
				GetTxtInput(gUDPAdrsASC);
				return;
				break;
			}
			return;
			break;
		}
		SaveInitFile();
		gMnu = MNU_LINK;
		gMnuPrev = -1;
		CreateMnu("", false);
		return;
		break;	
	case MNU_RADIOPORT:
		switch (gMnuSel)
		{
		case MNU_PORT_COPY:
			PrtStatus("Copy Radio Port [" + String(gRadioPort + 1) + "] to...", false);
			gMnuValue = gRadioPort + 1;
			gMnuValueMin = 1;
			gMnuValueMax = 16;
			hmiMode = HMI_VALUE;
			GetMnuValue(true, "COPY TO PORT", LONG_PRESS_NO);
			return;
			break;
		case MNU_PORT_ERASE:
			PrtStatus("Select [1] to Erase Port [" + String(gRadioPort + 1) + "]", false);
			gMnuValue = 0;
			gMnuValueMin = 0;
			gMnuValueMax = 1;
			hmiMode = HMI_VALUE;
			GetMnuValue(true, "ERASE PORT?", LONG_PRESS_NO);
			return;
			break;
		}
		break;
	case MNU_RADIO:
		mnuMainSel = gMnu - 2; // offset for first two menus
		switch (gRadioCAT)
		{
		case RADIO_CAT_FLEX:
			mnuFlexSel = gMnuSel;
			switch (gMnuSel)
			{
			case MNU_FLEX_IP:
				if (!gTCPCon)
				{
					gMnuSel = MNU_FLEX_IP_EDIT;
					txtInputSrcStart = 0;
					if (gIPAdrs == gIPAdrsDefault)
					{
						GetTxtInput(""); //don't make user delete the default label
					}
					else
					{
						GetTxtInput(gIPAdrs);
					}
				}
				else
				{
					hmiTagLineRefresh = 0;
					PrtStatus("Disconnect WiFi First", true);
				}
				break;
			case MNU_FLEX_PORT:
				gMnuSel = MNU_FLEX_PORT_EDIT;
				txtInputSrcStart = 0;
				GetTxtInput(gFlexPort);
				break;
			case MNU_FLEX_PAN:
				gFlexPan = !gFlexPan;
				if (!gFlexPan)
				{
					PrtStatus("Use Existing Panadapter", true);
				}
				else
				{
					PrtStatus("Create a New Panadapter", true);
				}
				CreateMnu("Toggled Panadapter Option", true);
				break;
			case MNU_FLEX_SLICE:
				gMnuSel = MNU_FLEX_SLICE_EDIT;					
				txtInputSrcStart = 0;
				GetTxtInput(gFlexSliceLbl[gFlexSlice]);				
				break;
			case MNU_FLEX_ANT:
				if (!gTCPCon)
				{
					PrtStatus("Select Antenna for Slice " + gFlexSliceLbl[gFlexSlice], false);
					gMnuValue = gFlexAnt;
					gMnuValueMin = 1;
					gMnuValueMax = 2;
					hmiMode = HMI_VALUE;
					GetMnuValue(true, "SLICE ANTENNA", LONG_PRESS_NO);
				}
				else
				{
					PrtStatus("Change Ant while Offline", true);
				}				
				break;
			case MNU_FLEX_TOGGLE:
				if (gTCPCon)
				{
					switch (gFlexATU)
					{
					case 0:
					case 1:
						Tx2WiFi(WIFI_FLEX, "|atu start");
						gMnuPrev = -1; //force menu refresh
						break;
					case 2:
						//tuning - bypass
					case 3:
						Tx2WiFi(WIFI_FLEX, "|atu bypass");
						gMnuPrev = -1; //force menu refresh
						break;
					case 0xff:
						PrtStatus("ATU Not Installed", true);
						break;
					}
				}
				else
				{
					goto FlexOffLine;
				}
				break;
			case MNU_FLEX_VOL:
				if (gTCPCon)
				{
					PrtStatus("Set Volumne Level", false);
					gMnuValue = gFlexVol;
					gMnuValueMin = 0;
					gMnuValueMax = 100;
					hmiMode = HMI_VALUE;
					GetMnuValue(true, "VOLUME", LONG_PRESS_YES);
				}
				else
				{
FlexOffLine:		PrtStatus("Must be Online to Change", true);
				}
				break;
			case MNU_FLEX_PO:
				if (gTCPCon)
				{
					PrtStatus("Set Power Output", false);
					gMnuValue = gFlexPO;
					gMnuValueMin = 0;
					gMnuValueMax = 100;
					hmiMode = HMI_VALUE;
					GetMnuValue(true, "POWER OUT", LONG_PRESS_YES);
				}
				else
				{
					goto FlexOffLine;
				}				
				break;
			case MNU_FLEX_TO:
				if (gTCPCon)
				{
					PrtStatus("Set Tune Power Output", false);
					gMnuValue = gFlexTO;
					gMnuValueMin = 0;
					gMnuValueMax = 100;
					hmiMode = HMI_VALUE;
					GetMnuValue(true, "TUNE POWER", LONG_PRESS_YES);
				}
				else
				{
					goto FlexOffLine;
				}				
				break;
			case MNU_FLEX_WNB:
				if (gTCPCon)
				{
					PrtStatus("Set Wide Noise Blanker", false);
					gMnuValue = gFlexWNB;
					gMnuValueMin = 0;
					gMnuValueMax = 100;
					hmiMode = HMI_VALUE;
					GetMnuValue(true, "WIDE NOISE BLANKER", LONG_PRESS_YES);
				}
				else
				{
					goto FlexOffLine;
				}				
				break;
			case MNU_FLEX_NB:
				if (gTCPCon)
				{
					PrtStatus("Set Noise Blanker", false);
					gMnuValue = gFlexNB;
					gMnuValueMin = 0;
					gMnuValueMax = 100;
					hmiMode = HMI_VALUE;
					GetMnuValue(true, "NOISE BLANKER", LONG_PRESS_YES);
				}
				else
				{
					goto FlexOffLine;
				}				
				break;
			case MNU_FLEX_NR:
				if (gTCPCon)
				{
					PrtStatus("Set Noise Reduction", false);
					gMnuValue = gFlexNR;
					gMnuValueMin = 0;
					gMnuValueMax = 100;
					hmiMode = HMI_VALUE;
					GetMnuValue(true, "NOISE REDUCTION", LONG_PRESS_YES);
				}
				else
				{
					goto FlexOffLine;
				}				
				break;
			case MNU_FLEX_AF:
				if (gTCPCon)
				{
					if (gRMode < RADIO_LSB)
					{
						PrtStatus("Set Auto Peaking Fltr", false);
						gMnuValue = gFlexAPF;
					}
					else
					{
						PrtStatus("Set Auto Notch Fltr", false);
						gMnuValue = gFlexANF;
					}					
					gMnuValueMin = 0;
					gMnuValueMax = 100;
					hmiMode = HMI_VALUE;
					GetMnuValue(true, "AUTO NOTCH FILTER", LONG_PRESS_YES);
				}
				else
				{
					goto FlexOffLine;
				}				
				break;
			case MNU_FLEX_PTT:
				if (!gEnabAmpSeq)
				{
					gFlexApiPTT = !gFlexApiPTT;
					CreateMnu("Toggled Flex PTT Mode", true);
				}
				else
				{
					gFlexApiPTT = false;
					PrtStatus("Turn OFF Amp Sequencing", true);
				}
				gMnuPrev = -1; //force refresh
				break;
			case MNU_FLEX_KEYER:
				if (!gEnabAmpSeq)
				{
					gFlexApiKey = !gFlexApiKey;
					CreateMnu("Toggled Flex Key Mode", true);
				}
				else
				{
					gFlexApiKey = false;
					PrtStatus("Turn OFF Amp Sequencing", true);
				}
				gMnuPrev = -1; //force refresh
				break;
			case MNU_FLEX_DELAY:
				//this can be set offline because we set it when we enable tx
				PrtStatus("Set CW Break-In Delay", false);
				gMnuValue = gFlexDelay;
				gMnuValueMin = 0;
				gMnuValueMax = 2000;
				hmiMode = HMI_VALUE;
				GetMnuValue(true, "BREAK-IN DELAY", LONG_PRESS_YES);
				break;
			}			
			break;								
		case RADIO_CAT_PCR1000:
			mnuPCRSel = gMnuSel;
			switch (gMnuSel)
			{
			case MNU_PCR_VOL:
				PrtStatus("Set Volume", false);
				gMnuValue = uint8_t(gPCRVol / 2.56); //downscale to 0-100			
				gMnuValueMin = 0;
				gMnuValueMax = 100; //scale this to 0-255 on exit
				hmiMode = HMI_VALUE;
				GetMnuValue(true, "VOLUME", LONG_PRESS_YES);
				break;
			case MNU_PCR_SQ:
				PrtStatus("Set Squelch", false);
				gMnuValue = uint8_t(gPCRSq / 2.56);
				gMnuValueMin = 0;
				gMnuValueMax = 100;
				hmiMode = HMI_VALUE;
				GetMnuValue(true, "SQUELCH", LONG_PRESS_YES);
				break;
			case MNU_PCR_IFSHIFT:
				PrtStatus("Set IF Shift", true);
				gMnuValue = uint8_t(gPCRIF / 2.56);
				gMnuValueMin = 0;
				gMnuValueMax = 100;
				hmiMode = HMI_VALUE;
				GetMnuValue(true, "IF SHIFT", LONG_PRESS_NO); //because 0 is = 50
				break;
			case MNU_PCR_ZEROIF:
				gPCRIF = 0x80; //midscale
				PCR_Update(MNU_PCR_IFSHIFT);
				gMnuPrev = -1; //force refresh
				CreateMnu("Zeroed IF", true);
				break;
			case MNU_PCR_DSP:
				if (gPCRHasDSP)
				{
					gPCRDSP = !gPCRDSP; //toggle
					if (gPCRDSP)
					{
						txt = "ON";
					}
					else
					{
						txt = "OFF";
					}
					txt = "DSP is " + txt;
				}
				else
				{
					gPCRDSP = false;
					txt = "DSP is not installed";
				}
				PCR_Update(MNU_PCR_DSPNR);
				gMnuPrev = -1; //force refresh
				CreateMnu(txt, true);				
				break;
			case MNU_PCR_DSPNR:
				if (gPCRHasDSP)
				{
					PrtStatus("Set Noise Reduction", false);
					gMnuValue = uint8_t(gPCRDSPNR);
					gMnuValueMin = 0;
					gMnuValueMax = 15;
					hmiMode = HMI_VALUE;
					GetMnuValue(true, "NOISE REDUCTION", LONG_PRESS_YES);
				}
				else
				{
					gPCRDSP = false;
					gMnuPrev = -1; //force refresh
					CreateMnu("DSP is not installed", true);					
				}
				break;
			case MNU_PCR_ANch:
				if (gPCRDSP)
				{
					gPCRANch = !gPCRANch; //toggle
					if (gPCRANch)
					{
						txt = "ON";
					}
					else
					{
						txt = "OFF";
					}
					PCR_Update(MNU_PCR_DSPNR);
					gMnuPrev = -1; //force refresh
					CreateMnu("DSP Auto Notch is " + txt, true);					
				}
				else
				{
					PrtStatus("DSP is OFF", true);
				}
				break;
			case MNU_PCR_BW:
				gPCRBW++;
				//SerialOut("gRMode= " + String(gRMode), true);
				//limit bw by mode
				switch (gRMode)
				{
				case RADIO_CW:
					//fall through
				case RADIO_CW_R:
				case RADIO_LSB:
				case RADIO_USB:
				case RADIO_DIG_L:
				case RADIO_DIG_H:
					//3 & 6 kHz only
					if (gPCRBW > 1)
					{
						gPCRBW = 0;
					}
					break;
				case RADIO_FM_N:
					//fall through
				case RADIO_AM:
					//3 to 15 kHz
					if (gPCRBW > 2)
					{
						gPCRBW = 0;
					}
					break;
				case RADIO_FM_W:
					//15 to 230 kHz only
					if (gPCRBW > 4)
					{
						gPCRBW = 2;
					}
					break;
				}
				gMnuPrev = -1; //force refresh
				PCR_Update(gMnuSel);
				CreateMnu("Bandwidth Changed", true);
				break;
			case MNU_PCR_AGC:
				gPCRAGC = !gPCRAGC; //toggle
				if (gPCRAGC)
				{
					txt = "ON";
				}
				else
				{
					txt = "OFF";
				}
				PCR_Update(0xfe);
				gMnuPrev = -1; //force refresh
				CreateMnu("AGC is " + txt, true);
				break;
			case MNU_PCR_ATTEN:
				gPCRAtten = !gPCRAtten; //toggle
				if (gPCRAtten)
				{
					txt = "ON";
				}
				else
				{
					txt = "OFF";
				}
				PCR_Update(0xfe);
				gMnuPrev = -1; //force refresh
				CreateMnu("Attenuation is " + txt, true);
				break;
			case MNU_PCR_NB:
				gPCRNB = !gPCRNB; //toggle
				if (gPCRNB)
				{
					txt = "ON";
				}
				else
				{
					txt = "OFF";
				}
				PCR_Update(0xfe);
				gMnuPrev = -1; //force refresh
				CreateMnu("Noise Blanker is " + txt, true);
				break;
			}	
			SendLinkParams();
			break;
		}		
		break;
	case MNU_AMPSEQ:
		switch (gMnuSel)
		{
		case 1: //pre delay
			PrtStatus("Set PTT Before Key Delay", false);
			gMnuValue = gSeqPre;
			gMnuValueMin = 0;
			gMnuValueMax = 0xffff;
			hmiMode = HMI_VALUE;
			GetMnuValue(true, "PRE-PTT DELAY", LONG_PRESS_YES);
			break;
		case 2: //post delay
			PrtStatus("Set PTT After Key Delay", false);
			gMnuValue = gSeqPost;
			gMnuValueMin = 0;
			gMnuValueMax = 0xffff;
			hmiMode = HMI_VALUE;
			GetMnuValue(true, "POST PTT DELAY", LONG_PRESS_YES);
			break;
		}
		break;
	}
}

void UpdateParam()
{
	//This routine sends the changed value with the SetParam window open so changes are immediate

	String txt = "";

	switch (gMnu)
	{
	case MNU_FREQ:
		if (gMnuSel == MNU_FREQ_RIT)
		{
			gRIT = gMnuValue;
			SendLinkParams();
			txt = "Updated RIT";			
		}
		break;
	case MNU_KEYER:
		switch (gMnuSel)
		{
		case MNU_KEYER_SPEED:
			gCWSpeed = gMnuValue;
			gFarnsSpeed = gMnuValue; //follow speed
			CalcDitDur();
			CalcDitFarnsDur();
			SendLinkParams();
			txt = "Updated Keyer Speed";
			gSaveRadio = true;
			break;
		case MNU_KEYER_FARNS:
			gFarnsSpeed = gMnuValue; //follow speed			
			CalcDitFarnsDur();
			SendLinkParams();
			txt = "Updated Farnsworth Spc";
			gSaveRadio = true;
			break;
		case MNU_KEYER_SN:
			mnuKeyerSN = gMnuValue;
			SaveInitFile();
			SendLinkParams();
			txt = "Updated Contest SN";
			break;
		}
		break;
	case MNU_LINK:				
		if (gMnuSel == MNU_LINK_OPTION6 && !gLinkMiniA)
		{
			//only on Mini-B
			gLink_ZeroBeat = gMnuValue;
			PrtWiFi(); //update icon
			TxFreq2Radio(gFreqRadio);
			gSaveRadio = true;
			txt = "Updated Radio-B Freq";
		}
		break;
	case MNU_RADIOPORT:
		switch (gMnuSel)
		{
		case MNU_PORT_COPY:
			txt = "Copy Radio Port [" + String(gRadioPort + 1) + "] to [" + String(gMnuValue) + "]";
			break;
		case MNU_PORT_ERASE:
			if (gMnuValue == 1)
			{
				txt = "Click [Ok] to Erase Port [" + String(gRadioPort + 1) + "]";
			}
			else
			{
				txt = "Select [1] to Erase Port [" + String(gRadioPort + 1) + "]";
			}
			break;
		}
	case MNU_RADIO:
		switch (gRadioCAT)
		{
		case RADIO_CAT_FLEX:
			//if we're connected to flex
			//Panadapter & Ant must be changed offline - they are used when we create a new slice
			if (gTCPCon)
			{
				Flex_Update(gMnuSel, gMnuValue, txt);
				txt = "Updated Flex parameter";
				SendLinkParams();
			}
			else
			{				
				switch (gMnuSel)
				{
				case MNU_FLEX_ANT:
					gFlexAnt = gMnuValue;
					txt = "Updated Antenna";
					break;
				default:
					txt = "Flex 6000 is Offline";
					break;
				}				
			}
			break;
		case RADIO_CAT_PCR1000:
			//if we're online
			switch (gMnuSel)
			{
			case MNU_PCR_VOL:
				gPCRVol = uint8_t(gMnuValue * 2.56); //used 0-100 scale, upscale to 0-255
				PCR_Update(gMnuSel);
				txt = "Updated Volume";
				break;
			case MNU_PCR_SQ:
				gPCRSq = uint8_t(gMnuValue * 2.56);
				PCR_Update(gMnuSel);
				txt = "Updated Squelch";
				break;
			case MNU_PCR_IFSHIFT:
				gPCRIF = uint8_t(gMnuValue * 2.56);
				PCR_Update(gMnuSel);
				txt = "Updated IF Shift";
				break;
			case MNU_PCR_DSPNR:
				if (gPCRHasDSP)
				{
					gPCRDSPNR = uint8_t(gMnuValue);
					PCR_Update(gMnuSel);
					txt = "Updated Noise Reduction";
				}
				break;
			}
			SendLinkParams();
			break;
		}
		break;
	case MNU_AMPSEQ:		
		switch (gMnuSel)
		{
		case 1:
			gSeqPre = gMnuValue;
			txt = "PTT->Key Pre Delay=" + String(gSeqPre) + "mS";
			break;
		case 2:
			gSeqPost = gMnuValue;
			txt = "PTT->Key Post Delay=" + String(gSeqPost) + "mS";
			break;
		}
		break;
	}
	PrtStatus(txt, true);
	gSaveRadio = true; //save radio when we can
}

void GetTxtInput(String txt)
{
	//edit text value
	txt.trim(); //remove spaces

	uint8_t len = txt.length();
	if (len > txtInputMax)
	{
		len = txtInputMax;
	}
	if (txtInputPtr > len - 1)
	{
		txtInputPtr = len - 1;
	}

	for (unsigned int i = 0; i < txtInputMax; i++)
	{
		if (i <= len)
		{
			txtInput[i] = txt.charAt(i);
		}
		else
		{
			txtInput[i] = ' '; //trim these when we save
		}
	}
	//SerialOut("txtInputSrc= " + String(txtInputSrc) + ", txt= " + txt, true);
	hmiMode = HMI_TXTINPUT; //txt input mode
	txtInputPtr = 0; //start with first letter
	txtInputSrcStart = 0;
	gKeyTypePrev = gKeyType; //save current and use Iambic-B for txt input
	gKeyType = KEYER_IAMBIC_B;
	PrtTxtInput(true); //open edit window
	TermOut("    : " + txt, false); //print text to edit
	//move cursor back to start of text
	for (int i = 0; i < txt.length(); i++)
	{
		TermOut(TERM_CURSOR_LEFT, false);
	}	
}

void GetMnuValue(bool refresh, String txt, uint8_t longPress)
{
	//open window to change the selected value
	bool useHex = false;
	String lead1 = ""; //leading 0
	String lead2 = "";
	static String lastTxt = " ";
	static uint8_t lastLongPress = LONG_PRESS_LAST; //if longPress = 0, use last setting

	if (!txt.length())
	{
		txt = lastTxt; //save header incase we refresh outside of SetMnuParam()
	}
	lastTxt = txt;

	if (longPress)
	{
		gLongPress = longPress; //for encoder
	}
	else
	{
		longPress = lastLongPress;
	}
	lastLongPress = longPress;
	

	if (gMnu == MNU_RADIOCAT && gMnuSel == MNU_ICOM_ADRS_EDIT)
	{
		useHex = true; //icom addresses are in hex
	}

	//dynamically relocate min/max and value if in long-press reset mode
	int valY = 100;
	int minMaxY = 170;
	if (longPress == LONG_PRESS_YES)
	{
		valY = 90;
		minMaxY = 150;
	}	

	if (refresh)
	{
		tft.fillRoundRect(10, 28, 300, 180, 5, hmiBase);
		tft.drawRoundRect(10, 28, 300, 180, 5, TFT_RED);
		tft.drawRoundRect(12, 30, 296, 176, 5, TFT_RED);
		
		//header
		SetFont(6, 1, hmiFGnd);
		tft.drawCentreString(txt, 160, 50, 4);

		SetFont(1, 1, hmiFGnd);		
		if (useHex)
		{
			if (gMnuValueMin < 0x10)
			{
				lead1 = "0";
			}
			if (gMnuValueMax < 0x10)
			{
				lead2 = "0";
			}
			tft.drawString("Min= 0x" + lead1 + String(gMnuValueMin, HEX) + "    Max= 0x" + lead2 + String(gMnuValueMax, HEX), 64, minMaxY);			
		}
		else
		{
			String spc = "      ";
			int offset = 0;
			if (gMnuValueMin < -999 && gMnuValueMax > 999)
			{
				spc = " "; //closer together for large #s
				offset = -15;
			}
			tft.drawString("Min: " + String(gMnuValueMin) + spc + " Max: " + String(gMnuValueMax), 85 + offset, minMaxY);
		}
		TermOut("Use cursor keys to change value then press [Enter]", true);		
	}

	if (longPress == LONG_PRESS_YES)
	{
		SetFont(1, 1, hmiFGnd);
		tft.fillRect(14, 170, 292, 25, hmiBase);
		if (!gMnuValue && !gMnuValuePrev)
		{
			//nothing set yet
			tft.drawCentreString("Set Level, Short-press=OK", 160, 175, 1);
		}
		else if (!gMnuValuePrev)
		{
			tft.drawCentreString("Short-press=OK,  Long=0", 160, 175, 1);
		}
		else
		{
			tft.drawCentreString("Short-press=OK, Long=Restore", 160, 175, 1);
		}
	}

	SetFont(6, 2, TFT_GREEN);
	tft.setTextColor(TFT_GREEN);
	tft.fillRect(23, 90, 274, 50, hmiBase); //erase value each time
	if (useHex)
	{
		lead1 = "";
		if (gMnuValue < 0x10)
		{
			lead1 = "0";
		}
		tft.drawCentreString("0x" + lead1 + String(gMnuValue, HEX), 160, valY, 4);
	}
	else
	{
		tft.drawCentreString(String(gMnuValue), 160, valY, 4);
	}
}

