// =====================================================================================================================================================================================================
// (c) 2022 Lynn Hansen, KU7Q															                                                                                                               |
// This Source Code Form is subject to the terms of the GNU GENERAL PUBLIC LICENSE, Version 3, 29 June 2007. A copy of this license can be found here: https://choosealicense.com/licenses/gpl-3.0/|
// =====================================================================================================================================================================================================


//This module contains radio service routines

String CalcBand()
{
	//calcualte the band we're in by gFreqRadio - update gAntBand if we're using band ant switching
	String temp = "";
	static uint8_t oldBand = 0;

	//update band and clock
	if (gFreqRadio >= 1800000 && gFreqRadio < 2000000)
	{
		gBand = 0;
		temp = "160m";
	}
	else if (gFreqRadio >= 3500000 && gFreqRadio < 4000000)
	{
		gBand = 1;
		temp = "80m";
	}
	else if (gFreqRadio >= 5330000 && gFreqRadio <= 5406500)
	{
		//we're in the 60m band - find USB channel
		gBand = 2;
		temp = "60m";
	}
	else if (gFreqRadio >= 7000000 && gFreqRadio < 7300000)
	{
		gBand = 3;
		temp = "40m";
	}
	else if (gFreqRadio >= 10100000 && gFreqRadio < 10150000)
	{
		gBand = 4;
		temp = "30m";
	}
	else if (gFreqRadio >= 14000000 && gFreqRadio < 14350000)
	{
		gBand = 5;
		temp = "20m";
	}
	else if (gFreqRadio >= 18068000 && gFreqRadio < 18168000)
	{
		gBand = 6;
		temp = "17m";
	}
	else if (gFreqRadio >= 21000000 && gFreqRadio < 21450000)
	{
		gBand = 7;
		temp = "15m";
	}
	else if (gFreqRadio >= 24089000 && gFreqRadio < 24990000)
	{
		gBand = 8;
		temp = "12m";
	}
	else if (gFreqRadio >= 28000000 && gFreqRadio < 29700000)
	{
		gBand = 9;
		temp = "10m";
	}
	else if (gFreqRadio >= 50000000 && gFreqRadio < 54000000)
	{
		gBand = 10;
		temp = "6m";
	}
	else if (gFreqRadio >= 144000000 && gFreqRadio < 148000000)
	{
		gBand = 11;
		temp = "2m";
	}
	else if (gFreqRadio >= 222000000 && gFreqRadio < 225000000)
	{
		gBand = 12;
		temp = "1.25m";
	}
	else if (gFreqRadio >= 420000000 && gFreqRadio < 450000000)
	{
		gBand = 13;
		temp = "70cm";
	}
	else if (gFreqRadio >= 902000000 && gFreqRadio < 928000000)
	{
		gBand = 14;
		temp = "33cm";
	}
	else
	{
		gBand = 15;
		temp = "Gen";
	}

	if (gAntSwSendBand && gBand != oldBand)
	{
		SendASCCmd(ASC_CMD_BAND, 0, 0); //send it twice
		SerialOut("Sending band to ASC", true);
		oldBand = gBand;
	}
			
	return temp;
}

void ChngBand(uint8_t newBand)
{
	if (!gLock && !gPlay)
	{
		//save current band info
		bandMemFreq[gBand] = gFreqRadio;
		bandMemRMode[gBand] = gRMode;
		bandMemBW[gBand] = gPCRBW;
		SaveBandList(gRadioPort);
		//load new band info
		gBand = newBand;
		gFreqRadio = bandMemFreq[gBand];
		gRMode = bandMemRMode[gBand];
		gPCRBW = bandMemBW[gBand];
		ChkFreqModeChng;
		gSaveRadio = true;
		SendLinkParams();
	}
}

void CalcKenwoodFreq(char* sFrq)
{
	//calculates freq string for kenwood reply
	uint32_t frq = gFreqRadio; //use gFreqHMI - gFreqRadio depends on a radio being selected
	uint32_t val = 0;
	//separate out individual freq values
	if (frq > 99999999)
	{
		val = frq / 100000000;
		sFrq[2] = val | 0x30;
		frq -= val * 100000000;
	}
	if (frq > 9999999)
	{
		val = frq / 10000000;
		sFrq[3] = val | 0x30;
		frq -= val * 10000000;
	}
	if (frq > 999999)
	{
		val = frq / 1000000;
		sFrq[4] = val | 0x30;
		frq -= val * 1000000;
	}
	if (frq > 99999)
	{
		val = frq / 100000;
		sFrq[5] = val | 0x30;
		frq -= val * 100000;
	}
	if (frq > 9999)
	{
		val = frq / 10000;
		sFrq[6] = val | 0x30;
		frq -= val * 10000;
	}
	if (frq > 999)
	{
		val = frq / 1000;
		sFrq[7] = val | 0x30;
		frq -= val * 1000;
	}
	if (frq > 99)
	{
		val = frq / 100;
		sFrq[8] = val | 0x30;
		frq -= val * 100;
	}
	if (frq > 9)
	{
		val = frq / 10;
		sFrq[9] = val | 0x30;
		frq -= val * 10;
	}
	sFrq[10] = frq | 0x30;
	sFrq[11] = 0;
}



bool TxFreq2Radio(uint32_t freq)
{
	//call this routine to tx new freq to radio using TxFreq() below
	//it will try twice then give up
	//ignore if in link rmt mode and we're mini-a
	bool success = true;
	if ((gBLEConnected || gUDPConnected) && (gSyncMode == SYNC_MODE_RMT && gLinkMiniA))
	{
		return success;
	}
	if (gRadioCAT != RADIO_CAT_NONE)
	{
		Chk4SyncZeroBeat(freq, true);
		success = TxFreq(freq);
		if (!success)
		{
			//try a second time if no reply
			success = TxFreq(freq);
		}
		if (!success)
		{			
			//PrtStatus("Radio Offline", true); //alert message - use red background
			return false;
		}
	}	
	return success;
}

bool TxFreq(uint32_t freq)
{
	//returns true if we successfully sent the data to the radio and didn't block it

	//WARNING: Don't use delay() in this routine or routines it calls as it can be called from the _ISR routines

	if (gRadioCAT == RADIO_CAT_FLEX)
	{
		Flex_TxFreq(freq);
		gFlexLastFreqUpdateTmr = millis(); //retransmit this after 1 second - sometimes it doesn't go through properly
		return true;
	}
	else if (gRadioCAT == RADIO_CAT_RIGCTLD)
	{
		return RigCtld_Set_Freq(freq);
	}

	while (RADIO.availableForWrite() == 0)
	{
		//wait for tx buffer to clear
	}

	if (radioCommBusy == true)
	{
		//SerialOut("Radio comm busy...", true);
		return false; //don't clear flags that sent this data
	}

	radioCommBusy = true; //block ISR from overwriting
	//gRadioLastFreqUpdateTime = millis(); // must have 2 seconds after one control changes freq before another one can

	//freq has changed, send it to the radio
	char cmdYaesuFTdx[13] = { 'F', 'A', '0', '0', '0', '0', '0', '0', '0', '0', '0', ';' , 0x00 };
	char cmdYaesu[5] = { 0, 0, 0, 0, 0x01 };
	char cmdKen[15] = { 'F', 'A', '0', '0', '0', '0', '0', '0', '0', '0', '0', '0', '0', ';', 0x00};
	char cmdIcom[11] = { 0xFE, 0xFE, gIcmAdr, 0xE0, 0x5, 0x0, 0x0, 0x0, 0x0, 0x0, 0xFD }; //0xE0 is the master "from" address
	char cmdPCR1000[20] = { 'K', '0', '0', '0', '0', '0', '0', '0', '0', '0', '0', '0', '0', '0', '0', '0', '0', '0',13,10 }; //K0GMMMKKKHHHmmff where G=Ghz, M=MHz, K=kHz, H=Hz, mm=mode, ff=bandwidth, padding, cr, lf
	char frq[9] = { 0,0,0,0,0,0,0,0,0 };

	uint32_t temp = freq;
	uint8_t val = 0;
	uint8_t numBytes = 0; //# of bytes to tx

	//make sure rx buffer is clear before we tx new msg
	while (RADIO.available() > 0)
	{
		RADIO.read();
	}

	//separate out individual freq values
	if (temp > 99999999)
	{
		//100 mhz digit
		val = temp / 100000000;
		frq[8] = val;
		temp -= val * 100000000;
	}
	if (temp > 9999999)
	{
		//10 mhz digit
		val = temp / 10000000;
		frq[7] = val;
		temp -= val * 10000000;
	}
	if (temp > 999999)
	{
		//1 mhz digit
		val = temp / 1000000;
		frq[6] = val;
		temp -= val * 1000000;
	}
	if (temp > 99999)
	{
		//100k digit
		val = temp / 100000;
		frq[5] = val;
		temp -= val * 100000;
	}
	if (temp > 9999)
	{
		//10k digit
		val = temp / 10000;
		frq[4] = val;
		temp -= val * 10000;
	}
	if (temp > 999)
	{
		//1k digit
		val = temp / 1000;
		frq[3] = val;
		temp -= val * 1000;
	}
	if (temp > 99)
	{
		//100s digit
		val = temp / 100;
		frq[2] = val;
		temp -= val * 100;
	}
	if (temp > 9)
	{
		//10s digit
		val = temp / 10;
		frq[1] = val;
		temp -= val * 10;
	}
	//1s digit
	frq[0] = temp;

	switch (gRadioCAT)
	{
	case RADIO_CAT_ICOM:
		//use long Icon command if not 731/735 - byte 9 is 100 mhz byte 

		cmdIcom[9] = frq[8]; //100mhz 
		cmdIcom[8] = frq[7] << 4; //10mhz in msb
		cmdIcom[8] += frq[6]; //1mhz in lsb
		cmdIcom[7] = frq[5] << 4; //100khz in msb
		cmdIcom[7] += frq[4]; //10khz in lsb
		cmdIcom[6] = frq[3] << 4; //1khz in msb
		cmdIcom[6] += frq[2]; //100hz in lsb
		cmdIcom[5] = frq[1] << 4; //10 hz in msb
		cmdIcom[5] += frq[0]; //1 hz in lsb
		numBytes = 11;//default to long command

		//if model 735, use short command (doesn't suport >99 mhz
		if (gIcmAdr == 4)
		{
			cmdIcom[9] = 0xfd; //move terminator over one
			cmdIcom[10] = 0x00;
			numBytes = 10;
		}
		//SerialOut("Tx to Icom", true);
		RADIO.write(cmdIcom, numBytes); //send it to the radio		
		break;
	case RADIO_CAT_KENWOOD1:
	case RADIO_CAT_KENWOOD2:
		cmdKen[4] = frq[8] | 0x30; //100mhz - convert to ascii
		cmdKen[5] = frq[7] | 0x30; //10mhz
		cmdKen[6] = frq[6] | 0x30; //1mhz
		cmdKen[7] = frq[5] | 0x30; //100khz
		cmdKen[8] = frq[4] | 0x30; //10khz
		cmdKen[9] = frq[3] | 0x30; //1khz
		cmdKen[10] = frq[2] | 0x30; //100hz
		cmdKen[11] = frq[1] | 0x30; //10hz
		cmdKen[12] = frq[0] | 0x30; //1hz	
		RADIO.write(cmdKen, 14);
		SerialOut(String(cmdKen), true);
		break;
	case RADIO_CAT_YAESUFTDX:
		cmdYaesuFTdx[2] = frq[8] | 0x30; //100mhz - convert to ascii
		cmdYaesuFTdx[3] = frq[7] | 0x30; //10mhz
		cmdYaesuFTdx[4] = frq[6] | 0x30; //1mhz
		cmdYaesuFTdx[5] = frq[5] | 0x30; //100khz
		cmdYaesuFTdx[6] = frq[4] | 0x30; //10khz
		cmdYaesuFTdx[7] = frq[3] | 0x30; //1khz
		cmdYaesuFTdx[8] = frq[2] | 0x30; //100hz
		cmdYaesuFTdx[9] = frq[1] | 0x30; //10hz
		cmdYaesuFTdx[10] = frq[0] | 0x30; //1hz	
		RADIO.write(cmdYaesuFTdx, 12);		
		//SerialOut("\nftdx= " + String(cmdYaesuFTdx), true);
		break;
	case RADIO_CAT_YAESU100:
		radioRxBytes = 5; //we're expecting 5 byts
	TxYaesu:
		cmdYaesu[0] = frq[8] << 4; //100mhz
		cmdYaesu[0] += frq[7]; //10mhz
		cmdYaesu[1] = frq[6] << 4; //1mhz
		cmdYaesu[1] += frq[5]; //100khz
		cmdYaesu[2] = frq[4] << 4; //10khz
		cmdYaesu[2] += frq[3]; //1khz
		cmdYaesu[3] = frq[2] << 4; //100hz
		cmdYaesu[3] += frq[1]; //10hz 
		//yaesu doesn't use 1hz digit
		//SerialOut("Tx: " + String(cmdYaesu[0], HEX) + " " + String(cmdYaesu[1], HEX) + " " + String(cmdYaesu[2], HEX) + " " + String(cmdYaesu[3], HEX), true);
		RADIO.write(cmdYaesu, 5);		
		break;
	case RADIO_CAT_YAESU757:
		goto TxYaesu;
		break;
	case RADIO_CAT_YAESU8x7:
		goto TxYaesu;
		break;
	case RADIO_CAT_YAESU890:
		goto TxYaesu;
		break;
	case RADIO_CAT_YAESU1000:
		goto TxYaesu;
		break;
	case RADIO_CAT_PCR1000:
		//send them as ascii strings
		cmdPCR1000[3] = frq[8] + 0x30;
		cmdPCR1000[4] = frq[7] + 0x30;
		cmdPCR1000[5] = frq[6] + 0x30;
		cmdPCR1000[6] = frq[5] + 0x30;
		cmdPCR1000[7] = frq[4] + 0x30;
		cmdPCR1000[8] = frq[3] + 0x30;
		cmdPCR1000[9] = frq[2] + 0x30;
		cmdPCR1000[10] = frq[1] + 0x30;
		cmdPCR1000[11] = frq[0] + 0x30;
		switch (gRMode)
		{
		case RADIO_CW: 
			cmdPCR1000[13] = '3';
			break;
		case RADIO_CW_R: 
			cmdPCR1000[13] = '3';
			break;
		case RADIO_LSB:
			cmdPCR1000[13] = '0';
			break;
		case RADIO_USB:
			cmdPCR1000[13] = '1';
			break;
		case RADIO_FM_N:
			cmdPCR1000[13] = '5';
			break;
		case RADIO_FM_W: //- only for 230 kHz bw			
			cmdPCR1000[13] = '6';
			break;
		case RADIO_AM:
			cmdPCR1000[13] = '2';
			break;
		case RADIO_DIG_L: //force to LSB
			cmdPCR1000[13] = '0';
			break;
		case RADIO_DIG_H: //- force to USB
			cmdPCR1000[13] = '1';
			break;
		}
		cmdPCR1000[15] = gPCRBW + 0x30; //convert to ascii - 0=3k, 1=6k, 2=15k, 3=50k, 4=230k		
		if (gPCRInit == true)
		{
			//SerialOut("-> tx freq: " + String(cmdPCR1000), true);
			RADIO.write(cmdPCR1000, 20);			
		}		
		break;
	}
	//SerialOut("--> Set Tx Freq" + String(freq), true);
	radioCommBusy = false;
	return true; //we sent it!
}


bool TxMode2Radio()
{
	//ignore if we're Mini-A and connected as remote
	if ((gBLEConnected || gUDPConnected) && (gSyncMode == SYNC_MODE_RMT && gLinkMiniA))
	{
		return true;
	}

	if (gRadioCAT == RADIO_CAT_FLEX)
	{
		Flex_TxMode();
		return true;
	}
	else if (gRadioCAT == RADIO_CAT_RIGCTLD)
	{
		RigCtld_Set_Mode();
		return true;
	}


	while (RADIO.availableForWrite() == 0)
	{
		//wait for tx buffer to clear
	}
	if (radioCommBusy == true)
	{
		return false; //return false if busy so we'll try again
	}
	if (gRadioCAT == RADIO_CAT_PCR1000)
	{
		//mode is set with freq		
		bool success = TxFreq2Radio(gFreqRadio);
		return success;
	}
	radioCommBusy = true; //blocks ISR from overwriting

	//gRMode has changed, send it to the radio

	//==============================================================================================================================
	//!!!!2/25/22: NOTE: FTdx101 CAT had a bug it in prior to Main v1.22 firmware. 
	//! You have to send MC instead of MD to control the radio mode with older firmware - users need to update radio's firmware
	char cmdYaesuFTdx[6] = { 'M', 'D', '0', '0', ';', 0x00 }; // digit 2 is always 0 to send to main radio, digit 3 holds mode
	//==============================================================================================================================


	char cmdYaesu[5] = { 0, 0, 0, 0, 0x07 };
	char cmdKen[4] = { 'M', 'D', '0', ';' }; //"MDx;" where x=radio mode
	char cmdIcom[8] = { 0xFE, 0xFE, gIcmAdr, 0xE0, 0x1, 0x0, 0x1, 0xFD }; //0xE0 is the master "from" address

	//make sure rx buffer is clear before we tx new msg
	while (RADIO.available() > 0)
	{
		RADIO.read();
	}

	switch (gRadioCAT)
	{
	case RADIO_CAT_ICOM:
		//convert CTR2 mode to Icom mode
		switch (gRMode)
		{
		case 0: //cw
			cmdIcom[5] = 3;
			break;
		case 1: //cwr
			cmdIcom[5] = 7;
			break;
		case 2: //lsb
			cmdIcom[5] = 0;
			break;
		case 3: //usb
			cmdIcom[5] = 1;
			break;
		case 4: //fmn
			cmdIcom[5] = 5;
			break;
		case 5: //fmw - force to fmn
			cmdIcom[5] = 6;
			break;
		case 6: //am
			cmdIcom[5] = 2;
			break;
		case 7: //digL - force to rtty reverse
			cmdIcom[5] = 8;
			break;
		case 8: //digH - force to rtty
			cmdIcom[5] = 4;
			break;
		}
		RADIO.write(cmdIcom, 8); //send it to the radio		
		break;
	case RADIO_CAT_KENWOOD1:
		//convert CTR2 mode to Kenwood mode
Kenwood:
		switch (gRMode)
		{
		case 0: //cw
			cmdKen[2] = '3'; //modes in ascii
			break;
		case 1: //cwr
			cmdKen[2] = '7';
			break;
		case 2: //lsb
			cmdKen[2] = '1';
			break;
		case 3: //usb
			cmdKen[2] = '2';
			break;
		case 4: //fmn
			cmdKen[2] = '4';
			break;
		case 5: //fmw - force to fmn
			cmdKen[2] = '4';
			break;
		case 6: //am
			cmdKen[2] = '5';
			break;
		case 7: //digL - force to fsk reverse
			cmdKen[2] = '6';
			break;
		case 8: //digH - force to fsk
			cmdKen[2] = '9';
			break;
		}
		RADIO.write(cmdKen, 4);		
		break;
	case RADIO_CAT_KENWOOD2:
		goto Kenwood;
		break;
	case RADIO_CAT_YAESUFTDX:
		switch (gRMode)
		{
		case 0: //cw
			cmdYaesuFTdx[3] = '3'; //modes in ascii
			break;
		case 1: //cwr
			cmdYaesuFTdx[3] = '7';
			break;
		case 2: //lsb
			cmdYaesuFTdx[3] = '1';
			break;
		case 3: //usb
			cmdYaesuFTdx[3] = '2';
			break;
		case 4: //fmn
			cmdYaesuFTdx[3] = 'B';
			break;
		case 5: //fmw - force to fmn
			cmdYaesuFTdx[3] = '4';
			break;
		case 6: //am
			cmdYaesuFTdx[3] = '5';
			break;
		case 7: //digL - data lower
			cmdYaesuFTdx[3] = '8';
			break;
		case 8: //digH - data upper
			cmdYaesuFTdx[3] = 'C';
			break;
		}
		RADIO.write(cmdYaesuFTdx, 5);		
		//SerialOut("\nftdx mode= " + String(cmdYaesuFTdx) ,true);	
		break;
	case RADIO_CAT_YAESU100:
	TxYaesu:
		radioRxBytes = 5; //we're expecting 5 byts
		//convert CTR2 mode to Yaesu mode
		switch (gRMode)
		{
		case 0: //cw
			cmdYaesu[0] = 2;
			break;
		case 1: //cwr
			cmdYaesu[0] = 3;
			break;
		case 2: //lsb
			cmdYaesu[0] = 0;
			break;
		case 3: //usb
			cmdYaesu[0] = 1;
			break;
		case 4: //fmn
			cmdYaesu[0] = 6;
			break;
		case 5: //fmw 
			cmdYaesu[0] = 7;
			break;
		case 6: //am
			cmdYaesu[0] = 4;
			break;
		case 7: //digL - force to packet low
			cmdYaesu[0] = 0x0A;
			break;
		case 8: //digH - force to packet
			cmdYaesu[0] = 0x0B;
			break;
		}
		RADIO.write(cmdYaesu, 5);		
		break;
	case RADIO_CAT_YAESU757:
		goto TxYaesu;
		break;
	case RADIO_CAT_YAESU8x7:
		radioRxBytes = 5; //we're expecting 5 byts
		//only send valid modes to radio otherwise it will CRASH!
		switch (gRMode)
		{
		case 0: //cw
			cmdYaesu[0] = 2;
			break;
		case 1: //cwr
			cmdYaesu[0] = 3;
			break;
		case 2: //lsb
			cmdYaesu[0] = 0;
			break;
		case 3: //usb
			cmdYaesu[0] = 1;
			break;
		case 4: //fmn
			cmdYaesu[0] = 8;
			break;
		case 5: //fmw 
			cmdYaesu[0] = 8; //same as fmn - radio will go to fmw if tuned to 88-108 mhz
			break;
		case 6: //am
			cmdYaesu[0] = 4;
			break;
		case 7: //digH - is dig on 817
			cmdYaesu[0] = 0x0C;
			break;
		case 8: //digL - is pkt on 8x7
			cmdYaesu[0] = 0x0A;
			break;
		default:
			cmdYaesu[0] = 3; //default to usb
			break;
		}
		RADIO.write(cmdYaesu, 5);		
		break;
	case RADIO_CAT_YAESU890:
		goto TxYaesu;
		break;
	case RADIO_CAT_YAESU1000:
		goto TxYaesu;
		break;
	}
	radioCommBusy = false;
	gRModePrev = gRMode;
	return true;

}

void SetRIT()
{
	//set the rx freq offset (RIT or clarifier depending on the CAT protocol)
	
	//ignore if we're Mini-A in remote mode and link is connected
	if ((gBLEConnected || gUDPConnected) && (gSyncMode == SYNC_MODE_RMT && gLinkMiniA))
	{
		return;
	}

	if (gRadioCAT == RADIO_CAT_FLEX)
	{
		Flex_RIT();
		return;
	}
	else if (gRadioCAT == RADIO_CAT_RIGCTLD)
	{
		RigCtld_Set_RIT();
		return;
	}

	while (RADIO.availableForWrite() == 0)
	{
		//wait for tx buffer to clear
	}
	if (radioCommBusy == true)
	{
		return; //return false if busy so we'll try again
	}
	if (gRadioCAT == RADIO_CAT_PCR1000)
	{
		//not used with PCR1000
		return;
	}
	radioCommBusy = true; //blocks ISR from overwriting

	//gRMode has changed, send it to the radio
	char cmdFTdx[8] = { 'R', 'x', '0', '0', '0', '0', ';', 0x00 };
	char cmdYaesu[5] = { 0, 0, 0, 0, 0 };
	char cmdKen[9] = { 'R', 'x', '0', '0', '0', '0', '0', ';', 0x00}; //kenwood uses 5 digis, FTdx uses 4 freq digits
	char cmdIcom[10] = { 0xFE, 0xFE, gIcmAdr, 0xE0, 0x21, 0x0, 0x0, 0x0, 0x0, 0xFD }; //0xE0 is the master "from" address

	uint8_t val = 0;
	int temp = gRIT; //used to calcualte digits
	

	//make sure rx buffer is clear before we tx new msg
	while (RADIO.available() > 0)
	{
		RADIO.read();
	}

	switch (gRadioCAT)
	{
	case RADIO_CAT_ICOM:		
		if (!temp)
		{
			cmdIcom[5] = 1; // rit control
			cmdIcom[6] = 0; // rit off
			cmdIcom[7] = 0xfd; //terminator			
			RADIO.write(cmdIcom, 8);
			/*
						for (int i = 0; i < 8; i++)
			{
				SerialOut(String(cmdIcom[i], HEX) + " ", false);
			}
			SerialOut(" ", true);
			*/
		}
		else
		{
			cmdIcom[5] = 1; // rit control
			cmdIcom[6] = 1; // rit on
			cmdIcom[7] = 0xfd; //terminator			
			RADIO.write(cmdIcom, 8);
			/*
			for (int i = 0; i < 8; i++)
			{
				SerialOut(String(cmdIcom[i], HEX) + " ", false);
			}
			SerialOut(" ", true);
			*/
			//calculate freq positions
			cmdIcom[5] = 0; //rit freq
			cmdIcom[6] = 0;
			cmdIcom[7] = 0;
			if (temp > 0)
			{
				cmdIcom[8] = 0; //positive
			}
			else
			{
				cmdIcom[8] = 1; //negative
				temp = abs(temp); //remove sign
			}
			if (temp > 999)
			{
				val = temp / 1000;
				cmdIcom[7] = val<<4; //msb
				temp -= val * 1000;
			}
			if (temp > 99)
			{
				val = temp / 100;
				cmdIcom[7] = (cmdIcom[7] & 0xf0) + val; //lsb
				temp -= val * 100;
			}
			if (temp > 9)
			{
				val = temp / 10;
				cmdIcom[6] = val<<4;
				temp -= val * 10;
			}
			cmdIcom[6] = (cmdIcom[6] & 0xf0) + temp;
			cmdIcom[9] = 0xfd;
			/*
			for (int i = 0; i < 10; i++)
			{
				SerialOut(String(cmdIcom[i], HEX) + " ", false);
			}
			SerialOut(" ", true);
			*/
			RADIO.write(cmdIcom, 10); //send it to the radio				
		}		
		break;
	case RADIO_CAT_KENWOOD1:
	case RADIO_CAT_KENWOOD2:
		/*
		    NOTE
			This code doesn't control the RIT freq on my TS-680 (enable/disable work) but the codes are going out right so I think the issue is with my radio
		*/
		//send enable/disable
		if (!temp)
		{

			cmdKen[1] = 'T'; //turn it off
			cmdKen[2] = '0'; 
			cmdKen[3] = ';';
			RADIO.write(cmdKen, 4);
		}
		else
		{
			cmdKen[1] = 'C'; //clear clarifier
			cmdKen[2] = ';';
			cmdKen[3] = 0;
			RADIO.write(cmdKen, 3);
			delay(50);

			cmdKen[1] = 'T'; //enable RIT
			cmdKen[2] = '1';
			cmdKen[3] = ';';
			cmdKen[4] = 0;
			RADIO.write(cmdKen, 4);			
			delay(50);
			if (temp > 0)
			{
				cmdKen[1] = 'U'; //move up				
			}
			else
			{
				cmdKen[1] = 'D'; //move down				
				temp = abs(temp); //remove sign
			}
			cmdKen[2] = '0'; //restore defaults
			cmdKen[3] = '0';
			cmdKen[4] = '0';	
			cmdKen[5] = '0';
			if (temp > 999)
			{
				val = temp / 1000;
				cmdKen[3] = val | 0x30;
				temp -= val * 1000;				
			}						
			if (temp > 99)
			{
				val = temp / 100;
				cmdKen[4] = val | 0x30;
				temp -= val * 100;
			}						
			if (temp > 9)
			{
				val = temp / 10;
				cmdKen[5] = val | 0x30;
				temp -= val * 10;
			}						
			cmdKen[6] = temp | 0x30;
			cmdKen[7] = ';';
			cmdKen[8] = 0;
			RADIO.write(cmdKen, 8);
			/*debug
			
			for (int i = 0; i < 8; i++)
			{
				SerialOut(String(char(cmdKen[i])), false);
			}
			SerialOut(" ", true);
			*/
		}
		break;
	case RADIO_CAT_YAESUFTDX:
		//send enable/disable
		if (!temp)
		{
			cmdFTdx[1] = 'T'; //turn off clarifier	
			cmdFTdx[2] = '0';
			cmdFTdx[3] = ';';
			cmdFTdx[4] = 0;
			RADIO.write(cmdFTdx, 4);
						
			cmdFTdx[1] = 'C'; //clear clarifier	
			cmdFTdx[2] = ';';
			cmdFTdx[3] = 0;
			RADIO.write(cmdFTdx, 3);			
		}
		else
		{
			cmdFTdx[1] = 'C'; //clear clarifier	
			cmdFTdx[2] = ';';
			cmdFTdx[3] = 0;
			RADIO.write(cmdFTdx, 3);
			
			if (temp > 0)
			{
				cmdFTdx[1] = 'U'; //move up
			}
			else
			{
				cmdFTdx[1] = 'D'; //move down
				temp = abs(temp); //remove sign
			}
			cmdFTdx[2] = '0'; //restore defaults
			cmdFTdx[3] = '0';
			cmdFTdx[4] = '0';			
			if (temp > 999)
			{
				val = temp / 1000;
				cmdFTdx[2] = val | 0x30;
				temp -= val * 1000;
			}			
			if (temp > 99)
			{
				val = temp / 100;
				cmdFTdx[3] = val | 0x30;
				temp -= val * 100;
			}			
			if (temp > 9)
			{
				val = temp / 10;
				cmdFTdx[4] = val | 0x30;
				temp -= val * 10;
			}			
			cmdFTdx[5] = temp | 0x30;
			cmdFTdx[6] = ';';
			cmdFTdx[7] = 0;
			RADIO.write(cmdFTdx, 7);			
			delay(50);
			//enable RIT
			cmdFTdx[1] = 'T';
			cmdFTdx[2] = '1';
			cmdFTdx[3] = ';';
			cmdFTdx[4] = 0;
			RADIO.write(cmdFTdx, 4);			
		}
		break;
	case RADIO_CAT_YAESU100:
	case RADIO_CAT_YAESU757:
	case RADIO_CAT_YAESU890:
	case RADIO_CAT_YAESU1000:
		radioRxBytes = 5; //we're expecting 5 byts		
		if (!temp)
		{
			cmdYaesu[2] = 0; //positive
			cmdYaesu[3] = 0; //off						
		}
		else
		{
			cmdYaesu[2] = 0xff; // negative
			cmdYaesu[3] = 01; //on			
		}
		if (temp > 999)
		{
			val = temp / 1000;
			cmdYaesu[1] = val<<4; //msb 
			temp -= val * 1000;
		}
		if (temp > 99)
		{
			val = temp / 100;
			cmdYaesu[1] = (cmdYaesu[1] & 0xf0) + val;
			temp -= val * 100;
		}
		if (temp > 9)
		{
			val = temp / 10;
			cmdYaesu[0] = val << 4;
			temp -= val * 10;
		}
		cmdYaesu[0] = (cmdYaesu[0] & 0xf0) + temp;
		cmdYaesu[4] = 9; //clarify cmd

		for (int i = 0; i < 5; i++)
		{
			SerialOut(String(cmdYaesu[i], HEX) + " ", false);
		}
		SerialOut(" ", true);
		RADIO.write(cmdYaesu, 5);
		break;
	case RADIO_CAT_YAESU8x7:
		radioRxBytes = 5; //we're expecting 5 byts
		if (!temp)
		{
			cmdYaesu[4] = 0x85; // rit off
			RADIO.write(cmdYaesu, 5);
		}
		else
		{
			cmdYaesu[4] = 0x05; // rit on					
			RADIO.write(cmdYaesu, 5);
			//calculate freq positions
			if (temp >= 0)
			{
				cmdYaesu[0] = 0; //positive
			}
			else
			{
				cmdYaesu[0] = 1; //negative
				temp = abs(temp); //remove sign
			}
			//yaesu uses 10k, 1k, 100, and 10 hz resolution
			if (temp > 9999)
			{
				val = temp / 10000;
				cmdYaesu[2] = val << 4; //msb
				temp -= val * 10000;
			}
			if (temp > 999)
			{
				val = temp / 1000;
				cmdYaesu[2] = (cmdYaesu[2] & 0xf0) + val; //lsb
				temp -= val * 1000;
			}
			if (temp > 99)
			{
				val = temp / 100;
				cmdYaesu[3] = val << 4;
				temp -= val * 100;
			}
			if (temp > 9)
			{
				val = temp / 10;
				cmdYaesu[3] = (cmdYaesu[3] & 0xf0) + val;
			}
			
			cmdYaesu[4] = 0xf5;
			
			for (int i = 0; i < 5; i++)
			{
				SerialOut(String(cmdYaesu[i], HEX) + " ", false);
			}
			SerialOut(" ", true);
			RADIO.write(cmdYaesu, 5);
		}
		break;
	}
	radioCommBusy = false;
}


void SetRadioBaud()
{
	int settings = 0;
	//set baud rate on radio port based on radio
	//flex radio uses Wifi API for control	
	switch (gRadioCAT)
	{
	case RADIO_CAT_FLEX:
		radioMsgTimeoutDefault = 500; //long timeout for wifi
		settings = SERIAL_8N1; //temp for flex test - may need to use API
		break;
	case RADIO_CAT_PCR1000:
		gRadioBaud[RADIO_CAT_PCR1000] = 9600; //force default baud
		settings = SERIAL_8N1;				
		gPCRInit = false; //force update
		break;
	case RADIO_CAT_ICOM:
		radioMsgTimeoutDefault = 80;
		settings = SERIAL_8N1;
		break;
	case RADIO_CAT_KENWOOD1:
		radioMsgTimeoutDefault = 200;;
		settings = SERIAL_8N1; //kenwood default 9600 1 stop bit (newer radios)
		break;
	case RADIO_CAT_KENWOOD2:
		radioMsgTimeoutDefault = 250;
		settings = SERIAL_8N2; //kenwood default 4800 2 stop bit (older radios)
		break;
	case RADIO_CAT_YAESUFTDX:
		radioMsgTimeoutDefault = 200;
		settings = SERIAL_8N2; //cat manual doesn't specify this - assume
		break;
	case RADIO_CAT_YAESU100:
		radioMsgTimeoutDefault = 100;
		settings = SERIAL_8N2;
		break;
	case RADIO_CAT_YAESU1000:
		radioMsgTimeoutDefault = 100;
		settings = SERIAL_8N2;
		break;
	case RADIO_CAT_YAESU757:
		radioMsgTimeoutDefault = 100;
		settings = SERIAL_8N2;
		break;
	case RADIO_CAT_YAESU890:
		radioMsgTimeoutDefault = 100;
		settings = SERIAL_8N2;
		break;
	case RADIO_CAT_YAESU8x7:
		radioMsgTimeoutDefault = 50; //short msgs - radio throws random 0 chrs - short timeout to ignore them
		settings = SERIAL_8N2;
		break;
	}
	
	if (gRadioBaud[gRadioCAT] != RADIO_CAT_NONE && gRadioCAT != RADIO_CAT_RIGCTLD)
	{
		RADIO.end();
		delay(100);
		RADIO.begin(gRadioBaud[gRadioCAT], settings); //change radio ctrl baud and settings	
		delay(250); //wait for radio to start
		SerialOut("baud= " + String(gRadioBaud[gRadioCAT]), true);
		
		if (gRadioCAT == RADIO_CAT_PCR1000)
		{
			//force auto update off
			//String reply = Tx2PCR1000("G300");
		}
	}
	else
	{
		RADIO.end(); //if no radio, just kill the radio process
	}
}


bool GetFreqMode()
{
	/*
	<< DON'T POLL FOR RADIO FREQ/MODE CHANGES IN TX MODE >>
	<<< CAT DATA IS CAUSING PULSING NOISE IN TX AUDIO >>
	*/
	
	if (gTxMode > RADIO_TX_ENAB  || gRadioCAT == RADIO_CAT_NONE)
	{
		return false;
	}

	//read the freq and mode from the selected radio
	//alternate between the two every other scan if icom

	//returns true if radio on line - false if not
	 if (gRadioCAT == RADIO_CAT_FLEX || gRadioCAT == RADIO_CAT_PCR1000)
	{
		//Flex will tell us if anything changed in the slice subscription packet - we don't ask here
		//On PCR1000 the "K0?" request is replied with "G000". We do get the set freq when we TxFreq() but that happens during an _ISR so we can't call it		
		radioCommBusy = false;
		return true;
	}
	else if (gRadioCAT == RADIO_CAT_RIGCTLD)
	{
		 radioCommBusy = false;
		 return RigCtld_Get_FreqMode();		
	}
	else if (radioCommBusy == true)
	 {
		 return false;
	 }

	radioCommBusy = true; // block ISR
	static bool toggle = false;
	int numChr = 5; //defaults to Yaesu len, change for icom and kenwood
	char cmdYaesuFTdx[4] = { 'I', 'F', ';' , 0x00 }; //
	char cmdYaesu[5] = { 0x03, 0x03, 0x03, 0x03, 0x03 };
	char cmdKenIF[3] = { 'I', 'F', ';' }; // 'IF;' command
	char cmdIcomFreq[6] = { 0xfe, 0xfe, gIcmAdr, 0xe0, 0x03, 0xfd };
	char cmdIcomMode[6] = { 0xfe, 0xfe, gIcmAdr, 0xe0, 0x04, 0xfd };

	//make sure rx buffer is clear before we tx new msg
	while (RADIO.available() > 0)
	{
		RADIO.read();
	}

	switch (gRadioCAT)
	{
	case RADIO_CAT_ICOM:
		numChr = 6;
		if (toggle == false)
		{
			radioRxBytes = 10;
			RADIO.write(cmdIcomFreq, numChr); //send it to the radio			
		}
		else
		{
			radioRxBytes = 7;
			RADIO.write(cmdIcomMode, numChr); //send it to the radio						
		}
		toggle = !toggle;
		break;
	case RADIO_CAT_KENWOOD1:
	case RADIO_CAT_KENWOOD2:
		numChr = 3;
		radioRxBytes = 37; //require 38 bytes from radio
		RADIO.write(cmdKenIF, numChr); //send it to the radio		
		break;
	case RADIO_CAT_YAESUFTDX:
		numChr = 3;
		radioRxBytes = 50; //this should be =28 but radio randomly sends additional chrs before the terminator (;) so look long 
		RADIO.write(cmdYaesuFTdx, numChr);		
		//SerialOut("\nftdx= " + String(cmdYaesuFTdx), true);
		break;
	case RADIO_CAT_YAESU100:
	case RADIO_CAT_YAESU757:
		//inExpectedRx = 64 '32 bytes of data in hex format - from WB8LGH
TxYaesu100:
		radioRxBytes = 15;
		cmdYaesu[0] = 0; //this value doesn't matter
		cmdYaesu[1] = 0;
		cmdYaesu[2] = 0;
		cmdYaesu[3] = 0;
		cmdYaesu[4] = 0x10; //read status command
		RADIO.write(cmdYaesu, numChr); //send it to the radio		
		break;
	case RADIO_CAT_YAESU890:
	case RADIO_CAT_YAESU1000:
		radioRxBytes = 15;
		cmdYaesu[0] = 0;
		cmdYaesu[1] = 0;
		cmdYaesu[2] = 0;
		cmdYaesu[3] = 3; //return 16 bytes with band [0], freq [1] through [4], clarifier [5] & [6], mode [7], IF filter offset [8], VFO/Mem flags [9], [10] through [15] not used
		cmdYaesu[4] = 0x10; //read status command
		RADIO.write(cmdYaesu, numChr); //send it to the radio		
		break;
	case RADIO_CAT_YAESU8x7:
		//use default msg		
		radioRxBytes = 4;
		RADIO.write(cmdYaesu, numChr); //send it to the radio		
		break;
	}

	//wait 100 mS for reply
	for (int i = 0; i < 100; i++)
	{
		delay(1);
		if (RADIO.available() > 0)
		{
			ChkRadioRx();
			radioCommBusy = false;
			return true;
		}
	}
	radioCommBusy = false;
	return false;
}



void ChkFreqModeChng()
{
	uint32_t val = 0;

	//chk for freq or mode change 
	//SerialOut("Checking FreqMode Change", true);

	if (gZeroDigits)
	{
		//zero digits below step digit
		gFreqRadio += (gUnit / 2); //round up to next significant digit
		gFreqRadio /= gUnit;		
		gFreqRadio *= gUnit;
	}

	if (gFreqRadio != lastStableFreq || gRMode != lastStableRMode)
	{
		if (freqChngTmrDefault < millis() - freqChngTmr)
		{
			// only save if we've been stable for over 10 seconds			
			freqChngTmr = millis(); //reset timer
			if (lastStableFreq == 0 && gFreqRadio >= 100000)
			{
				//we haven't been initialized yet. Do it now and return
				//SerialOut("Initializing FreqMode Change arrays", true);
				lastStableFreq = gFreqRadio;
				lastStableRMode = gRMode;
				for (int i = 0; i < 10; i++)
				{
					//preload all arrays to initial values
					prevFreq[i] = gFreqRadio;
					prevRMode[i] = gRMode;
				}
				goto ChkFreqMode;
			}
			//something has changed, push new freq/modes to stack
			for (int i = 10; i > 0; i--)
			{
				//move current values down one
				prevFreq[i] = prevFreq[i - 1];
				prevRMode[i] = prevRMode[i - 1];
			}
			prevFreq[0] = lastStableFreq; //put last stable info on top
			prevRMode[0] = lastStableRMode;			
			lastStableFreq = gFreqRadio;
			lastStableRMode = gRMode;		

			// debugg
			/*
			for (int i = 0; i < 10; i++)
			{
				SerialOut(String(i) + "= " + String(prevFreq[i]), true);
			}
			*/

		}
	}
ChkFreqMode:
	//send any changes to radio
	if (gFreqRadio != gFreqRadioPrev && gFreqRadio >= 100000)
	{			
		switch (gVFO)
		{
		case MNU_VFO_A:
		case MNU_VFO_B:
			gVFO_Freq[gVFO] = gFreqRadio;
			break;
		case MNU_VFO_SPLIT:
			if (gTxMode < RADIO_TX_KEY)
			{
				gVFO_Freq[MNU_VFO_A] = gFreqRadio;
			}
			else
			{
				gVFO_Freq[MNU_VFO_B] = gFreqRadio;
			}
			break;
		case MNU_VFO_MEM:
			if (!gMSelMode)
			{
				gVFO_Freq[MNU_VFO_A + 2] = gFreqRadio;
			}
			else
			{
				gVFO_Freq[MNU_VFO_B + 2] = gFreqRadio;
			}
			break;
		}
		
		gRIT = 0; //zero when tuning freq
		//SetRIT();
		if (TxFreq2Radio(gFreqRadio))
		{	
			if (gRadioCAT == RADIO_CAT_FLEX)
			{
				gFlexLastFreqUpdateTmr = millis(); //wait one second before allowing tx_freq to update gRadioFreq otherwise chk will overwrite it
			}
			if (gVFO != MNU_VFO_SPLIT)
			{
				gSaveRadio = true;
			}
			PrtFreq(false);
			PrtRIT();
			//SerialOut("Change Freq to " + String(gFreqRadio), true);						
			PrtMemLabel(" "); //prints memory label if it finds a match

			//if Link is connected, send new gFreqRadio to other Mini if we're not tracking			
			/*if ((gBLEConnected || gUDPConnected) && gFreqRadio != gFreqRadioPrev)
			{
				SendLinkParams();
			}
			*/
			gFreqRadioPrev = gFreqRadio; //set this after we've sent it to other end of BLE link
		}						
	}

	if (gRMode != gRModePrev)
	{		
		PrtRMode(true);
		if (gVFO != MNU_VFO_MEM)
		{
			gVFO_Mode[gVFO] = gRMode;			
		}				
		
		if (TxMode2Radio())
		{
			gRModePrev = gRMode;
		}	
	}
}

void PopPrevFreqMode()
{
	if (gLock)
	{
		return; //do nothing if locked
	}

	//pop the last freq & mode off the stack and move everything down one
	gFreqRadio = prevFreq[0];
	gRMode = prevRMode[0];
	//now move the others down
	for (int i = 0; i < 9; i++)
	{
		prevFreq[i] = prevFreq[i + 1];
		prevRMode[i] = prevRMode[i + 1];
	}

	//match up VFOs
	/*
	if (gVFO == MNU_VFO_B || (gVFO = MNU_VFO_MEM && gMSelMode))
	{
		gVFOB = gFreqRadio;
		gVFOB_Mode = gRMode;
		gVFOB_BW = gPCRBW;
	}
	else
	{
		gVFOA = gFreqRadio;
		gVFOA_Mode = gRMode;
		gVFOA_BW = gPCRBW;
	}
	*/
	gVFO_Freq[gVFO] = gFreqRadio;
	gVFO_Mode[gVFO] = gRMode;
	gVFO_BW[gVFO] = gPCRBW;

	lastStableFreq = gFreqRadio;
	lastStableRMode = gRMode;
	gFreqRadioPrev = gFreqRadio;
	gRModePrev = gRMode;
	//don't chk for change, just change radio
	TxFreq2Radio(gFreqRadio);
	TxMode2Radio();
	gSaveRadio = true;	
}

void ChkRadioRx()
{
	//wait for rx from radio	
	if (RADIO.available() > 0 && radioRxBytes > 0)
	{
		if ((gBLEConnected || gUDPConnected) && (gSyncMode == SYNC_MODE_RMT && gLinkMiniA))
		{
			while (RADIO.available())
			{
				byte junk = RADIO.read();
			}
			return;
		}
		//SerialOut("Rxing data", true);
		int radioMsgCtr = 0;

		byte rx = 0;

		long radioMsgTimeout = millis();//set start of rx window

Wait4Chr:
		while (radioMsgTimeoutDefault > millis() - radioMsgTimeout)
		{
			//wait for reply			
			if (RADIO.available() > 0)
			{
				rx = RADIO.read();
				//SerialOut(String(rx, HEX) + "-", false);

				//wait for start bytes
				switch (gRadioCAT)
				{
				case RADIO_CAT_ICOM:
					//must start with 0xfe 0xfe and end with 0xfd, and be the correct length
					if ((radioMsgCtr == 0) && (rx != 0xfe))
					{
						//SerialOut("Invalid first Icom byte", true);
						goto Wait4Chr;
					}
					else if ((radioMsgCtr == 1) && (rx != 0xfe))
					{
						//SerialOut("Invalid second Icom byte", true);
						radioMsgCtr = 0;
						goto Wait4Chr;
					}
					else if ((radioMsgCtr == 2) && (rx != 0xe0))
					{
						//msg with master "from" address - this is proably an echo of the tx cmd, wait for correct formt for reply
						//SerialOut("Not from Icom radio", true);
						radioMsgCtr = 0;
						goto Wait4Chr;
					}
					else if ((radioMsgCtr == radioRxBytes) && (rx != 0xfd))
					{
						//bad end chr, recync
						//SerialOut("Bad Icom Sync", true);
						radioMsgCtr = 0;
						goto Wait4Chr;
					}
					else if ((radioMsgCtr > 1 && radioMsgCtr < radioRxBytes - 1))
					{
						//check for sync chrs in the middle in case we join mid stream
						if (rx == 0xfe)
						{
							//SerialOut("Out of sync with Icom data", true);
							radioMsgCtr = 0; //accept this as the start of the next packet and continue							
						}
						else if (rx == 0xfd)
						{
							//SerialOut("Icom term byte rcvd mid-msg", true);
							radioMsgCtr = 0; //start over
							goto Wait4Chr;
						}
					}
					break;
				case RADIO_CAT_KENWOOD1:
Kenwood:
					//must start with "IF" and end with ";" and be the correct length
					if ((radioMsgCtr == 0) && (rx != 'I'))
					{
						//wait for a 'I'
						goto Wait4Chr;
					}
					else if ((radioMsgCtr == 1) && (rx != 'F'))
					{
						//reset if we don't get 'F' in the second position
						radioMsgCtr = 0;
						goto Wait4Chr;
					}
					else if ((radioMsgCtr != radioRxBytes) && (rx == ';'))
					{
						//reset if we get a ';' midstream
						radioMsgCtr = 0;
						goto Wait4Chr;
					}
					break;
				case RADIO_CAT_KENWOOD2:
					goto Kenwood;
					break;
				case RADIO_CAT_YAESUFTDX:
					//must start with IF and end with ";' and be correct length
					if ((radioMsgCtr == 0) && (rx != 'I'))
					{
						//wait for start of msg
						goto Wait4Chr;
					}
					else if ((radioMsgCtr == 1) && (rx != 'F'))
					{
						radioMsgCtr = 0;
						goto Wait4Chr;
					}
					else if ((radioMsgCtr + 1 != radioRxBytes) && (rx == ';'))
					{
						//IGNORE LOCATION OF TERMINATOR (;) - RADIO HAS A BUG THAT CAN ADD ADDITIONAL CHRS TO REPLY SO JUST LOOK FOR ; AT THE END OF THE MESSAGE
						//terminator in wrong bin
						//radioMsgCtr = 0;
						//goto Wait4Chr;
					}
					break;
				case RADIO_CAT_YAESU100:
					//no validation required for Yaesu
					break;
				}
				//SerialOut(String(radioMsgCtr) + ":" + String(rx, HEX) + " ", false);
				radioRx[radioMsgCtr] = rx;
				if (radioMsgCtr == radioRxBytes || (radioRx[radioMsgCtr] == ';' && gRadioCAT == RADIO_CAT_YAESUFTDX))
				{
					//SerialOut(" Got one!", true);
					DecRadioRx(radioMsgCtr);
					return; //we're done
				}
				radioMsgCtr++;
				goto Wait4Chr;
			}
			goto Wait4Chr;
		}
		//clear buffer if rx fails
		for (int i = 0; i < radioRxBufMax; i++)
		{
			radioRx[i] = 0;
		}
		//SerialOut("--> err - Radio Fail!", true);
		if (radioMsgCtr > 1)
		{
			//ignore single random chrs
			//PrtStatus("Radio Fail!", true);			
		}
		else
		{
			//SerialOut(" ignore junk from radio ", true);
		}
	}
}

void DecRadioRx(int msgCtr)
{
	//decodes msg rx from ChkRadioRx()	
	uint32_t  newFreq = 0;
	int RIT = 0;
	switch (gRadioCAT)
	{
	case RADIO_CAT_ICOM:
		if (radioRx[0] == 0xfe && radioRx[1] == 0xfe && radioRx[radioRxBytes] == 0xfd)
		{
			if (radioRxBytes == 10)
			{
				//decode freq				
				newFreq = ((radioRx[9] & 0xf) * 100000000) + ((radioRx[8] >> 4) * 10000000) + ((radioRx[8] & 0xf) * 1000000) + ((radioRx[7] >> 4) * 100000) + ((radioRx[7] & 0xf) * 10000) + ((radioRx[6] >> 4) * 1000) + ((radioRx[6] & 0xf) * 100) + ((radioRx[5] >> 4) * 10) + (radioRx[5] & 0xf);
				//SerialOut("gFreqRadio from Icom= " + String(newFreq), true);
				if (newFreq < 100000)
				{
					//invalid reply, ignore and exit
					SerialOut("Invalid freq rcvd = " + String(newFreq), true);					
					return;
				}
				Chk4SyncZeroBeat(newFreq, false);
				gFreqRadio = newFreq;
				gFreqRadioPrev = newFreq; //so we don't send it back to radio
				PrtFreq(false);
				gIcmAdr = radioRx[3]; //get "from" address				
			}
			else if (radioRxBytes == 7)
			{
				gRMode = radioRx[5];
				//match hmi mode values
				switch (gRMode)
				{
				case 0: //LSB
					gRMode = RADIO_LSB;
					break;
				case 1: //USB
					gRMode = RADIO_USB;
					break;
				case 2: //AM
					gRMode = RADIO_AM;
					break;
				case 3: //CW
					gRMode = RADIO_CW;
					break;
				case 4: //RTTY
					gRMode = RADIO_DIG_H; //DigU
					break;
				case 5: //FMn
					gRMode = RADIO_FM_N;
					break;
				case 6: //FMw
					gRMode = RADIO_FM_W;
					break;
				case 7: //CWr
					gRMode = RADIO_CW_R;
					break;
				case 8: //RTTY-r
					gRMode = RADIO_DIG_L; //DigL
					break;
				case 0x12: //Dig
					gRMode = RADIO_DIG_H; //DigU
					break;
				case 0x13:
					gRMode = RADIO_DIG_L; //DigL - not sure about this mode - DL40r.de shows it as 0x12, same as Dig normal
					break;
				}				
			}
		}
		break;
	case RADIO_CAT_KENWOOD1:
	case RADIO_CAT_KENWOOD2:
		if (radioRx[0] == 'I' && radioRx[1] == 'F' && radioRxBytes == 37 && radioRx[radioRxBytes] == ';')
		{

			//must start with "IF", be 38 bytes long, and end with ";" to be valid
			//not using the 10ghz or 1ghz digits in [2] and [3]


			
			//debug
			/*
			for (int i = 0; i < radioRxBytes; i++)
			{
				SerialOut(String(radioRx[i], HEX) + " ", false);
			}
			SerialOut("", true);
			SerialOut("# of Rx bytes= " + String(radioRxBytes), true);
			*/

			//info sent as ascii chars - strip off 4msb
			newFreq = ((radioRx[4] & 0x0f) * 100000000) + ((radioRx[5] & 0x0f) * 10000000) + ((radioRx[6] & 0x0f) * 1000000) + ((radioRx[7] & 0x0f) * 100000) + ((radioRx[8] & 0x0f) * 10000)
				+ ((radioRx[9] & 0x0f) * 1000) + ((radioRx[10] & 0x0f) * 100) + ((radioRx[11] & 0x0f) * 10) + (radioRx[12] & 0x0f);
			//SerialOut("gFreqRadio from Kenwood= " + String(newFreq), true);
			if (newFreq < 100000)
			{
				//invalid reply, ignore and exit
				SerialOut("Invalid freq rcvd = " + String(newFreq), true);				
				return;
			}
			Chk4SyncZeroBeat(newFreq, false);
			gFreqRadio = newFreq;
			gFreqRadioPrev = newFreq; //so we don't send it back to radio
			PrtFreq(false);
			gRMode = radioRx[29] & 0x0f;
			//match hmi mode values
			switch (gRMode)
			{
			case 1: //LSB
				gRMode = RADIO_LSB;
				break;
			case 2: //USB
				gRMode = RADIO_USB;
				break;
			case 3: //CW
				gRMode = RADIO_CW;
				break;
			case 4: //FM
				gRMode = RADIO_FM_W;
				break;
			case 5: //AM
				gRMode = RADIO_AM;
				break;
			case 6: //fsk
				gRMode = RADIO_DIG_L;
				break;
			case 7: //CWr
				gRMode = RADIO_CW_R;
				break;
			case 9: //FSKr
				gRMode = RADIO_DIG_H; //DigU
				break;
			}
			//RIT			
			if (radioRx[23] == '0')
			{
				//clarifier off
				RIT = 0;
			}
			else
			{
				//clarifier on, decode freq
				RIT = ((radioRx[19] & 0x0f) * 1000) + ((radioRx[20] & 0x0f) * 100) + ((radioRx[21] & 0x0f) * 10) + (radioRx[22] & 0x0f);
			}
			if (radioRx[18] == '-')
			{
				//negative value
				RIT *= -1;
			}
			if (gRIT != RIT)
			{
				gRIT = RIT;
				PrtRIT();
			}
		}
		break;
	case RADIO_CAT_YAESUFTDX:
		//allow for random length replies to IF; request - bug in radio causes random data appended to end of IF; reply but aLWAYS TERMINATED WITH A ;
		//SerialOut("\nmsgCtr=" + String(msgCtr), true);
		if (radioRx[0] == 'I' && radioRx[1] == 'F' && radioRx[msgCtr] == ';')
		{

			/*
			//debug
			for (int i = 0; i < radioRxBytes; i++)
			{
				SerialOut(String(radioRx[i], HEX) + " ", false);
			}
			SerialOut("", true);
			SerialOut("# of Rx bytes= " + String(radioRxBytes), true);
			*/

			//info sent as ascii chars - strip off 4msb
			newFreq = ((radioRx[5] & 0x0f) * 100000000) + ((radioRx[6] & 0x0f) * 10000000) + ((radioRx[7] & 0x0f) * 1000000) + ((radioRx[8] & 0x0f) * 100000) + ((radioRx[9] & 0x0f) * 10000)
				+ ((radioRx[10] & 0x0f) * 1000) + ((radioRx[11] & 0x0f) * 100) + ((radioRx[12] & 0x0f) * 10) + (radioRx[13] & 0x0f);
			//SerialOut("gFreqRadio from Yaesu FTdx " + String(newFreq), true);
			if (newFreq < 100000)
			{
				//invalid reply, ignore and exit
				SerialOut("Invalid freq rcvd = " + String(newFreq), true);				
				return;
			}
			Chk4SyncZeroBeat(newFreq, false);
			gFreqRadio = newFreq;
			gFreqRadioPrev = newFreq; //so we don't send it back to radio
			PrtFreq(false);
			gRMode = radioRx[21] & 0x0f;
			//match hmi mode values
			switch (gRMode)
			{
			case 1: //LSB
				gRMode = RADIO_LSB;
				break;
			case 2: //USB
				gRMode = RADIO_USB;
				break;
			case 3: //CW
				gRMode = RADIO_CW;
				break;
			case 4: //FM
				gRMode = RADIO_FM_W;
				break;
			case 5: //AM
				gRMode = RADIO_AM;
				break;
			case 6: //FSK
				gRMode = RADIO_DIG_L; //DigL
				break;
			case 7: //CWr
				gRMode = RADIO_CW_R;
				break;
			case 8: //Dig low
				gRMode = RADIO_DIG_L;
				break;
			case 0x0C:
				gRMode = RADIO_DIG_H;
				break;
			default:
				gRMode = RADIO_USB; //unknown mode
				break;
			}
			if (radioRx[19] == '0')
			{
				//clarifier off
				RIT = 0;
			}
			else
			{
				//clarifier on, decode freq
				RIT = ((radioRx[15] & 0x0f) * 1000) + ((radioRx[16] & 0x0f) * 100) + ((radioRx[17] & 0x0f) * 10) + (radioRx[18] & 0x0f);
			}
			if (radioRx[14] == '-')
			{
				//negative value
				RIT *= -1;
			}
			if (gRIT != RIT)
			{
				gRIT = RIT;
				PrtRIT();
			}			
		}
		break;
	case RADIO_CAT_YAESU100:
	case RADIO_CAT_YAESU757:
		//response from status update cmd
		newFreq = (radioRx[1] << 24) + (radioRx[2] << 16) + (radioRx[3] << 8) + radioRx[4]; //encoded as a 32 bit #
		//SerialOut("gFreqRadio from Yaesu100= " + String(newFreq), true);
		if (newFreq < 100000)
		{
			//invalid reply, ignore and exit
			SerialOut("Invalid freq rcvd = " + String(newFreq), true);			
			return;
		}
		Chk4SyncZeroBeat(newFreq, false);
		gFreqRadio = newFreq;
		gFreqRadioPrev = newFreq; //so we don't send it back to radio
		PrtFreq(false);
		gRMode = radioRx[5] & 0x0F;
		//match mode to hmi rMode
		switch (gRMode)
		{
		case 0: //LSB
			gRMode = RADIO_LSB;
			break;
		case 1: //USB
			gRMode = RADIO_USB;
			break;
		case 2: //CW
			gRMode = RADIO_CW;
			break;
		case 3: //CWr
			gRMode = RADIO_CW_R;
			break;
		case 4: //AM			
			gRMode = RADIO_AM;
			break;
		case 5: //Dig on FT100, not used on 8x7
			gRMode = RADIO_DIG_H;
			break;
		case 6: //FMn on FT100
			gRMode = RADIO_FM_N;
			break;
		case 7: //FMw on FT100
			gRMode = RADIO_FM_W;
			break;
		case 8: //FM
			gRMode = RADIO_FM_N;
			break;
		case 0xA: //Dig
			gRMode = RADIO_DIG_H; //DigU
			break;
		case 0xC: //Pkt
			gRMode = RADIO_DIG_L; //DigL
		}
		//============= this code hasn't been tested =================
		//decode clarifier
		//get freq 
		RIT = int16_t(radioRx[10] << 8 + radioRx[11]); //this is a 16-bit# using 2's compliment					 
		if (gRIT != RIT)
		{
			gRIT = RIT;
			PrtRIT();
		}
		break;
	case RADIO_CAT_YAESU890:
	case RADIO_CAT_YAESU1000: //return 16 bytes with band [0], freq [1] through [4], clarifier [5] & [6], mode [7], IF filter offset [8], VFO/Mem flags [9], [10] through [15] not used
		newFreq = (radioRx[1] << 24) + (radioRx[2] << 16) + (radioRx[3] << 8) + radioRx[4]; //encoded as a 32 bit #
		//SerialOut("gFreqRadio from Yaesu1000= " + String(newFreq), true);
		if (newFreq < 100000)
		{
			//invalid reply, ignore and exit
			SerialOut("Invalid freq rcvd = " + String(newFreq), true);			
			return;
		}
		Chk4SyncZeroBeat(newFreq, false);
		gFreqRadio = newFreq;
		gFreqRadioPrev = newFreq; //so we don't send it back to radio
		PrtFreq(false);
		gRMode = radioRx[7];
		//convert radio mode to hmi modes
		switch (gRMode)
		{
		case 0: //LSB
			gRMode = RADIO_LSB;
			break;
		case 1: //USB
			gRMode = RADIO_USB;
			break;
		case 2: //CW
			gRMode = RADIO_CW;
			break;
		case 3: //AM
			gRMode = RADIO_AM;
			break;
		case 4: //FM
			gRMode = RADIO_FM_N;
			break;
		case 5: //RTTY
			gRMode = RADIO_DIG_H; //DigU
			break;
		case 6: //Pkt
			gRMode = RADIO_DIG_L; //DigL
			break;
		}

		//============= this code hasn't been tested =================
		//decode clarifier
		if (radioRx[9] & 0x04)
		{
			//clarifier on - get freq 
			RIT = int16_t(radioRx[5]<< 8 + radioRx[6]); //this is a 16-bit# using 2's compliment			
		}
		else
		{
			RIT = 0;
		}
		if (gRIT != RIT)
		{
			gRIT = RIT;
			PrtRIT();
		}
		break;
	case RADIO_CAT_YAESU8x7:
		//chk mode first - if it's invalid, throw this packet away
		int val = radioRx[4];
		//match mode to hmi rMode
		switch (val)
		{
		case 0: //LSB
			gRMode = RADIO_LSB;
			break;
		case 1: //USB
			gRMode = RADIO_USB;
			break;
		case 2: //CW
			gRMode = RADIO_CW;
			break;
		case 3: //CWr
			gRMode = RADIO_CW_R;
			break;
		case 4: //AM			
			gRMode = RADIO_AM;
			break;
		case 6: //FMn on FT100
			gRMode = RADIO_FM_N;
			break;
		case 7: //FMw on FT100
			gRMode = RADIO_FM_W;
			break;
		case 8: //FM
			gRMode = RADIO_FM_N;
			break;
		case 0xA: //Dig
			gRMode = RADIO_DIG_H; //DigH
			break;
		case 0xC: //Pkt
			gRMode = RADIO_DIG_L; //DigL
			break;
		default:
			return; //bad mode = bad packet
			break;
		}
		//freq reply
		//SerialOut("\nRx: " + String(radioRx[0], HEX) + " " + String(radioRx[1], HEX) + " " + String(radioRx[2], HEX) + " " + String(radioRx[3], HEX)  + " " + String(radioRx[4], HEX), true);
		newFreq = ((radioRx[0] >> 4) * 100000000) + ((radioRx[0] & 0x0f) * 10000000) + ((radioRx[1] >> 4) * 1000000) + ((radioRx[1] & 0x0f) * 100000) + ((radioRx[2] >> 4) * 10000) + ((radioRx[2] & 0x0f) * 1000) + ((radioRx[3] >> 4) * 100) + ((radioRx[3] & 0x0f) * 10);
		//SerialOut("gFreqRadio from Yaesu8x7= " + String(newFreq), true);
		if (newFreq < 100000)
		{
			//invalid reply, ignore and exit
			SerialOut("Invalid freq rcvd = " + String(gFreqRadio), true);			
			return;
		}		
		Chk4SyncZeroBeat(newFreq, false);		
		gFreqRadio = newFreq;
		gFreqRadioPrev = newFreq; //so we don't send it back to radio
		PrtFreq(false);


	}

	//update global values if we're not memory tuning
	
	if (gVFO != MNU_VFO_MEM)
	{
		if (gVFO == MNU_VFO_B || (gVFO == MNU_VFO_SPLIT && gTxMode > RADIO_TX_ENAB))
		{
			gVFO_Freq[MNU_VFO_B] = gFreqRadio;
			gVFO_Mode[MNU_VFO_B] = gRMode;
		}
		else
		{
			gVFO_Freq[MNU_VFO_A] = gFreqRadio;
			gVFO_Mode[MNU_VFO_A] = gRMode;
		}
	}		
	
	PrtRMode(false);
}

void Chk4SyncZeroBeat(uint32_t& newFreq, bool tx)
{
	if (!gLinkMiniA && (gLink_ZeroBeat != 0) && (gSyncMode == SYNC_MODE_FREQ) && (gBLEConnected || gUDPConnected))
	{
		//offset freq by offset if Mini-B		
		//if tx=false, subtract offset from rcvr freq, if true, add it to tx freq
		if (!tx)
		{
			newFreq -= gLink_ZeroBeat;
		}
		else
		{
			newFreq += gLink_ZeroBeat;
		}
		
	}
}

//===================================================================================
// rigctld routines
//===================================================================================

bool RigCtld_Get_FreqMode()
{
	if (gTCPCon)
	{
		uint8_t junk = 0;
		uint32_t newFreq = 0;
		String txt = "";
		while (tcp.available())
		{
			junk = tcp.read();
		}
		tcp.print("f\n");
		delay(50);
		if (tcp.available())
		{
			txt = tcp.readStringUntil('\n');
			//SerialOut("freq=" + txt, true);
			if (txt.toInt() > 0)
			{
				newFreq = txt.toInt();
				Chk4SyncZeroBeat(newFreq, false);
				gFreqRadio = newFreq;
			}
		}
		delay(20);
		while (tcp.available())
		{
			junk = tcp.read();
		}
		tcp.print("m\n");
		delay(50);
		if (tcp.available())
		{
			txt = tcp.readStringUntil('\n');
			//SerialOut("mode=" + txt + ", len= " + String(txt.length()), true);
			if (txt == "LSB")
			{
				gRMode = RADIO_LSB;
			}
			else if (txt == "USB")
			{
				gRMode = RADIO_USB;
			}
			else if (txt == "CW")
			{
				gRMode = RADIO_CW;
			}
			else if (txt == "CWR")
			{
				gRMode = RADIO_CW_R;
			}
			else if (txt == "RTTY" || txt == "PKTLSB")
			{
				gRMode = RADIO_DIG_L;
			}
			else if (txt == "RTTYR" || txt == "PKTUSB")
			{
				gRMode = RADIO_DIG_H;
			}
			else if (txt == "AM")
			{
				gRMode = RADIO_AM;
			}
			else if (txt == "FM" || txt == "PKTFM")
			{
				gRMode = RADIO_FM_N;
			}
			else if (txt == "FMW")
			{
				gRMode = RADIO_FM_W;
			}
		}
		if (tcp.available())
		{
			txt = tcp.readStringUntil('\n');
			//SerialOut("bw=" + txt, true);
		}
		delay(20);
		RigCtld_Get_RIT();
	}
	return true;
}

bool RigCtld_Get_RIT()
{
	static bool ignoreRIT = false; //goes true if we get a 'RPRT -11' reply - not supported
	if (gTCPCon && !ignoreRIT)
	{
		uint8_t junk = 0;
		String txt = "";
		while (tcp.available())
		{
			junk = tcp.read();
		}
		tcp.print("j\n");
		delay(50);
		if (tcp.available())
		{
			txt = tcp.readStringUntil('\n');
			//SerialOut("rit=" + txt, true);
			if (txt == "RPRT -11")
			{
				ignoreRIT = true;
				return false;
			}			
			if (txt.toInt() > 0)
			{
				gRIT = txt.toInt();
			}
		}
	}
	return true;
}

bool RigCtld_Set_VFO()
{
	//set VFO to A when opening
	if (gTCPCon)
	{
		String txt = "";
		uint8_t junk = 0;
		while (tcp.available())
		{
			junk = tcp.read();
		}
		tcp.print("V VFOA \n");
		//SerialOut("VFO = VFOA", true);
		delay(50);
		if (tcp.available())
		{
			txt = tcp.readStringUntil('\n');
			//SerialOut("rxd v= " + txt, true);
		}
	}
	return true;
}

bool RigCtld_Set_Freq(uint32_t freq)
{
	if (gTCPCon)
	{
		String txt = "";
		uint8_t junk = 0;
		while (tcp.available())
		{
			junk = tcp.read();
		}
		tcp.print("F " + String(freq) + "\n");
		//SerialOut("F " + String(freq), true);
		delay(50);
		if (tcp.available())
		{
			txt = tcp.readStringUntil('\n');
			//SerialOut("rxd f= " + txt, true);
		}
	}
	return true;
}

bool RigCtld_Set_Mode()
{
	if (gTCPCon)
	{
		String txt = "";
		uint8_t junk = 0;
		while (tcp.available())
		{
			junk = tcp.read();
		}
		switch (gRMode)
		{
		case RADIO_LSB:
			txt = "LSB";
			break;
		case RADIO_USB:
			txt = "USB";
			break;
		case RADIO_CW:
			txt = "CW";
			break;
		case RADIO_CW_R:
			txt = "CWR";
			break;
		case RADIO_AM:
			txt = "AM";
			break;
		case RADIO_FM_N:
			txt = "FM";
			break;
		case RADIO_FM_W:
			txt = "WFM";
			break;
		case RADIO_DIG_L:
			txt = "PKTLSB";
			break;
		case RADIO_DIG_H:
			txt = "PKTUSB";
			break;
		}
		tcp.print("M " + txt + " 0 " + "\n"); //send 0 for no passband update
		//SerialOut("M " + txt, true);

		delay(50);
		if (tcp.available())
		{
			txt = tcp.readStringUntil('\n');
			//SerialOut("rxd m= " + txt, true);
		}
	}
	return true;
}

bool RigCtld_Set_RIT()
{
	if (gTCPCon)
	{
		String txt = "";
		uint8_t junk = 0;
		while (tcp.available())
		{
			junk = tcp.read();
		}
		tcp.print("3 " + String(gRIT) + "\n");
		//SerialOut("3 " + String(gRIT), true);
		delay(50);
		if (tcp.available())
		{
			txt = tcp.readStringUntil('\n');
			//SerialOut("rxd f= " + txt, true);
		}
	}
	return true;
}