// =====================================================================================================================================================================================================
// (c) 2022 Lynn Hansen, KU7Q															                                                                                                               |
// This Source Code Form is subject to the terms of the GNU GENERAL PUBLIC LICENSE, Version 3, 29 June 2007. A copy of this license can be found here: https://choosealicense.com/licenses/gpl-3.0/|
// =====================================================================================================================================================================================================




//This module contains the Setup code

void DoSetupTasks()
{
	Serial.begin(115200);	

	RADIO.begin(9600);
	tft.begin();

	tft.setRotation(3); //normal mode
	tft.fillScreen(TFT_BLACK);

	//top buttons
	pinMode(WIO_KEY_A, INPUT_PULLUP);
	pinMode(WIO_KEY_B, INPUT_PULLUP);
	pinMode(WIO_KEY_C, INPUT_PULLUP);

	//5 way switch 
	pinMode(WIO_5S_UP, INPUT_PULLUP);
	pinMode(WIO_5S_DOWN, INPUT_PULLUP);
	pinMode(WIO_5S_LEFT, INPUT_PULLUP);
	pinMode(WIO_5S_RIGHT, INPUT_PULLUP);
	pinMode(WIO_5S_PRESS, INPUT_PULLUP);

	//set buzzer pin as output
	pinMode(WIO_BUZZER, OUTPUT);

	pinMode(LCD_BACKLIGHT, OUTPUT);
	digitalWrite(LCD_BACKLIGHT, HIGH); // turn on the backlight

	//interface pins
	pinMode(SIDETONE, OUTPUT);
	pinMode(KEY_OUT, OUTPUT);
	pinMode(PTT_OUT, OUTPUT);
	pinMode(PDL_LEFT, INPUT_PULLUP);
	pinMode(PDL_RIGHT, INPUT_PULLUP);
	pinMode(ROTARY_A, INPUT_PULLUP);
	pinMode(ROTARY_B, INPUT_PULLUP);
	pinMode(ROTARY_SW, INPUT_PULLUP);
	pinMode(RMT_BTNS, INPUT_PULLUP);

	if (!SD.begin()) 
	{
		SerialOut("SD card failed", true);
		PrtSplash();
		PrtStatus("Bad or no SD card installed", true);
		while (1)
		{
			//just spin our wheels we're dead in the water
		}
	}

	delay(100);
	SerialOut("Initializing CTR2-Mini", true);
	LoadInitFile();
	delay(100);
	PrtSplash();	

	Chk4Folders();	 //gets gRadioPort from file before we load the radio
	LoadRadioTags();
	LoadRadio();	

	CalcDitDur();
	CalcDitFarnsDur();
	keyerDitDurPrev = keyerDitDur;
	keyerDitFarnsDurPrev = keyerDitFarnsDur;

	//Add this to output RTL8720 version to terminal
	//Serial.printf("RTL8720 Firmware Version: %s", rpc_system_version());
	

	//Display MAC adrs in serial terminal	
	/*
	WiFi.mode(WIFI_STA);
	String mac = WiFi.macAddress();		
	SerialOut("MAC Adrs = " + mac, true); //String(my_MAC, HEX), true);
	WiFi.disconnect(true); // turn it off until hmi asks for it	
	*/

	//Initialize hardware timer first - can't get sub timers to work as advertized

	if (ITimer0.attachInterruptInterval(HW_TIMER_INTERVAL_1MS, Do_Keyer_ISR))
	{
		//PrtStatus("Starting  ITimer0", false);
	}
	else
	{
		//PrtStatus("Can't set ITimer! FAIL", true);
	}
//Now setup the soft timers from the hardware timer - up to 16 soft timers can be added here
	//ISR_Timer.setInterval(TIMER_INTERVAL_2MS, Do_Keyer_ISR); //do this at 2mS intervals
	//ISR_Timer.setInterval(TIMER_INTERVAL_1000HZ, Do_SideTone_ISR); //do buzzer at 700hz intervals

	delay(200);

	if (!hmiTagLine.length())
	{
		hmiTagLine = "CTR2-Mini by KU7Q"; //sticky message
	}
	SerialOut("", true);
	SerialOut("Press [Enter] to start Terminal Mode", true);

	//if encoder is being held down during boot, ignore saved autoconnect state
	if (!digitalRead(ROTARY_SW))
	{
		gWiFi = false;	
		gBLEConnect = false;
	}
	else
	{
		//you can only connect to WiFi OR BLE, not both
		if (gLinkTypeUDP || gWiFiAutoConnect)
		{
			gWiFi = gWiFiAutoConnect; //we'll try to connect if this is true
		}
		else
		{
			gBLEConnect = gLinkAutoConnect;
		}		
	}

	PrtHomePg();

	
}
