// =====================================================================================================================================================================================================
// (c) 2022 Lynn Hansen, KU7Q															                                                                                                               |
// This Source Code Form is subject to the terms of the GNU GENERAL PUBLIC LICENSE, Version 3, 29 June 2007. A copy of this license can be found here: https://choosealicense.com/licenses/gpl-3.0/|
// =====================================================================================================================================================================================================


// Terminal routines for ChkPCCmd() 


void SendTermMenu()
{
	//We're in terminal mode on serial port - send starting menu		
	TermOut(TERM_CLS + " " + TERM_NORM + " " + TERM_HIDE_CURSOR, false);
	TermOut("CTR2-Mini Terminal Menu", true);	
	TermOut(" ", true);
	TermOut("---------------------------FREQUENCY/MENU CONTROL----------------------------", true);
	TermOut("     [Enter]  Encoder Click - Next Menu           [Tab]  Display this screen ", true);	
	TermOut("       [ESC]  Return to Home page            [/] or [?]  Display Menu Help   ", true);   
	TermOut("  [Up Arrow]  Increment Selection          [Down Arrow]  Decrement Selection ", true);
	TermOut("[Left Arrow]  Freq Step Up                [Right Arrow]  Freq Step Down      ", true);	
	TermOut("   [Page Up]  Open Frequency Menu           [Page Down]  Open User Menu      ", true);	
	TermOut("-------------------------------KEYER CONTROL---------------------------------", true);	
	TermOut("      [F1] Repeat last message                     [F2]  Contest Exchange    ", true);
	TermOut("      [F3] to [F9]  Send Message 3-9              [F10]  Send Call        ^  ", true);
	TermOut("Shift [F3] to [F7]  Send Message 10-14            [F11]  Send Contest SN  #  ", true);
	TermOut("  [Insert] Pause Tx Message                       [F12]  Send Exchange    %  ", true);
	TermOut("     [End] Delete Tx Message                                                 ", true);
	TermOut("    [Home] Toggle Tx Enable/Disable             [Enter]  Send Keybd Buffer   ", true);
	TermOut("-------------------------------TX MSG PROSIGNS-------------------------------", true);	
	TermOut("    [^] Call [#] SN   [%] Exch          [*] Speed Tgl                        ", true);
	TermOut("    [\"] RR   [$] SX   ['] WG   [(] KN   [)] KK   [+] AR   [-] DN   [/] DN   ", true);
	TermOut("    [:] KN   [;] AA   [<] AR   [=] BT   [>] SK   [@] AC   [[] AS   []] KN    ", true);
	TermOut(" ", true);
	TermOut("Type chrs then [Enter] to Tx, [Enter] opens Menu, [Tab] displays this screen ", true);			
	TermOut(TERM_SHOW_CURSOR, false);
}

void SendStatus2Term()
{	
	//send mini's status to terminal - include f-key and prosign mapping
	String txt = "";										
	
	TermOut("-----------------------------------------------------", true);
	TermOut("        Tx: ", false);
	if (!gTxMode)
	{
		TermOut("Disabled", true);
	}
	else
	{
		TermOut("ENABLED", true);
	}
	if (!gLock)
	{
		switch (gVFO)
		{
		case MNU_VFO_A:
			txt = "VFO-A";
			break;
		case MNU_VFO_B:
			txt = "VFO-B";
			break;
		case MNU_VFO_SPLIT:
			txt = "SPLIT";
			break;
		case MNU_VFO_MEM:
			if (!gMSelMode)
			{
				txt = "MEM-A";
			}
			else
			{
				txt = "MEM-B";
			}
			break;
		}
	}
	else
	{
		txt = "LOCKED";
	}
	TermOut("  VFO Mode: " + txt, true);
	txt = mnuRadioTag[gRadioPort];
	txt.trim();
	TermOut("Radio Port: " + txt, true);
	TermOut(" Radio CAT: " + mnuRadioCAT[gRadioCAT], true);
	TermOut("      Band: " + bandMemLabel[gBand], true);
	TermOut("      Mode: " + mnuMode[gRMode + 1], true);
	TermOut("      ", false); //for freq
	PrtHomePg();	
	TermOut("-----------------------------------------------------", true);
}

void TermOut(String txt, bool lfcr)
{
	//this routine is used by terminal mode
	if (gComMode == COM_MODE_TERM)
	{
		if (!lfcr)
		{
			Serial.print(txt);
		}
		else
		{
			Serial.print(txt + "\r\n");
		}
	}
}


void ProcessTerminalCmd(uint8_t & chrCtr)
{
	//process the terminal command received from ChkPCCmd()
	//we come here if chrs rx from pc start with 0x1b 0x5b, check remaining chrs
	//convert key sequence to keycode 	
	int kyCode = 0;

	//debug - uncomment to read keycodes in Status line
	/*
	String txt = "";
	for (int i = 0; i < chrCtr; i++)
	{
		txt += String(pcCmdRx[i], HEX) + " ";
	}
	PrtStatus("Keycode= " + txt, true);
	delay(3000);
	*/

	//chk for esc key 	
	if (pcCmdRx[0] == 27 && chrCtr == 1)
	{
		kyCode = pcCmdRx[0]; //return to home page
		goto ExitHr;
	}

	if (chrCtr == 3)
	{
		switch (pcCmdRx[2])
		{
		case 65: //up arrow
			kyCode = 218;
			break;
		case 66: //down arrow
			kyCode = 217;
			break;
		case 68: //left arrow
			kyCode = 216;
			break;
		case 67: //right arrow
			kyCode = 215;
			break;
		}
	}
	else if (chrCtr == 4 && pcCmdRx[3] == 126)
	{
		//chk for home key
		switch (pcCmdRx[2])
		{
		case 49: //home
			kyCode = 210;
			break;
		case 50: //insert			
			kyCode = 209;
			break;
		case 52: //end
			kyCode = 213;
			break;
		case 53: //pg up
			kyCode = 211;
			break;
		case 54: //pg down
			kyCode = 214;
			break;
		}
	}
	else if (chrCtr == 5 && pcCmdRx[4] == 126)
	{
		//Function Keys - longer esc strings
		switch (pcCmdRx[2])
		{
		case 49: //first chr of f1-f8 keys
			switch (pcCmdRx[3])
			{
			case 49: //f1 
				kyCode = 194;
				break;
			case 50: //f2 
				kyCode = 195;
				break;
			case 51: //f3
				kyCode = 196;
				break;
			case 52: //f4 
				kyCode = 197;
				break;
			case 53: //f5 
				kyCode = 198;
				break;
			case 55: //f6 
				kyCode = 199;
				break;
			case 56: //f7
				kyCode = 200;
				break;
			case 57: //f8
				kyCode = 201;
				break;
			}
			break;
		case 50: //first chr of f9-f12, and shift F3, F4, F5, & F6
			switch (pcCmdRx[3])
			{
			case 48: //f9
				kyCode = 202;
				break;
			case 49: //f10
				kyCode = 203;
				break;
			case 51: //f11 
				kyCode = 204;
				break;
			case 52: //f12 
				kyCode = 205;
				break;
			case 53: //shift F3
				kyCode = 303;
				break;
			case 54: //shift F4
				kyCode = 304;
				break;
			case 55: //shift F5
				kyCode = 305;
				break;
			case 56: //shift F6
				kyCode = 306;
				break;
			}
			break;
		case 51: //shift f7, f8, f9, f10
			switch (pcCmdRx[3])
			{
			case 49: //shift f7
				kyCode = 307;
				break;
			case 50: //shift f8
				kyCode = 308;
				break;
			case 51: //shift f9
				kyCode = 309;
				break;
			case 52: //shift f10
				kyCode = 310;
				break;
			}
		}
	}
	else if (chrCtr > 5)
	{
		chrCtr = 0; //too long, reset
	}
ExitHr:
	if (kyCode)
	{
		chrCtr = 0;
		keyboardChr = kyCode;		
		//process key input from terminal
	}

}
