// =====================================================================================================================================================================================================
// (c) 2022 Lynn Hansen, KU7Q															                                                                                                               |
// This Source Code Form is subject to the terms of the GNU GENERAL PUBLIC LICENSE, Version 3, 29 June 2007. A copy of this license can be found here: https://choosealicense.com/licenses/gpl-3.0/|
// =====================================================================================================================================================================================================



//This module contains all the WiFi service routines

//========================================================================================
// WIFI Station Routines
//========================================================================================

void ConnectToWiFi()
{
	blockISR = true;
	PrtStatus("Connecting to " + String(gSSID), true);
	if (gUDPConnected)
	{
		udp.stop();
		gUDPConnected = false;
	}
	if (gTCPCon)
	{
		tcp.stop();
		gTCPCon = TCP_CON_OFF;
	}
	WiFi.disconnect(true);
	WiFi.persistent(false);
	
	delay(100);
	WiFi.mode(WIFI_STA);

	char ssid[30];
	char pw[30];	
	gSSID.trim();
	gStaKey.trim();
	gSSID.toCharArray(ssid, gSSID.length()+1);
	gStaKey.toCharArray(pw, gStaKey.length() + 1);	
	const char* adrs = ssid; //must be const
	const char* key = pw;
	
	WiFi.begin(adrs, key);	
	delay(500);

	/*
	for (int i = 0; i < 60; i++)
	{
		if (WiFi.status() != WL_CONNECTED)
		{
			delay(100);
			SerialOut(".", false);
		}
	}
	*/

	if (WiFi.status() == WL_CONNECTED)
	{

		hmiTagLineRefresh = 0;
		PrtStatus("Connected: " + gSSID, true);
		gWiFi = true;
		gStaAdrs = DecIP(WiFi.localIP());
		gSubnet = DecIP(WiFi.subnetMask());
		gGW = DecIP(WiFi.gatewayIP());
		;
		SerialOut("WiFi connected!", true);
		SerialOut("Mac Address: ", false);
		Serial.println(WiFi.macAddress()); //println will accept IP address formats
		SerialOut("IP address: " + gStaAdrs, true);
		SerialOut("Subnet Mask: " + gSubnet, true);
		SerialOut("Gateway IP: " + gGW, true);

		if (gRadioCAT == RADIO_CAT_RIGCTLD)
		{
			char adrs[20];

			gIPAdrs.toCharArray(adrs, gIPAdrs.length() + 1);
			const char* fAdrs = adrs; //must be a const
			//Serial.println("IP adrs= " + gIPAdrs + ", " + String(adrs) + ", " + String(fAdrs));
						
			delay(100);
			tcp.setTimeout(10); //in seconds
			tcp = server.available();			
			delay(1000);

			if (!tcp.connected())
			{
				hmiTagLineRefresh = 0;
				PrtStatus("Connect: " + gIPAdrs, true);				
				hmiTagLineRefresh = 0;		
				// The Wio will lockup at the next line if the IP is wrong and the connection fails
				// ---> press and hold the encoder during boot to ignore the auto-connect setting
				if (!tcp.connect(fAdrs,4532,10))
				{					
					delay(1000);
					PrtStatus("rigctld Connection FAILED", true);
					SerialOut("Failed to connect to " + String(fAdrs), true);
					tcp.stop();
					gTCPCon = TCP_CON_OFF;
					return;
				}
				else
				{
					PrtStatus("Connected: " + gIPAdrs, true);
					SerialOut("Connected to " + String(fAdrs), true);
					gTCPCon = TCP_CON_ACTIVE;
					RigCtld_Set_VFO(); //default it to VFO-A
				}
			}			
		}

		if (gLinkAutoConnect)
		{
			ConnectUDP(); //connects Link UDP so we can rx udp msgs			
		}
	}
	else
	{
		hmiTagLineRefresh = 0;
		PrtStatus("Failed: "+ gSSID, true);
		SerialOut("WiFi failed to connect", true);
		gWiFi = false;
	}	
	blockISR = false;
}

String DecIP(const IPAddress & ipAddress)
{
	//decode IPAdress and return with string		
	return String(ipAddress[0]) + String(".") + \
		String(ipAddress[1]) + String(".") + \
		String(ipAddress[2]) + String(".") + \
		String(ipAddress[3]);	
}

void ConnectUDP()
{	
	gUDPAdrs.trim();
	gUDPAdrsASC.trim();
	if (!gUDPAdrs.length() && !gUDPAdrsASC.length())
	{
		return; //nothing to connect to
	}
	if (gLinkTypeUDP && (gUDPAdrs.indexOf('.') > -1) || gUDPAdrsASC.indexOf('.' > -1))
	{				
		if (gUDPAdrs.length())
		{

			int ptr = 0;
			//chk for 3 '.', up to the second to last digit
			for (int i = 0; i < gUDPAdrs.length() - 1; i++)
			{
				if (gUDPAdrs.charAt(i) == '.')
				{
					ptr++;
				}
			}
			if (ptr != 3)
			{
				goto ErrOut;
			}

			//gUDPAdrs is address of remote Mini
			ptr = gUDPAdrs.indexOf('.');
			udpAdrs[0] = gUDPAdrs.substring(0, ptr).toInt();
			ptr++;
			udpAdrs[1] = gUDPAdrs.substring(ptr, gUDPAdrs.indexOf(ptr, '.')).toInt();
			ptr = gUDPAdrs.indexOf('.', ptr + 1);
			ptr++;
			udpAdrs[2] = gUDPAdrs.substring(ptr, gUDPAdrs.indexOf(ptr, '.')).toInt();
			ptr = gUDPAdrs.indexOf('.', ptr + 1);
			ptr++;
			udpAdrs[3] = gUDPAdrs.substring(ptr).toInt();
		}

		delay(1000);
		udp.begin(WiFi.localIP(), udpPort); //open udp on this unit
		delay(250);
		gUDPConnected = true;	
		ResetLink();
		if (gUDPAdrs.length())
		{
			SendLinkParams(); //send current data to start
		}
		if (gUDPAdrsASC.length())
		{
			gAntSwAPrev = -1; //force display update on first reply from asc
			gAntSwBPrev = -1;
			SendASCCmd(ASC_CMD_STATUS, 0, 0); //get update from ASC if we have an address for it	
			SendASCCmd(ASC_CMD_STATUS, 0, 0); //send it twice
		}
		//override if we're Mini-A in remote link mode
		if ((gBLEConnected || gUDPConnected) && (gSyncMode == SYNC_MODE_RMT && gLinkMiniA))
		{
			PrtHomePg(); //refresh home page with new lables for Tag and CAT
		}
		hmiTagLineRefresh = 0;
		PrtStatus("Linked to " + gUDPAdrs, true);
		//SerialOut("started udp " + String(udpAdrs[0]) + '.' + String(udpAdrs[1]) + '.' + String(udpAdrs[2]) + '.' + String(udpAdrs[3]), true);
	}
	return;

ErrOut:
	//if we get here, display error
	hmiTagLineRefresh = 0;
	PrtStatus("IP " + gUDPAdrs + " is Invalid", true);
	delay(1500);
}

void Tx2WiFi(int dest, String txt)
{
	char msg[200];
	String cmd = "C";

	if (gWiFi && dest == WIFI_FLEX)
	{
		cmd += String(gFlexCmdNum) + txt + "\r\n\0";
		gFlexCmdNum++;
		cmd.toCharArray(msg, cmd.length());
		tcp.write(msg, cmd.length());

#ifdef DEBUG_FLEX_API
		SerialOut("\n\r------------------------------------", true);
		SerialOut("<==== Flex TCP Tx: " + String(msg), true);
		SerialOut("------------------------------------\n\r", true);
#endif		
	}
}

